/*
 *++
 *  FACILITY:
 *
 *	breakdic.c
 *	
 *  ABSTRACT:
 *
 *	This program filters an input ASCII sorted dictionary
 *	(one word per line) and writes a temporary output dictionary
 *	containing the words which are Cword compatible.
 *
 *	All words containing non alphabetic characters are discarded.
 *	All acceptable words are changed to lower case, if necessary.
 *
 *	It then breaks up this temporary dictionary (one word per
 *	line) into a series of similar length word files (also
 *	compatible with Cword). i.e.
 *
 *                  --> WORDS.01 (sorted file with words of 1 letter)
 *                  |
 *                  --> WORDS.02 (sorted file with words of 2 letters)
 *      input       |
 *      dictionary  --> WORDS.03 (sorted file with words of 3 letters)
 *      WORDS.DIC   |
 *		    --> WORDS.04 (sorted file with words of 4 letters)
 *		    |
 *		    --> ...
 *
 *	N.B. This is a standalone program and has no function
 *	dependencies on any other Cword modules, except for a few
 *	constants from main.h.
 *
 *  AUTHORS:
 *
 *      James A. Lundon
 *
 *  CREATION DATE:
 *
 *      6 December 1993 (original program functionality is lost)
 *--
 */

/*
 *  Include Files
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>

#if defined(VMS) || defined(__VMS)

#include <unixio.h>

#endif

#if defined(__unix) || defined(__unix__)

#include <unistd.h>

#endif

#if defined(__MSDOS__) || defined(MSDOS)

#include <io.h>

#endif

/*
 *  Include useful Cword header files.
 */

#include "main.h"
#include "breakdic.h"

#define PREFIX_NEWDICT_FILE  "clean"

/*
 *  Module global used to store the word file prefix.
 */

static char breakdic_l_prefix_str[CWORD_SIZEOF_FILESPEC + 1];



/*
 *  Private function prototypes.
 */

PRIVATE int breakdic__access_file (
	char			*file_spec);

PRIVATE int breakdic__write_word_files (
	char			*dict_word);

PRIVATE int breakdic__read_dic_and_break (
	const char		*dict_str);



PUBLIC BOOLEAN breakdic_remove_file (
	char			*fatal_str)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *      Funcvtion to remove the inputted fatal_str file.
 *
 *  FORMAL PARAMETERS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      TRUE			File deleted.
 *	FALSE			Problem deleting file.
 *--
 */

{

    if (fatal_str[0] == (char) 0)
    {
	return (FALSE);
    }

    if (remove (fatal_str) == 0)
    {
	printf (
	    "BRKDIC-I-DELETE, Temporary dictionary file %s deleted\n",
	    fatal_str);

	return (TRUE);
    }
    else
    {
	printf (
	    "BRKDIC-E-ERROR, Error deleting temporary dictionary file %s \n",
	    fatal_str);

	return (FALSE);
    }

}



PRIVATE int breakdic__access_file (
	char			*file_spec)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Portable way of checking for the existence of a specified
 *	input file.
 *
 *  FORMAL PARAMETERS:
 *
 *      *file_spec		(r) File spec to check for.
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      TRUE			Access to file allowed.
 *	FALSE			Access to file is disallowed.
 *--
 */

{

    /*
     *  OS dependent!
     */

#if defined(__MSDOS__) || defined(MSDOS)
    if (_access (file_spec, 0) != 0)
#else
    if (access (file_spec, 0) != 0)
#endif
    {
	return (FALSE);
    }
    else
    {
	return (TRUE);
    }

}



PUBLIC int breakdic_cleanup_dic (
	char			*indict_str,
	char			*outdict_str)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function that takes an input dictionary (*indict_str) and
 *	outputs a temporary, clean dictionary file (*outdict_str)
 *	which will be then used to create valid Cword word files
 *	with.  When this is completed this temporary file is deleted.
 *
 *  FORMAL PARAMETERS:
 *
 *	*indict_str		(r) Input dictionary spec.
 *	*outdict_str		(r) Output temporary dictionary spec.
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      TRUE			Output temporary dictionary written successfully.
 *	FALSE			Output temporary dictionary not written successfully.
 *--
 */

{

    FILE *input_fp,
    	 *output_fp;

    char reference_str[CWORD_SIZEOF_WORD + 1];

    register long words_total = 0,
		  words_ok = 0,
		  words_failed = 0;

    unsigned int count;

    BOOLEAN is_word_ok;

    if (breakdic__access_file (indict_str) == FALSE)
    {
	printf ("BRKDIC-E-ACCESS, Access to input dictionary not allowed\n");

	return (FALSE);
    }

    errno = 0;
    input_fp = fopen (
		    indict_str, 
		    "r");

    if (input_fp == (FILE *) 0)
    {
    	perror ("BRKDIC-E-ERROR, Input dictionary file error");

    	return (FALSE);
    }

    /*
     *  Temporary dictionary file specification
     */

    sprintf (
	outdict_str,
	"%s%s",
	PREFIX_NEWDICT_FILE,
	CWORD_SUFFIX_DICT_FILE);

    errno = 0;
    output_fp = fopen (outdict_str, "w");

    if (output_fp == (FILE *) 0)
    {
    	perror ("BRKDIC-E-ERROR, Output dictionary file error");

    	return (FALSE);
    }
    else
	printf (
	    "BRKDIC-I-CREATE, Creating temporary dictionary: %s\n",
	    outdict_str);

    while (fscanf (
	    input_fp, 
	    "%s\n",
	    reference_str) != EOF)
    {
    	is_word_ok = TRUE;

        for (count = 0; count < strlen (reference_str) - 1; count++)
        {
	    /*
	     *	First make all letters lower case and then check whether they
	     *	are within the ASCII range 97 to 122.  If they aren't then
	     *	eject the word from the list to be written to the output file
	     *	defined by NEW_WORDS_FILE.
	     */

            reference_str[count] = (char) tolower(reference_str[count]);

    	    if ( (reference_str[count] < 97) || (reference_str[count] > 122) )
    	    {
    	        is_word_ok = FALSE;
    		break;
    	    }
        }

    	if (is_word_ok == TRUE)
    	{
            if (fprintf (
		    output_fp, 
		    "%s\n", 
		    reference_str) == EOF)
	    {
		printf (
		    "BRKDIC-E-ERROR, Error writing to output dictionary %s\n",
		    outdict_str);

                return (FALSE);
	    }

	    words_ok++;
    	}
	else
	{
	    words_failed++;
	}

	words_total++;

	if ( (words_total % 10000) == 0)
	{
	    printf (
		"BRKDIC-I-PRCSS, Processed %6lu words - current word is \"%s\"\n", 
		words_total,
		reference_str);
	}
    }

    if (words_failed != 0)
    {
	printf (
	    "BRKDIC-W-PRCSS, %lu words written successfully *but* %lu other words unsuccessfully written\n", 
	    words_ok, 
	    words_failed);
    }
    else
    {
	printf (
	    "BRKDIC-I-PRCSS, All %lu words written successfully\n", 
	    words_ok);
    }

    fclose (input_fp);
    fclose (output_fp);

    return (TRUE);

}



/*
 *  ?? MS-DOS problem whereby it's not possible to have more than 14
 *  files open at any one time using some compilers.  I will probably
 *  rewrite this function to handle this eventuality, at a slight
 *  overhead at some stage in the future ??
 *
 *  ?? Check limit on size of input word ??
 */

PRIVATE int breakdic__write_word_files (
	char			*dict_word)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function to write the dictionary words to individual word
 *	files.
 *
 *  FORMAL PARAMETERS:
 *
 *      *dict_word		(r) Word to write to word file from dictionary.
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      TRUE			Word written successfully.
 *	FALSE			Word not written successfully.
 *--
 */

{

    static FILE *wordfile_ptr_array[CWORD_SIZEOF_WORD];

    static BOOLEAN is_first_time = TRUE;

    char temp_spec[CWORD_SIZEOF_FILESPEC + 1];

    unsigned short count,
		   word_len;

    word_len = strlen (dict_word);
	
    if (is_first_time != TRUE)
    {
	if (wordfile_ptr_array[word_len] != (FILE *) 0)
	{
	    errno = 0;
	    if (fprintf (
		    wordfile_ptr_array[word_len],
		    "%s\n",
		    dict_word) == EOF)
	    {
		perror ("BRKDIC-E-ERROR, Word file write error");

		return (FALSE);
	    }
	}
	else
	{
	    sprintf (
		temp_spec,
		WORDFILE_PREFIX,
		breakdic_l_prefix_str,
		word_len);

	    errno = 0;
	    wordfile_ptr_array[word_len] = fopen (
						temp_spec, 
						"w");

	    if (wordfile_ptr_array[word_len] == (FILE *) 0)
	    {
		perror ("BRKDIC-E-ERROR, Error creating word file");

		return (FALSE);
	    }
	    else
	    {
		printf ("BRKDIC-I-INFO, %s word file created\n", temp_spec);

		errno = 0;
		if (fprintf (
			wordfile_ptr_array[word_len],
			"%s\n",
			dict_word) == EOF)
		{
		    perror ("BRKDIC-E-ERROR, Word file write error");

		    return (FALSE);
		}
	    }
	}
    }
    else
    {
	is_first_time = FALSE;
	
	for (count = 0; count < CWORD_SIZEOF_WORD; count++)
	{
	    wordfile_ptr_array[count] = (FILE *) 0;
	}

	if (wordfile_ptr_array[word_len] != (FILE *) 0)
	{
	    errno = 0;
	    if (fprintf (
		    wordfile_ptr_array[word_len],
		    "%s\n",
		    dict_word) == EOF)
	    {
		perror ("BRKDIC-E-ERROR, Word file write error");

		return (FALSE);
	    }
	}
	else
	{
	    sprintf (
		temp_spec,
		WORDFILE_PREFIX,
		breakdic_l_prefix_str,
		word_len);

	    errno = 0;
	    wordfile_ptr_array[word_len] = fopen (
						temp_spec,
						"w");

	    if (wordfile_ptr_array[word_len] == (FILE *) 0)
	    {
		perror ("BRKDIC-E-ERROR, Error creating word file");

		return (FALSE);
	    }
	    else
	    {
		printf (
		    "BRKDIC-I-INFO, %s word file created\n",
		    temp_spec);

		errno = 0;
		if (fprintf (
			wordfile_ptr_array[word_len],
			"%s\n",
			dict_word) == EOF)
		{
		    perror ("BRKDIC-E-ERROR, Word file write error");

		    return (FALSE);
		}
	    }
	}
    }

    return (TRUE);

}



PRIVATE int breakdic__read_dic_and_break (
	const char		*dict_str)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function to read the dictionary file and feed words into the
 *	routine breakdic__write_word_files for writing into its own
 *	individual word file for quicker access!
 *
 *  FORMAL PARAMETERS:
 *
 *      *dict_str		(r) Dictionary spec to break
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      TRUE			Successfully broke dictionary into series of word files.
 *	FALSE			Unsuccessfully broke dictionary into series of word files.
 *--
 */

{

    FILE *dict_ptr;

    char dict_word[CWORD_SIZEOF_WORD + 1];

    register long word_num = 0;

    int	status;

    errno = 0;
    dict_ptr = fopen (
		    dict_str,
		    "r");

    if (dict_ptr == (FILE *) 0)
    {
	perror ("BRKDIC-E-ERROR, Dictionary file error reported");

	return (FALSE);
    }

    while (feof (dict_ptr) == 0)
    {
	fscanf (dict_ptr, "%s\n", dict_word);

    	word_num++;

    	if ( (word_num % 10000) == 0 )
    	{
    	    printf (
    		"BRKDIC-I-PRCSS, Processed %6lu words - current word is \"%s\"\n",
    		word_num,
    		dict_word);
    	}

	status = breakdic__write_word_files (dict_word);

	if (status == FALSE)
	    return (FALSE);
    }

    printf (
	"BRKDIC-I-PRCSS, Total number of words processed was %lu\n",
	word_num);

    fclose (dict_ptr);

    return (TRUE);

}



PUBLIC int breakdic_process_input_dict (
	char			*dict_str)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function that can be called from both Cword and in standalone
 *	mode to execute the cleanup and break up of an input
 *	dictionary.
 *
 *  FORMAL PARAMETERS:
 *
 *      *dict_str		(r) INput dictionary file specification
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      TRUE			Successfully processed dictionary.
 *	FALSE			Error processing dictionary.
 *--
 */

{

    char file_spec[CWORD_SIZEOF_FILESPEC + 1],
	 outdict_str[CWORD_SIZEOF_FILESPEC + 1],
	 *str_ptr; /* Not used */

    int	status,
	count;

    if (breakdic__access_file (dict_str) == FALSE)
    {
	printf ("BRKDIC-E-ERROR, Input dictionary not found\n");

	return (FALSE);
    }

    /*
     *	?? Should check whether prefix.%% exists before continuing as
     *	processing will stop if they exist so what's the point cleaning up the
     *	input dictionary if the breaking if this dictionary cannot then take
     *	place ??
     */

    /*
     *	Clean up the input dictionary creating a temporary file outdict_str.
     */

    status = breakdic_cleanup_dic (
		dict_str,
		outdict_str);

    if (status == FALSE)
    {
	breakdic_remove_file (outdict_str);

	return (FALSE);
    }

    /*
     *	Get the prefix of the input dictionary file and use it for outputting
     *	newly create word files.
     */

    /*
     *	?? Bug concerns giving breakdic a full file specification - this
     *	probably makes breakdic segment violate or exit with error ??
     *
     *	?? Find the final '.' instead in the input file specification ??
     */

    strcpy (file_spec, dict_str);

    count = strlen (file_spec) - 1;
    while (count > 0)
    {
	/*
	 *  Latch onto last '.' in input dict_str and insert an EOL to create
	 *  output word file specification.
	 */

	if (file_spec[count] == (char) '.')
	{
	    file_spec[count] = (char) 0;

	    break;
	}
	else
	{
	    count--;
	}
    }

    /*
     *	breakdic_l_prefix_str is a static global used as a prefix into what the
     *	output word files will be called.
     */

    strcpy (breakdic_l_prefix_str, file_spec);

#if 0
    /*
     *	Gets first '.' and uses all upto this in output word file prefix spec.
     */

    str_ptr = strtok (file_spec, ".");

    if (str_ptr != (char *) 0)
	strcpy (breakdic_l_prefix_str, str_ptr);
    else
	strcpy (breakdic_l_prefix_str, dict_str);
#endif

    printf (
	"BRKDIC-I-INFO, Word file prefix is %s\n",
	breakdic_l_prefix_str);

    /*
     *	Check for existing word files in the default directory.  If there are
     *	any then, if we continued, they would be written over [possibly :-(].
     *	An error is reported and it's upto the user to decide what to do with
     *	them.  This program will not run as long as one file bearing the file
     *	specification defined below remains in the directory.
     */

    for (count = 1; count < CWORD_SIZEOF_WORD; count++)
    {
	sprintf (
	    file_spec,
	    WORDFILE_PREFIX,
	    breakdic_l_prefix_str,
	    count);

	/*
	 *  If the file exists then it's an error otherwise continue.
	 */

	if (breakdic__access_file (file_spec) == TRUE)
	{
	    printf (
		"BRKDIC-E-ERROR, Output word file %s already exists\n", 
		file_spec);

	    breakdic_remove_file (outdict_str);

	    return (FALSE);
	}
    }

    status = breakdic__read_dic_and_break (outdict_str);

    /*
     *	Delete the temporary dictionary file as it's of no use anymore.
     */

    breakdic_remove_file (outdict_str);

    if (status == TRUE)
	return (TRUE);
    else
	return (FALSE);

}



/*
 *  If MAIN is defined during the compilation stage it is assumed
 *  that breakdic will be executed in standalone mode only.
 *
 *  If MAIN is not defined then breakdic will act as simply another
 *  module of Cword and breakdic_process_input_dict can be called as
 *  any other PUBLIC function will be.
 */

#ifdef MAIN

PUBLIC int main (
	int			argc, 
	char			*argv[])

{

    int	status;

    if (argc != 2)
    {
	printf ("Usage: breakdic [dictionary file]\n");

    	exit (EXIT_FAILURE | NO_SCREEN);
    }

    printf ("breakdic: processes inputted dictionary into valid Cword word files\n");
    fflush (stdout);

    status = breakdic_process_input_dict (argv[1]);

    if (status == TRUE)
    {
	printf (
	    "BRKDIC-I-INFO, Successfully executed breakdic on %s\n",
	    argv[1]);

	exit (EXIT_SUCCESS | NO_SCREEN);
    }
    else
    {
	printf (
	    "BRKDIC-E-ERROR, Error executing breakdic on %s\n",
	    argv[1]);

	exit (EXIT_FAILURE | NO_SCREEN);
    }

}

#endif
