/*
 *  ?? General comments about the content of both the crossword
 *  and grid files include:
 *
 *  1. The crossword file format doesn't need either the number
 *  of the clue or the length of the light as both these pieces
 *  of information can be obtained firstly from the overall
 *  crossword configuration and secondly from the string length
 *  of the light in question.
 *
 *  2. The grid file format is slightly incorrect in the sense
 *  the directions of the lights therein are given as 'A' or
 *  'D'.  This is the "English" way of doing things and maybe
 *  not be altogether the best way of handling this.  Something
 *  like 1/0 (used internally in Cword) might be more
 *  appropriate ??
 *
 *  ?? Both these possible changes will affect negatively on the
 *  manner in which either file can be easily understood by the
 *  user from outside Cword ??
 */

/*
 *++
 *  FACILITY:
 *
 *      fileio.c
 *
 *  ABSTRACT:
 *
 *	Cword input/output routines used for reading and writing crossword/grid/seed
 *	definitions to/from ordinary text files.
 *
 *  AUTHORS:
 *
 *      James A. Lundon
 *
 *  CREATION DATE:     
 *
 *	19 February 1993
 *--
 */

/*
 *  Include Files
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <ctype.h>
#include <errno.h>

#if defined(__unix) || defined(__unix__)

#include <sys/file.h>
#include <sys/uio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <limits.h>
#include <unistd.h>

#endif

#if defined(VMS) || defined (__VMS)

#include <unixio.h>
#include <file.h>

#endif

#if defined(__MSDOS__) || defined(MSDOS)

#include <io.h>
#include <conio.h>
#include <fcntl.h>
#include <sys\stat.h>

#endif

#include "main.h"
#include "print.h"
#include "grid.h"
#include "memman.h"
#include "messages.h"
#include "fileio.h"
#include "utils.h"
#include "word.h"
#include "stats.h"
#include "verify.h"



#ifndef FILEIO

/*
 *  Macro Definitions - though possibly not the best way to handle the clean up
 *  when a file is in error.
 *
 *  No static type checking possible using macros.
 *
 *  Remember these definitions are only macros and have to inserted into
 *  existing code very carefully.
 */

#define CWORD_FILEIO_XRD_READ_ERROR(message, file_name, crossword_record, file_id) \
	    cword_messages_output_message ( \
			message, \
			(TRACE2 | SCREEN0), \
			file_name); \
	    memset (crossword_record, 0, sizeof (CROSSWORD_RECORD)); \
	    close (file_id); \
	    return (CWORD_S_FAILURE)

#define CWORD_FILEIO_XRD_WRITE_ERROR(message, file_name, file_id) \
	    cword_messages_output_message ( \
			message, \
			(TRACE2 | SCREEN0), \
			file_name); \
	    close (file_id); \
	    return (CWORD_S_FAILURE)

#define CWORD_FILEIO_GRD_READ_ERROR(message, file_name, crossword_record, file_id) \
	    cword_messages_output_message ( \
			message, \
			(TRACE2 | SCREEN0), \
			file_name); \
	    memset (crossword_record, 0, sizeof (CROSSWORD_RECORD)); \
	    close (file_id); \
	    return (CWORD_S_FAILURE)

#define CWORD_FILEIO_GRD_WRITE_ERROR(message, file_name, file_id) \
	    cword_messages_output_message ( \
			message, \
			(TRACE2 | SCREEN0), \
			file_name); \
	    close (file_id); \
	    return (CWORD_S_FAILURE)



PRIVATE int cword_fileio__parse_format (
	char			*input_str,
	int			*x_dim,
	int			*y_dim,
	int			*clue_sum,
	int			*grid_type,
	int			*symmetry_type,
	BOOLEAN			*is_old_format);



/*
 *  ?? This function is really a big piece of shit and should be changed ASAP ??
 */

PUBLIC int cword_fileio_read_line (
	int			file_handle,
	char			*file_line,
	int			sizeof_file_line)
/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function which will automate the use of the low level I/O read function
 *	to read from a text file. Hopefully this will work over all platforms to
 *	be implemented :
 *
 *	1.  OpenVMS
 *	2.  ULTRIX
 *	3.  MSDOS
 *
 *  FORMAL PARAMETERS:
 *
 *	file_handle		(r) file handle to be read
 *	*file_line		(w) string read from file
 *
 *  FUNCTION VALUE:
 *
 *	Positive value		Number of bytes read
 *	0			File EOF reached
 *	-1			File error
 *--
 */

{

    char *ptr_chr;
    
    int	count1,
	status;

    memset (file_line, 0, sizeof_file_line);

    ptr_chr = (char *) malloc (1);

    count1 = 0;
    status = read (file_handle, ptr_chr, 1);

    if (status == 0)
    {
	free (ptr_chr);

	return (0);
    }
    else if (status == -1)  
    {
	free (ptr_chr);

	return (-1);
    }

    if (*ptr_chr == '\n')
    {
	count1++;
    }

    while ( (*ptr_chr != '\n') && (status != 0) )
    {
	file_line[count1] = *ptr_chr;	    
	count1++;
	status = read (file_handle, ptr_chr, 1);	

	if (status == 0)
        {
            free (ptr_chr);

	    return (0);
        }
	else if (status == -1)
        {
            free (ptr_chr);

	    return (-1);
        }
    }

    cword_word_truncate_str (file_line);

    free (ptr_chr);

    return (count1);

}



PRIVATE int cword_fileio__parse_format (
	char			*input_str,
	int			*x_dim,
	int			*y_dim,
	int			*clue_sum,
	int			*grid_type,
	int			*symmetry_type,
	BOOLEAN			*is_old_format)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *      [@tbs@]
 *
 *  FORMAL PARAMETERS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      [@description or none@]
 *--
 */

{

    *x_dim = *y_dim = *clue_sum = 0;
    *is_old_format = FALSE;

    *grid_type = NOTDEFINED_GRID;
    *symmetry_type = NOTDEFINED_SYMMETRY;

    if (5 != sscanf (
	    input_str, 
	    "%d %d %d %d %d", 
	    x_dim,
	    y_dim,
	    clue_sum,
	    grid_type,
	    symmetry_type))
    {
	/*
	 *  New style grids created by Cword which contain both the symmetry and
	 *  grid type (5 parameters on this line).
	 */

	if (3 != sscanf (
		input_str, 
		"%d %d %d", 
		x_dim,
		y_dim,
		clue_sum))
	{
	    return (CWORD_S_FAILURE);
	}

	*is_old_format = TRUE;
    }

    return (CWORD_S_NORMAL);

}



PUBLIC int cword_fileio_read_crossword (
	char			*input_file_spec,
	CROSSWORD_RECORD	*crossword_record)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Reads in the details of the crossword from a file defined by the user
 *	at run time.  No checking of the crossword grid is done here, only
 *	simple I/O to a structure to be used later in the application.
 *
 *  FORMAL PARAMETERS:
 *
 *      *input_file_spec	(r) File name inputted from command line as arg
 *	*crossword_record	(w) Pointer to crossword structure containing the clues
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL		Read crossword definition successfully into data structure
 *	CWORD_S_FAILURE		File error
 *--
 */

{

    char input_str[CWORD_SIZEOF_MESSAGE + 1],
	 temp_char,
	 *str_ptr = (char *) 0;

    int	input_fd,
    	read_status = 1,
	count1,
	temp_int, 
	temp_int1, /* ?? Array instead ?? */ 
	temp_int2,
	temp_int3,
	temp_int4,
	temp_int5,
	status;

    BOOLEAN is_old_format;

    crossword_record->is_completely_specified = FALSE;

    status = cword_utils_is_file_ok (
		    READ_FILE,
		    CWORD_SUFFIX_XWORD_FILE,
		    input_file_spec);

    if (CWORD_S_NORMAL != status)
    {
	return (CWORD_S_FAILURE);
    }

    input_fd = open (
		    input_file_spec,	
		    O_RDONLY,
 		    0700);

    if (input_fd == EOF)
    {
	CWORD_FILEIO_XRD_READ_ERROR(
		    CWORD_FILEREADERROR, 
		    input_file_spec, 
		    crossword_record, 
		    input_fd);
    }

    /*
     *	Use a function to do the low level reading of the file rather than the
     *	straight read function which seems only to work that way I want in OVMS.
     */

    read_status = cword_fileio_read_line (
			    input_fd,
			    input_str,
			    sizeof (input_str));

    if (read_status < 1) 
    {
	CWORD_FILEIO_XRD_READ_ERROR(
		    CWORD_FILEREADERROR, 
		    input_file_spec, 
		    crossword_record, 
		    input_fd);
    }

    strcpy (crossword_record->name, input_str);

    read_status = cword_fileio_read_line (
			    input_fd,
			    input_str,
			    sizeof (input_str));

    if (read_status < 1) 
    {
	CWORD_FILEIO_XRD_READ_ERROR(
		    CWORD_FILEREADERROR, 
		    input_file_spec, 
		    crossword_record, 
		    input_fd);
    }

    strcpy (crossword_record->remarks, input_str);

    read_status = cword_fileio_read_line (
			    input_fd,
			    input_str,
			    sizeof (input_str));

    if (read_status < 1) 
    {
	CWORD_FILEIO_XRD_READ_ERROR(
		    CWORD_FILEREADERROR, 
		    input_file_spec, 
		    crossword_record, 
		    input_fd);
    }

    if (cword_fileio__parse_format (
	    input_str,
	    &temp_int1,
	    &temp_int2,
	    &temp_int3,
	    &temp_int4,
	    &temp_int5,
	    &is_old_format) != CWORD_S_NORMAL)
    {
	CWORD_FILEIO_XRD_READ_ERROR(
		    CWORD_PARSEFILEERROR, 
		    input_file_spec, 
		    crossword_record, 
		    input_fd);
    }

    crossword_record->x_dimensions	= temp_int1;
    crossword_record->y_dimensions	= temp_int2;
    crossword_record->clue_sum		= temp_int3;

    if ( (crossword_record->x_dimensions > CWORD_SIZEOF_CELLDIM) ||
	  (crossword_record->y_dimensions > CWORD_SIZEOF_CELLDIM) ||
           (crossword_record->clue_sum > CWORD_SIZEOF_CLUENUM) )
    {
	CWORD_FILEIO_XRD_READ_ERROR(
		    CWORD_FILEDIMSTOOBIG, 
		    input_file_spec, 
		    crossword_record, 
		    input_fd);
    }

    /*
     *	Execute this loop for each clue/light in the crossword definition.
     */

    for (count1 = 0; count1 < crossword_record->clue_sum; count1++)
    {
	read_status = cword_fileio_read_line (
			    input_fd,
			    input_str,
			    sizeof (input_str));

	if (read_status < 1) 
	{
	    CWORD_FILEIO_XRD_READ_ERROR(
			CWORD_FILEREADERROR, 
			input_file_spec, 
			crossword_record, 
			input_fd);
	}

	if (5 != sscanf (
	    input_str,
	    "%d %c %d %d %d", 
	    &temp_int,
	    &temp_char,
	    &temp_int1,
	    &temp_int2,
	    &temp_int3))
	{
	    CWORD_FILEIO_XRD_READ_ERROR(
			CWORD_PARSEFILEERROR, 
			input_file_spec, 
			crossword_record, 
			input_fd);
	}

	crossword_record->clue_details[count1].clue_number	    = temp_int;
	crossword_record->clue_details[count1].x_coordinate	    = temp_int1;
	crossword_record->clue_details[count1].y_coordinate	    = temp_int2;
	crossword_record->clue_details[count1].answer_length	    = temp_int3;

	if (temp_char == CWORD_A_CHAR)
	    crossword_record->clue_details[count1].clue_direction = ACROSS; 
	else
	    crossword_record->clue_details[count1].clue_direction = DOWN; 

	read_status = cword_fileio_read_line (
				input_fd,
				input_str,
				sizeof (input_str));

	if (read_status < 1) 
	{
	    CWORD_FILEIO_XRD_READ_ERROR(
			CWORD_FILEREADERROR, 
			input_file_spec, 
			crossword_record, 
			input_fd);
	}

	strcpy (crossword_record->clue_details[count1].clue_description, input_str); 

	read_status = cword_fileio_read_line (
				input_fd,
				input_str,
				sizeof (input_str));

	if (read_status < 1) 
	{
	    CWORD_FILEIO_XRD_READ_ERROR(
			CWORD_FILEREADERROR, 
			input_file_spec, 
			crossword_record, 
			input_fd);
	}

	strcpy (crossword_record->clue_details[count1].clue_answer, input_str);
    }

    close (input_fd);

    cword_messages_output_message (
		CWORD_XRDREAD,
		TRACE2,
		input_file_spec);

    DEBUG_TRACE (
	    "Crossword dimensions: X: %2d Y: %2d num: %2d\n", 
	    crossword_record->x_dimensions,
	    crossword_record->y_dimensions,
	    crossword_record->clue_sum);

    for (count1 = 0; count1 < crossword_record->clue_sum; count1++)
    {
	DEBUG_TRACE (
	    "Clue details: %2d %2d %2d %2d %2d %s\n",
	    crossword_record->clue_details[count1].clue_number,
	    crossword_record->clue_details[count1].x_coordinate,
	    crossword_record->clue_details[count1].y_coordinate,
	    crossword_record->clue_details[count1].answer_length,
	    crossword_record->clue_details[count1].clue_direction,
	    crossword_record->clue_details[count1].clue_answer);
    }

    /*
     *	?? All other things being equal it's now time to get the string to
     *	insert into file_name which will subsequently be used for outputting
     *	purposes ??
     *
     *  ?? Code in multiple locations => function ??
     */

    strcpy (input_str, input_file_spec);

    /*
     *  ?? What happens if there is no '.' in the input_file_spec ??
     *
     *  if ((str_ptr = strtok (input_str, ".")) != 0)
     *  {
     *      strcpy (crossword_record->file_name, str_ptr);
     *  }
     *  else
     *  {
     *      strcpy (crossword_record->file_name, input_str);
     *  }
     */

    str_ptr = strtok (input_str, ".");

    strcpy (crossword_record->file_name, str_ptr);

    /*
     *	Get the grid_type and symmetry parameters if they aren't specified in
     *	the crossword file itself.
     */

    if (is_old_format == FALSE)
    {
	crossword_record->grid_type = temp_int4;
	crossword_record->symmetry  = temp_int5;
    }
    else
    {
	/*
	 *  No symmetry or grid type defined in the crossword file itself.
	 *  Once the crossword file is read then we can check what the
	 *  prevailing grid type and symmetry are!
	 *
	 *  If the grid type, symmetry, or most likely both are undefined at
	 *  this stage then we have to obtain a valid grid type and symmetry
	 *  before continuing.
	 */

	if (crossword_record->is_completely_specified == FALSE)
	    cword_utils_complete_definition (crossword_record);

	crossword_record->grid_type = cword_verify_get_grid_type (crossword_record);

	crossword_record->symmetry = cword_verify_get_grid_symmetry (crossword_record);
    }

    return (CWORD_S_NORMAL);

}



PUBLIC int cword_fileio_write_crossword (
	char			*output_file_spec,
	CROSSWORD_RECORD	*crossword_record)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function to write a completed crossword to a flat text file. This file
 *	can then be used for many of the other operation possible in cword.
 *
 *  FORMAL PARAMETERS:
 *
 *      *output_file_spec	(r) file name to where the definition is to go
 *	*crossword_record	(w) record into which the info about the grid 
 *                                  will be inputted
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL		Crossword definition written to file successfully
 *	CWORD_S_FAILURE		File error
 *--
 */

{

    char output_str[CWORD_SIZEOF_MESSAGE + 1],
	 temp_direction;

    int	output_fd,
	count,
    	write_status = 0,
	status;

    status = cword_utils_is_file_ok (
		    WRITE_FILE,
		    CWORD_SUFFIX_XWORD_FILE,
		    output_file_spec);

    if (CWORD_S_NORMAL != status)
    {
	return (CWORD_S_FAILURE);
    }

    /*
     *  ?? OpenVMS - variable length file created ??
     */

    output_fd = open (
		    output_file_spec,	
		    O_WRONLY | O_CREAT | O_TRUNC,
#if defined(VMS) || defined (__VMS)
 		    0700,
		    "rfm=var",
		    "rat=cr");
#else
 		    0700);
#endif

    if (output_fd == EOF)
    {
	CWORD_FILEIO_XRD_WRITE_ERROR(
		CWORD_FILEWRITEERROR, 
		output_file_spec, 
		output_fd);
    }

    sprintf (
	    output_str,
	    "%s\n",
	    crossword_record->name);

    write_status = write (
			output_fd,
			output_str,
			strlen(output_str));    

    if (write_status == EOF) 
    {
	CWORD_FILEIO_XRD_WRITE_ERROR(
		CWORD_FILEWRITEERROR, 
		output_file_spec, 
		output_fd);
    }

    sprintf (
	    output_str,
	    "%s\n",
	    crossword_record->remarks);

    write_status = write (
			output_fd,
			output_str,
			strlen(output_str));    

    if (write_status < 1) 
    {
	CWORD_FILEIO_XRD_WRITE_ERROR(
		CWORD_FILEWRITEERROR, 
		output_file_spec, 
		output_fd);
    }

    /*
     *	All *new* crosswords created should have grid_type and symmetry
     *	included.
     */

    sprintf (
	output_str,
	"%d %d %d %d %d\n",
	crossword_record->x_dimensions,
	crossword_record->y_dimensions,
	crossword_record->clue_sum,
	crossword_record->grid_type,
	crossword_record->symmetry);

    write_status = write (
			output_fd,
			output_str,
			strlen(output_str));    

    if (write_status < 1) 
    {
	CWORD_FILEIO_XRD_WRITE_ERROR(
		CWORD_FILEWRITEERROR, 
		output_file_spec, 
		output_fd);
    }

    for (count = 0; count < crossword_record->clue_sum; count++)
    {
	if (crossword_record->clue_details[count].clue_direction == ACROSS)
	    temp_direction = CWORD_A_CHAR;
	else
	    temp_direction = CWORD_D_CHAR;

	sprintf (
		output_str,
		"%d %c %d %d %d\n%s\n%s\n",
		crossword_record->clue_details[count].clue_number,
		temp_direction,
		crossword_record->clue_details[count].x_coordinate,
		crossword_record->clue_details[count].y_coordinate,
		crossword_record->clue_details[count].answer_length,
		crossword_record->clue_details[count].clue_description,
		crossword_record->clue_details[count].clue_answer);
	
	write_status = write (
			    output_fd,
			    output_str,
			    strlen(output_str));    

	if (write_status < 1) 
	{
	    CWORD_FILEIO_XRD_WRITE_ERROR(
		    CWORD_FILEWRITEERROR, 
		    output_file_spec, 
		    output_fd);
	}
    }
    
    close (output_fd);

    cword_messages_output_message (
		CWORD_XRDWRITTEN,
		(TRACE0 | SCREEN0),
		output_file_spec);

    return (CWORD_S_NORMAL);

}



PUBLIC int cword_fileio_read_grid ( 
        char			*input_file_spec,
	CROSSWORD_RECORD	*crossword_record)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function to read a grid definition file and insert the available values
 *	into a CROSSWORD_RECORD.
 *
 *  FORMAL PARAMETERS:
 *
 *	*input_file_spec	(r) name of file containing grid information
 *	*crossword_record	(w) crossword record definition
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL		Successfully read grid file
 *	CWORD_S_FAILURE		Error processing grid file
 *--
 */

{

    char input_str[CWORD_SIZEOF_MESSAGE + 1],
	 temp_char,
	 *str_ptr = (char *) 0;

    int	input_fd,
	count1,
	temp_int1 = 0,
	temp_int2 = 0,
	temp_int3 = 0,
	temp_int4 = 0,
	temp_int5 = 0,
	read_status,
	status;

    BOOLEAN is_old_format = FALSE;

    crossword_record->is_completely_specified = FALSE;

    status = cword_utils_is_file_ok (
		    READ_FILE,
		    CWORD_SUFFIX_GRID_FILE,
		    input_file_spec);

    if (CWORD_S_NORMAL != status)
    {
	return (CWORD_S_FAILURE);
    }

    input_fd = open (
		    input_file_spec,	
		    O_RDONLY,
		    0700);

    if (input_fd == EOF) 
    {
	CWORD_FILEIO_GRD_READ_ERROR (
		CWORD_FILEREADERROR, 
		input_file_spec, 
		crossword_record, 
		input_fd);
    }

    read_status = cword_fileio_read_line (
			    input_fd,
			    input_str,
			    sizeof (input_str));

    if (read_status < 1) 
    {
	CWORD_FILEIO_GRD_READ_ERROR (
		CWORD_FILEREADERROR, 
		input_file_spec, 
		crossword_record, 
		input_fd);
    }

    strcpy (crossword_record->name, input_str);

    read_status = cword_fileio_read_line (
			    input_fd,
			    input_str,
			    sizeof (input_str));

    if (read_status < 1) 
    {
	CWORD_FILEIO_GRD_READ_ERROR (
		CWORD_FILEREADERROR, 
		input_file_spec, 
		crossword_record, 
		input_fd);
    }

    /*
     *	New style grids created by Cword which contain both the symmetry and
     *	grid type (5 parameters on this line).
     */

    if (cword_fileio__parse_format (
	    input_str,
	    &temp_int1,
	    &temp_int2,
	    &temp_int3,
	    &temp_int4,
	    &temp_int5,
	    &is_old_format) != CWORD_S_NORMAL)
    {
	CWORD_FILEIO_GRD_READ_ERROR (
		CWORD_PARSEFILEERROR, 
		input_file_spec, 
		crossword_record, 
		input_fd);
    }

    crossword_record->x_dimensions  = temp_int1;
    crossword_record->y_dimensions  = temp_int2;
    crossword_record->clue_sum	    = temp_int3;

    for (count1 = 0; count1 < crossword_record->clue_sum; count1++)
    {
	read_status = cword_fileio_read_line (
			    input_fd,
			    input_str,
			    sizeof (input_str));

	if (read_status < 1) 
	{
	    CWORD_FILEIO_GRD_READ_ERROR (
		    CWORD_FILEREADERROR, 
		    input_file_spec, 
		    crossword_record, 
		    input_fd);
	}

	if (4 != sscanf (
	    input_str, 
	    "%c %d %d %d", 
	    &temp_char,
	    &temp_int1,
	    &temp_int2,
	    &temp_int3))
	{
	    CWORD_FILEIO_GRD_READ_ERROR (
		    CWORD_PARSEFILEERROR, 
		    input_file_spec, 
		    crossword_record, 
		    input_fd);
	}

	if (temp_char == CWORD_A_CHAR)
	    crossword_record->clue_details[count1].clue_direction = ACROSS; 
	else
	    crossword_record->clue_details[count1].clue_direction = DOWN; 

	crossword_record->clue_details[count1].x_coordinate = temp_int1;
	crossword_record->clue_details[count1].y_coordinate = temp_int2;
	crossword_record->clue_details[count1].answer_length = temp_int3;	
    }

    close (input_fd);

    cword_messages_output_message (
	CWORD_GRDREAD,
	TRACE0,
	input_file_spec);

    DEBUG_TRACE (
	"%2d %2d %2d\n", 
	crossword_record->x_dimensions,
	crossword_record->y_dimensions,
	crossword_record->clue_sum);

    for (count1 = 0; count1 < crossword_record->clue_sum; count1++)
	DEBUG_TRACE (
	    "%2d %2d %2d %2d %2d %s\n",
	    crossword_record->clue_details[count1].x_coordinate,
	    crossword_record->clue_details[count1].y_coordinate,
	    crossword_record->clue_details[count1].answer_length,
	    crossword_record->clue_details[count1].clue_direction);

    /*
     *	?? All other things being equal it's not time to get the string to
     *	insert into file_name which will subsequently be used for outputting
     *	purposes ??
     *
     *  ?? Code in multiple locations ??
     */

    strcpy (input_str, input_file_spec);

    /*
     *  ?? What happens if there is no '.' in the input_file_spec ??
     *
     *  if ((str_ptr = strtok (input_str, ".")) != 0)
     *  {
     *      strcpy (crossword_record->file_name, str_ptr);
     *  }
     *  else
     *  {
     *      strcpy (crossword_record->file_name, input_str);
     *  }
     */

    str_ptr = strtok (input_str, ".");

    strcpy (crossword_record->file_name, str_ptr);

    /*
     *	Get the grid_type and symmetry parameters if they aren't specified in
     *	the grid file itself.
     */

    if (is_old_format == FALSE)
    {
	crossword_record->grid_type = temp_int4;
	crossword_record->symmetry  = temp_int5;
    }
    else
    {
	/*
	 *  No symmetry or grid type defined in the grid file itself.  Once the
	 *  grid file is read then we can check what the prevailing grid type
	 *  and symmetry are!
	 *
	 *  If the grid type, symmetry, or most likely both are undefined at
	 *  this stage then we have to obtain a valid grid type and symmetry
	 *  before continuing.
	 */

	if (crossword_record->is_completely_specified == FALSE)
	    cword_utils_complete_definition (crossword_record);

	crossword_record->grid_type = cword_verify_get_grid_type (crossword_record);

	crossword_record->symmetry = cword_verify_get_grid_symmetry (crossword_record);
    }

    return (CWORD_S_NORMAL);

}



PUBLIC int cword_fileio_write_grid ( 
	char			*output_file_spec,
        CROSSWORD_RECORD	*crossword_record)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function to write a grid definition to an output flat text file.
 *
 *  FORMAL PARAMETERS:
 *
 *	*output_file_spec	(r) name of file to contain grid information
 *	*crossword_record	(w) crossword record definition
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL		Successfully wrote grid file
 *	CWORD_S_FAILURE		Error writing grid file
 *--
 */

{

    char output_str[CWORD_SIZEOF_MESSAGE + 1],
	 temp_char;

    int	output_fd,
	count1,
	write_status,
	status;

    status = cword_utils_is_file_ok (
		    WRITE_FILE,
		    CWORD_SUFFIX_GRID_FILE,
		    output_file_spec);

    if (CWORD_S_NORMAL != status)
    {
	return (CWORD_S_FAILURE);
    }

    /*
     *  ?? OpenVMS - variable length file created ??
     */

    output_fd = open (
		    output_file_spec,	
		    O_WRONLY | O_CREAT | O_TRUNC,
#if defined(VMS) || defined (__VMS)
 		    0700,
		    "rfm=var",
		    "rat=cr");
#else
 		    0700);
#endif

    if (output_fd == EOF)
    {
	CWORD_FILEIO_GRD_WRITE_ERROR(
		CWORD_FILEWRITEERROR, 
		output_file_spec, 
		output_fd);
    }

    sprintf (
	    output_str,
	    "%s\n",
	    crossword_record->name);

    write_status = write (
			output_fd,
			output_str,
			strlen(output_str));    

    if (write_status < (int) strlen (output_str))
    {
	CWORD_FILEIO_GRD_WRITE_ERROR(
		CWORD_FILEWRITEERROR, 
		output_file_spec, 
		output_fd);
    }

    /*
     *	All *new* grids created should have grid_type and symmetry included.
     */

    sprintf (
	output_str, 
	"%d %d %d %d %d\n", 
	crossword_record->x_dimensions,
	crossword_record->y_dimensions,
	crossword_record->clue_sum,
	crossword_record->grid_type,
	crossword_record->symmetry);

    write_status = write (
			output_fd,
			output_str,
			strlen (output_str));    

    if (write_status < (int) strlen (output_str))
    {
	CWORD_FILEIO_GRD_WRITE_ERROR(
		CWORD_FILEWRITEERROR, 
		output_file_spec, 
		output_fd);
    }

    for (count1 = 0; count1 < crossword_record->clue_sum; count1++)
    {
	if (crossword_record->clue_details[count1].clue_direction == ACROSS)
	    temp_char = CWORD_A_CHAR;
	else
	    temp_char = CWORD_D_CHAR;
	
	sprintf (
	    output_str, 
	    "%c %d %d %d\n", 
	    temp_char,
	    crossword_record->clue_details[count1].x_coordinate,
	    crossword_record->clue_details[count1].y_coordinate,
	    crossword_record->clue_details[count1].answer_length);

	write_status = write (
			    output_fd,
			    output_str,
			    strlen(output_str));    

	if (write_status < (int) strlen (output_str)) 
	{
	    CWORD_FILEIO_GRD_WRITE_ERROR(
		    CWORD_FILEWRITEERROR, 
		    output_file_spec, 
		    output_fd);
	}
    }

    close (output_fd);

    cword_messages_output_message (
	    CWORD_GRDWRITTEN,
	    (TRACE0 | SCREEN0),
	    output_file_spec);

    return (CWORD_S_NORMAL);

}



PUBLIC int cword_fileio_write_seeds (
	char			*output_file_spec,
	CROSSWORD_RECORD	*crossword_record)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function to write the seeds file created by seeing during light seeding
 *	before light filling.
 *
 *  FORMAL PARAMETERS:
 *
 *	*output_file_spec	(r) Output seeds file spec
 *	*crossword_record	(r) Crossword definition containing seeds info
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL		Successful completion
 *      CWORD_S_FAILURE		Failed to write seeds file correctly
 *--
 */

{

    char output_str[CWORD_SIZEOF_MESSAGE + 1];

    int	output_fd,
	count,
	write_status;

    /*
     *  ?? OpenVMS - variable length file created ??
     */

    output_fd = open (
		    output_file_spec,	
		    O_WRONLY | O_CREAT | O_TRUNC,
#if defined(VMS) || defined (__VMS)
 		    0700,
		    "rfm=var",
		    "rat=cr");
#else
 		    0700);
#endif

    if (output_fd == EOF)
    {
	CWORD_FILEIO_XRD_WRITE_ERROR(
		CWORD_FILEWRITEERROR, 
		output_file_spec, 
		output_fd);
    }

    sprintf (
	output_str, 
	"%s %d %d %d\n", 
	crossword_record->name,
	crossword_record->x_dimensions,
	crossword_record->y_dimensions,
	crossword_record->clue_sum);

    write_status = write (
			output_fd,
			output_str,
			strlen(output_str));    

    if (write_status < (int) strlen (output_str))
    {
	CWORD_FILEIO_GRD_WRITE_ERROR(
		CWORD_FILEWRITEERROR, 
		output_file_spec, 
		output_fd);
    }

    for (count = 0; count < crossword_record->clue_sum; count++)
    {
	/*
	 *  Check through the crossword record for seeds and aby found are written
	 *  to the seeds file
	 */

	if (crossword_record->clue_details[count].clue_answer[0] != ' ')
	{
	    sprintf (
		    output_str, 
		    "%d %c %d %d %d %s\n",
		    crossword_record->clue_details[count].clue_number,
		    crossword_record->clue_details[count].clue_direction ? CWORD_A_CHAR : CWORD_D_CHAR,
		    crossword_record->clue_details[count].x_coordinate,
		    crossword_record->clue_details[count].y_coordinate,
		    crossword_record->clue_details[count].answer_length,
		    crossword_record->clue_details[count].clue_answer);

	    write_status = write (
				output_fd,
				output_str,
				strlen(output_str));    

	    if (write_status < 1) 
	    {
		CWORD_FILEIO_XRD_WRITE_ERROR(
			CWORD_FILEWRITEERROR, 
			output_file_spec, 
			output_fd);
	    }
	}
    }

    close (output_fd);

    cword_messages_output_message (
		CWORD_S_SEEDFILEWRITTEN,
		(TRACE0 | SCREEN0),
		output_file_spec);

    return (CWORD_S_NORMAL);

}



PUBLIC int cword_fileio_read_seeds (
	char			*input_file_spec,
        CROSSWORD_RECORD	*crossword_record)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *      Function to read seeds file and input results into crossword definition.
 *
 *  FORMAL PARAMETERS:
 *
 *	*input_file_spec	(r) Input seeds file spec
 *	*crossword_record	(w) Crossword definition which will contain seeds info
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL		Successful completion
 *      CWORD_S_FAILURE		Failed to read seeds file correctly
 *--
 */

{

    char input_str[CWORD_SIZEOF_MESSAGE + 1],
	 temp_str[CWORD_SIZEOF_MESSAGE + 1],
	 temp_char;

    int	input_fd,
	count,
	read_status,
	temp_int1,
	temp_int2,
	temp_int3,
	temp_int4;

    memset (crossword_record, 0, sizeof (CROSSWORD_RECORD));

    input_fd = open (
		    input_file_spec,	
		    O_RDONLY,
		    0700);

    if (input_fd == EOF) 
    {
	close (input_fd);

	return (CWORD_S_FAILURE);
    }

    read_status = cword_fileio_read_line (
			    input_fd,
			    input_str,
			    sizeof (input_str));

    if (read_status < 1) 
    {
	close (input_fd);

	return (CWORD_S_FAILURE);
    }

    if (4 != sscanf (
	    input_str, 
	    "%s %d %d %d",
	    temp_str,
	    &temp_int1,
	    &temp_int2,
	    &temp_int3))
    {
	close (input_fd);

	return (CWORD_S_FAILURE);
    }

    strcpy (crossword_record->name, temp_str);
    crossword_record->x_dimensions = temp_int1;
    crossword_record->y_dimensions = temp_int2;
    crossword_record->clue_sum = temp_int3;

    count = 0;
    while (cword_fileio_read_line (
			    input_fd,
			    input_str,
			    sizeof (input_str)) > 0)
    {
	if (6 != sscanf (
		input_str, 
		"%d %c %d %d %d %s",
		&temp_int1,
		&temp_char,
		&temp_int2,
		&temp_int3,
		&temp_int4,
		temp_str))
	{
	    memset (crossword_record, 0, sizeof (CROSSWORD_RECORD));

	    close (input_fd);

	    return (CWORD_S_FAILURE);
	}

	crossword_record->clue_details[count].clue_number = temp_int1;

	if (temp_char == CWORD_A_CHAR)
	    crossword_record->clue_details[count].clue_direction = ACROSS; 
	else
	    crossword_record->clue_details[count].clue_direction = DOWN; 

	crossword_record->clue_details[count].x_coordinate = temp_int2;
	crossword_record->clue_details[count].y_coordinate = temp_int3;
	crossword_record->clue_details[count].answer_length = temp_int4;

	strcpy (crossword_record->clue_details[count].clue_answer, temp_str);

	count++;
    }

    close (input_fd);

    cword_messages_output_message (
		CWORD_S_SEEDFILEREAD,
		(TRACE0 | SCREEN0),
		input_file_spec);

    return (CWORD_S_NORMAL);

}



/*
 *  ?? Doesn't handle embedded spaces in the name str ??
 */

PUBLIC int cword_fileio_read_crossword_top (
	char			*input_file_spec,
	char			*xrd_name_str,
	int			*x_dims,
	int			*y_dims,
	int			*clue_num)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function to read the top of a crossword file to obtain some cursory
 *	info about the crossword definition held within.
 *
 *  FORMAL PARAMETERS:
 *
 *	*input_file_spec	(r) Input crossword file spec
 *	*xrd_name_str		(w) Name of crossword
 *	*x_dims			(w) X dimensions of crossword
 *	*y_dims			(w) Y dimensions of crossword
 *	*clue_num		(w) CLue number of crossword
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL		Successful completion
 *      CWORD_S_FAILURE		Failed to read crossword file details correctly
 *--
 */

{

    char input_str[CWORD_SIZEOF_MESSAGE + 1];

    int status,
	read_status,
	input_fd,
	temp_int1,
	temp_int2,
	temp_int3;;

    *x_dims = 0;
    *y_dims = 0;
    *clue_num = 0;
    xrd_name_str[0] = (char) 0;

    status = cword_utils_is_file_ok (
		    READ_FILE,
		    CWORD_SUFFIX_XWORD_FILE,
		    input_file_spec);

    if (CWORD_S_NORMAL != status)
    {
	return (CWORD_S_FAILURE);
    }

    input_fd = open (
		    input_file_spec,	
		    O_RDONLY,
		    0700);

    if (input_fd == EOF)
    {
	return (CWORD_S_FAILURE);
    }

    read_status = cword_fileio_read_line (
			    input_fd,
			    input_str,
			    sizeof (input_str));

    if (read_status < 1) 
    {
	close (input_fd);

	return (CWORD_S_FAILURE);
    }

    strcpy (xrd_name_str, input_str);

    read_status = cword_fileio_read_line (
			    input_fd,
			    input_str,
			    sizeof (input_str));

    if (read_status < 1) 
    {
	xrd_name_str[0] = (char) 0;

	close (input_fd);

	return (CWORD_S_FAILURE);
    }

    read_status = cword_fileio_read_line (
			    input_fd,
			    input_str,
			    sizeof (input_str));

    if (read_status < 1) 
    {
	xrd_name_str[0] = (char) 0;

	close (input_fd);

	return (CWORD_S_FAILURE);
    }

    if (3 != sscanf (
	    input_str, 
	    "%d %d %d", 
	    &temp_int1,
	    &temp_int2,
	    &temp_int3))
    {
	xrd_name_str[0] = (char) 0;

	close (input_fd);

	return (CWORD_S_FAILURE);
    }

    *x_dims = temp_int1;
    *y_dims = temp_int2;
    *clue_num = temp_int3;

    close (input_fd);

    return (CWORD_S_NORMAL);

}



/*
 *  ?? Doesn't handle embedded spaces in the name str ??
 */

PUBLIC int cword_fileio_read_grid_top (
	char			*input_file_spec,
	char			*grd_name_str,
	int			*x_dims,
	int			*y_dims,
	int			*clue_num,
	int			*grid_type,
	int			*symmetry)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function to read the top of a grid file to obtain some cursory
 *	info about the grid definition held within.
 *
 *  FORMAL PARAMETERS:
 *
 *	*input_file_spec	(r) Input grid file spec
 *	*xrd_name_str		(w) Name of grid
 *	*x_dims			(w) X dimensions of grid
 *	*y_dims			(w) Y dimensions of grid
 *	*clue_num		(w) Clue number of grid
 *	*grid_type		(w) Grid type
 *	*symmetry		(w) Symmetry of grid
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL		Successful completion
 *      CWORD_S_FAILURE		Failed to read grid file details correctly
 *--
 */

{

    char input_str[CWORD_SIZEOF_MESSAGE + 1];

    int status,
	read_status,
	input_fd,
	temp_int1,
	temp_int2,
	temp_int3,
	temp_int4,
	temp_int5;

    BOOLEAN is_old_format = FALSE;

    *x_dims = *y_dims = *clue_num = 0;
    *grid_type = INVALID_GRID;
    *symmetry = INVALID_SYMMETRY;
    grd_name_str[0] = (char) 0;

    status = cword_utils_is_file_ok (
		    READ_FILE,
		    CWORD_SUFFIX_GRID_FILE,
		    input_file_spec);

    if (CWORD_S_NORMAL != status)
    {
	return (CWORD_S_FAILURE);
    }

    input_fd = open (
		    input_file_spec,	
		    O_RDONLY,
		    0700);

    if (input_fd == EOF)
    {
	return (CWORD_S_FAILURE);
    }

    /*
     *	????? Problems running piece of code BELOW on HP-UX (using cc) and
     *	Ultix (using c89) ?????
     *
     *  Only happens with crosswrd.grd.
     *
     *	Works fine with gcc on both Ultrix and HP-UX but not
     *	with c89 on either platform - weird!
     */

    read_status = cword_fileio_read_line (
			    input_fd,
			    input_str,
			    sizeof (input_str));

    if (read_status < 1) 
    {
	close (input_fd);

	return (CWORD_S_FAILURE);
    }

    strcpy (grd_name_str, input_str);

    read_status = cword_fileio_read_line (
			    input_fd,
			    input_str,
			    sizeof (input_str));

    if (read_status < 1) 
    {
	grd_name_str[0] = (char) 0;

	close (input_fd);

	return (CWORD_S_FAILURE);
    }

    /*
     *	????? Problems running piece of code ABOVE on HP-UX (using cc) and
     *	Ultix (using c89) ?????
     */

    if (cword_fileio__parse_format (
	    input_str,
	    &temp_int1,
	    &temp_int2,
	    &temp_int3,
	    &temp_int4,
	    &temp_int5,
	    &is_old_format) != CWORD_S_NORMAL)
    {
	grd_name_str[0] = (char) 0;

	close (input_fd);

	return (CWORD_S_FAILURE);
    }

    *x_dims = temp_int1;
    *y_dims = temp_int2;
    *clue_num = temp_int3;

    if (is_old_format == FALSE)
    {
	*grid_type = temp_int4;
	*symmetry = temp_int5;
    }
    else
    {
	/*
	 *  If the grid type or symmetry - or most likely both - are undefined
	 *  at this stage then we have to obtain a valid grid type and symmetry
	 *  before continuing.
	 */

	*grid_type = cword_stats_get_grid_type (input_file_spec);

	*symmetry = cword_stats_get_symmetry (input_file_spec);
    }

    close (input_fd);

    return (CWORD_S_NORMAL);

}



#else

extern int main (
		int argc, 
		char *argv[])	

{

    CROSSWORD_RECORD	*crossword_record;
    CROSSWORD_GRID	crossword_grid;
    char		output_file_spec[100];     
    int			status;
    
    crossword_record = (CROSSWORD_RECORD *) malloc (sizeof (CROSSWORD_RECORD));
    memset (crossword_record, 0, sizeof (crossword_record));

    /*
     *  read in the name of a crossword file from the command line
     */

    if (argc < 2)
    {
	printf ("please input file name (.input) next time\n");
	exit (EXIT_FAILURE);
    }

    cword_fileio_read_defaults (argv[1]);

#if 0
    status = cword_fileio_read_crossword (
			    argv[1],
			    crossword_record);

    status = cword_print_verify_grid (
			    crossword_record,
			    crossword_grid);

    /*
     *  write it out again
     */

    strcpy (output_file_spec, "my_crossword.input");

    status = cword_fileio_write_crossword (
			    output_file_spec,
			    crossword_record);

    status = cword_fileio_write_grid ( 
	"grid1.grid",
	crossword_record);

    if (status != CWORD_S_NORMAL)
	exit (EXIT_FAILURE);

    crossword_record = (CROSSWORD_RECORD *) malloc (sizeof (CROSSWORD_RECORD));
    memset (crossword_record, 0, sizeof (crossword_record));

    status = cword_fileio_read_grid ( 
	"grid1.grd",
	crossword_record);

    if (status != CWORD_S_NORMAL)
	exit (EXIT_FAILURE);
#endif

    exit (EXIT_SUCCESS);

}

#endif
