/*
 *++
 *  FACILITY:
 *
 *      getstats.c
 *    
 *  ABSTRACT:
 *
 *	This program analyses an inputted text dictionary and writes a valid
 *	C header file containing very useful statistics which the light
 *	filler will find useful when filling certain types of grids.
 *
 *      Statistics in the outputted header file include:
 *
 *      1. Total word frequency.
 *      2. Total letter frequencies.
 *      3. Total words starting with each letter.
 *      4. Total first and second letter adjacency frequencies.
 *
 *	This is a standalone program and needs only to be linked with
 *	breakdic.
 *
 *  AUTHORS:
 *
 *      James A. Lundon
 *
 *  CREATION DATE:
 *
 *      December 1992 and 29 April 1994
 *--
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#if defined(VMS) || defined(__VMS)

#include <file.h>
#include <unixio.h>

#endif

#if defined(__unix) || defined(__unix__)

#include <strings.h>

#endif

#include "main.h"
#include "breakdic.h"

/*
 *  Re-define precision to space out the adjacency_freq data
 *  making it more readable!
 */

#define LOW_PRECISION    4
#define HIGH_PRECISION   5

#define SIZEOF_ALPHABET  26

#define OUTPUT_H_FILE    "dicstats.h"

#define DIGITEND_FORMAT   "%*u\n"
#define DIGITCOMMA_FORMAT "%*u,"

#define OUTPUT_FORMAT     "#define %.30s %7lu\n"



PUBLIC int getstats__create_header (
    char            *dic_str)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *      [@tbs@]
 *
 *  FORMAL PARAMETERS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  {@function value or completion codes@}
 *
 *      [@description or none@]
 *--
 */

{

    FILE *dict_fp,
         *header_fp;

    long letter_freq[SIZEOF_ALPHABET],
         starting_freq[SIZEOF_ALPHABET],
         first_adjac_freq[SIZEOF_ALPHABET][SIZEOF_ALPHABET],
         second_adjac_freq[SIZEOF_ALPHABET][SIZEOF_ALPHABET];

    register long total_words = 0,
                  total_letters = 0;

    char word_str[CWORD_SIZEOF_WORD + 1],
         temp_str[CWORD_SIZEOF_LINE + 1],
         starting_char = 'a';

    unsigned int count1,
             count2,
             char_index = 0;

    printf (
        "GTSTTS-I-START, Starting creation of the header file: %s\n",
        OUTPUT_H_FILE);

    for (count1 = 0; count1 < SIZEOF_ALPHABET; count1++)
    {
        letter_freq[count1] = starting_freq[count1] = 0;

        for (count2 = 0; count2 < SIZEOF_ALPHABET; count2++)
        {
            first_adjac_freq[count1][count2] = 0;
            second_adjac_freq[count1][count2] = 0;
        }
    }

    dict_fp = fopen (
                  dic_str, 
                  "r");

    if (dict_fp == (FILE *) 0)
    {
        perror ("GTSTTS-E-ERROR, Error reading input dictionary file");

        return (FALSE);
    }

    while (feof (dict_fp) == 0)
    {
        fscanf (
            dict_fp, 
            "%s\n", 
            word_str);

        total_words++;

        if ( (total_words % 10000) == 0 )
        {
            printf (
                "GTSTTS-I-PRCSS, Processed %6lu words - current word is \"%s\"\n",
                total_words,
                word_str);
        }

        /*
         *  Get starting character frequency
         */

        if (word_str[0] > starting_char)
        {
            char_index++;
            starting_char++;
            starting_freq[char_index]++;
        }
        else
        {
            starting_freq[char_index]++;
        }

        /*
         *  Get simple word frequency statistics.
         */

        for (count1 = 0; count1 < strlen (word_str); count1++)
        {
            letter_freq[(int) (word_str[count1] - 'a')]++;
            total_letters++;
        }

        /*
         *  Get second letter adjacent frequency statistics - also useful for the
         *  light filler.
         */

        if (strlen (word_str) > 2)
        {
            for (count1 = 0; count1 < strlen (word_str) - 2; count1++)
                second_adjac_freq[(int) (word_str[count1] - 'a')]
                     [(int) (word_str[count1 + 2] - 'a')]++;
        }

        /*
         *  Get first letter adjacent frequency statistics - useful for the light
         *  filler.
         */

        if (strlen (word_str) > 1)
        {
            for (count1 = 0; count1 < (strlen (word_str) - 1); count1++)
                first_adjac_freq[(int) (word_str[count1] - 'a')]
                 [(int) (word_str[count1 + 1] - 'a')]++;
        }
    }

    /*
     *  Output word/letter statistics to valid header file
     */

    header_fp = fopen (
                    OUTPUT_H_FILE, 
                    "w"
#if defined(VMS) || defined(__VMS)
                    /*
                     *  Create a variable length record format file - it's
                     *  easier to copy and manipulate on OpenVMS.
                     */
                    ,"rfm=var");
#else
                    );
#endif

    fprintf (
        header_fp,
        "#ifndef DICSTATS_LOADED\n#define DICSTATS_LOADED 1\n\n");

    fprintf (
        header_fp,
        "/* This header file was generated by running getstats */\n\n",
        dic_str);

    fprintf (
        header_fp,
        "/* Total number of words */\n\n");

    fprintf (
        header_fp,
    	OUTPUT_FORMAT,
    	"FREQ_WORDS",
    	total_words);

    fprintf (
        header_fp,
        "\n/* Total number of letters */\n\n");

    fprintf (
        header_fp,
    	OUTPUT_FORMAT,
    	"FREQ_LETTERS",
    	total_letters);

    fprintf (
        header_fp,
        "\n/* Letter frequency */\n\n");

    for (count1 = 0; count1 < SIZEOF_ALPHABET; count1++)
    {
    	sprintf (
    	    temp_str,
    	    "FREQ_LETTER_%c",
    	    toupper((char) (count1 + 'a')));

        fprintf (
            header_fp,
    	    OUTPUT_FORMAT,
            temp_str,
            letter_freq[count1]);
    }

    fprintf (
        header_fp,
        "\n/* Starting letter frequency */\n\n");

    for (count1 = 0; count1 < SIZEOF_ALPHABET; count1++)
    {
    	sprintf (
    	    temp_str,
    	    "FREQ_WORDS_STARTING_WITH_%c",
    	    toupper((char) (count1 + 'a')));

        fprintf (
            header_fp,
    	    OUTPUT_FORMAT,
            temp_str,
            starting_freq[count1]);
    }

    fprintf (
        header_fp,
        "\n/* First letter adjacency frequency -\n   read data by row then column */\n\n");

    fprintf (
        header_fp,
        "static long first_adjacency_array[%d][%d] =\n{\n",
        SIZEOF_ALPHABET,
        SIZEOF_ALPHABET); 

    for (count1 = 0; count1 < SIZEOF_ALPHABET; count1++)
    {
        fprintf (
            header_fp,
            "/*\'%c\'*/ ", 
            (char) ('a' + count1));
    
        for (count2 = 0; count2 < SIZEOF_ALPHABET; count2++)
        {
            if ( ((count2 % 13) == 0) && (count2 != 0) )
            {
                fprintf (
                    header_fp,
                    "\n        ");
            }

            if ( (count2 != SIZEOF_ALPHABET - 1) ||
                  (count1 != SIZEOF_ALPHABET - 1) )
            {
                fprintf (
                    header_fp,
                    DIGITCOMMA_FORMAT,
                    total_words < 99999 ? LOW_PRECISION : HIGH_PRECISION,
                    first_adjac_freq[count1][count2]);
            }
            else
            {
                fprintf (
                    header_fp,
                    DIGITEND_FORMAT,
                    total_words < 99999 ? LOW_PRECISION : HIGH_PRECISION,
                    first_adjac_freq[count1][count2]);
            }
        }

        if (count1 != SIZEOF_ALPHABET - 1)
        {
            fprintf (
                header_fp,
                "\n");
        }
        else
        {
            fprintf (
                header_fp,
                "};\n");
        }
    }

    fprintf (
        header_fp,
        "\n/* Second letter adjacency frequency -\n   read data by row then column */\n\n");

    fprintf (
        header_fp,
        "static long second_adjacency_array[%d][%d] =\n{\n",
        SIZEOF_ALPHABET,
        SIZEOF_ALPHABET); 

    for (count1 = 0; count1 < SIZEOF_ALPHABET; count1++)
    {
        fprintf (
            header_fp,
            "/*\'%c\'*/ ", 
            (char) ('a' + count1));
    
        for (count2 = 0; count2 < SIZEOF_ALPHABET; count2++)
        {
            if ( ((count2 % 13) == 0) && (count2 != 0) )
            {
                fprintf (
                    header_fp,
                    "\n        ");
            }

            if ( (count2 != SIZEOF_ALPHABET - 1) ||
                  (count1 != SIZEOF_ALPHABET - 1) )
            {
                fprintf (
                    header_fp,
                    DIGITCOMMA_FORMAT,
                    total_words < 99999 ? LOW_PRECISION : HIGH_PRECISION,
                    second_adjac_freq[count1][count2]);
            }
            else
            {
                fprintf (
                    header_fp,
                    DIGITEND_FORMAT,
                    total_words < 99999 ? LOW_PRECISION : HIGH_PRECISION,
                    second_adjac_freq[count1][count2]);
            }
        }

        if (count1 != SIZEOF_ALPHABET - 1)
        {
            fprintf (
                header_fp,
                "\n");
        }
        else
        {
            fprintf (
                header_fp,
                "};\n\n");
        }
    }

    fprintf (
        header_fp,
        "#endif\n");

    printf (
        "GTSTTS-I-TOTALS, Total number of letters processed was %7lu\n", 
        total_letters);

    fclose (dict_fp);
    fclose (header_fp);

    printf (
        "GTSTTS-I-FINISH, Finished successfully creating the header file: %s\n",
        OUTPUT_H_FILE);

    return (TRUE);

}



extern main (
        int             argc, 
        char            *argv[])

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *        TBD!
 *
 *  FORMAL PARAMETERS:
 *
 *        argc                   (r) Number of parameters.
 *        *argv[]                (r) Details of parameters.
 *
 *  COMPLETION CODES:
 *
 *	  Compilable header file defined by OUTPUT_H_FILE containing statistics
 *	  useful during Cword light filling.
 *--
 */

{

    char output_file_str[CWORD_SIZEOF_FILESPEC + 1];

    if (argc != 2)
    {
        printf ("Usage: getstats [input dictionary file]\n");

        exit (EXIT_FAILURE | NO_SCREEN);
    }

    printf ("getstats: creates header file of stats from input dictionary file\n");
    fflush (stdout);

    if (breakdic_cleanup_dic (
            argv[1],
            output_file_str) != TRUE)
    {
        exit (EXIT_FAILURE | NO_SCREEN);
    }

    if (getstats__create_header (output_file_str) != TRUE)
    {
        exit (EXIT_FAILURE | NO_SCREEN);
    }

    if (breakdic_remove_file (output_file_str) != TRUE)
    {
        exit (EXIT_FAILURE | NO_SCREEN);
    }

    exit (EXIT_SUCCESS | NO_SCREEN);

}
