/*
 *++
 *  FACILITY:
 *
 *	main.c
 *
 *  ABSTRACT:
 *
 *	This module contains *only* the function main ().  This module should
 *	contain no other function definitions!  Everything is set up in main ()
 *	and then the character cell interface functions is called.  Once
 *	control is returned everything is nicely tidied up and exits by to the
 *	calling process.
 *
 *  AUTHORS:
 *
 *      James A. Lundon
 *
 *  CREATION DATE:     
 *
 *	31 December 1992
 *--
 */

/*
 *  Include files
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <signal.h>
#include <locale.h>

#include "main.h"
#include "print.h"
#include "match.h"
#include "grid.h"
#include "charui.h"
#include "messages.h"
#include "utils.h"
#include "fileio.h"
#include "defaults.h"
#include "rdb.h"
#include "cl.h"

/*
 *  Global defaults data structure declared here.  All other instances look for
 *  this definition to hook onto.
 */

CWORD_G_DEFAULTS_RECORD	cword_g_defaults_record;



PUBLIC int main (
	int			argc,
	char			*argv[])	

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *      Controlling function of the Cword application.
 *
 *  FORMAL PARAMETERS:
 *
 *      argc			(r) Number of command line parameters entered.
 *	*argv[]			(r) Contains the command line parameters.
 *--
 */

{

    int status;

    /*
     *  Use the local locale!
     */

    setlocale (LC_ALL, "");

    cword_defaults_init_defaults ();

    /*
     *	Parse the command line and handle return conditions apprepriately.
     *
     *	If this function returns CWORD_S_FAILURE then call exit() otherwise
     *	continue onto the user interface.
     */

    status = cword_cl_parse_command_line (
    		argc, 
    		argv);

    if (status == CWORD_S_FAILURE)
    {
	exit (EXIT_SUCCESS | NO_SCREEN);
    }

    /*
     *	?? If we get down here then the UI wil be brought up and Cword proper
     *	will be executed ??
     */

    /*
     *  Define some atexit() functions:
     *  
     *	1. Close the database if on openVMS
     *	2. Close the log file if it's not already closed
     */

#if defined(VMS) || defined(__VMS)
    atexit (cword_utils_close_trace_file);

#ifdef DB
#if defined(VMS) || defined(__VMS) || defined(__osf__)
    atexit (cword_rdb_close_db);
#endif
#endif
#endif

    atexit (cword_defaults_free_defaults);

    /*
     *	The signal handlers defined here are not really that powerful or
     *	un-intuitive.  They just allow Cword to exit gracefully in an undefined
     *	error is signaled.  Each will call exit() and this will allow the
     *	functions defined by atexit() to be executed.
     *
     *	?? Define these three signals.  None of which work on Windows, one of
     *	which isn't defined on DOS (SIGBUS) ??
     */

#if !defined(_WINDOWS)

    cword_utils_manage_sigint (ENABLE);
    signal (SIGSEGV, cword_utils_grab_signal);

#if !(defined(__MSDOS__) || defined(MSDOS))
    /*
     *  This signal is not defined on DOS
     */

    signal (SIGBUS, cword_utils_grab_signal);
#endif

#endif

    /*
     *	Read the defaults file which accompanies cword.  This function first
     *	defines default values for many Cword user configurable operations.  It
     *	then reads the values that are user defined in the defaults file, if it
     *	exists.
     */

    cword_defaults_read_defaults ();

    /* 
     *	?? Have to have trace file opened before reading defaults file - this I
     *	think probably means defining a constant string containing Cword ??
     */

    status = OPEN_TRACE (cword_g_defaults_record.log_file_prefix);

    if (status == CWORD_S_FAILURE)
    {
	exit (EXIT_FAILURE | NO_SCREEN);
    }

    /*
     *  ?? ANSI C only ??
     */

    cword_messages_output_message (
	    CWORD_APPBUILTWHEN,
	    (SCREEN0 | TRACE0),
	    CWORD_VERSION_ID,
	    __TIME__,
	    __DATE__);

    cword_defaults_output_defaults ();

    cword_messages_output_message (
	    CWORD_CHARCELLUIMSG,
	    TRACE0);

    cword_messages_output_message (
	CWORD_S_ISDBBUILD,
	(SCREEN0 | TRACE0),
	CWORD_RDB_C_DB_BUILT == 1 ? ENABLED_STR : NOT_ENABLED_STR);

    cword_charui_pause_ui (CWORD_S_PRESSRETURN, FALSE);

    /*
     *	This version of main () calls a simple character cell UI defined in
     *	cword_charui_display_main - this may change in the future.
     */

    status = cword_charui_display_main ();

    /*
     *  Explicitly close the all important trace log file befre ending!
     */

    CLOSE_TRACE ();

    /*
     *	exit () serves the same purpose as return ().  Some compilers may
     *	complain about not having a return value returned out of this function.
     *	That's C++ for you!
     */

    if (status != CWORD_S_NORMAL)
    {
	cword_messages_output_message (
		CWORD_EXITERRMSG,
		TRACE0);

	exit (EXIT_FAILURE | NO_SCREEN);
    }
    else
    {
	cword_messages_output_message (
		CWORD_EXITOKMSG,
		TRACE0);

	exit (EXIT_SUCCESS | NO_SCREEN);
    }

}
