!  FACILITY :
!
!	makedb.sql
!
!  ABSTRACT :
!
!	SQL script to create CWORD_DATABASE.
!
!  AUTHORS :	
!
!	James A. Lundon.
!
!  CREATION DATE :
!
!	4 May 1995.
!
!
!  ?? Rename column names to correspond more closely with relation
!  names i.e. CLUE_NUMBER in CLUE_DESCRIPTION will be renamed
!  CD_CLUE_NUMBER ??
!
!  Set the SQL dialect to SQL-92.
!
SET DIALECT 'SQL92';
!SET ANSI QUOTING ON;
!
!  Creates Cword's Rdb database in the location pointed to by the logical
!  CWORD_DATABASE.
!
CREATE DATABASE FILENAME CWORD_DATABASE DICTIONARY IS NOT REQUIRED;
!
!  Save the database definition before continuing.
! 
DISCONNECT ALL;
!
!  Re-invoke the database to add relations and other definitions.
!
ATTACH 'FILENAME CWORD_DATABASE';
!
!  Set transaction handle before continuing.
!
SET TRANSACTION READ WRITE;
!
!  Define the DOMAINS.
!
CREATE DOMAIN	CROSS_DESC  CHAR(151);
CREATE DOMAIN	CROSS_DIM   INTEGER;
CREATE DOMAIN   CROSS_NAME  CHAR(151);
CREATE DOMAIN   CROSS_SUM   INTEGER;
!
CREATE DOMAIN   CLUE_NUM    INTEGER;
CREATE DOMAIN   CLUE_DIR    INTEGER;
CREATE DOMAIN   CLUE_LOC    INTEGER;
CREATE DOMAIN   CLUE_ANS    CHAR(40);
CREATE DOMAIN   CLUE_DESC   CHAR(151);
!
CREATE DOMAIN   LIST_DESC   CHAR(101);
!
!  Create CROSSWORD_DESCRIPTION relation.
!
CREATE TABLE CROSSWORD_DESCRIPTION
(
!
!  ?? Check whether NULL is acceptable ??
!
	CROSSWORD_NAME		CROSS_NAME UNIQUE DEFERRABLE,
    	    CONSTRAINT CROSSWORD_NAME_PRIMARY
    	        PRIMARY KEY (CROSSWORD_NAME) DEFERRABLE,
	CROSSWORD_REMARKS	CROSS_DESC,
	CROSSWORD_LENGTH	CROSS_DIM NOT NULL DEFERRABLE,  
	CROSSWORD_WIDTH		CROSS_DIM NOT NULL DEFERRABLE,
	CROSSWORD_CLUES		CROSS_SUM NOT NULL DEFERRABLE,
	CROSSWORD_GRID_TYPE	CROSS_DIM NOT NULL DEFERRABLE,
	CROSSWORD_GRID_SYMMETRY CROSS_DIM NOT NULL DEFERRABLE
);
!
COMMENT ON TABLE CROSSWORD_DESCRIPTION IS -
'Holds reference information about crossword definitions stored in the database';
!
!  Create index so that no crossword with the same name can be entered
!  into the database.
!
CREATE UNIQUE INDEX CROSSWORD_INDEX ON CROSSWORD_DESCRIPTION
    (CROSSWORD_NAME);
!
!  Create CLUE_DESCRIPTION relation.
!
CREATE TABLE CLUE_DESCRIPTION
(
	CROSSWORD_NAME	CROSS_NAME NOT NULL DEFERRABLE,
	CLUE_NUMBER	CLUE_NUM
	    CONSTRAINT CLUE_NUMBER_VALUES
	    CHECK (
		CLUE_NUMBER <= 255 AND
    		CLUE_NUMBER >= 1 AND
		CLUE_NUMBER IS NOT NULL
		) DEFERRABLE,
	CLUE_DIRECTION	CLUE_DIR
	    CONSTRAINT CLUE_DIRECTION_VALUES
	    CHECK (
		CLUE_DIRECTION >= 0 AND
		CLUE_DIRECTION <= 1 AND
		CLUE_DIRECTION IS NOT NULL
		) DEFERRABLE,
	CLUE_X_COORD	CLUE_LOC NOT NULL DEFERRABLE,
	CLUE_Y_COORD	CLUE_LOC NOT NULL DEFERRABLE,
!
!  ?? NOT NULL doesn't seem to mean that the field cannot then NULL - 
!  VMS/VAX V6.1 ??
!
	CLUE_QUESTION	CLUE_DESC NOT NULL DEFERRABLE,
	CLUE_ANSWER	CLUE_ANS NOT NULL DEFERRABLE
);
!
COMMENT ON TABLE CLUE_DESCRIPTION IS -
'Relation holding data about how the stored crossword is defined';
!
!  Non-unique index on CLUE_DESCRIPTION
!
CREATE INDEX CLUE_INDEX ON CLUE_DESCRIPTION
    (CROSSWORD_NAME);
!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!  Create GRID_DESCRIPTION relation
!
CREATE TABLE GRID_DESCRIPTION
(
	GRID_NAME	CROSS_NAME UNIQUE DEFERRABLE,
    	    CONSTRAINT GRID_NAME_PRIMARY
    	        PRIMARY KEY (GRID_NAME) DEFERRABLE,
	GRID_LENGTH	CROSS_DIM NOT NULL DEFERRABLE,  
	GRID_WIDTH	CROSS_DIM NOT NULL DEFERRABLE,
	GRID_CLUES	CROSS_SUM NOT NULL DEFERRABLE,
	GRID_TYPE	CROSS_DIM NOT NULL DEFERRABLE,
	GRID_SYMMETRY	CROSS_DIM NOT NULL DEFERRABLE
);
!
COMMENT ON TABLE GRID_DESCRIPTION IS -
'Holds reference information about grid definitions stored in the local database';
!
!  Create index on grid_name.
!
CREATE UNIQUE INDEX GRID_INDEX ON GRID_DESCRIPTION
    (GRID_NAME);
!
!  Create GRID_DESCRIPTION relation.
!
CREATE TABLE LIGHT_DESCRIPTION
(
	GRID_NAME	    CROSS_NAME NOT NULL DEFERRABLE,
	LIGHT_DIRECTION	    CLUE_DIR
	    CONSTRAINT LIGHT_DIRECTION_VALUES
	    CHECK (
		LIGHT_DIRECTION >= 0 AND
		LIGHT_DIRECTION <= 1 AND
		LIGHT_DIRECTION IS NOT NULL
		) DEFERRABLE,
	LIGHT_X_COORD	    CLUE_LOC NOT NULL DEFERRABLE,
	LIGHT_Y_COORD	    CLUE_LOC NOT NULL DEFERRABLE,
	LIGHT_LENGTH	    CLUE_LOC NOT NULL DEFERRABLE
);
!
COMMENT ON TABLE LIGHT_DESCRIPTION IS -
'Relation holding data about how the stored grid is defined';
!
!  Unique index on LIGHT_DESCRIPTION
!
CREATE UNIQUE INDEX LIGHT_INDEX ON LIGHT_DESCRIPTION 
    (GRID_NAME, LIGHT_DIRECTION,
     LIGHT_X_COORD, LIGHT_Y_COORD);
!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!
!  Triggers are needed so that whenver a CROSSWORD_DESCRIPTION or
!  GRID_DESCRIPTION column is deleted the corresponding data in the
!  CLUE_DESCRIPTION or LIGHT_DESCRIPTION relations is also automatically
!  deleted.
!

!
!  Trigger on delete of CROSSWORD_DESCRIPTION.
!

CREATE TRIGGER CROSSWORD_DELETE
    BEFORE DELETE ON CROSSWORD_DESCRIPTION
	(DELETE FROM CLUE_DESCRIPTION T1
    	WHERE (T1.CROSSWORD_NAME = CROSSWORD_DESCRIPTION.CROSSWORD_NAME))
	    FOR EACH ROW;

!
!  Trigger on delete of GRID_DESCRIPTION.
!

CREATE TRIGGER GRID_DELETE
    BEFORE DELETE
    ON GRID_DESCRIPTION
	(DELETE FROM LIGHT_DESCRIPTION T2
    	WHERE (T2.GRID_NAME = GRID_DESCRIPTION.GRID_NAME))
	    FOR EACH ROW;

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!  Other tables can be defined here for future use.
!
!  Synonym description table.
!
CREATE TABLE SYNONYM_DESCRIPTION
(
	S_INDICATOR		LIST_DESC NOT NULL DEFERRABLE,
	S_DESCRIPTION		LIST_DESC NOT NULL DEFERRABLE
);
!
COMMENT ON TABLE SYNONYM_DESCRIPTION IS 'relation storing list words and its synonyms';
!
CREATE UNIQUE INDEX SYNONYM_INDEX ON SYNONYM_DESCRIPTION
    (S_INDICATOR, S_DESCRIPTION);
!
!  Homophone description table.
!
CREATE TABLE HOMOPHONE_DESCRIPTION
(
	H_INDICATOR		LIST_DESC NOT NULL DEFERRABLE,
	H_DESCRIPTION		LIST_DESC NOT NULL DEFERRABLE
);
!
COMMENT ON TABLE HOMOPHONE_DESCRIPTION IS 'relation storing list words and its homophones';
!
CREATE UNIQUE INDEX HOMOPHONE_INDEX ON HOMOPHONE_DESCRIPTION
    (H_INDICATOR, H_DESCRIPTION);
!
!  Anagram indicator table.
!
CREATE TABLE ANAGRAM_DESCRIPTION
(
	A_DESCRIPTION		LIST_DESC NOT NULL DEFERRABLE
);
!
COMMENT ON TABLE ANAGRAM_DESCRIPTION IS 'relation storing words which indicate an anagram';
!
CREATE UNIQUE INDEX ANAGRAM_INDEX ON ANAGRAM_DESCRIPTION
    (A_DESCRIPTION);
!
!  Save database description to the file pointed to by CWORD_DATABASE.
!
COMMIT;
!
EXIT;
