 /*
 *++
 *  FACILITY:
 *
 *      print.c
 *
 *  ABSTRACT:
 *
 *	This module creates the familiar crossword grid output in either
 *	postscript or text format from a Cword formatted crossword/grid file
 *	input.
 *
 *  AUTHORS:
 *
 *      James A. Lundon.
 *
 *  CREATION DATE:     
 *
 *	5 November 1992
 *--
 */

/*
 *  Include files
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <ctype.h>
#include <math.h>

#if defined(__unix) || defined(__unix__)

#include <sys/file.h>
#include <unistd.h>
#include <sys/uio.h>

#endif

#if defined(VMS) || defined(__VMS)

#include <unixio.h>
#include <file.h>

#endif

#if defined(__MSDOS__) || defined(MSDOS)

#include <io.h>
#include <fcntl.h>
#include <sys\stat.h>

#endif

#include "main.h"
#include "utils.h"
#include "fileio.h"
#include "verify.h"
#include "memman.h"
#include "print.h"
#include "stats.h"

/*
 *  The identifier MESSAGES_DEFINED is defined if we are using the print test
 *  harness.
 */

#ifdef PRINT
#define MESSAGES_DEFINED 1
#endif

#include "messages.h"
#include "os.h"
#include "defaults.h"

extern CWORD_G_DEFAULTS_RECORD	cword_g_defaults_record;



#ifndef PRINT

/*
 *  Macro Definitions
 */

#define CWORD_PRINT_WRITE_LINE(output_fd, output_str)  \
		    if ( (write ( \
			    output_fd, \
			    output_str, \
			    strlen(output_str))) != strlen(output_str) ) \
		    { \
			cword_messages_output_message ( \
				CWORD_WRITEERROR, \
				(TRACE0 | SCREEN0)); \
			return (CWORD_WRITEERROR); \
		    }

#define PS_PRINT_PAGECOMMENT(output, page_number) \
		    sprintf ( \
			output, \
			"%%%%Page: %d %d\n\n", \
			page_number, \
			page_number);

/*
 *  Allowable width of line on output.
 */

#define CWORD_C_PRINT_PS_LINE_WIDTH    50
#define CWORD_C_PRINT_TXT_LINE_WIDTH   60

#define CWORD_C_SEPERATOR_CHAR	    "\277"  /* Octal for ASCII character "" */
#define CWORD_C_SEPERATOR	    191	    /* Decimal for ASCII character "" */

/*
 *  Reference co-ordinates (bottom left corner X-Y) of the eventual crossword
 *  grid.  These co-ordinates determine most of the other parameters of the
 *  postscript output for both crossword and grid printing options.  They can
 *  be changed by tweaking the values of these identifiers.  
 */

/* For grids of between 0-19 in dimensions */

#define CWORD_C_USUAL_X    150
#define CWORD_C_USUAL_Y    410

/*
 *  Centre co-ordinates of PS output page to be used when centring the
 *  crossword on its own page.
 */

#define CWORD_C_CENTRE_X   295
#define CWORD_C_CENTRE_Y   400

/*
 *  These are the postscript page co-ordinates where the headers and footers
 *  are to be located
 */

#define CWORD_C_HEADER_X       50
#define CWORD_C_HEADER_Y      750

#define CWORD_C_FOOTER_X       25
#define CWORD_C_FOOTER_Y       25


/*
 *  Maximum grid side of crossword before three pages are to be used when
 *  outputting a completed crossword definition.
 */

#define CWORD_C_MAXGRID_SIZE   18

/*
 *  Size (in points) of the individual cells in the grid, together with the
 *  point size of the characters and clue indexes used during postscript
 *  printing for normally sized and large grids.
 */

#define CWORD_C_CELL_POINT_SIZE		20
#define CWORD_C_FONT_LETTER_SIZE	12
#define CWORD_C_FONT_NUM_SIZE		7

/*
 *  These constants used for grids of CWORD_C_MAXGRID_SIZE and greater.
 */

#define CWORD_C_SMALL_CELL_POINT_SIZE	14
#define CWORD_C_SMALL_FONT_LETTER_SIZE	9
#define CWORD_C_SMALL_FONT_NUM_SIZE	5

/*
 *  These constants don't change no matter what grid size is chosen
 */

#define CWORD_C_OUTSTROKE_POINT_SIZE	3
#define CWORD_C_STROKE_POINT_SIZE	1

/*
 *  This is the most important constant when printing postscript
 *  barred grids as it determines the width of the black bar in 
 *  the grid.
 */

#define CWORD_C_BARRED_POINT_WIDTH_SIZE 2

#define CWORD_C_WIDTH_BETWEEN_CLUES	10

/*
 *  Size of text cell box 
 */

#define CWORD_C_TXT_CELLSIZE_X      4
#define CWORD_C_TXT_CELLSIZE_Y      3
/*
 *  Whether to tell the offset function that we are printing a grid 
 *  or a set of answers.
 */

#define GRID_PRINT		    2
#define CLUES_PRINT		    1
#define NEWPAGE_CLUES_PRINT	    0

/*
 *  Whether to print the postscript file preamble or not!
 */

#define PREAMBLE		    1
#define NOPREAMBLE		    0

/*
 *  Four types of possible grid/crossword combinations possible!
 */

#define BLOCK_GRID_FILE		    1
#define BARRED_GRID_FILE	    2
#define BLOCK_CROSSWORD_FILE	    3
#define BARRED_CROSSWORD_FILE	    4

/* 
 *  Pre-defined output text file constants
 */

#define CORNER_STR		    "+"
#define HOR_STR			    "---+"
#define VER_STR			    "|"
#define INNER_BLACK_STR		    "###|"
#define INNER_BLANK1_STR	    "   |"
#define INNER_BLANK2_STR	    "   #"
#define BLACK4_STR		    "####"
#define BLACK5_STR		    "#####"

/*
 *  ?? DASH3_STR, DASH4_STR, DASH5_STR doesn't not encapsulate 100% how
 *  to create barred grids correctly ??
 */

#define DASH3_STR		    "*--" 
#define DASH4_STR		    "*---"
#define DASH5_STR		    "*----"

#define END_STR			    "\n"
#define CHAR1_STR		    " %c |"
#define CHAR2_STR		    " %c #"
#define NUM1_STR		    "%-3d|"
#define NUM2_STR		    "%-3d#"

/*
 *  Fonts to be used for printing PS grids/crosswords.
 *
 *  N.B. It's possible to change these fonts to ones of the user's own choice
 *  but the fonts must be known to the printer that will be used to print the
 *  result.
 */

#define HEADER_FONT		    "Times-Bold"
#define TEXT_FONT		    "Times-Roman"

#define NULL_CLUE_NUM		    0



/*
 *  Definition of the struct hold the offsets and other important printing
 *  information to be used during postscript printing.
 */

typedef struct 
{
    /*
     *	These fields should give full information about how to output the
     *	postscript details of the clues and/or grid.
     */

    int x;
    int y;

    int cell_point_size;
    int font_letter_size;
    int font_num_size;

    int cell_numoff_x;
    int cell_numoff_y;
    int cell_charoff_x;
    int cell_charoff_y;

    BOOLEAN new_page;
} COORDS_STRUCT;

/*
 *  Forward function prototypes - PRIVATE only.
 */

PRIVATE void cword_print__check_size (
	int			x_dimensions,
	int			y_dimensions);

PRIVATE int cword_print__cal_ps_offset (
	int			clues_or_grid,
	CROSSWORD_RECORD	*crossword_record,
	COORDS_STRUCT		*offset_vals);

PRIVATE int cword_print__cal_txt_offset (
	CROSSWORD_RECORD	*crossword_record,
	COORDS_STRUCT		*offset_vals);

PRIVATE void cword_print__convert_ps_line (
	char			*break_str);

PRIVATE void cword_print__break_up_line (
	int			text_type,
	char			*break_str);
	
PRIVATE int cword_print__header_ps_comments (
	int			file_type,
	char			*crossword_name,
	int			output_fd);

PRIVATE int cword_print__preamble (
	int			output_fd);

PRIVATE int cword_print__ps_new_page (
	int			output_fd);

PRIVATE int cword_print__block_ps_grid (
	int			preamble,
	int			output_fd,
	CROSSWORD_RECORD	*crossword_record,
	CROSSWORD_GRID		crossword_grid);

PRIVATE int cword_print__common_ps_clues (
	int			output_fd,
	CROSSWORD_RECORD	*crossword_record);

PRIVATE int cword_print__block_ps_answers (
	int			output_fd,
	CROSSWORD_RECORD	*crossword_record,
	CROSSWORD_GRID		crossword_grid);

PRIVATE int cword_print__block_txt_grid (
	int			preamble,
	int			output_fd,
	CROSSWORD_RECORD	*crossword_record,
	CROSSWORD_GRID		crossword_grid);

PRIVATE int cword_print__common_txt_clues (
	int			output_fd,
	CROSSWORD_RECORD	*crossword_record);

PRIVATE int cword_print__block_txt_answers (
	int			output_fd,
	CROSSWORD_RECORD	*crossword_record,
	CROSSWORD_GRID		crossword_grid);

PRIVATE int cword_print__barred_ps_answers (
	int			output_fd,
	CROSSWORD_RECORD	*crossword_record,
	CROSSWORD_GRID		crossword_grid);

PRIVATE int cword_print__barred_ps_grid (
	int			preamble,
	int			output_fd,
	CROSSWORD_RECORD	*crossword_record,
	CROSSWORD_GRID		crossword_grid);

PRIVATE int cword_print__barred_txt_answers (
	int			output_fd,
	CROSSWORD_RECORD	*crossword_record,
	CROSSWORD_GRID		crossword_grid);

PRIVATE int cword_print__barred_txt_grid (
	int			preamble,
	int			output_fd,
	CROSSWORD_RECORD	*crossword_record,
	CROSSWORD_GRID		crossword_grid);

PRIVATE cword_print__clue_num_offset (
	int			clue_number);

/*
 *  Keep a track of the page number of PS output currently being outputted via
 *  this module global.
 */

static int cword_print_l_page_number;



PRIVATE void cword_print__check_size (
	int			x_dimensions,
	int			y_dimensions)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function to inform user if there is a problem printing out a test grid
 *	or crossword.  Nothing is done once the user is informed of a problem
 *	BTW!
 *
 *  FORMAL PARAMETERS:
 *
 *	x_dimensions		(r) X dimensions of crossword definition
 *	y_dimensions		(r) Y dimensions of crossword definition
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      None
 *--
 */

{

    if ((x_dimensions * CWORD_C_TXT_CELLSIZE_X) + 1 > CWORD_SIZEOF_BIGPAGE_X)
    {
	cword_messages_output_message (
	    CWORD_S_PRINTOVER132,
	    (TRACE3 | SCREEN0));
    }
    else if ((x_dimensions * CWORD_C_TXT_CELLSIZE_X) + 1 > CWORD_SIZEOF_PAGE_X)
    {
	cword_messages_output_message (
	    CWORD_S_PRINTOVER80,
	    (TRACE3 | SCREEN0));
    }

    if ((y_dimensions * CWORD_C_TXT_CELLSIZE_Y) + 1 > CWORD_SIZEOF_PAGE_Y)
    {
	printf ("# Output grid/crossword will be over one page long.  Print appropriately.\n");
    }

    return;

}



PRIVATE int cword_print__cal_ps_offset (
	int			clues_or_grid,
	CROSSWORD_RECORD	*crossword_record,
	COORDS_STRUCT		*offset_vals)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function to calculate an adequate series of offsets with which to print
 *	out either a postscript barred or postscript blocked grid or crossword.
 *
 *  FORMAL PARAMETERS:
 *
 *	clues_or_grid		(r) Have a value of GRID_PRINT or CLUES_PRINT
 *      *crossword_record	(r) Pointer to crossword definition
 *      offset_vals		(w) Structure containing important offsets 
 *
 *  IMPLICIT INPUTS:
 *
 *      cword_g_defaults_record.x_dimensions
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *	CWORD_S_NORMAL		Normal Completion
 *--
 */

{

    /*
     *  Initialise all to 0!
     */

    offset_vals->x = 0;
    offset_vals->y = 0;
    offset_vals->cell_point_size = 0;
    offset_vals->font_letter_size = 0;
    offset_vals->font_num_size = 0;
    offset_vals->cell_numoff_x = 0;
    offset_vals->cell_numoff_y = 0;
    offset_vals->cell_charoff_x = 0;
    offset_vals->cell_charoff_y = 0;
    offset_vals->new_page = FALSE;

    /*
     *	If either dimensions are greater than CWORD_C_MAXGRID_SIZE alter
     *	the offsets accordingly!
     */

    if ( (crossword_record->x_dimensions > CWORD_C_MAXGRID_SIZE) ||
          (crossword_record->x_dimensions > CWORD_C_MAXGRID_SIZE) )
    {
	/*
	 *  Create a new page for the clues as the page the grid is going in
	 *  won't be big enough.
	 */

	offset_vals->new_page = TRUE;

	if (clues_or_grid == CLUES_PRINT)
	{
	    /*
	     *	From this set of co-ordinates start printing the clues on
	     *	the top of the *new* page.
	     */

	    offset_vals->x = 50;
	    offset_vals->y = CWORD_C_HEADER_Y - 40;
	}
	else
	{
	    /*
	     *	For this grid calculate the offsets from the centre of the
	     *	output page.
	     */

	    offset_vals->x = CWORD_C_CENTRE_X - ((int) (crossword_record->x_dimensions / 2)) * 
                                            CWORD_C_SMALL_CELL_POINT_SIZE;

	    offset_vals->y = CWORD_C_CENTRE_Y - ((int) (crossword_record->y_dimensions / 2)) * 
                                            CWORD_C_SMALL_CELL_POINT_SIZE;
	}

        offset_vals->cell_point_size = CWORD_C_SMALL_CELL_POINT_SIZE;
	offset_vals->font_letter_size = CWORD_C_SMALL_FONT_LETTER_SIZE;
	offset_vals->font_num_size = CWORD_C_SMALL_FONT_NUM_SIZE;

	/* Cell clue number offset */

	/* 
	 *  Just one point to the right of the left wall of cell 
	 */

	offset_vals->cell_numoff_x = 1;

	/*
	 *  Cell size - number size
	 */

	offset_vals->cell_numoff_y = CWORD_C_SMALL_CELL_POINT_SIZE - CWORD_C_SMALL_FONT_NUM_SIZE;

	/* Cell clue character offset */

	offset_vals->cell_charoff_x = CWORD_C_SMALL_FONT_NUM_SIZE;
	offset_vals->cell_charoff_y = CWORD_C_SMALL_CELL_POINT_SIZE - CWORD_C_SMALL_FONT_LETTER_SIZE - 2;
    }
    else
    {
	if (clues_or_grid == CLUES_PRINT)
	{
	    /*
	     *  From this set of Co-ordinates start printing out the clues
	     */

	    offset_vals->x = 50;
	    offset_vals->y = CWORD_C_USUAL_Y + ((int) (CWORD_C_CELL_POINT_SIZE / 2)) * 
			      (cword_g_defaults_record.y_dimensions - crossword_record->y_dimensions) -
                               50;
	}
	else
	{
	    offset_vals->x = CWORD_C_USUAL_X + ((int) (CWORD_C_CELL_POINT_SIZE / 2)) * 
			      (cword_g_defaults_record.x_dimensions - crossword_record->x_dimensions);

	    offset_vals->y = CWORD_C_USUAL_Y + ((int) (CWORD_C_CELL_POINT_SIZE / 2)) * 
			      (cword_g_defaults_record.y_dimensions - crossword_record->y_dimensions);
	}

        offset_vals->cell_point_size = CWORD_C_CELL_POINT_SIZE;
	offset_vals->font_letter_size = CWORD_C_FONT_LETTER_SIZE;
	offset_vals->font_num_size = CWORD_C_FONT_NUM_SIZE;

	/* Cell clue number offset */

	offset_vals->cell_numoff_x = 1;
	offset_vals->cell_numoff_y = CWORD_C_CELL_POINT_SIZE - CWORD_C_FONT_NUM_SIZE;

	/* Cell clue character offset */

	offset_vals->cell_charoff_x = 8;
	offset_vals->cell_charoff_y = CWORD_C_CELL_POINT_SIZE - CWORD_C_FONT_LETTER_SIZE - 3;
    }

    /*
     *  Disregard previous suggestions if NEWPAGE_CLUES_PRINT is selected!
     */

    if (clues_or_grid == NEWPAGE_CLUES_PRINT)
    {
	/*
	 *  From this set of co-ordinates start printing the clues on
	 *  the top of the *new* page.
	 */

	offset_vals->x = 50;
	offset_vals->y = CWORD_C_HEADER_Y - 40;
    }

    cword_messages_output_message (
	    CWORD_S_PRINTOFFSETSMSG,
	    (TRACE3 | SCREEN3),
	    offset_vals->x,
	    offset_vals->y,
	    clues_or_grid ? CWORD_GRID_STR : CWORD_CLUES_STR);

    return (CWORD_S_NORMAL);

}



PRIVATE int cword_print__cal_txt_offset (
	CROSSWORD_RECORD	*crossword_record,
	COORDS_STRUCT		*offset_vals)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function to calculate an adequate series of offsets with which to print
 *	out either a text barred or text blocked grid or crossword.
 *
 *  FORMAL PARAMETERS:
 *
 *	*crossword_record	(r) Pointer to crossword definition
 *	*offset_vals		(w) Structure containing offsets
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *	CWORD_S_NORMAL		Normal Completion
 *--
 */

{

    int length;

    /*
     *  Initialise all to 0 but only two awill be used -- x and y!
     */

    offset_vals->x = 0;
    offset_vals->y = 0;
    offset_vals->cell_point_size = 0;
    offset_vals->font_letter_size = 0;
    offset_vals->font_num_size = 0;
    offset_vals->cell_numoff_x = 0;
    offset_vals->cell_numoff_y = 0;
    offset_vals->cell_charoff_x = 0;
    offset_vals->cell_charoff_y = 0;
    offset_vals->new_page = FALSE;

    /*
     *  Two real options:
     *
     *	1. Less than CWORD_SIZEOF_PAGE_X characters in length
     *	2. Greater than CWORD_SIZEOF_PAGE_X and less than CWORD_SIZEOF_BIGPAGE_X characters 
     *  3. greater than CWORD_SIZEOF_BIGPAGE_X -- you mind yourself at the moment!
     */

    /*
     *  Fix up the X offset
     */

    length = crossword_record->x_dimensions * CWORD_C_TXT_CELLSIZE_X;

    if (((crossword_record->x_dimensions * CWORD_C_TXT_CELLSIZE_X) + 1) < CWORD_SIZEOF_PAGE_X)
    {
	offset_vals->x = (int) ((float) (CWORD_SIZEOF_PAGE_X - 1 - length) /(float) 2);

	cword_messages_output_message (
	    CWORD_S_TXTOFFSMALL,
	    (SCREEN3 | TRACE3),
    	    CWORD_X_CHAR,
	    CWORD_SIZEOF_PAGE_X,
	    offset_vals->x);
    }
    else if ( ((crossword_record->x_dimensions * CWORD_C_TXT_CELLSIZE_X) + 1 > CWORD_SIZEOF_PAGE_X) &&
	       ((crossword_record->x_dimensions * CWORD_C_TXT_CELLSIZE_X) + 1 < CWORD_SIZEOF_BIGPAGE_X) )
    {
	offset_vals->x = (int) ((float) (CWORD_SIZEOF_BIGPAGE_X - 1 - length) /(float) 2);

	cword_messages_output_message (
	    CWORD_S_TXTOFFLARGEX,
	    (SCREEN3 | TRACE3),
    	    CWORD_X_CHAR,
	    CWORD_SIZEOF_PAGE_X,
	    CWORD_SIZEOF_BIGPAGE_X,
	    offset_vals->x);
    }
    else
    {
	/*
	 *  Do nothing for the time being
	 */

	cword_messages_output_message (
	    CWORD_S_TXTOFFNO,
	    (SCREEN3 | TRACE3),
    	    CWORD_X_CHAR);
    }

    /*
     *  Fix up the Y offset
     */

    length = crossword_record->y_dimensions * CWORD_C_TXT_CELLSIZE_Y;

    if ((crossword_record->y_dimensions * CWORD_C_TXT_CELLSIZE_Y) + 1 < CWORD_SIZEOF_PAGE_Y)
    {
	offset_vals->y = (int) ((float) (CWORD_SIZEOF_PAGE_Y - 3 - length) /(float) 2);

	cword_messages_output_message (
	    CWORD_S_TXTOFFSMALL,
	    (SCREEN3 | TRACE3),
    	    CWORD_Y_CHAR,
	    CWORD_SIZEOF_PAGE_Y,
	    offset_vals->y);
    }
    else
    {
	/*
	 *  Just offset the display of the crossword grid by two units for
	 *  viewability sake!
	 */

	offset_vals->y = 2;

	/*
	 *  ?? What happens if it's too big ??
	 */

	cword_messages_output_message (
	    CWORD_S_TXTOFFLARGEY,
	    (SCREEN3 | TRACE3),
    	    CWORD_Y_CHAR,
	    CWORD_SIZEOF_PAGE_Y,
	    offset_vals->y);
    }

    return (CWORD_S_NORMAL);

}



PRIVATE void cword_print__convert_ps_line (
	char			*break_str)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function to insert forward slashes, if necessary, in a line of text to
 *	be printed to a postscript file.  It searches for characters such as
 *	'(', ')' and '\'.  When found these characters are prefixed by a '\' to
 *	ensure that the text in the line is correctly printed.  If these
 *	characters aren't prefixed, they will cause problems for most
 *	postscript interpreters upon printing.
 *
 *  FORMAL PARAMETERS:
 *
 *	*break_str		(r/w) String to check and alter if needs be
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      None
 *--
 */

{

    char temp_str[CWORD_SIZEOF_CLUE + 1];

    int	count = 0,
	count1 = 0;

    while (break_str[count] != (char) 0)
    {
	if ( (break_str[count] == (char) '(') ||
	      (break_str[count] == (char) ')') ||
	       (break_str[count] == (char) '\\') )
	{
	    temp_str[count1++] = (char) '\\';
	    temp_str[count1++] = break_str[count++];
	}
	else
	{
	    temp_str[count1++] = break_str[count++];
	}
    }

    temp_str[count1] = (char) 0;

    DEBUG_TRACE (
	"Clue after PS conversion is \"%s\"\n",
	temp_str);

    strcpy (break_str, temp_str);

    return;

}



PRIVATE void cword_print__break_up_line (
	int			text_type,
	char			*break_str)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function to help in the task of breaking up a line of output to a
 *	postscript or text print file.
 *
 *  FORMAL PARAMETERS:
 *
 *	text_type		(r) Tells the function whether the line's being printed 
 *				    to a PS or text file
 *      *break_str		(r/w) Line of text to check
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL		Normal completion
 *--
 */

{

    /*
     *	This function will help decide where a sentence should be parsed before
     *	printing.  This will take into account the following:
     * 
     *	1. Splitting of words isn't allowed but will take place if it cannot be 
     *	   avoided.
     */

    char *temp_ptr = (char *) 0,
	 *free_ptr = (char *) 0,
	 total_str[CWORD_SIZEOF_CLUE + 1];

    unsigned int count = 0,
		 count1 = 0,
		 width = 0,
		 size_break_string = 0;

    width = (text_type == POSTSCRIPT) ? CWORD_C_PRINT_PS_LINE_WIDTH : CWORD_C_PRINT_TXT_LINE_WIDTH;

    size_break_string = strlen (break_str);

    DEBUG_TRACE ("Before line breaks :\"%s\"\n", break_str);

    memset (total_str, 0, sizeof (total_str));

    if (strlen (break_str) >= (width + 1))
    {
	temp_ptr = (char *) calloc (1, CWORD_SIZEOF_CLUE + 1);
	free_ptr = temp_ptr;

	while (strlen (break_str) > width)
	{
	    if (break_str[width - 1] == (char) ' ')
	    {
		/*
		 *  Stick a seperator in here as we've struck lucky on our
		 *  first go.
		 */

		break_str[width - 1] = (char) CWORD_C_SEPERATOR;
	    }
	    else if ( (break_str[width - 1] == (char) '(') ||
	               (break_str[width - 1] == (char) ')') )
	    {
		/*
		 *  If we hit a '(' or a ')' then continue to go backwards
		 *  until we find a space.  When we do find a space stick a
		 *  seperator in!
		 */

		count = width - 1;

		while ( (break_str[count] != (char) ' ') && (count != 0) )
		    count--;

		if (count == 0)
		{
		    while ( (break_str[count] != (char) ' ') && (break_str[count] != (char) 0) )
			count++;

		    if (break_str[count] == (char) 0)
		    {
			printf ("# Error correctly parsing clues - print problem possible\n");
		    }
		}

		break_str[count] = (char) CWORD_C_SEPERATOR;
	    }
	    else
	    {
	        count = width - 1;

		while ( (break_str[count] != (char) ' ') && (break_str[count] != (char) 0) )
		    count++;

		if (break_str[count] == (char) 0)
		{
		    while ( (break_str[count] != (char) ' ') && (count != 0) )
			count--;

		    if (count == 0)
		    {
			printf ("# Error correctly parsing clues - print problem possible\n");
		    }
		}

		break_str[count] = (char) CWORD_C_SEPERATOR;
	    }

	    count = 0;
	    while ( (break_str[count] != (char) CWORD_C_SEPERATOR) && 
                      (break_str[count] != (char) 0) )
	    {
		total_str[count1] = break_str[count];
		count1++;
		count++;
	    }
	    total_str[count1++] = break_str[count++];

	    /*
	     *	Doing some juggling to get break_str properly set up to
	     *	continue in this loop.
	     */

	    strcpy (temp_ptr, break_str);
	    temp_ptr += count;
	    strcpy (break_str, temp_ptr);
	}

	strcat (total_str, break_str);
	strcpy (break_str, total_str);

	free (free_ptr);
    }

    /*
     *	Check the output str for unpopular postscript characters when it's been
     *	already split for printing.
     */

    if (text_type == POSTSCRIPT)
    {
	cword_print__convert_ps_line (break_str);
    }

    DEBUG_TRACE ("After line breaks :\"%s\"\n", break_str);

    if (size_break_string > strlen (break_str))
    {
	printf ("# Error converting clue for printing - please check output file\n");
    }

    return;

}



PRIVATE int cword_print__header_ps_comments (
	int			file_type,
	char			*crossword_name,
	int			output_fd)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function to output the postscript informational header to the beginning
 *	of the output postscript file.
 *
 *  FORMAL PARAMETERS:
 *
 *	file_type		(r) Type of output postscript file being written 
 *	*crossword_name		(r) Crossword/grid name
 *	output_fd		(r) Output file descriptor
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL
 *--
 */

{

    char output_str[CWORD_SIZEOF_MESSAGE + 1];

    time_t  temp_time;

#if !(defined(__MSDOS__) || defined(MSDOS))

    sprintf (
	output_str,
	"%%!PS-Adobe-1.0\n%%%%Creator: %s\n%%%%Title: %s\n",
	cuserid (0),
	crossword_name);
    CWORD_PRINT_WRITE_LINE (output_fd, output_str);

    time (&temp_time);

    sprintf (
	output_str,
	"%%%%BoundingBox: 0 0 612 792\n%%%%CreationDate: %s%%%%EndComments\n%%!\n\n",
	asctime (localtime (&temp_time)));
    CWORD_PRINT_WRITE_LINE (output_fd, output_str);

#else

    sprintf (
	output_str,
	"%%!PS-Adobe-1.0\n%%%%Creator: %s\n%%%%Title: %s\n",
	cword_g_defaults_record.username,
	crossword_name);
    CWORD_PRINT_WRITE_LINE (output_fd, output_str);

    /*
     *  ?? Is this portable to DOS ??
     */

    sprintf (
	output_str,
	"%%%%BoundingBox: 0 0 612 792\n%%%%CreationDate: %s%%%%EndComments\n%%!\n\n",
	asctime (localtime (&temp_time)));
    CWORD_PRINT_WRITE_LINE (output_fd, output_str);

#endif

    switch (file_type)
    {
    case (BLOCK_GRID_FILE):
	sprintf (
	    output_str,
	    "%%%% %s file\n\n",
	    BLOCK_GRID_STR);
	CWORD_PRINT_WRITE_LINE (output_fd, output_str);

	break;

    case (BARRED_GRID_FILE):
	sprintf (
	    output_str,
	    "%%%% %s file\n\n",
	    BARRED_GRID_STR);
	CWORD_PRINT_WRITE_LINE (output_fd, output_str);

	break;

    case (BLOCK_CROSSWORD_FILE):
	sprintf (
	    output_str,
	    "%%%% %s file\n\n",
	    BLOCK_CROSSWORD_STR);
	CWORD_PRINT_WRITE_LINE (output_fd, output_str);

	break;

    case (BARRED_CROSSWORD_FILE):
	sprintf (
	    output_str,
	    "%%%% %s file\n\n",
	    BARRED_CROSSWORD_STR);
	CWORD_PRINT_WRITE_LINE (output_fd, output_str);

	break;

    default:

	break;

    }

    return (CWORD_S_NORMAL);

}



PRIVATE int cword_print__preamble (
	int			output_fd)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function to write to the PS output file the preamble details.
 *
 *  FORMAL PARAMETERS:
 *
 *	output_fd		(r) Output file descriptor
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL
 *--
 */

{

    /*
     *	Write out the details of the six types of pre-defined postscript macros
     *	being used during PS printing:
     *
     *	1.  DrawBoxLines
     *	2.  BlackBox1
     *  3.  BlackBox2
     *	4.  ClueIndex
     *	5.  Clue
     *	6.  ClueAnswer
     */

    char output_str[CWORD_SIZEOF_MESSAGE + 1];

    strcpy (output_str, "%% Preamble defining some PS functions used further down.\n\n");
    CWORD_PRINT_WRITE_LINE (output_fd, output_str);

    strcpy (output_str, "/DrawBoxLines\t{\n\t\tnewpath\n\t\t4 2 roll\n\t\tmoveto\n\t\tlineto\n\t\tstroke\n\t\t} def\n\n");
    CWORD_PRINT_WRITE_LINE (output_fd, output_str);

    strcpy (output_str, "/BlackBox1\t{\n\t\t/b_side exch def\n\t\tnewpath\n\t\tmoveto\n\t\t0 b_side rlineto\n\t\tb_side 0 rlineto\n\t\t0 b_side neg rlineto\n\t\tclosepath fill\n\t\t} def\n\n");
    CWORD_PRINT_WRITE_LINE (output_fd, output_str);

    strcpy (output_str, "/BlackBox2\t{\n\t\t/c_side exch def\n\t\t/b_side exch def\n\t\tnewpath\n\t\tmoveto\n\t\t0 b_side rlineto\n\t\tc_side 0 rlineto\n\t\t0 b_side neg rlineto\n\t\tclosepath fill\n\t\t} def\n\n");
    CWORD_PRINT_WRITE_LINE (output_fd, output_str);

    strcpy (output_str, "/ClueIndex\t{\n\t\t3 1 roll\n\t\tmoveto show\n\t\t} def\n\n");
    CWORD_PRINT_WRITE_LINE (output_fd, output_str);

#if 0
    /*
     *  New definition of clue better 
     */

    strcpy (output_str, "/Clue\t\t{\n\t\t/answer exch def\n\t\t/index exch def\n\t\tmoveto currentpoint\n\t\tindex show\n\t\texch\n\t\t15 add\n\t\texch\n\t\tmoveto\n\t\tanswer show\n\t\t} def\n\n");
    CWORD_PRINT_WRITE_LINE (output_fd, output_str);
#endif

    strcpy (output_str, "/Clue\t\t{\n\t\t/answer exch def\n\t\tmoveto currentpoint\n\t\tanswer show\n\t\t} def\n\n");
    CWORD_PRINT_WRITE_LINE (output_fd, output_str);

    strcpy (output_str, "/ClueAnswer\t{\n\t\t3 1 roll\n\t\tmoveto show\n\t\t} def\n\n");
    CWORD_PRINT_WRITE_LINE (output_fd, output_str);

    strcpy (output_str, "%%EndProlog\n\n");
    CWORD_PRINT_WRITE_LINE (output_fd, output_str);

    return (CWORD_S_NORMAL);

}



PRIVATE int cword_print__ps_new_page (
	int			output_fd)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *      Function to simplify the creation of a new page for postscript output.
 *
 *  FORMAL PARAMETERS:
 *
 *	output_fd		(r) File descriptor of output file
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      None!
 *--
 */

{

    char output_str[CWORD_SIZEOF_MESSAGE + 1];

    CWORD_PRINT_WRITE_LINE (output_fd, "\nshowpage\n");

    cword_print_l_page_number++;
    PS_PRINT_PAGECOMMENT (output_str, cword_print_l_page_number);
    CWORD_PRINT_WRITE_LINE (output_fd, output_str);

    return (CWORD_S_NORMAL);

}



PRIVATE int cword_print__block_ps_grid (
	int			preamble,
	int			output_fd,
	CROSSWORD_RECORD	*crossword_record,
	CROSSWORD_GRID		crossword_grid)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function which writes out the empty crossword grid to the output file
 *	plus clue numbers.
 *
 *  FORMAL PARAMETERS:
 *
 *	preamble,		(r) Whether to print the preamble or not?
 *      output_fd		(r) Output file descriptor
 *	*crossword_record	(r) structure containing the crossword details
 *	crossword_grid		(r) structure containing the completed crossword grid
 *
 *  IMPLICIT INPUTS:
 *
 *      cword_g_defaults_record.x_dimensions
 *	cword_print_l_page_number
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL
 *--
 */

{

    char output_str[CWORD_SIZEOF_MESSAGE + 1];

    COORDS_STRUCT offset_vals;

    int	count1,
	count2;

    cword_messages_output_message (
	    CWORD_PSTXTMSG,
	    TRACE2,
	    "block postscript grid", 
	    crossword_record->name);

    if (preamble == PREAMBLE)
    {
	cword_print__preamble (output_fd);

	/* Writing out the name of the grid to the top of the page */

	strcpy (output_str, "%% Setting up the crossword name 18 points wide\n");
	CWORD_PRINT_WRITE_LINE (output_fd, output_str);

	if (crossword_record->name[0] == (char) 0)
	{
	    sprintf (
		    output_str, 
		    "/%s findfont 18 scalefont setfont\n\t%3d %3d moveto (Grid) show\n\n",
		    HEADER_FONT,
		    CWORD_C_HEADER_X,
		    CWORD_C_HEADER_Y);
	}
	else
	{
	    cword_print__convert_ps_line (crossword_record->name);

	    sprintf (
		    output_str, 
		    "/%s findfont 18 scalefont setfont\n\t%3d %3d moveto (Grid : %s) show\n\n",
		    HEADER_FONT,
		    CWORD_C_HEADER_X,
		    CWORD_C_HEADER_Y,
		    crossword_record->name);
	}

	CWORD_PRINT_WRITE_LINE (output_fd, output_str);
    }

    /* Calculate the offsets to make the postscript look better when printed */

    cword_print__cal_ps_offset (
		    GRID_PRINT,
		    crossword_record,
		    &offset_vals);

    sprintf (
	output_str, 
	"\n%% Setting up the overall box shape %d points wide\n",
	CWORD_C_OUTSTROKE_POINT_SIZE);
    CWORD_PRINT_WRITE_LINE (output_fd, output_str);

    sprintf (
	    output_str, 
	    "newpath\n %3d %3d moveto\n 0 %3d rlineto\n %3d 0 rlineto\n 0 -%.3d rlineto\n closepath\n",
	    offset_vals.x - 1,
	    offset_vals.y - 1,
	    (offset_vals.cell_point_size * crossword_record->y_dimensions) + 1,
	    (offset_vals.cell_point_size * crossword_record->x_dimensions) + 1,
	    (offset_vals.cell_point_size * crossword_record->y_dimensions) + 1);
    CWORD_PRINT_WRITE_LINE (output_fd, output_str);

    /*
     *  Outside boundary box stroke width
     */

    sprintf (
	output_str,
	" %-d setlinewidth\n stroke\n\n",
	CWORD_C_OUTSTROKE_POINT_SIZE);

    CWORD_PRINT_WRITE_LINE (output_fd, output_str);

    /* Writing out the separating lines in the crossword box */

    sprintf (
	output_str, 
	"%-d setlinewidth\n\n",
	CWORD_C_STROKE_POINT_SIZE);
    CWORD_PRINT_WRITE_LINE (output_fd, output_str);

    strcpy (output_str, "%% Setting up the moving X coordinates\n");
    CWORD_PRINT_WRITE_LINE (output_fd, output_str);

    for ( count1 = offset_vals.x + offset_vals.cell_point_size; 
	   count1 < (offset_vals.x + (offset_vals.cell_point_size * crossword_record->x_dimensions)); 
	    count1 += offset_vals.cell_point_size)
    {
	sprintf (
		output_str,
		"\t%3d %3d %3d %3d DrawBoxLines\n",
		count1,
		offset_vals.y,
		count1,
		offset_vals.y + (offset_vals.cell_point_size * crossword_record->y_dimensions));
	CWORD_PRINT_WRITE_LINE (output_fd, output_str);
    }

    strcpy (output_str, "\n%% Setting up the moving Y coordinates\n");
    CWORD_PRINT_WRITE_LINE (output_fd, output_str);

    for ( count1 = offset_vals.y + offset_vals.cell_point_size; 
           count1 < offset_vals.y + (offset_vals.cell_point_size * crossword_record->y_dimensions);
	    count1 += offset_vals.cell_point_size)
    {
	sprintf (
		output_str,
		"\t%3d %3d %3d %3d DrawBoxLines\n",
		offset_vals.x,
		count1,
		offset_vals.x + (offset_vals.cell_point_size * crossword_record->x_dimensions),
		count1);
	CWORD_PRINT_WRITE_LINE (output_fd, output_str);
    }

    /* Setting up the BLACK/WHITE boxes for the clues */

    strcpy (output_str, "\n%% Setting up the BLACK/WHITE boxes in the grid\n");
    CWORD_PRINT_WRITE_LINE (output_fd, output_str);

    for (count1 = 0; count1 < crossword_record->x_dimensions; count1++)
	for (count2 = 0; count2 < crossword_record->y_dimensions; count2++)
	{
	    if (crossword_grid[count1][count2] == CWORD_UNFILLED_CHAR)
	    {
		sprintf (
			output_str, 
			"\t%3d %3d %3d BlackBox1\n",
			offset_vals.x + (count1 * offset_vals.cell_point_size),
			offset_vals.y + (count2 * offset_vals.cell_point_size),
			offset_vals.cell_point_size);
		CWORD_PRINT_WRITE_LINE (output_fd, output_str);
	    }
	}

    /* Setting up the clue numbers for the clues */

    strcpy (output_str, "\n%% Setting up the clue numbers in the grid\n");
    CWORD_PRINT_WRITE_LINE (output_fd, output_str);

    sprintf (
	    output_str, 
	    "/%s findfont %2d scalefont setfont\n",
	    TEXT_FONT,
	    offset_vals.font_num_size);
    CWORD_PRINT_WRITE_LINE (output_fd, output_str);

    for (count1 = 0; count1 < crossword_record->clue_sum; count1++)
    {
	if (crossword_record->clue_details[count1].y_coordinate != crossword_record->y_dimensions)
	{
	    sprintf (
		output_str,
		"\t%3d %3d (%1d) ClueIndex\n",
		offset_vals.x + offset_vals.cell_numoff_x +
		 ((crossword_record->clue_details[count1].x_coordinate - 1) * offset_vals.cell_point_size),
		offset_vals.y + offset_vals.cell_numoff_y +
		 ((crossword_record->clue_details[count1].y_coordinate - 1) * offset_vals.cell_point_size),
		crossword_record->clue_details[count1].clue_number);
	}
	else
	{
	    /*
	     *	Allow the printing of the clue numbers on the top row of the crossword
	     *	to be more flush with the outer boundary than those located anywhere
	     *	else!
	     */

	    sprintf (
		output_str,
		"\t%3d %3d (%1d) ClueIndex\n",
		offset_vals.x + offset_vals.cell_numoff_x +
		 ((crossword_record->clue_details[count1].x_coordinate - 1) * offset_vals.cell_point_size),
		offset_vals.y + offset_vals.cell_numoff_y +
		 ((crossword_record->clue_details[count1].y_coordinate - 1) * offset_vals.cell_point_size) - 1,
		crossword_record->clue_details[count1].clue_number);
	}
	CWORD_PRINT_WRITE_LINE (output_fd, output_str);
    }

    if (preamble == PREAMBLE)
    {
	cword_print__ps_new_page (output_fd);
    }

    return (CWORD_S_NORMAL);

}



PRIVATE cword_print__clue_num_offset (
	int			clue_number)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function to return an offset of an offset for printing the clue numbers
 *	in a postscript list of clues.
 *
 *  FORMAL PARAMETERS:
 *
 *      clue_number		(r) Clue number.
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      offset to be used to print clue number on postscript page.
 *--
 */

{

    if (clue_number < 10)
	return (+2);
    else if ( (clue_number > 9) && (clue_number < 100) )
	return (0);
    else if (clue_number > 99)
	return (-2);

}



PRIVATE int cword_print__common_ps_clues (
	int			output_fd,
	CROSSWORD_RECORD	*crossword_record)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *      Function which writes out the clues to the output file.
 *
 *  FORMAL PARAMETERS:
 *
 *      output_fd		(r) Output file descriptor
 *	*crossword_record	(r) Structure containing the crossword details
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL
 *--
 */

{

    char output_str[CWORD_SIZEOF_MESSAGE + 1],
	 temp_str[CWORD_SIZEOF_MESSAGE + 1],
    	 *string_ptr;

    COORDS_STRUCT offset_vals;

    int	count_across,
	count_down,
	index_across,
	index_down,
	sum_across,
	sum_down,
	temp_val,
	temp_y;

    BOOLEAN are_all_clues_written = FALSE;

    cword_messages_output_message (
	CWORD_PSTXTMSG,
	TRACE2,
	"common postscript clues", 
	crossword_record->name);

    /*
     *	Calculate the initial offsets to make the postscript look better when
     *	printed.
     */

    cword_print__cal_ps_offset (
	CLUES_PRINT,
	crossword_record,
	&offset_vals);

    /* 
     *	Calculate the offset to make the postscript look better when printed.
     *
     *	Check if the grid for this crossword is greater than 19 in width or
     *	length.  If so then execute a showpage and print the clues on a new
     *	page.
     */

    if (offset_vals.new_page == TRUE)
    {
	/*
	 *  Put the clues on a new page
	 */

	cword_print__ps_new_page (output_fd);

	/* Writing out the crossword clues header to the top of the page */

	strcpy (output_str, "%% Setting up the crossword clues header 18 points wide\n");
	CWORD_PRINT_WRITE_LINE (output_fd, output_str);

	if (crossword_record->name[0] == (char) 0)
	{
	    sprintf (
		output_str, 
		"/%s findfont 18 scalefont setfont\n\t%3d %3d moveto (Clues) show\n\n",
		HEADER_FONT,
		CWORD_C_HEADER_X,
		CWORD_C_HEADER_Y);
	}
	else
	{
	    sprintf (
		output_str, 
		"/%s findfont 18 scalefont setfont\n\t%3d %3d moveto (Clues : %s) show\n\n",
		HEADER_FONT,
		CWORD_C_HEADER_X,
		CWORD_C_HEADER_Y,
		crossword_record->name);
	}

	CWORD_PRINT_WRITE_LINE (output_fd, output_str);
    }

    /* Setting up the descriptions of the clues */

    strcpy (output_str, "\n%% Setting up the clue descriptions\n");
    CWORD_PRINT_WRITE_LINE (output_fd, output_str);

    sprintf (
	output_str, 
	"/%s findfont 12 scalefont setfont %3d %3d moveto (   %s) show\n",
	HEADER_FONT,
	offset_vals.x + 25,
	offset_vals.y,
	CWORD_ACROSS_STR);
    CWORD_PRINT_WRITE_LINE (output_fd, output_str);

    sprintf (
	output_str, 
	"/%s findfont 12 scalefont setfont %3d %3d moveto (   %s) show\n",
	HEADER_FONT,
	offset_vals.x + 325,
	offset_vals.y,
	CWORD_DOWN_STR);
    CWORD_PRINT_WRITE_LINE (output_fd, output_str);
 
    sprintf (
	output_str, 
	"/%s findfont 10 scalefont setfont\n",
	HEADER_FONT);
    CWORD_PRINT_WRITE_LINE (output_fd, output_str);

    /*
     *  ?? Delete once I'm happy that the new impelementation works fine ??
     */

#if 0

    /*
     *	Not strong enough to allow multi-page postscript output of the clues as
     *	there is not enough of information passing between the across and down
     *	clues in the given definition.
     */

    temp_val = 0;
    for (count1 = 0; count1 < crossword_record->clue_sum; count1++)
    {
	/*
	 *  Nail the ACROSS clues first
	 */

	if (crossword_record->clue_details[count1].clue_direction == ACROSS)
	{
	    sprintf (
		output_str, 
		"\t%3d %3d (%2d.) Clue\n",
		(crossword_record->clue_details[count1].clue_number < 100) ? 
		    offset_vals.x - 30 : offset_vals.x - 35,
		offset_vals.y - 25 - (temp_val * CWORD_C_WIDTH_BETWEEN_CLUES),
		crossword_record->clue_details[count1].clue_number);
	    CWORD_PRINT_WRITE_LINE (output_fd, output_str);

	    strcpy (temp_str, crossword_record->clue_details[count1].clue_description);

	    /*
	     *  Function to split the clue for proper outputting.
	     */

	    cword_print__break_up_line (
			POSTSCRIPT,
			temp_str);

	    string_ptr = strtok (temp_str, CWORD_C_SEPERATOR_CHAR);

	    do
	    {
		temp_y = offset_vals.y - 25 - (temp_val * CWORD_C_WIDTH_BETWEEN_CLUES);

		if (temp_y < CWORD_C_FOOTER_Y)
		{
		    temp_val = 0;

		    cword_print__ps_new_page (output_fd);

#if 0
		    /*
		     *	Just print a message stating overflow is occuring --
		     *	nothing else.
		     */

		    cword_messages_output_message (
			    CWORD_S_PRINTPSOVERFLOW,
			    SCREEN2);
#endif
		}

		sprintf (
			output_str, 
			"\t%3d %3d (%s) Clue\n",
			offset_vals.x - 10,
			temp_y,
			string_ptr);
		CWORD_PRINT_WRITE_LINE (output_fd, output_str);

		temp_val++;
	    }
	    while ( (string_ptr = strtok (0, CWORD_C_SEPERATOR_CHAR)) 
    				 != (char *) 0 );
	}
    }

    temp_val = 0;
    for (count1 = 0; count1 < crossword_record->clue_sum; count1++)
    {    
	if (crossword_record->clue_details[count1].clue_direction == DOWN)
	{
	    sprintf (
		output_str, 
		"\t%3d %3d (%2d.) Clue\n",
		(crossword_record->clue_details[count1].clue_number < 100) ? 
		    offset_vals.x + 250 : offset_vals.x + 245,
		offset_vals.y - 25 - (temp_val * CWORD_C_WIDTH_BETWEEN_CLUES),
		crossword_record->clue_details[count1].clue_number);
	    CWORD_PRINT_WRITE_LINE (output_fd, output_str);

	    strcpy (temp_str, crossword_record->clue_details[count1].clue_description);

	    /*
	     *  Function to split the clue for proper outputting.
	     */

	    cword_print__break_up_line (
			POSTSCRIPT,
			temp_str);

	    string_ptr = strtok (temp_str, CWORD_C_SEPERATOR_CHAR);

	    do
	    {
		temp_y = offset_vals.y - 25 - (temp_val * CWORD_C_WIDTH_BETWEEN_CLUES);

		if (temp_y < CWORD_C_FOOTER_Y)
		{
		    temp_val = 0;

		    cword_print__ps_new_page (output_fd);

#if 0
		    /*
		     *	Just print a message stating overflow is occuring --
		     *	nothing else.
		     */

		    cword_messages_output_message (
			    CWORD_S_PRINTPSOVERFLOW,
			    SCREEN2);
#endif
		}

		sprintf (
			output_str, 
			"\t%3d %3d (%s) Clue\n",
			offset_vals.x + 270,
			temp_y,
			string_ptr);
		CWORD_PRINT_WRITE_LINE (output_fd, output_str);

		temp_val++;
	    }
	    while ( (string_ptr = strtok (0, CWORD_C_SEPERATOR_CHAR))
    				 != (char *) 0 );
	}	
    }
#endif

    sum_across =  cword_stats_get_clue_sum_for (
			ACROSS,
			crossword_record);

    sum_down =  cword_stats_get_clue_sum_for (
			DOWN,
			crossword_record);

    /*
     *	This scheme will allow the clues be printed in a manner that will allow
     *	extra pages be appended in an overflow situation?
     */

    count_across = count_down = 0;
    index_across = index_down = 0;
    while (are_all_clues_written == FALSE)
    {
	temp_val = 0;
	if (count_across < sum_across)
	{
	    /*
	     *	Write ACROSS clues until end of page, incrementing count_across
	     *	and index_count as we go.
	     */

	    while ( (index_across < crossword_record->clue_sum)
                                   &&
		     (count_across < sum_across) )
	    {
		if (crossword_record->clue_details[index_across].clue_direction == ACROSS)
		{
		    temp_y = offset_vals.y - 25 - (temp_val * CWORD_C_WIDTH_BETWEEN_CLUES);

		    /*
		     *	The extra 25 is to make sure that the clues don't bump
		     *	into possible remarks field and also that the clues
		     *	don't get too close to the end of the page :-)
		     */

		    if (temp_y < CWORD_C_FOOTER_Y + 35)
		    {
			DEBUG_TRACE ("New page needed for across clues\n");

			break;
		    }

		    sprintf (
			output_str, 
			"\t%3d %3d (%3d.) Clue\n",
#if 0
			(crossword_record->clue_details[index_across].clue_number < 100) ? 
			    offset_vals.x - 30 : offset_vals.x - 33,
#endif
			cword_print__clue_num_offset (
			    crossword_record->clue_details[index_across].clue_number) + 
				offset_vals.x - 30,			
			offset_vals.y - 25 - (temp_val * CWORD_C_WIDTH_BETWEEN_CLUES),
			crossword_record->clue_details[index_across].clue_number);
		    CWORD_PRINT_WRITE_LINE (output_fd, output_str);

		    strcpy (temp_str, crossword_record->clue_details[index_across].clue_description);

		    /*
		     *  Function to split the clue for proper outputting.
		     */

		    cword_print__break_up_line (
				POSTSCRIPT,
				temp_str);

		    string_ptr = strtok (temp_str, CWORD_C_SEPERATOR_CHAR);
		    do
		    {
			/*
			 *  Make sure there is enough room at the end of the
			 *  page to print out the contents of this do/while
			 *  loop.
			 */

			temp_y = offset_vals.y - 25 - (temp_val * CWORD_C_WIDTH_BETWEEN_CLUES);

			sprintf (
				output_str, 
				"\t%3d %3d (%s) Clue\n",
				offset_vals.x - 10,
				temp_y,
				string_ptr);
			CWORD_PRINT_WRITE_LINE (output_fd, output_str);

			temp_val++;
		    }
		    while ( (string_ptr = strtok (0, CWORD_C_SEPERATOR_CHAR)) 
					 != (char *) 0 );

		    /*
		     *  Increment count_across before continuing
		     */

		    count_across++;
		}

		index_across++;
	    }
	}

	temp_val = 0;
	if (count_down < sum_down)
	{
	    /*
	     *	Write DOWN clues until end of page, incrementing count_down as
	     *	we go along.
	     */

	    while ( (index_down < crossword_record->clue_sum)
                                   &&
		     (count_down < sum_down) )
	    {
		if (crossword_record->clue_details[index_down].clue_direction == DOWN)
		{
		    temp_y = offset_vals.y - 25 - (temp_val * CWORD_C_WIDTH_BETWEEN_CLUES);

		    /*
		     *	The extra 25 is to make sure that the clues don't bump
		     *	into possible remarks field and also that the clues
		     *	don't get too close to the end of the page :-)
		     */

		    if (temp_y < CWORD_C_FOOTER_Y + 35)
		    {
			DEBUG_TRACE ("New page needed for down clues\n");

			break;
		    }

		    sprintf (
			output_str, 
			"\t%3d %3d (%3d.) Clue\n",
#if 0
			(crossword_record->clue_details[index_down].clue_number < 100) ? 
			    offset_vals.x + 250 : offset_vals.x + 247,
#endif
			cword_print__clue_num_offset (
			    crossword_record->clue_details[index_down].clue_number) + 
				offset_vals.x + 250,			
			offset_vals.y - 25 - (temp_val * CWORD_C_WIDTH_BETWEEN_CLUES),
			crossword_record->clue_details[index_down].clue_number);
		    CWORD_PRINT_WRITE_LINE (output_fd, output_str);

		    strcpy (temp_str, crossword_record->clue_details[index_down].clue_description);

		    /*
		     *  Function to split the clue for proper outputting.
		     */

		    cword_print__break_up_line (
				POSTSCRIPT,
				temp_str);

		    string_ptr = strtok (temp_str, CWORD_C_SEPERATOR_CHAR);
		    do
		    {
			temp_y = offset_vals.y - 25 - (temp_val * CWORD_C_WIDTH_BETWEEN_CLUES);

			sprintf (
				output_str, 
				"\t%3d %3d (%s) Clue\n",
				offset_vals.x + 270,
				temp_y,
				string_ptr);
			CWORD_PRINT_WRITE_LINE (output_fd, output_str);

			temp_val++;
		    }
		    while ( (string_ptr = strtok (0, CWORD_C_SEPERATOR_CHAR))
					 != (char *) 0 );

		    /*
		     *  Increment count_down before continuing
		     */

		    count_down++;
		}	

		index_down++;
	    }
	}

	if ( (count_across < sum_across) || (count_down < sum_down) )
	{
	    /*
	     *	Calculate the offsets of a second page of clues.  These values
	     *	could be different given what kind of grid was initially chosen.
	     */

	    cword_print__cal_ps_offset (
			    NEWPAGE_CLUES_PRINT,
			    crossword_record,
			    &offset_vals);

	    /*
	     *  Write newpage, print ACROSS/DOWN, and continue.
	     */

	    cword_print__ps_new_page (output_fd);

	    sprintf (
		    output_str, 
		    "/%s findfont 12 scalefont setfont %3d %3d moveto (   %s) show\n",
		    HEADER_FONT,
		    offset_vals.x + 25,
		    offset_vals.y,
		    CWORD_ACROSS_STR);
	    CWORD_PRINT_WRITE_LINE (output_fd, output_str);

	    sprintf (
		    output_str, 
		    "/%s findfont 12 scalefont setfont %3d %3d moveto (   %s) show\n",
		    HEADER_FONT,
		    offset_vals.x + 325,
		    offset_vals.y,
		    CWORD_DOWN_STR);
	    CWORD_PRINT_WRITE_LINE (output_fd, output_str);

	    /*
	     *	Re-set the default font to be used when printing.
	     */

	    sprintf (
		output_str, 
		"/%s findfont 10 scalefont setfont\n",
		HEADER_FONT);
	    CWORD_PRINT_WRITE_LINE (output_fd, output_str);
	}
	else
	{
	    /*
	     *	Finished.  Set are_all_clues_written to TRUE and exit from this
	     *	loop!
	     */

	    are_all_clues_written = TRUE;
	}
    }

    return (CWORD_S_NORMAL);

}



PRIVATE int cword_print__block_ps_answers (
	int			output_fd,
	CROSSWORD_RECORD	*crossword_record,
	CROSSWORD_GRID		crossword_grid)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function which writes out the answers in a grid to the output file.
 *
 *  FORMAL PARAMETERS:
 *
 *      output_fd		(r) Output file descriptor
 *	*crossword_record	(r) Structure containing the crossword details
 *	crossword_grid		(r) Structure containing the completed crossword grid
 *
 *  IMPLICIT INPUTS:
 *
 *      cword_g_defaults_record.x_dimensions
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL
 *	other
 *--
 */

{

    char output_str[CWORD_SIZEOF_MESSAGE + 1];

    COORDS_STRUCT offset_vals;

    int	status,
	count1,
	count2;

    cword_messages_output_message (
	    CWORD_PSTXTMSG,
	    TRACE2,
	    "block postscript answers", 
	    crossword_record->name);

    /*
     *  Call the print grid before putting answers into place
     */

    status = cword_print__block_ps_grid (
		NOPREAMBLE,
		output_fd,
		crossword_record,
		crossword_grid);

    if (status != CWORD_S_NORMAL)
    {    
	return (status);
    }

    /* Calculate the offsets to make the postscript look better when printed */

    cword_print__cal_ps_offset (
		    GRID_PRINT,
		    crossword_record,
		    &offset_vals);

    /* Setting up the header to the answers */

    strcpy (output_str, "\n%% Setting up the clue answers header\n");
    CWORD_PRINT_WRITE_LINE (output_fd, output_str);

    if (crossword_record->name[0] == (char) 0)
    {
	sprintf (
		output_str, 
		"/%s findfont 18 scalefont setfont\n\t%3d %3d moveto (Answers) show\n",
		HEADER_FONT,
		CWORD_C_HEADER_X,
		CWORD_C_HEADER_Y);
    }
    else
    {
	sprintf (
		output_str, 
		"/%s findfont 18 scalefont setfont\n\t%3d %3d moveto (Answers : %s) show\n",
		HEADER_FONT,
		CWORD_C_HEADER_X,
		CWORD_C_HEADER_Y,
		crossword_record->name);
    }
 
    CWORD_PRINT_WRITE_LINE (output_fd, output_str);

    /* Setting up the clue answers */

    strcpy (output_str, "\n%% Setting up the clue answers\n");
    CWORD_PRINT_WRITE_LINE (output_fd, output_str);

    sprintf (
	    output_str,
	    "/%s findfont %2d scalefont setfont\n",
	    HEADER_FONT,
	    offset_vals.font_letter_size);
    CWORD_PRINT_WRITE_LINE (output_fd, output_str);

    for (count1 = 0; count1 < crossword_record->clue_sum; count1++)
    {
	if (crossword_record->clue_details[count1].clue_direction == DOWN)
	{
	    for (count2 = 0; count2 < strlen (crossword_record->clue_details[count1].clue_answer); count2++)
	    {
		sprintf (
		    output_str, 
		    "\t%3d %3d (%c) ClueAnswer\n",
		    offset_vals.x + offset_vals.cell_charoff_x + 
		     (crossword_record->clue_details[count1].x_coordinate - 1) * offset_vals.cell_point_size,
		    offset_vals.y + offset_vals.cell_charoff_y + 
		     (crossword_record->clue_details[count1].y_coordinate - 1) * offset_vals.cell_point_size - 
		      (count2 * offset_vals.cell_point_size),
		    crossword_record->clue_details[count1].clue_answer[count2]);
		CWORD_PRINT_WRITE_LINE(output_fd, output_str);
	    }
        }
	else
	{
	    for (count2 = 0; count2 < strlen (crossword_record->clue_details[count1].clue_answer); count2++)
	    {
		sprintf (
		    output_str, 
		    "\t%3d %3d (%c) ClueAnswer\n",
		    offset_vals.x + offset_vals.cell_charoff_x + 
		     (crossword_record->clue_details[count1].x_coordinate - 1) * offset_vals.cell_point_size + 
		      (count2 * offset_vals.cell_point_size),
		    offset_vals.y + offset_vals.cell_charoff_y + 
		     (crossword_record->clue_details[count1].y_coordinate - 1) * offset_vals.cell_point_size,
		    crossword_record->clue_details[count1].clue_answer[count2]);
		CWORD_PRINT_WRITE_LINE(output_fd, output_str);
	    } 
        }	
    }

    return (CWORD_S_NORMAL);

}



PUBLIC int cword_print_block_ps (
	CROSSWORD_RECORD	*crossword_record)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function to print the completed verified grid and clues to a postscript
 *	file with a few comments thrown in
 *
 *  FORMAL PARAMETERS:
 *
 *      *crossword_record	(r) pointer to structure containing verified clues
 *
 *  IMPLICIT INPUTS:
 *
 *      cword_g_defaults_record.username
 *	cword_print_l_page_number
 *
 *  COMPLETION CODES:
 *
 *	CWORD_S_NORMAL		Normal Completion - print file completed
 *	other			Error status - print file aborted
 *--
 */
{

    CROSSWORD_GRID crossword_grid;

    char output_str[CWORD_SIZEOF_MESSAGE + 1],
	 output_file_name[CWORD_SIZEOF_FILESPEC + 1];

    int	output_fd,
	status;

    /*
     *	Will assume that the correct type of grid has already been found before
     *	coming in here!
     */

    status = cword_verify_get_block_props (
		crossword_record,
		crossword_grid);

    if (CWORD_S_NORMAL != status)
    {
	return (status);
    }

    sprintf (
	output_file_name,
	"%s%s",
	crossword_record->file_name,
	CWORD_SUFFIX_PRNPS_FILE);

    status = cword_utils_is_file_ok (
		WRITE_FILE,
		CWORD_SUFFIX_PRNPS_FILE,
		output_file_name);

    if (CWORD_S_NORMAL != status)
    {
	return (CWORD_S_FAILURE);
    }

    output_fd = open (
		    output_file_name, 
		    O_WRONLY | O_CREAT | O_TRUNC, 
		    0700);
    
    if (output_fd == -1)
    {
	cword_messages_output_message (
	    CWORD_FILEERROR,
	    (TRACE0 | SCREEN0),
	    output_file_name);

	return (CWORD_WRITEERROR);
    }

    /*
     *  Print out PS comments before continuing
     */

    cword_print__header_ps_comments (
	BLOCK_CROSSWORD_FILE,
	crossword_record->name,
	output_fd);

    status = cword_print__preamble (output_fd);

    if (status != CWORD_S_NORMAL)
    {
	return (status);
    }

    /* Writing out the name of the crossword to the top of the page */

    strcpy (output_str, "%% Setting up the crossword name 18 points wide\n");
    CWORD_PRINT_WRITE_LINE (output_fd, output_str);

    if (crossword_record->name[0] == (char) 0)
	sprintf (
	    output_str, 
	    "/%s findfont 18 scalefont setfont\n\t%3d %3d moveto (Crossword) show\n\n",
	    HEADER_FONT,
	    CWORD_C_HEADER_X,
	    CWORD_C_HEADER_Y);
    else
    {
	/*
	 *  Only called once as this will do for clues and answers too!
	 */

	cword_print__convert_ps_line (crossword_record->name);

	sprintf (
	    output_str, 
	    "/%s findfont 18 scalefont setfont\n\t%3d %3d moveto (Crossword : %s) show\n\n",
	    HEADER_FONT,
	    CWORD_C_HEADER_X,
	    CWORD_C_HEADER_Y,
	    crossword_record->name);
    }

    CWORD_PRINT_WRITE_LINE (output_fd, output_str);

    /* Setting up the remarks for the crossword */

    if (crossword_record->remarks[0] != (char) 0)
    {
	strcpy (output_str, "%% Setting up the crossword remarks\n");
	CWORD_PRINT_WRITE_LINE (output_fd, output_str);

	cword_print__convert_ps_line (crossword_record->remarks);

	sprintf (
	    output_str, 
	    "/%s findfont 14 scalefont setfont\n\t%3d %3d moveto (Remarks : %s) show\n",
	    HEADER_FONT,
	    CWORD_C_FOOTER_X,
	    CWORD_C_FOOTER_Y,
	    crossword_record->remarks);
        CWORD_PRINT_WRITE_LINE (output_fd, output_str);
    }

    status = cword_print__block_ps_grid (
		NOPREAMBLE,
		output_fd,
		crossword_record,
		crossword_grid);

    if (status != CWORD_S_NORMAL)
    {    
	return (status);
    }

    status = cword_print__common_ps_clues (
		output_fd,
		crossword_record);

    if (status != CWORD_S_NORMAL)
    {    
	return (status);
    }

    cword_print__ps_new_page (output_fd);

    status = cword_print__block_ps_answers (
		output_fd,
		crossword_record,
		crossword_grid);

    if (status != CWORD_S_NORMAL)
    {    
	return (status);
    }

    cword_print__ps_new_page (output_fd);

    close (output_fd);

    cword_messages_output_message (
	CWORD_PSWRITTEN,
	(TRACE0 | SCREEN0),
	output_file_name);

    return (CWORD_S_NORMAL);

}



PRIVATE int cword_print__block_txt_grid (
	int			preamble,
	int			output_fd,
	CROSSWORD_RECORD	*crossword_record,
	CROSSWORD_GRID		crossword_grid)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function to write out the crossword grid formation for a text file.
 *
 *  FORMAL PARAMETERS:
 *
 *	preamble		(r) Whether to print preamble or not?
 *      output_fd		(r) Output file descriptor
 *	*crossword_record	(r) Structure containing the crossword details
 *	crossword_grid		(r) Structure containing the completed crossword grid
 *
 *  FUNCTION VALUE:
 *
 *	CWORD_S_NORMAL		Normal completion
 *--
 */

{

    CROSSWORD_GRID clue_number_grid;

    COORDS_STRUCT offset_vals;

    char output_str[CWORD_SIZEOF_MESSAGE + 1],
	 horiz_str[CWORD_SIZEOF_MESSAGE + 1],
	 verti1_str[CWORD_SIZEOF_MESSAGE + 1],
	 verti2_str[CWORD_SIZEOF_MESSAGE + 1],
	 temp1_str[51],
	 temp2_str[11];

    int	 count1, 
	 count2;

    cword_messages_output_message (
	    CWORD_PSTXTMSG,
	    TRACE2,
	    "block text grid", 
	    crossword_record->name);

    if (preamble == PREAMBLE)
    {
	if (crossword_record->name[0] == (char) 0)
	{
	    strcpy (output_str, "\n\nGrid\n\n");
	}
	else
	{
	    sprintf (output_str,
		     "\n\nGrid : %s\n\n",
		     crossword_record->name);
	}

	CWORD_PRINT_WRITE_LINE (output_fd, output_str);
    }

    cword_print__cal_txt_offset (
	    crossword_record,
	    &offset_vals);

    for (count1 = 0; count1 < offset_vals.y; count1++)
    {
	strcpy (output_str, "\n");
	CWORD_PRINT_WRITE_LINE (output_fd, output_str);
    }    

    /*
     *  Have to write out the clue numbers but not the answers yet so construct an
     *  array of temporary clue numbers to use with the crossword.
     */

    for (count1 = 0; count1 < CWORD_SIZEOF_CELLDIM; count1++)
	for (count2 = 0; count2 < CWORD_SIZEOF_CELLDIM; count2++)
	    clue_number_grid[count1][count2] = NULL_CLUE_NUM;

    for (count1 = 0; count1 < crossword_record->clue_sum; count1++)
    {
	clue_number_grid[crossword_record->clue_details[count1].x_coordinate - 1]
			    [crossword_record->clue_details[count1].y_coordinate - 1] =
				crossword_record->clue_details[count1].clue_number;
    }

    /*
     *  Using the X offset to move grid into the middle of page
     */

    if (offset_vals.x > 0)
    {
	memset (horiz_str, 0, sizeof (horiz_str));

	for (count1 = 0; count1 < offset_vals.x; count1++)
	    horiz_str[count1] = (char) ' ';

	strcat (horiz_str, CORNER_STR);
    }
    else
    {
	strcpy (horiz_str, CORNER_STR);
    }

    for (count1 = 0; count1 < crossword_record->x_dimensions; count1++)
    {
	strcat (horiz_str, HOR_STR);
    }
    strcat (horiz_str, END_STR);

    CWORD_PRINT_WRITE_LINE (output_fd, horiz_str);

    /*
     *	Find out whether the biggest number clue is > 99.  If so then sprintf
     *  clue numbers appropriately!
     */

    for (count2 = crossword_record->y_dimensions - 1; count2 > - 1; count2--)
    {
	/*
	 *  Using the X offset to move grid into the middle of page
	 */

	if (offset_vals.x > 0)
	{
	    memset (verti1_str, 0, sizeof (verti1_str));
	    for (count1 = 0; count1 < offset_vals.x; count1++)
		    verti1_str[count1] = (char) ' ';

	    memset (verti2_str, 0, sizeof (verti2_str));
	    for (count1 = 0; count1 < offset_vals.x; count1++)
		    verti2_str[count1] = (char) ' ';

	    strcat (verti1_str, VER_STR);
	    strcat (verti2_str, VER_STR);
	}
	else
	{
	    strcpy (verti1_str, VER_STR);
	    strcpy (verti2_str, VER_STR);
	}

	for (count1 = 0; count1 < crossword_record->x_dimensions; count1++)
	{
	    /* Outputs the number of the clue to print out to the file */

	    if (clue_number_grid[count1][count2] != NULL_CLUE_NUM)
	    {
		DEBUG_TRACE (
		    "clue_number_grid[%2d][%2d] == %2d\n",
		    count1,
		    count2,
		    clue_number_grid[count1][count2]);		

		sprintf (
		    temp1_str, 
		    NUM1_STR,
		    clue_number_grid[count1][count2]);
	    }
	    else
	    if (crossword_grid[count1][count2] != CWORD_UNFILLED_CHAR)
	    {
		strcpy (temp1_str, INNER_BLANK1_STR);
	    }
	    else
	    {
		strcpy (temp1_str, INNER_BLACK_STR);
	    }

	    /* Taken or not taken */

	    if (crossword_grid[count1][count2] != CWORD_UNFILLED_CHAR)
	    {
		strcpy (temp2_str, INNER_BLANK1_STR);
	    }
	    else
	    {
		strcpy (temp2_str, INNER_BLACK_STR);
	    }
	    
	    strcat (verti1_str, temp1_str);
	    strcat (verti2_str, temp2_str);
	}

	strcat (verti1_str, END_STR);
	strcat (verti2_str, END_STR);

	CWORD_PRINT_WRITE_LINE (output_fd, verti1_str);
	CWORD_PRINT_WRITE_LINE (output_fd, verti2_str);
	CWORD_PRINT_WRITE_LINE (output_fd, horiz_str);
    }

    return (CWORD_S_NORMAL);

}



PRIVATE int cword_print__common_txt_clues (
	int			output_fd,
	CROSSWORD_RECORD	*crossword_record)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *      Function to write out the clues for the crossword to a text file
 *
 *  FORMAL PARAMETERS:
 *
 *      output_fd		(r) Output file descriptor
 *	*crossword_record	(r) Structure containing the crossword details
 *
 *  FUNCTION VALUE:
 *
 *	CWORD_S_NORMAL		Normal completion
 *--
 */

{

    char output_str[CWORD_SIZEOF_MESSAGE + 1],
    	 temp_str[CWORD_SIZEOF_MESSAGE + 1],
    	 *string_ptr;

    int count1;

    BOOLEAN is_first_pass,
	    is_over_100 = FALSE;

    cword_messages_output_message (
	    CWORD_PSTXTMSG,
	    TRACE2,
	    "block text clues", 
	    crossword_record->name);

    strcpy (output_str, "\t\t Across\n");
    CWORD_PRINT_WRITE_LINE (output_fd, output_str);

    strcpy (output_str, "\t\t ------\n\n");
    CWORD_PRINT_WRITE_LINE (output_fd, output_str);

    /*
     *	Find out whether the biggest number clue is > 99.  If so then sprintf
     *	with %3d. instead of %2d.
     */

    if (cword_stats_get_high_clue_num (
	    BOTH,
	    crossword_record) > 99)
	is_over_100 = TRUE;

    for (count1 = 0; count1 < crossword_record->clue_sum; count1++)
    {
	if (crossword_record->clue_details[count1].clue_direction == ACROSS)
	{
	    sprintf (
		    output_str,
		    "%*d. ",
		    is_over_100 ? 3 : 2,
		    crossword_record->clue_details[count1].clue_number);
	    CWORD_PRINT_WRITE_LINE (output_fd, output_str);

	    strcpy (temp_str, crossword_record->clue_details[count1].clue_description);

	    /*
	     *  Function to split the clue for proper outputting.
	     */

	    cword_print__break_up_line (
			TEXT,
			temp_str);

	    is_first_pass = TRUE;

	    string_ptr = strtok (temp_str, CWORD_C_SEPERATOR_CHAR);
	    do
	    {
		if (is_first_pass == TRUE)
		    sprintf (
			output_str, 
			"%s\n",
			string_ptr);
		else
		    sprintf (
			output_str, 
			"    %s\n",
			string_ptr);

		CWORD_PRINT_WRITE_LINE (output_fd, output_str);

		is_first_pass = FALSE;	    
	    }
	    while ( (string_ptr = strtok (0, CWORD_C_SEPERATOR_CHAR)) 
    				 != (char *) 0);
	}
    }

    strcpy (output_str, "\n\n\t\t Down\n");
    CWORD_PRINT_WRITE_LINE (output_fd, output_str);

    strcpy (output_str, "\t\t ----\n\n");
    CWORD_PRINT_WRITE_LINE (output_fd, output_str);
    
    for (count1 = 0; count1 < crossword_record->clue_sum; count1++)
    {
	if ( crossword_record->clue_details[count1].clue_direction == DOWN)
	{
	    sprintf (
		    output_str, 
		    "%*d. ",
		    is_over_100 ? 3 : 2,
		    crossword_record->clue_details[count1].clue_number);
	    CWORD_PRINT_WRITE_LINE (output_fd, output_str);

	    strcpy (temp_str, crossword_record->clue_details[count1].clue_description);

	    /*
	     *  Function to split the clue for proper outputting.
	     */

	    cword_print__break_up_line (
			TEXT,
			temp_str);

	    is_first_pass = TRUE;

	    string_ptr = strtok (temp_str, CWORD_C_SEPERATOR_CHAR);
	    do
	    {
		if (is_first_pass == TRUE)
		    sprintf (
			output_str, 
			"%s\n",
			string_ptr);
		else
		    sprintf (
			output_str, 
			"    %s\n",
			string_ptr);

		CWORD_PRINT_WRITE_LINE (output_fd, output_str);

		is_first_pass = FALSE;	    
	    }
	    while ( (string_ptr = strtok (0, CWORD_C_SEPERATOR_CHAR))
    				 != (char *) 0 );
	}	
    }

    return (CWORD_S_NORMAL);

}



PRIVATE int cword_print__block_txt_answers (
	int			output_fd,
	CROSSWORD_RECORD	*crossword_record,
	CROSSWORD_GRID		crossword_grid)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *      Function to writes out the answers of the crossword to a text file.
 *
 *  FORMAL PARAMETERS:
 *
 *      output_fd		(r) Output file descriptor
 *	*crossword_record	(r) Structure containing the crossword details
 *	crossword_grid		(r) Structure containing the completed crossword grid
 *
 *  FUNCTION VALUE:
 *
 *	CWORD_S_NORMAL		Normal completion
 *--
 */

{

    CROSSWORD_GRID clue_number_grid,
		   letter_grid;

    COORDS_STRUCT offset_vals;

    char horiz_str[CWORD_SIZEOF_MESSAGE + 1],
	 verti1_str[CWORD_SIZEOF_MESSAGE + 1],
	 verti2_str[CWORD_SIZEOF_MESSAGE + 1],
	 temp1_str[51],
	 temp2_str[11];

    int	count1,
	count2;

    cword_messages_output_message (
	    CWORD_PSTXTMSG,
	    TRACE2,
	    "block text answers", 
	    crossword_record->name);

    cword_print__cal_txt_offset (
	    crossword_record,
	    &offset_vals);

    for (count1 = 0; count1 < CWORD_SIZEOF_CELLDIM; count1++)
	for (count2 = 0; count2 < CWORD_SIZEOF_CELLDIM; count2++)
	{
	    clue_number_grid[count1][count2] = NULL_CLUE_NUM;
	    letter_grid[count1][count2] = NOT_TAKEN;
	}

    /* Obtains the clue_number_grid values */

    for (count1 = 0; count1 < crossword_record->clue_sum; count1++)
    {
	DEBUG_TRACE (
	    "clue_grid[%2d][%2d] == %2d\n",
	    crossword_record->clue_details[count1].x_coordinate - 1,
	    crossword_record->clue_details[count1].y_coordinate - 1,
	    crossword_record->clue_details[count1].clue_number);

	clue_number_grid[crossword_record->clue_details[count1].x_coordinate - 1]
			    [crossword_record->clue_details[count1].y_coordinate - 1] =
				crossword_record->clue_details[count1].clue_number;
    }

    /* Obtains the letter_grid values */

    for (count1 = 0; count1 < crossword_record->clue_sum; count1++)
    {
	if (crossword_record->clue_details[count1].clue_direction == ACROSS)
	{
	    for (count2 = 0; count2 < strlen (crossword_record->clue_details[count1].clue_answer); count2++)
	    {
		letter_grid[crossword_record->clue_details[count1].x_coordinate - 1 + count2]
			   [crossword_record->clue_details[count1].y_coordinate - 1] = 
				(int) crossword_record->clue_details[count1].clue_answer[count2];
	    }
	}
	else
	{
	    for (count2 = 0; count2 < strlen (crossword_record->clue_details[count1].clue_answer); count2++)
	    {
		letter_grid[crossword_record->clue_details[count1].x_coordinate - 1]
			   [crossword_record->clue_details[count1].y_coordinate - 1 - count2] = 
				(int) crossword_record->clue_details[count1].clue_answer[count2];
	    }
	}	
    }

    /*
     *  Using the X offset to move grid into the middle of page
     */

    if (offset_vals.x > 0)
    {
	memset (horiz_str, 0, sizeof (horiz_str));
	for (count1 = 0; count1 < offset_vals.x; count1++)
	    horiz_str[count1] = (char) ' ';

	strcat (horiz_str, CORNER_STR);
    }
    else
    {
	strcpy (horiz_str, CORNER_STR);
    }

    for (count1 = 0; count1 < crossword_record->x_dimensions; count1++)
    {
	strcat (horiz_str, HOR_STR);
    }
    strcat (horiz_str, END_STR);

    CWORD_PRINT_WRITE_LINE (output_fd, horiz_str);

    for (count2 = crossword_record->y_dimensions - 1; count2 > - 1; count2--)
    {
	/*
	 *  Using the X offset to move grid into the middle of page
	 */

	if (offset_vals.x > 0)
	{
	    memset (verti1_str, 0, sizeof (verti1_str));
	    for (count1 = 0; count1 < offset_vals.x; count1++)
		    verti1_str[count1] = (char) ' ';

	    memset (verti2_str, 0, sizeof (verti2_str));
	    for (count1 = 0; count1 < offset_vals.x; count1++)
		    verti2_str[count1] = (char) ' ';

	    strcat (verti1_str, VER_STR);
	    strcat (verti2_str, VER_STR);
	}
	else
	{
	    strcpy (verti1_str, VER_STR);
	    strcpy (verti2_str, VER_STR);
	}


	for (count1 = 0; count1 < crossword_record->x_dimensions; count1++)
	{
	    /* Outputs the number of the clue to print out to the file */

	    if (clue_number_grid[count1][count2] != NULL_CLUE_NUM)
	    {
		DEBUG_TRACE (
		    "clue_grid[%2d][%2d] == %2d\n",
		    count1,
		    count2,
		    clue_number_grid[count1][count2]);

		sprintf (
		    temp1_str, 
		    NUM1_STR,
		    clue_number_grid[count1][count2]);
	    }
	    else 
	    if (crossword_grid[count1][count2] != CWORD_UNFILLED_CHAR)
	    {
		strcpy (temp1_str, INNER_BLANK1_STR);
	    }
	    else
	    {
		strcpy (temp1_str, INNER_BLACK_STR);
	    }

	    /* Taken or not taken */

	    if (crossword_grid[count1][count2] != CWORD_UNFILLED_CHAR)
	    {
		sprintf (
		    temp2_str,
		    CHAR1_STR,
		    (char) letter_grid[count1][count2]);
	    }
	    else
	    {
		strcpy (temp2_str, INNER_BLACK_STR);
	    }
	    
	    strcat (verti1_str, temp1_str);
	    strcat (verti2_str, temp2_str);
	}

	strcat (verti1_str, END_STR);
	strcat (verti2_str, END_STR);

	CWORD_PRINT_WRITE_LINE (output_fd, verti1_str);
	CWORD_PRINT_WRITE_LINE (output_fd, verti2_str);
	CWORD_PRINT_WRITE_LINE (output_fd, horiz_str);
    }

    return (CWORD_S_NORMAL);

}



PUBLIC int cword_print_block_txt (
	CROSSWORD_RECORD	*crossword_record)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function which prints the completed verified grid and clues to an
 *	output text file with a few comments thrown in.
 *
 *  FORMAL PARAMETERS:
 *
 *      *crossword_record	(r) Pointer to structure containing verified clues
 *
 *  FUNCTION VALUE:
 *
 *	CWORD_S_NORMAL
 *	CWORD_PSTYPEERRPR
 *      Other
 *--
 */

{

    CROSSWORD_GRID crossword_grid;

    COORDS_STRUCT offset_vals;

    char output_str[CWORD_SIZEOF_MESSAGE + 1],
	 output_file_name[CWORD_SIZEOF_FILESPEC + 1];

    int	output_fd,
	status,
	count1;

    /*
     *	Will assume that the correct type of grid has already been found before
     *	coming in here!
     */

    status = cword_verify_get_block_props (
		crossword_record,
		crossword_grid);

    if (status != CWORD_S_NORMAL)
    {
	return (status);
    }

    sprintf (
	output_file_name,
	"%s%s",
	crossword_record->file_name,
	CWORD_SUFFIX_PRNDAT_FILE);

    status = cword_utils_is_file_ok (
		WRITE_FILE,
		CWORD_SUFFIX_PRNDAT_FILE,
		output_file_name);

    if (CWORD_S_NORMAL != status)
    {
	return (CWORD_S_FAILURE);
    }

    output_fd = open (
	    output_file_name, 
	    O_WRONLY | O_CREAT | O_TRUNC, 
	    0700);
    
    if (output_fd == -1)
    {
	cword_messages_output_message (
		    CWORD_FILEERROR,
		    (TRACE0 | SCREEN0),
		    output_file_name);

	return (CWORD_WRITEERROR);
    }

    /* 
     *  Writing out the comments to the text file.
     */

    if (crossword_record->name[0] == (char) 0)
    {
	strcpy (output_str, "Crossword\n");
    }
    else
    {
	sprintf (
		output_str,
		"Crossword : %s\n",
		crossword_record->name);
    }

    CWORD_PRINT_WRITE_LINE (output_fd, output_str);

    cword_print__cal_txt_offset (
	    crossword_record,
	    &offset_vals);

    status = cword_print__block_txt_grid (
		    NOPREAMBLE,
	    	    output_fd,
		    crossword_record,
	    	    crossword_grid);

    if (status != CWORD_S_NORMAL)
    {
	return (status);
    }

    for (count1 = 0; count1 < offset_vals.y; count1++)
    {
	strcpy (output_str, "\n");
	CWORD_PRINT_WRITE_LINE (output_fd, output_str);
    }

    if (crossword_record->remarks[0] != (char) 0)
    {
	sprintf (
		output_str,
		"Remarks : %s",
		crossword_record->remarks);
    }
    CWORD_PRINT_WRITE_LINE (output_fd, output_str);

    strcpy (output_str, "\f\n");
    CWORD_PRINT_WRITE_LINE (output_fd, output_str);
    
    status = cword_print__common_txt_clues (
	    	    output_fd,
		    crossword_record);

    if (status != CWORD_S_NORMAL)
    {
	return (status);
    }

    if (crossword_record->name[0] == (char) 0)
    {
	strcpy (output_str, "\f\nAnswers\n");
    }
    else
    {
	sprintf (
	    output_str, 
	    "\f\nAnswers : %s\n",
	    crossword_record->name);
    }

    CWORD_PRINT_WRITE_LINE (output_fd, output_str);

    for (count1 = 0; count1 < offset_vals.y; count1++)
    {
	strcpy (output_str, "\n");
	CWORD_PRINT_WRITE_LINE (output_fd, output_str);
    }

    status = cword_print__block_txt_answers (
	    	    output_fd,
		    crossword_record,
		    crossword_grid);

    if (status != CWORD_S_NORMAL)
    {
	return (status);
    }

    close (output_fd);

    cword_messages_output_message (
		CWORD_TXTWRITTEN,
		(TRACE0 | SCREEN0),
		output_file_name);

    return (CWORD_S_NORMAL);

}



PRIVATE int cword_print__barred_txt_grid (
	int			preamble,
	int			output_fd,
	CROSSWORD_RECORD	*crossword_record,
	CROSSWORD_GRID		crossword_grid)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function to write the barred grid without answers to the output file.
 *
 *  FORMAL PARAMETERS:
 *
 *	preamble		(r) Whether to output preamble or not
 *      output_fd		(r) Output file descriptor
 *	*crossword_record	(r) Structure containing the crossword details
 *	crossword_grid		(r) Structure containing the completed crossword grid
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL	    normal completion
 *      write error possibly
 *--
 */

{

    CROSSWORD_GRID clue_number_grid;

    COORDS_STRUCT offset_vals;

    char    output_str[CWORD_SIZEOF_MESSAGE + 1],
	    horiz_str[CWORD_SIZEOF_MESSAGE + 1],
	    verti1_str[CWORD_SIZEOF_MESSAGE + 1],
	    verti2_str[CWORD_SIZEOF_MESSAGE + 1],
	    temp1_str[51],
	    temp2_str[11];

    int	    count1, 
	    count2,
	    count3;

    cword_messages_output_message (
	    CWORD_PSTXTMSG,
	    TRACE2,
	    "barred text grid", 
	    crossword_record->name);

    if (preamble == PREAMBLE)
    {
	if (crossword_record->name[0] == (char) 0)
	{
	    strcpy (output_str, "\n\nGrid\n\n");
	}
	else
	{
	    sprintf (output_str,
		     "\n\nGrid : %s\n\n",
		     crossword_record->name);
	}

	CWORD_PRINT_WRITE_LINE (output_fd, output_str);
    }

    cword_print__cal_txt_offset (
	    crossword_record,
	    &offset_vals);

    for (count1 = 0; count1 < offset_vals.y; count1++)
    {
	strcpy (output_str, "\n");
	CWORD_PRINT_WRITE_LINE (output_fd, output_str);
    }

    /*
     *  Output the grid minus answers to a text file
     */

    for (count1 = 0; count1 < CWORD_SIZEOF_CELLDIM; count1++)
	for (count2 = 0; count2 < CWORD_SIZEOF_CELLDIM; count2++)
	    clue_number_grid[count1][count2] = NULL_CLUE_NUM;

    for (count1 = 0; count1 < crossword_record->clue_sum; count1++)
	clue_number_grid[crossword_record->clue_details[count1].x_coordinate - 1]
			    [crossword_record->clue_details[count1].y_coordinate - 1] =
				crossword_record->clue_details[count1].clue_number;

    for (count2 = crossword_record->y_dimensions - 1; count2 > - 1; count2--)
    {
	/*
	 *  Have to figure out the printing out of the horizontal bar in the grid
	 */

	if ((crossword_record->y_dimensions - 1) == count2)
	{
	    if (offset_vals.x > 0)
	    {
		memset (temp1_str, 0, sizeof (temp1_str));
		for (count3 = 0; count3 < offset_vals.x; count3++)
		    temp1_str[count3] = (char) ' ';

		/* First horizontal line output */
		sprintf (
			horiz_str,
			"%s%s%s",
			temp1_str,
			CORNER_STR,
			DASH4_STR);
            }
	    else
	    {
		/* First horizontal line output */
		sprintf (
			horiz_str,
			"%s%s",
			CORNER_STR,
			DASH4_STR);
	    }

	    for (count1 = 1; count1 < crossword_record->x_dimensions - 1; count1++)
		strcat (horiz_str, DASH4_STR);
	    strcat (horiz_str, HOR_STR);
	}
	else
	{
	    for (count1 = 0; count1 < crossword_record->x_dimensions; count1++)
	    {
		/*
		 * Even more complicated than this with the addition of logic to
		 * figure out what left and right cells are adjacent to the top 
		 */

		/*
		 *  ?? Add in '+' into corner and intersecting parts of the
		 *  grid definition ??
		 *
		 *  ?? This could mean a complete re-write of this function to
		 *  encapsulate this detail ??
		 */

		if (0 != count1)
		{
		    if ( !(TOP & crossword_grid[count1][count2]) && 
                          !( TOP & crossword_grid[count1 - 1][count2] ) )
			strcat (horiz_str, BLACK4_STR);
		    else if ( !(TOP & crossword_grid[count1][count2]) &&
			     (TOP & crossword_grid[count1 - 1][count2]) )
			strcat (horiz_str, BLACK5_STR);
		    else if ( !( TOP & crossword_grid[count1 - 1][count2]) &&
			       (count1 != (crossword_record->x_dimensions - 1)) )
			strcat (horiz_str, DASH3_STR);
		    else if ( !( TOP & crossword_grid[count1 - 1][count2]) &&
			       (count1 == (crossword_record->x_dimensions - 1)) )
			strcat (horiz_str, DASH4_STR);
		    else if ( count1 == (crossword_record->x_dimensions - 1) ) 
			strcat (horiz_str, DASH5_STR);
		    else
			strcat (horiz_str, DASH4_STR);
		}
		else
		{
		    if (offset_vals.x > 0)
		    {
			memset (horiz_str, 0, sizeof (horiz_str));
			for (count3 = 0; count3 < offset_vals.x; count3++)
			    horiz_str[count3] = (char) ' ';

			if ( !(TOP & crossword_grid[count1][count2]) )
			    strcat (horiz_str, BLACK5_STR);
			else
			    strcat (horiz_str, DASH4_STR);
		    }
		    else
		    {
			if ( !(TOP & crossword_grid[count1][count2]) )
			    strcpy (horiz_str, BLACK5_STR);
			else
			    strcpy (horiz_str, DASH4_STR);
		    }
		}
	    }
	}

	for (count1 = 0; count1 < crossword_record->x_dimensions; count1++)
	{
	    /*
	     *  Figure out the clues now  
	     */

	    if (0 == count1)
	    {
		if (offset_vals.x > 0)
		{
		    memset (verti1_str, 0, sizeof (verti1_str));
		    for (count3 = 0; count3 < offset_vals.x; count3++)
			verti1_str[count3] = (char) ' ';

		    memset (verti2_str, 0, sizeof (verti2_str));
		    for (count3 = 0; count3 < offset_vals.x; count3++)
			verti2_str[count3] = (char) ' ';

		    strcat (verti1_str, VER_STR);
		    strcat (verti2_str, VER_STR);
		}
		else
		{
		    strcpy (verti1_str, VER_STR);
		    strcpy (verti2_str, VER_STR);
		}

	    }  

	    /* Output the number of the clue to the file */

	    if (!(RIGHT & crossword_grid[count1][count2]))
	    {
		if (clue_number_grid[count1][count2] != NULL_CLUE_NUM)
		{
		    if ((crossword_record->x_dimensions - 1) != count1)
		    {
			sprintf (
			    temp1_str, 
			    NUM2_STR,
			    clue_number_grid[count1][count2]);

			strcpy (temp2_str, INNER_BLANK2_STR);	    
		    }
		    else
		    {
			sprintf (
			    temp1_str, 
			    NUM1_STR,
			    clue_number_grid[count1][count2]);

			strcpy (temp2_str, INNER_BLANK1_STR);	    
		    }
		}
		else
		{	
		    /*
		     *  The cell isn't a clue number but still could have a letter in it 
		     */
	  
		    if (crossword_grid[count1][count2] != CWORD_UNFILLED_CHAR)
		    {	
			if ((crossword_record->x_dimensions - 1) != count1)
			{
			    strcpy (temp1_str, INNER_BLANK2_STR);
			    strcpy (temp2_str, INNER_BLANK2_STR);
			}
			else
			{
			    strcpy (temp1_str, INNER_BLANK1_STR);
			    strcpy (temp2_str, INNER_BLANK1_STR);
			}
		    }
		    else
		    {
			if ((crossword_record->x_dimensions - 1) != count1)
			{
			    strcpy (temp1_str, INNER_BLANK2_STR);
			    strcpy (temp2_str, INNER_BLANK2_STR);
			}
			else
			{
			    strcpy (temp1_str, INNER_BLANK1_STR);
			    strcpy (temp2_str, INNER_BLANK1_STR);
			}
		    }
		}
	    }
	    else
	    {
		if (clue_number_grid[count1][count2] != NULL_CLUE_NUM)
		{
		    sprintf (
			temp1_str, 
			NUM1_STR,
			clue_number_grid[count1][count2]);

		    strcpy (temp2_str, INNER_BLANK1_STR);
		}
		else
		{	
		    /*
		     *  The cell isn't a clue number but still could have a letter in it 
		     */
	  
		    if (crossword_grid[count1][count2] != CWORD_UNFILLED_CHAR)
		    {
			strcpy (temp1_str, INNER_BLANK1_STR);
			strcpy (temp2_str, INNER_BLANK1_STR);
		    }
		    else
		    {
			strcpy (temp1_str, INNER_BLACK_STR);
			strcpy (temp2_str, INNER_BLACK_STR);
		    }
		}
	    }

	    strcat (verti1_str, temp1_str);
	    strcat (verti2_str, temp2_str);
	}

	strcat (verti1_str, END_STR);
	strcat (verti2_str, END_STR);
	strcat (horiz_str, END_STR);

	CWORD_PRINT_WRITE_LINE (output_fd, horiz_str);
	CWORD_PRINT_WRITE_LINE (output_fd, verti1_str);
	CWORD_PRINT_WRITE_LINE (output_fd, verti2_str);
	verti1_str[0] = verti2_str[0] = (char) 0;
    }

    /* Last line output outside the loop */

    if (offset_vals.x > 0)
    {
	memset (temp1_str, 0, sizeof (temp1_str));
	for (count1 = 0; count1 < offset_vals.x; count1++)
	    temp1_str[count1] = (char) ' ';

	sprintf (
		horiz_str,
		"%s%s%s",
		temp1_str,
		CORNER_STR,
		DASH4_STR);
    }
    else
    {
	sprintf (
		horiz_str,
		"%s%s",
		CORNER_STR,
		DASH4_STR);
    }

    for (count1 = 1; count1 < crossword_record->x_dimensions - 1; count1++)
	strcat (horiz_str, DASH4_STR);
    strcat (horiz_str, HOR_STR);

    strcat (horiz_str, END_STR);

    CWORD_PRINT_WRITE_LINE (output_fd, horiz_str);

    return (CWORD_S_NORMAL);

}



PRIVATE int cword_print__barred_txt_answers (
	int			output_fd,
	CROSSWORD_RECORD	*crossword_record,
	CROSSWORD_GRID		crossword_grid)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *      Prints the barred grid plus answers to the output file
 *
 *	The ** tells the differences between the function barred_grid and
 *	barred_answers.
 *
 *  FORMAL PARAMETERS:
 *
 *      output_fd	    (r) Output file descriptor
 *	*crossword_record   (r) structure containing the crossword details
 *	crossword_grid	    (r) structure containing the completed crossword grid
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL	    normal completion
 *      write error possibly
 *--
 */

{

    CROSSWORD_GRID clue_number_grid,
/**/		   grid_contents;

    COORDS_STRUCT offset_vals;

    char horiz_str[CWORD_SIZEOF_MESSAGE + 1],
	 verti1_str[CWORD_SIZEOF_MESSAGE + 1],
	 verti2_str[CWORD_SIZEOF_MESSAGE + 1],
	 temp1_str[51],
	 temp2_str[11];

    int	count1, 
	count2,
	count3;

    cword_messages_output_message (
	    CWORD_PSTXTMSG,
	    TRACE2,
	    "barred text answers", 
	    crossword_record->name);

    cword_print__cal_txt_offset (
	    crossword_record,
	    &offset_vals);

    for (count1 = 0; count1 < CWORD_SIZEOF_CELLDIM; count1++)
	for (count2 = 0; count2 < CWORD_SIZEOF_CELLDIM; count2++)
	{
	    clue_number_grid[count1][count2] = NULL_CLUE_NUM;
/**/	    grid_contents[count1][count2] = NOT_TAKEN;
	}

    /*
     *	If barred grid print is chosen then all the cells in grid_contents will
     *	be occupied by alphabetic characters.
     */

/**/cword_verify_get_letters_grid (
	    crossword_record,
	    grid_contents);

    for (count1 = 0; count1 < crossword_record->clue_sum; count1++)
	clue_number_grid[crossword_record->clue_details[count1].x_coordinate - 1]
			    [crossword_record->clue_details[count1].y_coordinate - 1] =
				crossword_record->clue_details[count1].clue_number;

    for (count2 = crossword_record->y_dimensions - 1; count2 > -1; count2--)
    {
	if ((crossword_record->y_dimensions - 1) == count2)
	{
	    /*
	     *  Using the X offset to move grid into the middle of page
	     */

	    if (offset_vals.x > 0)
	    {
		memset (temp1_str, 0, sizeof (temp1_str));
		for (count3 = 0; count3 < offset_vals.x; count3++)
		    temp1_str[count3] = (char) ' ';

		/* First horizontal line output */
		sprintf (
		    horiz_str,
		    "%s%s%s",
		    temp1_str,
		    CORNER_STR,
		    DASH4_STR);
	    }
	    else
	    {
		/* First horizontal line output */
		sprintf (
		    horiz_str,
		    "%s%s",
		    CORNER_STR,
		    DASH4_STR);
	    }

	    for (count3 = 1; count3 < crossword_record->x_dimensions - 1; count3++)
		strcat (horiz_str, DASH4_STR);
	    strcat (horiz_str, HOR_STR);
	}
	else
	{
	    for (count1 = 0; count1 < crossword_record->x_dimensions; count1++)
	    {
		/*
		 * Even more complicated than this with the addition of logical to
		 * figure out what left and right cells are adjhacent to the top 
		 */

		if (0 != count1)
		{
		    if ( !(TOP & crossword_grid[count1][count2]) && 
                          !( TOP & crossword_grid[count1 - 1][count2] ) )
			strcat (horiz_str, BLACK4_STR);
		    else if ( !(TOP & crossword_grid[count1][count2]) &&
			     (TOP & crossword_grid[count1 - 1][count2]) )
			strcat (horiz_str, BLACK5_STR);
		    else if ( !( TOP & crossword_grid[count1 - 1][count2]) &&
			       (count1 != (crossword_record->x_dimensions - 1)) )
			strcat (horiz_str, DASH3_STR);
		    else if ( !( TOP & crossword_grid[count1 - 1][count2]) &&
			       (count1 == (crossword_record->x_dimensions - 1)) )
			strcat (horiz_str, DASH4_STR);
		    else if ( count1 == (crossword_record->x_dimensions - 1) ) 
			strcat (horiz_str, DASH5_STR);
		    else
			strcat (horiz_str, DASH4_STR);
		}
		else
		{
		    if (offset_vals.x > 0)
		    {
			memset (horiz_str, 0, sizeof (horiz_str));
			for (count3 = 0; count3 < offset_vals.x; count3++)
			    horiz_str[count3] = (char) ' ';

			if (!(TOP & crossword_grid[count1][count2]))
			    strcat (horiz_str, BLACK5_STR);
			else
			    strcat (horiz_str, DASH4_STR);
		    }
		    else
		    {
			if (!(TOP & crossword_grid[count1][count2]))
			    strcpy (horiz_str, BLACK5_STR);
			else
			    strcpy (horiz_str, DASH4_STR);
		    }
		}
	    }
	}

	for (count1 = 0; count1 < crossword_record->x_dimensions; count1++)
	{
	    /*
	     *  Figure out the clues now  
	     */

	    if (0 == count1)
	    {
		if (offset_vals.x > 0)
		{
		    memset (verti1_str, 0, sizeof (verti1_str));
		    for (count3 = 0; count3 < offset_vals.x; count3++)
			verti1_str[count3] = (char) ' ';

		    memset (verti2_str, 0, sizeof (verti2_str));
		    for (count3 = 0; count3 < offset_vals.x; count3++)
			verti2_str[count3] = (char) ' ';

		    strcat (verti1_str, VER_STR);
		    strcat (verti2_str, VER_STR);
		}
		else
		{
		    strcpy (verti1_str, VER_STR);
		    strcpy (verti2_str, VER_STR);
		}

	    }  

	    /* Output the number of the clue to the file */

	    if (!(RIGHT & crossword_grid[count1][count2]))
	    {
		if (clue_number_grid[count1][count2] != NULL_CLUE_NUM)
		{
		    if ((crossword_record->x_dimensions - 1) != count1)
		    {
			sprintf (
			    temp1_str, 
			    NUM2_STR,
			    clue_number_grid[count1][count2]);

/**/			sprintf (
			    temp2_str,
			    CHAR2_STR,
			    (char) grid_contents[count1][count2]);
		    }
		    else
		    {
			sprintf (
			    temp1_str, 
			    NUM1_STR,
			    clue_number_grid[count1][count2]);

/**/		        sprintf (
			    temp2_str,
			    CHAR1_STR,
			    (char) grid_contents[count1][count2]);
		    }
		}
		else
		{	
		    /*
		     *	The cell isn't a clue number but still could have a
		     *	letter in it.
		     */
	  
		    if (crossword_grid[count1][count2] != CWORD_UNFILLED_CHAR)
		    {	
			if ((crossword_record->x_dimensions - 1) != count1)
			{
			    strcpy (temp1_str, INNER_BLANK2_STR);
/**/			    sprintf (
				temp2_str,
				CHAR2_STR,
				(char) grid_contents[count1][count2]);
			}
			else
			{
			    strcpy (temp1_str, INNER_BLANK1_STR);
/**/			    sprintf (
				temp2_str,
				CHAR1_STR,
				(char) grid_contents[count1][count2]);
			}
		    }
		    else
		    {
			if ((crossword_record->x_dimensions - 1) != count1)
			{
			    strcpy (temp1_str, INNER_BLANK2_STR);
/**/			    sprintf (
				temp2_str,
				CHAR2_STR,
				(char) grid_contents[count1][count2]);
			}
			else
			{
			    strcpy (temp1_str, INNER_BLANK1_STR);
/**/			    sprintf (
				temp2_str,
				CHAR1_STR,
				(char) grid_contents[count1][count2]);
			}
		    }
		}
	    }
	    else
	    {
		if (clue_number_grid[count1][count2] != NULL_CLUE_NUM)
		{
		    sprintf (
			temp1_str, 
			NUM1_STR,
			clue_number_grid[count1][count2]);

/**/		    sprintf (
			temp2_str,
			CHAR1_STR,
			(char) grid_contents[count1][count2]);
		}
		else
		{	
		    /*
		     *	The cell isn't a clue number but still could have a
		     *	letter in it.
		     */
	  
		    if (crossword_grid[count1][count2] != CWORD_UNFILLED_CHAR)
		    {
			strcpy (temp1_str, INNER_BLANK1_STR);
/**/			sprintf (
			    temp2_str,
			    CHAR1_STR,
			    (char) grid_contents[count1][count2]);
		    }
		    else
		    {
			strcpy (temp1_str, "???|");
			strcpy (temp2_str, "???|");
		    }
		}
	    }

	    strcat (verti1_str, temp1_str);
	    strcat (verti2_str, temp2_str);
	}

	strcat (verti1_str, END_STR);
	strcat (verti2_str, END_STR);
	strcat (horiz_str, END_STR);

	CWORD_PRINT_WRITE_LINE (output_fd, horiz_str);
	CWORD_PRINT_WRITE_LINE (output_fd, verti1_str);
	CWORD_PRINT_WRITE_LINE (output_fd, verti2_str);
	verti1_str[0] = verti2_str[0] = (char) 0;
    }

    /* Last line output outside the loop */

    if (offset_vals.x > 0)
    {
	memset (temp1_str, 0, sizeof (temp1_str));
	for (count3 = 0; count3 < offset_vals.x; count3++)
	    temp1_str[count3] = (char) ' ';

	sprintf (
		horiz_str,
		"%s%s%s",
		temp1_str,
		CORNER_STR,
		DASH4_STR);
    }
    else
    {
	sprintf (
		horiz_str,
		"%s%s",
		CORNER_STR,
		DASH4_STR);
    }

    for (count3 = 1; count3 < crossword_record->x_dimensions - 1; count3++)
	strcat (horiz_str, DASH4_STR);
    strcat (horiz_str, HOR_STR);
    strcat (horiz_str, END_STR);

    CWORD_PRINT_WRITE_LINE (output_fd, horiz_str);

    return (CWORD_S_NORMAL);

}



PUBLIC int cword_print_barred_txt (
	CROSSWORD_RECORD	*crossword_record)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *      Function to print out a barred grid to an output text file.
 *
 *  FORMAL PARAMETERS:
 *
 *	*crossword_record	(r) structure containing the crossword details
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL		Normal completion
 *	CWORD_WRITEERROR	Error writing to output file
 *--
 */

{
    
    /*
     *	Possible to use the same machanism to print out the clues then used in
     *	the blocked option but not the same grid printing facility 
     */

    CROSSWORD_GRID crossword_grid;

    COORDS_STRUCT offset_vals;

    char output_str[CWORD_SIZEOF_MESSAGE + 1],
	 output_file_name[CWORD_SIZEOF_FILESPEC + 1];

    int	output_fd,
        status,
	count1;

    /*
     *	Will assume that the correct type of grid has already been found before
     *	coming in here!
     */

    status = cword_verify_get_barred_props (
		crossword_record,
		crossword_grid);

    if (status != CWORD_S_NORMAL)
    {
	return (status);
    }

    sprintf (
	output_file_name,
	"%s%s",
	crossword_record->file_name,
	CWORD_SUFFIX_PRNDAT_FILE);

    status = cword_utils_is_file_ok (
		WRITE_FILE,
		CWORD_SUFFIX_PRNDAT_FILE,
		output_file_name);

    if (CWORD_S_NORMAL != status)
    {
	return (CWORD_S_FAILURE);
    }

    output_fd = open (
	    output_file_name, 
	    O_WRONLY | O_CREAT | O_TRUNC, 
	    0700);
    
    if (output_fd == -1)
    {
	cword_messages_output_message (
		    CWORD_FILEERROR,
		    (TRACE0 | SCREEN0),
		    output_file_name);

	return (CWORD_WRITEERROR);
    }

    /* 
     * Writing out the comments to the text file.
     */

    if (crossword_record->name[0] == (char) 0)
    {
	strcpy (output_str, "Crossword\n");
    }
    else
    {
	sprintf (output_str,
		 "Crossword : %s\n",
		 crossword_record->name);
    }

    CWORD_PRINT_WRITE_LINE (output_fd, output_str);

    cword_print__cal_txt_offset (
	crossword_record,
	&offset_vals);

    status = cword_print__barred_txt_grid (
		NOPREAMBLE,
		output_fd,
		crossword_record,
		crossword_grid);

    if (status != CWORD_S_NORMAL)
    {
	return (status);
    }

    for (count1 = 0; count1 < offset_vals.y; count1++)
    {
	strcpy (output_str, "\n");
	CWORD_PRINT_WRITE_LINE (output_fd, output_str);
    }

    if (crossword_record->remarks[0] != (char) 0)
    {
	sprintf (output_str,
		 "Remarks : %s",
		 crossword_record->remarks);
    }
    CWORD_PRINT_WRITE_LINE (output_fd, output_str);

    strcpy (output_str, "\f\n");
    CWORD_PRINT_WRITE_LINE (output_fd, output_str);
    
    status = cword_print__common_txt_clues (
	    	    output_fd,
		    crossword_record);

    if (status != CWORD_S_NORMAL)
    {
	return (status);
    }

    if (crossword_record->name[0] == (char) 0)
    {
	strcpy (output_str, "\f\nAnswers\n");
    }
    else
    {
	sprintf (
	    output_str, 
	    "\f\nAnswers : %s\n",
	    crossword_record->name);
    }

    CWORD_PRINT_WRITE_LINE (output_fd, output_str);

    for (count1 = 0; count1 < offset_vals.y; count1++)
    {
	strcpy (output_str, "\n");
	CWORD_PRINT_WRITE_LINE (output_fd, output_str);
    }

    status = cword_print__barred_txt_answers (
	    	    output_fd,
		    crossword_record,
		    crossword_grid);

    if (status != CWORD_S_NORMAL)
    {
	return (status);
    }

    close (output_fd);

    cword_messages_output_message (
		CWORD_TXTWRITTEN,
		(TRACE0 | SCREEN0),
		output_file_name);

    return (CWORD_S_NORMAL);

}



PRIVATE int cword_print__barred_ps_grid (
	int			preamble,
	int			output_fd,
	CROSSWORD_RECORD	*crossword_record,
	CROSSWORD_GRID		crossword_grid)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *      Prints a barred grid to the output file descriptor.
 *
 *  FORMAL PARAMETERS:
 *
 *	preamble		(r) Whether to print PS preamble before hand
 *	output_fd		(r) Output file descriptor
 *	*crossword_record	(r) Pointer to CROSSWORD_RECORD structure
 *	crossword_grid		(r) Array of CROSSWORD_RECORD
 *
 *  IMPLICIT INPUTS:
 *
 *      cword_g_defaults_record.x_dimensions
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL		Normal completion
 *	CWORD_S_FAILURE		Error in completion
 *--
 */

{

    char output_str[CWORD_SIZEOF_MESSAGE + 1];

    COORDS_STRUCT offset_vals;

    int	count1,
	count2;

    cword_messages_output_message (
	    CWORD_PSTXTMSG,
	    TRACE2,
	    "barred postscript grid", 
	    crossword_record->name);

    if (preamble == PREAMBLE)
    {
	cword_print__preamble (output_fd);

	/* Writing out the name of the crossword to the top of the page */

	strcpy (output_str, "%% Setting up the crossword name 18 points wide\n");
	CWORD_PRINT_WRITE_LINE (output_fd, output_str);

	if (crossword_record->name[0] == (char) 0)
	{
	    sprintf (
		    output_str, 
		    "/%s findfont 18 scalefont setfont\n\t%3d %3d moveto (Grid) show\n\n",
		    HEADER_FONT,
		    CWORD_C_HEADER_X,
		    CWORD_C_HEADER_Y);
	}
	else
	{
	    cword_print__convert_ps_line (crossword_record->name);

	    sprintf (
		    output_str, 
		    "/%s findfont 18 scalefont setfont\n\t%3d %3d moveto (Grid : %s) show\n\n",
		    HEADER_FONT,
		    CWORD_C_HEADER_X,
		    CWORD_C_HEADER_Y,
		    crossword_record->name);
	}

	CWORD_PRINT_WRITE_LINE (output_fd, output_str);
    }

    /* Calculate the offsets to make the postscript look better when printed */

    cword_print__cal_ps_offset (
		    GRID_PRINT,
		    crossword_record,
		    &offset_vals);

    /* Writing out the heavy outside line of the crossword grid */

    sprintf (
	output_str, 
	"\n%% Setting up the overall box shape %-d points wide\n",
	CWORD_C_STROKE_POINT_SIZE);
    CWORD_PRINT_WRITE_LINE (output_fd, output_str);

    sprintf (
	    output_str, 
	    "newpath\n %3d %3d moveto\n 0 %3d rlineto\n %3d 0 rlineto\n 0 -%.3d rlineto\n closepath\n",
	    offset_vals.x,
	    offset_vals.y,
	    (offset_vals.cell_point_size * crossword_record->y_dimensions),
	    (offset_vals.cell_point_size * crossword_record->x_dimensions) + 1,
	    (offset_vals.cell_point_size * crossword_record->y_dimensions) + 1);
    CWORD_PRINT_WRITE_LINE (output_fd, output_str);

    sprintf (
	output_str, 
	" %-d setlinewidth\n stroke\n\n",
	CWORD_C_STROKE_POINT_SIZE);
    CWORD_PRINT_WRITE_LINE (output_fd, output_str);

    /* Writing out the separating lines in the crossword box */

    strcpy (output_str, "%% Setting up the moving X coordinates\n"); 
    CWORD_PRINT_WRITE_LINE (output_fd, output_str);

    for (count1 = offset_vals.x + offset_vals.cell_point_size; 
	  count1 < (offset_vals.x + (offset_vals.cell_point_size * crossword_record->x_dimensions));
	   count1 += offset_vals.cell_point_size)
    {
	sprintf (
		output_str, 
		"\t%3d %3d %3d %3d DrawBoxLines\n",
		count1,
		offset_vals.y,
		count1,
		offset_vals.y + (offset_vals.cell_point_size * crossword_record->y_dimensions));
	CWORD_PRINT_WRITE_LINE (output_fd, output_str);
    }

    strcpy (output_str, "\n%% Setting up the moving Y coordinates\n");
    CWORD_PRINT_WRITE_LINE (output_fd, output_str);

    for (count1 = offset_vals.y + offset_vals.cell_point_size; 
          count1 < offset_vals.y + (offset_vals.cell_point_size * crossword_record->y_dimensions);
	   count1 += offset_vals.cell_point_size)
    {
	/*
	 *  Add two to the X co-ordinate print 
	 */

	sprintf (
		output_str,
		"\t%3d %3d %3d %3d DrawBoxLines\n",
		offset_vals.x,
		count1,
		offset_vals.x + (offset_vals.cell_point_size * crossword_record->x_dimensions),
		count1);
	CWORD_PRINT_WRITE_LINE (output_fd, output_str);
    }

    /* Setting up the barred boxes in the grid */

    strcpy (output_str, "\n%% Setting up the barred boxes in the grid\n");
    CWORD_PRINT_WRITE_LINE (output_fd, output_str);

    for (count1 = 0; count1 < crossword_record->x_dimensions; count1++)
    {
	for (count2 = 0; count2 < crossword_record->y_dimensions; count2++)
	{
	    if ( !(crossword_grid[count1][count2] & LEFT) && (count1 != 0) )
	    {
		sprintf (
		    output_str, 
		    "\t%3d %3d %2d %2d BlackBox2\n",
		    offset_vals.x + (count1 * offset_vals.cell_point_size) - 
			CWORD_C_BARRED_POINT_WIDTH_SIZE,
		    offset_vals.y + (count2 * offset_vals.cell_point_size),
		    offset_vals.cell_point_size,
		    CWORD_C_BARRED_POINT_WIDTH_SIZE);
		CWORD_PRINT_WRITE_LINE (output_fd, output_str);
	    }

#if 0
	    if ( !(crossword_grid[count1][count2] & RIGHT) && 
	          (count1 != crossword_record->x_dimensions - 1) )
	    {
		sprintf (
		    output_str, 
		    "\t%3d %3d %2d %2d BlackBox2\n",
		    offset_vals.x + ((count1 + 1) * offset_vals.cell_point_size) - CWORD_C_BARRED_POINT_WIDTH_SIZE,
		    offset_vals.y + (count2 * offset_vals.cell_point_size),
		    offset_vals.cell_point_size,
		    CWORD_C_BARRED_POINT_WIDTH_SIZE);
		CWORD_PRINT_WRITE_LINE (output_fd, output_str);
	    }
#endif

	    if ( !(crossword_grid[count1][count2] & TOP) && 
                  (count2 != crossword_record->y_dimensions - 1) )
	    {
		/*
		 *  If this cell also possesses a LEFT bit mask then the output
		 *  bar is slightly longer to conform with barred grid
		 *  outputting conventions.
		 */

		if ( !(crossword_grid[count1][count2] & LEFT) && (count1 != 0) )
		{
		    sprintf (
			output_str, 
			"\t%3d %3d %2d %2d BlackBox2\n",
			offset_vals.x + (count1 * offset_vals.cell_point_size) - 
			    CWORD_C_BARRED_POINT_WIDTH_SIZE,
			offset_vals.y + ((count2 + 1) * offset_vals.cell_point_size),
			CWORD_C_BARRED_POINT_WIDTH_SIZE,
			offset_vals.cell_point_size + CWORD_C_BARRED_POINT_WIDTH_SIZE);
		    CWORD_PRINT_WRITE_LINE (output_fd, output_str);
		}
		else
		{
		    sprintf (
			output_str, 
			"\t%3d %3d %2d %2d BlackBox2\n",
			offset_vals.x + (count1 * offset_vals.cell_point_size),
			offset_vals.y + ((count2 + 1) * offset_vals.cell_point_size),
			CWORD_C_BARRED_POINT_WIDTH_SIZE,
			offset_vals.cell_point_size);
		    CWORD_PRINT_WRITE_LINE (output_fd, output_str);
		}
	    }

#if 0
	    if ( !(crossword_grid[count1][count2] & BOTTOM) && (count2 != 0) )
	    {
		sprintf (
			output_str, 
			"\t%3d %3d %2d %2d BlackBox2\n",
			offset_vals.x + (count1 * offset_vals.cell_point_size),
			offset_vals.y + (count2 * offset_vals.cell_point_size),
			CWORD_C_BARRED_POINT_WIDTH_SIZE,
			offset_vals.cell_point_size);
		CWORD_PRINT_WRITE_LINE (output_fd, output_str);
	    }
#endif
	}
    }

    /* Setting up the clue numbers for the clues */

    strcpy (output_str, "\n%% Setting up the clue numbers in the grid\n");
    CWORD_PRINT_WRITE_LINE (output_fd, output_str);

    sprintf (
	    output_str,
	    "/%s findfont %2d scalefont setfont\n",
	    TEXT_FONT,
	    offset_vals.font_num_size);
    CWORD_PRINT_WRITE_LINE (output_fd, output_str);

    for (count1 = 0; count1 < crossword_record->clue_sum; count1++)
    {
	if (crossword_record->clue_details[count1].y_coordinate != crossword_record->y_dimensions)
	{
	    sprintf (
		    output_str, 
		    "\t%3d %3d (%d) ClueIndex\n",
		    offset_vals.x + offset_vals.cell_numoff_x + 
		     ((crossword_record->clue_details[count1].x_coordinate - 1) * offset_vals.cell_point_size),
		    offset_vals.y + offset_vals.cell_numoff_y + 
		     ((crossword_record->clue_details[count1].y_coordinate - 1) * offset_vals.cell_point_size),
		    crossword_record->clue_details[count1].clue_number);
	}
	else
	{
	    /*
	     *	Allow the printing of the clue numbers on the top row of the crossword
	     *	to be more flush with the outer boundary than those located anywhere
	     *	else!
	     */

	    sprintf (
		    output_str, 
		    "\t%3d %3d (%d) ClueIndex\n",
		    offset_vals.x + offset_vals.cell_numoff_x + 
		     ((crossword_record->clue_details[count1].x_coordinate - 1) * offset_vals.cell_point_size),
		    offset_vals.y + offset_vals.cell_numoff_y + 
		     ((crossword_record->clue_details[count1].y_coordinate - 1) * offset_vals.cell_point_size) - 1,
		    crossword_record->clue_details[count1].clue_number);
	}

	CWORD_PRINT_WRITE_LINE (output_fd, output_str);
    }

    if (preamble == PREAMBLE)
    {
	cword_print__ps_new_page (output_fd);
    }

    return (CWORD_S_NORMAL);

}



PRIVATE int cword_print__barred_ps_answers (
	int			output_fd,
	CROSSWORD_RECORD	*crossword_record,
	CROSSWORD_GRID		crossword_grid)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *      Function to print out the answers to a postscript output file
 *
 *  FORMAL PARAMETERS:
 *
 *	output_fd		    (r) Output file descriptor
 *	*crossword_record	    (r)	Pointer to CROSSWORD_RECORD structure
 *	crossword_grid		    (r)	Array of CROSSWORD_RECORD
 *
 *  IMPLICIT INPUTS:
 *
 *      cword_g_defaults_record.x_dimensions
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL		Normal completion
 *	other error status
 *--
 */

{

    char output_str[CWORD_SIZEOF_MESSAGE + 1];

    COORDS_STRUCT offset_vals;

    int	status,
	count1,
	count2;

    cword_messages_output_message (
	    CWORD_PSTXTMSG,
	    TRACE2,
	    "barred postscript answers", 
	    crossword_record->name);

    status = cword_print__barred_ps_grid (
		    NOPREAMBLE,
		    output_fd,
		    crossword_record,
		    crossword_grid);

    if (CWORD_S_NORMAL != status)
    {
	return (status);
    }

    /* Calculate the offsets to make the postscript look better when printed */

    cword_print__cal_ps_offset (
		    GRID_PRINT,
		    crossword_record,
		    &offset_vals);

    /* Setting up the header to the answers */

    strcpy (output_str, "\n%% Setting up the clue answers header\n");
    CWORD_PRINT_WRITE_LINE (output_fd, output_str);

    if (crossword_record->name[0] == (char) 0)
    {
	sprintf (
		output_str, 
		"/%s findfont 18 scalefont setfont\n\t%3d %3d moveto (Answers) show\n",
		HEADER_FONT,
		CWORD_C_HEADER_X,
		CWORD_C_HEADER_Y);
    }
    else
    {
	sprintf (
		output_str, 
		"/%s findfont 18 scalefont setfont\n\t%3d %3d moveto (Answers : %s) show\n",
		HEADER_FONT,
		CWORD_C_HEADER_X,
		CWORD_C_HEADER_Y,
		crossword_record->name);
    }

    CWORD_PRINT_WRITE_LINE (output_fd, output_str);
    
    /* Setting up the clue answers */

    strcpy (output_str, "\n%% Setting up the clue answers\n");
    CWORD_PRINT_WRITE_LINE (output_fd, output_str);

    sprintf (
	    output_str, 
	    "/%s findfont %2d scalefont setfont\n",
	    HEADER_FONT,
	    offset_vals.font_letter_size);
    CWORD_PRINT_WRITE_LINE (output_fd, output_str);

    for (count1 = 0; count1 < crossword_record->clue_sum; count1++)
    {
	if (crossword_record->clue_details[count1].clue_direction == DOWN)
	{
	    for (count2 = 0; count2 < strlen (crossword_record->clue_details[count1].clue_answer); count2++)
	    {
		sprintf (
		    output_str, 
		    "\t%3d %3d (%c) ClueAnswer\n",
		    offset_vals.x + offset_vals.cell_charoff_x + 
		     (crossword_record->clue_details[count1].x_coordinate - 1) * offset_vals.cell_point_size,
		    offset_vals.y + offset_vals.cell_charoff_y + 
		     (crossword_record->clue_details[count1].y_coordinate - 1) * offset_vals.cell_point_size - 
		      (count2 * offset_vals.cell_point_size ),
		    crossword_record->clue_details[count1].clue_answer[count2]);
		CWORD_PRINT_WRITE_LINE (output_fd, output_str);
	    }
        }
	else
	{
	    for (count2 = 0; count2 < strlen (crossword_record->clue_details[count1].clue_answer); count2++)
	    {
		sprintf (
		    output_str, 
		    "\t%3d %3d (%c) ClueAnswer\n",
		    offset_vals.x + offset_vals.cell_charoff_x +
		     (crossword_record->clue_details[count1].x_coordinate - 1) * offset_vals.cell_point_size+ 
		      (count2 * offset_vals.cell_point_size),
		    offset_vals.y + offset_vals.cell_charoff_y +
		     (crossword_record->clue_details[count1].y_coordinate - 1) * offset_vals.cell_point_size,
		    crossword_record->clue_details[count1].clue_answer[count2]);
		CWORD_PRINT_WRITE_LINE (output_fd, output_str);
	    }
        }	
    }

    return (CWORD_S_NORMAL);

}



PUBLIC int cword_print_barred_ps (
	CROSSWORD_RECORD	*crossword_record)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function to print out a fully barred grid in PS using a series of
 *	developed macros.
 *
 *  FORMAL PARAMETERS:
 *
 *      *crossword_record	(r)	pointer to an instance of a CROSSWORD_RECORD
 *
 *  IMPLICIT INPUTS:
 *
 *	cword_g_defaults_record.username
 *	cword_print_l_page_number
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL		Successfully written to file
 *	CWORD_S_FAILURE		Error writing crossword to file
 *--
 */

{

    /*
     *	?? Implemented the macro style of printing out in this function.  No
     *	other functions have yet done this ??
     */

    CROSSWORD_GRID crossword_grid;

    char output_file_name[CWORD_SIZEOF_FILESPEC + 1],
	 output_str[CWORD_SIZEOF_MESSAGE + 1];

    int	output_fd,
	status;

    /*
     *	Will assume that the correct type of grid has already been found before
     *	coming in here!
     */

    /*
     *  Check that the grid is completely barred off
     */

    status = cword_verify_get_barred_props (
		crossword_record,
		crossword_grid);

    if (CWORD_S_NORMAL != status)
    {
	return (status);
    }

    sprintf (
	output_file_name,
	"%s%s",
	crossword_record->file_name,
	CWORD_SUFFIX_PRNPS_FILE);

    status = cword_utils_is_file_ok (
		WRITE_FILE,
		CWORD_SUFFIX_PRNPS_FILE,
		output_file_name);

    if (CWORD_S_NORMAL != status)
    {
	return (CWORD_S_FAILURE);
    }

    output_fd = open (
		    output_file_name, 
		    O_WRONLY | O_CREAT | O_TRUNC, 
		    0700);
    
    if (output_fd == -1)
    {
	cword_messages_output_message (
	    CWORD_FILEERROR,
	    (TRACE0 | SCREEN0),
	    output_file_name);

	return (CWORD_WRITEERROR);
    }

    /*
     *  Print out PS comments before continuing
     */

    cword_print__header_ps_comments (
	BARRED_CROSSWORD_FILE,
	crossword_record->name,
	output_fd);

    status = cword_print__preamble (output_fd);

    if (CWORD_S_NORMAL != status)
    {
	return (status);
    }

    /* Writing out the name of the crossword to the top of the page */

    strcpy (
	output_str,
	"%%%% Setting up the crossword name 18 points wide\n\n");
    CWORD_PRINT_WRITE_LINE (output_fd, output_str);

    if (crossword_record->name[0] == (char) 0)
	sprintf (
	    output_str, 
	    "/%s findfont 18 scalefont setfont\n\t%3d %3d moveto (Crossword) show\n\n",
	    HEADER_FONT,
	    CWORD_C_HEADER_X,
	    CWORD_C_HEADER_Y);
    else
    {
	/*
	 *  Only called once as this will do for clues and answers too!
	 */

	cword_print__convert_ps_line (crossword_record->name);

	sprintf (
	    output_str, 
	    "/%s findfont 18 scalefont setfont\n\t%3d %3d moveto (Crossword : %s) show\n\n",
	    HEADER_FONT,
	    CWORD_C_HEADER_X,
	    CWORD_C_HEADER_Y,
	    crossword_record->name);
    }

    CWORD_PRINT_WRITE_LINE (output_fd, output_str);

    /* Setting up the remarks for the crossword */

    if (crossword_record->remarks[0] != (char) 0)
    {
	strcpy (
	    output_str,
	    "%%%% Setting up the crossword remark\n");
	CWORD_PRINT_WRITE_LINE (output_fd, output_str);

	cword_print__convert_ps_line (crossword_record->remarks);

	sprintf (
	    output_str, 
	    "/%s findfont 14 scalefont setfont\n\t%3d %3d moveto (Remarks : %s) show\n",
	    HEADER_FONT,
	    CWORD_C_FOOTER_X,
	    CWORD_C_FOOTER_Y,
	    crossword_record->remarks);
        CWORD_PRINT_WRITE_LINE (output_fd, output_str);
    }

    status = cword_print__barred_ps_grid (
		NOPREAMBLE,
		output_fd,
		crossword_record,
		crossword_grid);

    if (CWORD_S_NORMAL != status)
    {
	return (status);
    }

    status = cword_print__common_ps_clues (
		output_fd,
		crossword_record);

    if (CWORD_S_NORMAL != status)
    {
	return (status);
    }

    cword_print__ps_new_page (output_fd);

    status = cword_print__barred_ps_answers (
		output_fd,
		crossword_record,
		crossword_grid);

    if (CWORD_S_NORMAL != status)
    {
	return (status);
    }

    cword_print__ps_new_page (output_fd);

    strcpy (
	output_str,
	"%%%% Trailer\n");
    CWORD_PRINT_WRITE_LINE (output_fd, output_str);

    close (output_fd);

    cword_messages_output_message (
	CWORD_PSWRITTEN,
	(TRACE0 | SCREEN0),
	output_file_name);

    return (CWORD_S_NORMAL);

}



PUBLIC int cword_print_crossword (
	int			print_option,
	char			*crossword_file)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Generic print function which will accepy a filespec containing a
 *	crossword definition and then print it out according to the
 *	print_option (POSTSCRIPT/TEXT) inputted through print_option.
 *
 *  FORMAL PARAMETERS:
 *
 *      print_option		(r) Output option POSTSCRIPT/TEXT to print crossword file as.
 *	crossword_file		(r) file name containing crossword definition
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL		Normal completion
 *	CWORD_S_FAILURE		Valid print option not selected
 *	CWORD_WRITEERROR	Output file error
 *--
 */

{

    CROSSWORD_RECORD *crossword_record;

    int status;

    cword_print_l_page_number = 0;
    
    if ( !((print_option == POSTSCRIPT) || (print_option == TEXT)) ) 
    {
	cword_messages_output_message (
	    CWORD_PRINTOPTERR,
	    (TRACE2 | SCREEN0));

	return (CWORD_S_FAILURE);
    }

    cword_messages_output_message (
		CWORD_XRDOUTFILE,
		(TRACE0 | SCREEN0),
		crossword_file);    

    crossword_record = cword_memman_make_crossword_rec ();

    status = cword_fileio_read_crossword (
			crossword_file,
			crossword_record);

    if (CWORD_S_NORMAL != status)
    {
	cword_memman_free_cross_rec (crossword_record);

	return (status);
    }

    /*
     *  Find the grid type - BLOCK/BARRED for printing
     */

    /*
     *  ?? Should already be present but... ??
     */

    if (crossword_record->grid_type == NOTDEFINED_GRID)
    {
	/*
	 *  Get grid_type into crossword_record
	 */

	crossword_record->grid_type = cword_verify_get_grid_type (crossword_record);
    }

    if (status != CWORD_S_NORMAL) 
    {
	cword_memman_free_cross_rec (crossword_record);

	return (status);
    }

    /*
     *  Value for grid_type possibly found 
     */

    if (crossword_record->grid_type == INVALID_GRID)
    {
	cword_messages_output_message (
		    CWORD_GRDDEFNOTFOUND,
		    (TRACE2 | SCREEN0));

	cword_memman_free_cross_rec (crossword_record);

	return (CWORD_S_FAILURE);
    }

    switch (print_option)
    {
    case POSTSCRIPT:

	switch (crossword_record->grid_type)
	{
	case BLOCK_GRID:
	case AMERICAN_GRID:
	case NOVELTY_GRID:
	    cword_messages_output_message (
			CWORD_SELECTPSTYPE,
			(TRACE3 | SCREEN0),
			CWORD_BLOCK_GRID_STR);

	    status = cword_print_block_ps (crossword_record);

	    break;

	case BARRED_GRID:
	    cword_messages_output_message (
			CWORD_SELECTPSTYPE,
			(TRACE3 | SCREEN0),
			CWORD_BARRED_GRID_STR);

	    status = cword_print_barred_ps (crossword_record);

	    break;

	default:
	    printf ("# Error no grid type selected\n");

	    return (CWORD_S_FAILURE);  
	}

	break;

    case TEXT:

	switch (crossword_record->grid_type)
	{
	case BLOCK_GRID:
	case AMERICAN_GRID:
	case NOVELTY_GRID:
	    cword_messages_output_message (
			CWORD_SELECTPSTYPE,
			(TRACE3 | SCREEN0),
			CWORD_BLOCK_GRID_STR);

	    cword_print__check_size (
			crossword_record->x_dimensions,
			crossword_record->y_dimensions);

	    status = cword_print_block_txt (crossword_record);

	    break;

	case BARRED_GRID:
	    cword_messages_output_message (
			CWORD_SELECTPSTYPE,
			(TRACE3 | SCREEN0),
			CWORD_BARRED_GRID_STR);

	    cword_print__check_size (
			crossword_record->x_dimensions,
			crossword_record->y_dimensions);

	    status = cword_print_barred_txt (crossword_record);

	    break;

	default:
	    printf ("# Error no grid type selected\n");

	    return (CWORD_S_FAILURE);  
	}

	break;

    default:
    	cword_messages_output_message (
		    CWORD_INVPRINTOPT,
		    (TRACE2 | SCREEN0));

	status = CWORD_S_FAILURE;

	break;
    }

    cword_memman_free_cross_rec (crossword_record);

    return (status);

}



PUBLIC int cword_print_grid (
	int			print_option,
	char			*grid_file)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	This function will print to either a postscript or a text file, the
 *	grid defined in the file specification defined in grid_file.  The
 *	variable print_option is used to decide how to print the file:
 *	POSTSCRIPT or TEXT.
 *
 *  FORMAL PARAMETERS:
 *
 *      print_option		(r) Output option POSTSCRIPT/TEXT to print grid file as.
 *	grid_file		(r) file spec containing grid definition
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL		Normal completion
 *	CWORD_S_FAILURE		Error in completion
 *--
 */

{

    CROSSWORD_RECORD *crossword_record;

    CROSSWORD_GRID crossword_grid;

    char output_file_name[CWORD_SIZEOF_FILESPEC + 1];

    int	status,
	output_fd;

    cword_print_l_page_number = 0;

    if ( !((print_option == POSTSCRIPT) || (print_option == TEXT)) ) 
    {
	cword_messages_output_message (
	    CWORD_GRIDOPTERR,
	    (TRACE2 | SCREEN0));

	return (CWORD_S_FAILURE);
    }

    cword_messages_output_message (
	CWORD_GRDOUTFILE,
	(TRACE0 | SCREEN0),
	grid_file);    

    crossword_record = cword_memman_make_crossword_rec ();

    status = cword_fileio_read_grid (
		grid_file,
		crossword_record);

    if (CWORD_S_NORMAL != status)
    {
	cword_memman_free_cross_rec (crossword_record);

	return (status);
    }

    /*
     *  Tidy up grid definition before printing.
     */

    if (crossword_record->is_completely_specified == FALSE)
    {
	cword_stats_gen_clue_nums (crossword_record);

	cword_utils_complete_definition (crossword_record);
    }

    if (crossword_record->grid_type == NOTDEFINED_GRID)
    {
	crossword_record->grid_type = cword_verify_get_grid_type (crossword_record);
    }

    /*
     *  Value for grid_type found
     */

    if (crossword_record->grid_type != INVALID_GRID)
    {
	/*
	 *  Valid grid type found so now get grid properties.
	 */

	if (crossword_record->grid_type == BARRED_GRID)
	{
	    status = cword_verify_get_barred_props (
			crossword_record,
			crossword_grid);
	}
	else
	{
	    status = cword_verify_get_block_props (
			crossword_record,
			crossword_grid);
	}

	if (status != CWORD_S_NORMAL)
	{
	    cword_messages_output_message (
		CWORD_GRDINVALIDERR,
		(TRACE2 | SCREEN0));

	    cword_memman_free_cross_rec (crossword_record);

	    return (CWORD_S_FAILURE);
	}
    }
    else
    {
	cword_messages_output_message (
	    CWORD_GRDINVALIDERR,
	    (TRACE2 | SCREEN0));

	cword_memman_free_cross_rec (crossword_record);

	return (CWORD_S_FAILURE);
    }

    /*
     *  Output message before we go to writing the output file
     */

    switch (crossword_record->grid_type)
    {
    case BLOCK_GRID:
    case AMERICAN_GRID:
    case NOVELTY_GRID:
	cword_messages_output_message (
	    CWORD_SELECTPSTYPE,
	    (TRACE3 | SCREEN0),
	    CWORD_BLOCK_GRID_STR);

	break;

    case BARRED_GRID:
	cword_messages_output_message (
	    CWORD_SELECTPSTYPE,
	    (TRACE3 | SCREEN0),
	    CWORD_BARRED_GRID_STR);

	break;

    default:
	printf ("# Error no grid type selected\n");

	return (CWORD_S_FAILURE);  
    }

    /*
     *  *Now* print grid to a postscript/text file
     */

    switch (print_option)
    {
    case POSTSCRIPT:
	sprintf (
	    output_file_name,
	    "%s%s",
	    crossword_record->file_name,
	    CWORD_SUFFIX_PRNPS_FILE);

	status = cword_utils_is_file_ok (
		    WRITE_FILE,
		    CWORD_SUFFIX_PRNPS_FILE,
		    output_file_name);

	if (CWORD_S_NORMAL != status)
	{
	    return (CWORD_S_FAILURE);
	}

	output_fd = open (
			output_file_name, 
			O_WRONLY | O_CREAT | O_TRUNC, 
			0700);

	if (output_fd == -1)
	{
	    cword_messages_output_message (
		CWORD_FILEERROR,
		(TRACE2 | SCREEN0),
		output_file_name);

	    cword_memman_free_cross_rec (crossword_record);

	    return (CWORD_WRITEERROR);
	}

	switch (crossword_record->grid_type)
	{
	case BLOCK_GRID:
	case AMERICAN_GRID:
	case NOVELTY_GRID:
	    /*
	     *  Print out PS comments before continuing
	     */

	    cword_print__header_ps_comments (
		BLOCK_GRID_FILE,
		crossword_record->name,
		output_fd);

	    /*
	     *  Print out PS block grid
	     */

	    status = cword_print__block_ps_grid (
			PREAMBLE,
			output_fd,
			crossword_record,
			crossword_grid);
	    break;

	case BARRED_GRID:
	    /*
	     *  Print out PS comments before continuing
	     */

	    cword_print__header_ps_comments (
		BARRED_GRID_FILE,
		crossword_record->name,
		output_fd);

	    /*
	     *  Print out PS barred grid
	     */

	    status = cword_print__barred_ps_grid (
			PREAMBLE,
			output_fd,
			crossword_record,
			crossword_grid);
	    break;

	default:
	    printf ("# Error no grid type selected\n");

	    return (CWORD_S_FAILURE);  
	}

	break;

    case TEXT:
	sprintf (
	    output_file_name,
	    "%s%s",
	    crossword_record->file_name,
	    CWORD_SUFFIX_PRNDAT_FILE);

	status = cword_utils_is_file_ok (
		    WRITE_FILE,
		    CWORD_SUFFIX_PRNDAT_FILE,
		    output_file_name);

	if (CWORD_S_NORMAL != status)
	{
	    return (CWORD_S_FAILURE);
	}

	output_fd = open (
		output_file_name, 
		O_WRONLY | O_CREAT | O_TRUNC, 
		0700);

	if (output_fd == -1)
	{
	    cword_messages_output_message (
		CWORD_FILEERROR,
		(TRACE2 | SCREEN0),
		output_file_name);

	    cword_memman_free_cross_rec (crossword_record);

	    return (CWORD_WRITEERROR);
	}

	cword_print__check_size (
		crossword_record->x_dimensions,
		crossword_record->y_dimensions);

	switch (crossword_record->grid_type)
	{
	case BLOCK_GRID:
	case AMERICAN_GRID:
	case NOVELTY_GRID:
	    /*
	     *  Print out TEXT block grid
	     */

	    status = cword_print__block_txt_grid (
			PREAMBLE,
			output_fd,
			crossword_record,
			crossword_grid);

	    break;

	case BARRED_GRID:
	    /*
	     *  Print out TEXT barred grid
	     */

	    status = cword_print__barred_txt_grid (
			PREAMBLE,
			output_fd,
			crossword_record,
			crossword_grid);

	    break;

	default:
	    printf ("# Error no grid type selected\n");

	    return (CWORD_S_FAILURE);  
	}

	break;

    default:
	cword_messages_output_message (
	    CWORD_INVPRINTOPT,
	    (TRACE2 | SCREEN0));

	status = CWORD_S_FAILURE;

	break;
    }

    if (status == CWORD_S_NORMAL)
    {
	cword_messages_output_message (
	    CWORD_GRIDOUTPUTMSG,
	    (TRACE0 | SCREEN0),
	    output_file_name);
    }

    close (output_fd);

    cword_memman_free_cross_rec (crossword_record);

    return (status);

}



#else

extern int main (int argc, char *argv[])

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *      Main function driving the program to completion - PRINT HARNESS
 *
 *  FORMAL PARAMETERS:
 *
 *      argc		(r) number of parameters inputted
 *	*argv[]		(r) contains the parameters inputted at command line
 *--
 */
{

    CROSSWORD_RECORD	    *crossword_record;

    char		    file_spec[CWORD_SIZEOF_FILESPEC + 1];

    int			    status;

    status = CWORD_OS_GET_FILES (
		"*.input",
		file_spec);

    while (CWORD_S_NORMAL == status)
    {
	printf ("%s\n", (strstr (file_spec, "]") + 1));

	status = CWORD_OS_GET_FILES (
		    (char) 0,
		    file_spec);    
    }

    if (argc < 3)
    {
	printf ("Correct options not entered\n");
	exit (0);
    }

    /* Take all four options */

    if (!(strcmp (argv[2], "barred_ps")))
    {
	status = cword_print_grid (
			BARRED_PS,
			argv[1]);
    }
    else if ( !(strcmp (argv[2], "barred_text")) )
    {
	status = cword_print_grid (
			BARRED_TEXT,
			argv[1]);
    }
    else if ( !(strcmp (argv[2], "block_ps")) )
    {
	status = cword_print_grid (
			BLOCK_PS,
			argv[1]);
    }
    else if ( !(strcmp (argv[2], "block_text")) )
    {
	status = cword_print_grid (
			BLOCK_TEXT,
			argv[1]);
    }


    exit (EXIT_SUCCESS);

}

#endif
