#ifdef DB

/*
 *++
 *  FACILITY:
 *
 *      rdb.sc
 *
 *  ABSTRACT:
 *
 *	Module containing all database functions to interact with the Cword rdb
 *	database.
 *
 *	This database can contain both complete crosswords, complete grids or
 *	just individual clue definitions that can be used during grid filling.
 *
 *  AUTHORS:
 *
 *      James A. Lundon
 *
 *  CREATION DATE:     
 *
 *	23 December 1992
 *--
 */

/*
 *  Include files
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <ctype.h>

#if defined(VMS) || defined(__VMS)

/*
 *  OpenVMS only includes!
 */

#include <descrip.h>
#include <ssdef.h>
#include <file.h>
#include <unixio.h>

/*
 *  Prototype for SQL$GET_ERROR_TEXT!
 */

void SQL$GET_ERROR_TEXT (struct dsc$descriptor_s *text_descriptor);

#endif

/*
 *  SQL compile time declarations.
 */

#if defined(VMS) || defined(__VMS)

/*
 *  ?? Not needed on Digital Unix (OSF/1) why ??
 */

#include <sqlsrvca.h>

#endif

#include <sqlsrv.h>

/*
 *  Prevent "multiply defined symbol during link.
 */

EXEC SQL INCLUDE SQLCA;

/* 
 *  The code for Rdb V4.0 or less:
 *
 *  EXEC SQL DECLARE SCHEMA AUTHORIZATION FOR FILENAME 'CWORD_DATABASE';
 */

EXEC SQL DECLARE ALIAS FOR FILENAME CWORD_DATABASE;

#include "main.h"
#include "memman.h"
#include "print.h"
#include "os.h"
#include "rdb.h"
#include "utils.h"
#include "messages.h"
#include "verify.h"
#include "word.h"
#include "strll.h"
#include "defaults.h"

/*
 *  ?? Set up proper mechanism for naming undefined crosswords in the database.
 *  At the moment only one unnamed crossword can be stored in the database ??
 */

#define XRD_NO_NAME	"crossword_0"
#define GRD_NO_NAME	"grid_0"

#define UNDEFINED_XRD	"Undefined Crossword"

/*
 *  Whether to read or write table.
 */

#define READ_TABLE		1
#define WRITE_TABLE		2

/*
 *  Which table to lock.
 */

#define CROSSWORD_DESCRIPTION   1
#define CLUE_DESCRIPTION	2
#define GRID_DESCRIPTION	3
#define LIGHT_DESCRIPTION	4
#define ANAGRAM_DESCRIPTION	5
#define HOMOPHONE_DESCRIPTION	6
#define SYNONYM_DESCRIPTION     7
#define XRD_OPERATION		8
#define GRD_OPERATION		9
#define ALL			10

#define CROSSWORD_DESC_STR	"CROSSWORD_DESCRIPTION"
#define CLUE_DESC_STR		"CLUE_DESCRIPTION"
#define GRID_DESC_STR		"GRID_DESCRIPTION"
#define LIGHT_DESC_STR		"LIGHT_DESCRIPTION"

/*
 *  Module global to determine whether the database has been bound to or not.
 */

static BOOLEAN cword_rdb_l_db_opened = FALSE;



/*
 *  PRIVATE function prototypes
 */

PRIVATE int cword_rdb__set_transaction (
	int			rw_status,
	int			table);

PRIVATE int cword_rdb__check_status (
	int			sql_status);

PRIVATE int cword_rdb__possible_fits (
	int			no_of_letters, 
	char			*string_to_match,
	CWORD_T_STR_LL		**word_ll);

/* 
 *  ????? Not happy with all ROLLBACK mechanisms ?????
 */



PRIVATE int cword_rdb__check_status (
	int			sql_status)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function to generically check the inputted SQL status value.
 *
 *  FORMAL PARAMETERS:
 *
 *      sql_status		(r) Variable containing SQL return status
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL		Normal Completion
 *      other			Error
 *--
 */

{

#if defined(VMS) || defined(__VMS)
    struct dsc$descriptor_s text_descriptor;
#endif

    char error_str[CWORD_SIZEOF_MESSAGE + 1];

    int	status,
        error_len;

    if (sql_status == SQL_SUCCESS)
        return (CWORD_S_NORMAL);

    /*
     *  Error status #defines are to be found in <sqlsrv.h>.
     */

    switch (sql_status)
    {
    case SQL_EOS :
	/*
	 *  This is usually OK!
	 */

	status = CWORD_S_RDBENDOFSTREAM;

	break;

    case SQL_DEADLOCK :
	cword_messages_output_message (
	    CWORD_RDBDEADLOCK,
	    (TRACE0 | SCREEN0));

	status = CWORD_RDBDEADLOCK;

	break;

    case SQL_LOCK_CONFLICT :
	cword_messages_output_message (
	    CWORD_RDBCONFLICT,
	    (TRACE0 | SCREEN0));

	status = CWORD_RDBCONFLICT;

	break;

    case SQL_BAD_TXN_STATE :
	cword_messages_output_message (
	    CWORD_RDBINVTRANX,
	    (TRACE0 | SCREEN0));

	status = CWORD_RDBINVTRANX;

	break;

    case SQL_NO_TXN :
	/*
	 *  No transaction pending is not an OK situation, I think?
	 */

	cword_messages_output_message (
	    CWORD_RDBNOTRANX,
	    (TRACE0 | SCREEN0));

	status = CWORD_S_NORMAL;

	break;

    case SQL_NO_DUP :
	/*
	 *  Breaks index constraints in datbase.
	 */

	status = CWORD_RDBINVINDEX;

	break;

    case SQL_BAD_VERSION :
	cword_messages_output_message (
	    CWORD_RDBBADTRANX,
	    (TRACE0 | SCREEN0));

	status = CWORD_RDBBADTRANX;

	break;

    case SQL_INTEG_FAIL :
	cword_messages_output_message (
	    CWORD_RDBINXCONSTR,
	    (TRACE0 | SCREEN0));

	status = CWORD_RDBINXCONSTR;

	break;

    case SQL_RDBERR :
	cword_messages_output_message (
	    CWORD_RDBERR,
	    (TRACE0 | SCREEN0));

        /*
         *  Get the actual text associated with the error found!
         */

	status = CWORD_RDBERR;

	break;

    case SQL_OUTOFRAN :
	cword_messages_output_message (
	    CWORD_RDBOUTOFRANGE,
	    (TRACE0 | SCREEN0));

	status = CWORD_RDBOUTOFRANGE;

	break;

    case SQL_ATTACHERR:
	cword_messages_output_message (
	    CWORD_S_RDBATTACHERR,
	    (TRACE0 | SCREEN0));

	status = CWORD_S_FAILURE;

	break;

    default :
	/*
	 *  Any other error states are covered by the general case and the
	 *  value of sql_status is printed.
	 */

	cword_messages_output_message (
	    CWORD_S_RDBUNDEFINEDERR,
	    (TRACE0 | SCREEN0));

	status = CWORD_RDBUNDEFERR;

	break;
    }

#if defined(VMS) || defined(__VMS)

    text_descriptor.dsc$b_class = DSC$K_CLASS_S;
    text_descriptor.dsc$b_dtype = DSC$K_DTYPE_T;
    text_descriptor.dsc$w_length = CWORD_SIZEOF_MESSAGE;
    text_descriptor.dsc$a_pointer = &error_str;

    SQL$GET_ERROR_TEXT (&text_descriptor);

    error_str[CWORD_SIZEOF_MESSAGE] = (char) 0;
    cword_utils_trim_string (error_str);

#else

    sql_get_error_text (
        error_str,
        sizeof (error_str),
        &error_len);

    error_str[error_len] = (char) 0;

#endif

    cword_messages_output_message (
        CWORD_S_RDBERRORTEXT ,
	(TRACE0 | SCREEN0),
        error_str);

    return (status);

}



PRIVATE int cword_rdb__set_transaction (
	int			rw_status,
	int			table)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function that will set the transaction state before any other
 *	processing takes place.  By using this function we can guarantee that
 *	the database will not be locked to any other Cword processed wanting to
 *	gain access.
 *
 *  FORMAL PARAMETERS:
 *
 *      rw_status		(r) READ_TABLE/WRITE_TABLE
 *	table			(r) Table(s) on which SET TRANSACTION is/are to work.
 *
 *  IMPLICIT INPUTS:
 *
 *      cword_rdb_l_db_opened
 *
 *  IMPLICIT OUTPUTS:
 *
 *      cword_rdb_l_db_opened
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL		Normal completion.
 *	CWORD_S_FAILURE		Error using SET TRANSACTION.
 *--
 */

{

    int status;

    if (cword_rdb_l_db_opened == FALSE)
    {
	/*
	 *  When connecting to the database for the first time, there is
	 *  usually a little delay.  Better tell the user this :-).
	 */

	cword_messages_output_message (
	    CWORD_RDBOPENDB,
	    (TRACE0 | SCREEN0));
    }

    if (rw_status == READ_TABLE)
    {
	/*
	 *  Start READ ONLY transaction for any selected Cword table(s)!
	 */

	EXEC SQL SET TRANSACTION READ ONLY;
    }
    else if (rw_status == WRITE_TABLE)
    {
	if (table == CROSSWORD_DESCRIPTION)
	{
	    EXEC SQL SET TRANSACTION READ WRITE RESERVING
		CROSSWORD_DESCRIPTION FOR EXCLUSIVE WRITE;
	}
	else if (table == CLUE_DESCRIPTION)
	{
	    EXEC SQL SET TRANSACTION READ WRITE RESERVING
		CLUE_DESCRIPTION FOR EXCLUSIVE WRITE;
	}
	else if (table == GRID_DESCRIPTION)
	{
	    EXEC SQL SET TRANSACTION READ WRITE RESERVING
		GRID_DESCRIPTION FOR EXCLUSIVE WRITE;
	}
	else if (table == LIGHT_DESCRIPTION)
	{
	    EXEC SQL SET TRANSACTION READ WRITE RESERVING
		LIGHT_DESCRIPTION FOR EXCLUSIVE WRITE;
	}
	else if (table == ANAGRAM_DESCRIPTION)
	{
	}
	else if (table == HOMOPHONE_DESCRIPTION)
	{
	}
	else if (table == SYNONYM_DESCRIPTION)
	{
	}
	else if (table == XRD_OPERATION)
	{
	    EXEC SQL SET TRANSACTION READ WRITE RESERVING
		CROSSWORD_DESCRIPTION FOR EXCLUSIVE WRITE,
		CLUE_DESCRIPTION FOR EXCLUSIVE WRITE;
	}
	else if (table == GRD_OPERATION)
	{
	    EXEC SQL SET TRANSACTION READ WRITE RESERVING
		GRID_DESCRIPTION FOR EXCLUSIVE WRITE,
		LIGHT_DESCRIPTION FOR EXCLUSIVE WRITE;
	}
	else if (table == ALL)
	{
	    EXEC SQL SET TRANSACTION READ WRITE RESERVING
		CROSSWORD_DESCRIPTION FOR EXCLUSIVE WRITE,
		CLUE_DESCRIPTION FOR EXCLUSIVE WRITE,
		GRID_DESCRIPTION FOR EXCLUSIVE WRITE,
		LIGHT_DESCRIPTION FOR EXCLUSIVE WRITE;
	}
	else
	{
	    printf ("# Error inputting valid table\n");
	}
    }
    else
    {
	printf ("# Error inputting valid rw_status\n");
    }

    status = cword_rdb__check_status (SQLCA.SQLCODE);

    if (status == CWORD_S_NORMAL)
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICSUC,
	    TRACE1,
	    SET_TRANSACTION_STR);
    }
    else
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICERR,
	    (TRACE0 | SCREEN0),
	    SET_TRANSACTION_STR);

	return (CWORD_S_FAILURE);
    }

    cword_rdb_l_db_opened = TRUE;

    return (CWORD_S_NORMAL);

}



/*
 *  ?? Do I need to "FINISH/DISCONNECT ALL" at all before exiting - check in
 *  the Rdb notesfiles/existing code for a definitive solution ??
 */

PUBLIC void cword_rdb_close_db ()

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function called when closing and finishing with the Cword database.
 *
 *  FORMAL PARAMETERS:
 *
 *      None
 *
 *  IMPLICIT INPUTS:
 *
 *      cword_rdb_l_db_opened
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL		Normal Completion
 *	other			Rdb error
 *--
 */

{

    int status;

    if (cword_rdb_l_db_opened == FALSE)
    {
	return;
    }

#if 0
    /* 
     *  The code for V4.0 or less:
     *
     *  EXEC SQL FINISH;
     *
     *  An automatic ROLLBACK is executed when DISCONNECT is executed!
     *
     *  EXEC SQL FINISH;
     *
     *	?? Have to put the FINISH inside quotes or else it will be use by
     *	sqlpre ??
     */

    status = cword_rdb__check_status (SQLCA.SQLCODE);

    if (status == CWORD_S_NORMAL)
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICSUC,
	    TRACE1,
	    DISCONNECT_STR);
    }
    else
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICERR,
	    (TRACE1 | SCREEN1),
	    DISCONNECT_STR);
    }
#endif

    return;

}



PRIVATE int cword_rdb__possible_fits (
	int			no_of_letters, 
	char			*string_to_match,
	CWORD_T_STR_LL		**word_ll)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	This function emulates its counterpart cword_match__possible_fits but
 *	instead of querying a word file, it queries the local rdb database.
 *	The function will return a list of matching words found in the rdb
 *	database.
 *
 *  FORMAL PARAMETERS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  {@function value or completion codes@}
 *
 *      [@description or none@]
 *--
 */

{

    char rdb_str[41];

    register long word_count = 0;

    int	status,
	count;

    BOOLEAN is_word_ok = TRUE,
	    is_first_letter = FALSE;

    *word_ll = (CWORD_T_STR_LL *) 0;

    cword_messages_output_message (
	CWORD_CLOCKBEFORE,
	TRACE2,
	clock ());

    if (string_to_match[0] != (char) ' ')
	is_first_letter = TRUE;

    /*
     *	Go away to the database, grab the matches, and return in a linked list.
     */

    /*
     *	Declare a cursor with which to fetch words from database with.  This
     *	cursor will return in ascending order the words to be matched.
     */

    /*
     *  ?? Add in WORDS.CLUE_QUESTION to the sort ??
     *
     *  ?? Is sorting needed at all ??
     */

    /*
     *  ?? Singleton sort ??
     */

    EXEC SQL DECLARE WORDS_CURSOR CURSOR FOR
	SELECT
	    WORDS.CLUE_ANSWER
	FROM CLUE_DESCRIPTION WORDS
	ORDER BY WORDS.CLUE_ANSWER ASC;

    status = cword_rdb__check_status (SQLCA.SQLCODE);

    if (status == CWORD_S_NORMAL)
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICSUC,
	    TRACE3,
	    DECLARE_CURSOR_STR);
    }
    else
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICERR,
	    (TRACE1 | SCREEN1),
	    DECLARE_CURSOR_STR);

	EXEC SQL ROLLBACK;

	return (CWORD_S_RDBENDOFSTREAM);
    }

    EXEC SQL OPEN WORDS_CURSOR;

    status = cword_rdb__check_status (SQLCA.SQLCODE);

    if (status == CWORD_S_NORMAL)
    {
	cword_messages_output_message (
		CWORD_RDBGENERICSUC,
		TRACE3,
		OPEN_CURSOR_STR);
    }
    else
    {
	cword_messages_output_message (
		CWORD_RDBGENERICERR,
		(TRACE1 | SCREEN1),
		OPEN_CURSOR_STR);

	EXEC SQL ROLLBACK;

	return (CWORD_S_RDBENDOFSTREAM);
    }

    EXEC SQL FETCH WORDS_CURSOR INTO
	:rdb_str;

    while (SQLCA.SQLCODE != SQL_EOS)
    {
	cword_word_truncate_str (rdb_str);

	/*
	 *  Check does the word match the input template
	 */

	if (strlen (string_to_match) == strlen (rdb_str))
	{
	    count = 0;
	    is_word_ok = TRUE;

	    while (count < no_of_letters)
	    {
		if (string_to_match[count] == ' ')
		    count++;
		else
		{
		    if (string_to_match[count] != rdb_str[count])
		    {
			is_word_ok = FALSE;
			break;
		    }
		    else
			count++;
		}
	    }

	    if (is_word_ok == TRUE)
	    {
		cword_strll_add_str (
		    SORTED | UNIQUE,
		    rdb_str,
		    word_ll);
		
		if (*word_ll == (CWORD_T_STR_LL *) 0)
		{
		    EXEC SQL CLOSE WORDS_CURSOR;

		    EXEC SQL ROLLBACK;

		    cword_messages_output_message (
			    CWORD_CLOCKAFTER,
			    TRACE2,
			    clock ());

		    return (CWORD_S_FAILURE);
		}

		word_count++;

		/* 
		 *  Not implementing max matches check as it'll probably never
		 *  be reached.
		 */
	    }

	    if (is_first_letter == TRUE)
	    {
		if ((word_count % CWORD_SIZEOF_CHECK_EVERY) == 0)
		{
		    if ((int) string_to_match[0] < (int) rdb_str[0])
		    {
			cword_messages_output_message (
				CWORD_CLOCKAFTER,
				TRACE2,
				clock ());

			EXEC SQL CLOSE WORDS_CURSOR;

			EXEC SQL ROLLBACK;

			/*
			 *  Assume that closing the cursor and rolling back
			 *  went OK before returning.
			 */

			return (CWORD_S_NORMAL);
		    }
		}
	    }
	}

	EXEC SQL FETCH WORDS_CURSOR INTO
	    :rdb_str;
    }

    cword_messages_output_message (
	    CWORD_RDBGENERICSUC,
	    TRACE3,
	    FETCH_CURSOR_STR);

    EXEC SQL CLOSE WORDS_CURSOR;

    status = cword_rdb__check_status (SQLCA.SQLCODE);

    if (status == CWORD_S_NORMAL)
    {
	cword_messages_output_message (
		CWORD_RDBGENERICSUC,
		TRACE3,
		CLOSE_CURSOR_STR);
    }
    else
    {
	cword_messages_output_message (
		CWORD_RDBGENERICERR,
		(TRACE1 | SCREEN1),
		CLOSE_CURSOR_STR);

	EXEC SQL ROLLBACK;

	return (CWORD_S_FAILURE);
    }

    cword_messages_output_message (
	    CWORD_CLOCKAFTER,
	    TRACE2,
	    clock ());

    EXEC SQL ROLLBACK;

    return (CWORD_S_NORMAL);

}



PUBLIC int cword_rdb_match_template_ll (
	char			*search_str,
	CWORD_T_STR_LL		**word_ll)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function to return a list of words that match the input template from
 *	the adjoining rdb database.  Whichever word is chosen will then be
 *	re-queried for the clue text which is related to it in the database.
 *	This function is modelled on cword_match_template_ll which just
 *	queries the accompanying word file.
 *
 *  FORMAL PARAMETERS:
 *
 *	*search_str		    (r) search string
 *	**word_ll 		    (w)	link list of words to be returned
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  {@function value or completion codes@}
 *
 *      [@description or none@]
 *--
 */

{

    /*
     *	Very similar to cword_match_template_ll but I felt it was more
     *	intuitive to have it dealt with in a separate function.
     */

    int status;

    *word_ll = (CWORD_T_STR_LL *) 0;

    if (strlen (search_str) > CWORD_SIZEOF_WORD)
    {
	cword_messages_output_message (
	    CWORD_WORDTOOBIG,
	    (TRACE0 | SCREEN0));

	return (CWORD_S_FAILURE);
    }

    /*
     *  Format it correctly before going to select matches off it. 
     */

    cword_word_preproc_match_str (search_str);

    cword_messages_output_message (
	CWORD_S_TEMPSEARCHDB,
	(TRACE3 | SCREEN3),
	search_str);

    /*
     *  Call the rdb function that will get the matches from the database.
     */

    status = cword_rdb__possible_fits (
		strlen (search_str),
		search_str,
		word_ll);

    if (status != CWORD_S_NORMAL)
    {
	cword_strll_free_str_ll (word_ll);

	return (CWORD_S_FAILURE);
    }

    if (*word_ll == (CWORD_T_STR_LL *) 0)
    {
	cword_messages_output_message (
	    CWORD_S_NOTEMPMATCHDB,
	    (TRACE0 | SCREEN0));

	return (CWORD_S_NORMAL);
    }

    return (CWORD_S_NORMAL);

}



/*
 *  ?? I should check whether the crossword is complete before trying to save
 *  it, or at least inform the user what the status of the crossword is ??
 */

PUBLIC int cword_rdb_add_xrd (
	CROSSWORD_RECORD	*crossword_record)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function to add a crossword definition to the local Cword database.
 *
 *  FORMAL PARAMETERS:
 *
 *	*crossword_record	(r) structure containing the crossword details
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL		Crossword definition successfully written.
 *	other			Crossword definition not successfully written.
 *--
 */

{

    /*
     *	The sizes of the character buffers *must equal* with those as defined
     *	in the database definition.
     *
     *	One cannot use constant identifiers (such as CWORD_SIZEOF_FILESPEC)
     *	when defining the strings to be used here.  This is because the
     *	precompiler isn't smart enough to substitute the values for the
     *	constants into the defitionof each of the string lengths.
     */

    char temp1_str[151],
	 temp2_str[151],
	 temp3_str[41]; /* CWORD_SIZEOF_WORD or equivalent */

    int status,
	count1,
	temp1_int,
	temp2_int,
	temp3_int,
	temp4_int,
	temp5_int;

    status = cword_rdb__set_transaction (
		WRITE_TABLE,
		XRD_OPERATION);

    if (CWORD_S_NORMAL != status)
    {
	return (status);
    }

    /*
     *  Make sure that a valid name is given to the crossword in question!
     *
     *	This is a bit non-intuitive but it's the best I can probably do given
     *	the constraints of the problem.
     */

    if (crossword_record->name[0] == (char) 0)
    {
	if (crossword_record->file_name[0] == (char) 0)
	{
	    strcpy (crossword_record->name, XRD_NO_NAME);
	}
	else
	{
	    strcpy (crossword_record->name, crossword_record->file_name);
	}

	printf (
	    "# Using crossword name of %s since none provided in input definition\n",
	    crossword_record->name);
    }

    /* 
     *  temp1_str same for all the INSERTS in this module.
     */

    strcpy (temp1_str, crossword_record->name);

    strcpy (temp2_str, crossword_record->remarks);
    temp1_int = crossword_record->clue_sum;
    temp2_int = crossword_record->x_dimensions;
    temp3_int = crossword_record->y_dimensions;
    temp4_int = crossword_record->grid_type;
    temp5_int = crossword_record->symmetry;

    EXEC SQL INSERT INTO CROSSWORD_DESCRIPTION
    (
	CROSSWORD_NAME,
	CROSSWORD_REMARKS,
	CROSSWORD_CLUES,
	CROSSWORD_LENGTH,
	CROSSWORD_WIDTH,
	CROSSWORD_GRID_TYPE,
	CROSSWORD_GRID_SYMMETRY
    )
    VALUES
    (
	:temp1_str,
	:temp2_str,
	:temp1_int,
	:temp2_int,
	:temp3_int,
	:temp4_int,
	:temp5_int
    );

    status = cword_rdb__check_status (SQLCA.SQLCODE);

    if (CWORD_S_NORMAL == status)
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICSUC,
	    TRACE3,
	    INSERT_STR,
	    CROSSWORD_DESC_STR);
    }
    else if (CWORD_RDBINVINDEX == status)
    {
	cword_messages_output_message (
	    CWORD_S_RDBDUPLICDEFN,
	    (TRACE0 | SCREEN0),
	    CROSSWORD_STR,
	    temp1_str);

	EXEC SQL ROLLBACK;

	return (CWORD_S_RDBDUPLICDEFN);
    }
    else
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICERR,
	    (TRACE1 | SCREEN1),
	    INSERT_STR,
	    CROSSWORD_DESC_STR);

	EXEC SQL ROLLBACK;

	return (CWORD_S_FAILURE);
    }

    cword_messages_output_message (
	CWORD_RDBADDCROSSSUC,
	(TRACE2 | SCREEN0));

    for (count1 = 0; count1 < crossword_record->clue_sum; count1++)
    {
	temp1_int = crossword_record->clue_details[count1].clue_number;
	temp2_int = crossword_record->clue_details[count1].clue_direction;
	temp3_int = crossword_record->clue_details[count1].x_coordinate;
	temp4_int = crossword_record->clue_details[count1].y_coordinate;
	strcpy (temp2_str, crossword_record->clue_details[count1].clue_description);
	strcpy (temp3_str, crossword_record->clue_details[count1].clue_answer);

	EXEC SQL INSERT INTO CLUE_DESCRIPTION
	(
	    CROSSWORD_NAME,
	    CLUE_NUMBER,
	    CLUE_DIRECTION,
	    CLUE_X_COORD,
	    CLUE_Y_COORD,
	    CLUE_QUESTION,
	    CLUE_ANSWER
	)
	VALUES	
	(
	    :temp1_str,
	    :temp1_int,
	    :temp2_int,
	    :temp3_int,
	    :temp4_int,
	    :temp2_str,
	    :temp3_str
	);

	status = cword_rdb__check_status (SQLCA.SQLCODE);

	if (status != CWORD_S_NORMAL)
	{
	    cword_messages_output_message (
		CWORD_RDBGENERICERR,
		(TRACE3 | SCREEN3),
		INSERT_STR);

	    cword_messages_output_message (
		CWORD_S_RDBDUPDEFN,
		(TRACE1 | SCREEN1),
		CROSSWORD_STR);

	    EXEC SQL ROLLBACK;

	    return (CWORD_S_FAILURE);
	}
    }

    cword_messages_output_message (
	CWORD_RDBGENERICSUC,
	TRACE3,
	INSERT_STR,
	CLUE_DESC_STR);

    cword_messages_output_message (
	CWORD_RDBADDCLUESUC,
	(TRACE2 | SCREEN0));

    EXEC SQL COMMIT;

    status = cword_rdb__check_status (SQLCA.SQLCODE);

    if (status == CWORD_S_NORMAL)
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICSUC,
	    TRACE3,
	    COMMIT_STR);
    }
    else
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICERR,
	    (TRACE1 | SCREEN1),
	    COMMIT_STR);

	EXEC SQL ROLLBACK;

	return (CWORD_S_FAILURE);
    }

    return (CWORD_S_NORMAL);

}



/*
 *  ?? Get a crossword from the database and print directly to an output file
 *  type: POSTSCRIPT/TEXT ??
 */

PUBLIC int cword_rdb_get_xrd (
	char			*xrd_name_str,
	CROSSWORD_RECORD	*crossword_record)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function to obtain from the Cword database a crossword description
 *	given its name name.
 *
 *  FORMAL PARAMETERS:
 *
 *      *xrd_name_str		(r) name of crossword to obtain
 *	*crossword_record	(r) record into which to put the crossword details
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL		Obtained the crossword details properly.
 *	other			Failed to obtain crossword definition from database.
 *--
 */

{

    char temp1_str[151],
	 temp2_str[151],
	 temp_name[21];

    int status,
	count1,
	temp1_int,
	temp2_int,
	temp3_int,
	temp4_int,
	temp5_int;

    if (xrd_name_str[0] == (char) 0)
    {
	printf ("# Error inputting valid crossword name\n");

	return (CWORD_S_FAILURE); 
    }

    status = cword_rdb__set_transaction (
		READ_TABLE,
		XRD_OPERATION);

    if (CWORD_S_NORMAL != status)
    {
	return (status);
    }

    /*
     *  Getting the crossword definition/statistics.
     */

    EXEC SQL DECLARE DEFN_CURSOR CURSOR FOR
	SELECT
	    DEFN.CROSSWORD_REMARKS,
	    DEFN.CROSSWORD_CLUES,
	    DEFN.CROSSWORD_LENGTH,
	    DEFN.CROSSWORD_WIDTH,
	    DEFN.CROSSWORD_GRID_TYPE,
	    DEFN.CROSSWORD_GRID_SYMMETRY
	FROM CROSSWORD_DESCRIPTION DEFN
	WHERE
	    DEFN.CROSSWORD_NAME = :xrd_name_str;

    status = cword_rdb__check_status (SQLCA.SQLCODE);

    if (status == CWORD_S_NORMAL)
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICSUC,
	    TRACE3,
	    DECLARE_CURSOR_STR);
    }
    else
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICERR,
	    (TRACE3 | SCREEN3),
	    DECLARE_CURSOR_STR);

	EXEC SQL ROLLBACK;

	memset (crossword_record, 0, sizeof (CROSSWORD_RECORD));

	return (CWORD_S_RDBENDOFSTREAM);
    }

    EXEC SQL OPEN DEFN_CURSOR;

    status = cword_rdb__check_status (SQLCA.SQLCODE);

    if (status == CWORD_S_NORMAL)
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICSUC,
	    TRACE3,
	    OPEN_CURSOR_STR);
    }
    else
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICERR,
	    (TRACE1 | SCREEN1),
	    OPEN_CURSOR_STR);

	EXEC SQL ROLLBACK;

	memset (crossword_record, 0, sizeof (CROSSWORD_RECORD));

	return (CWORD_S_RDBENDOFSTREAM);
    }

    EXEC SQL FETCH DEFN_CURSOR INTO 
	:temp1_str,
	:temp1_int,
	:temp2_int,
	:temp3_int,
	:temp4_int,
	:temp5_int;

    status = cword_rdb__check_status (SQLCA.SQLCODE);

    if (status == CWORD_S_NORMAL)
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICSUC,
	    TRACE3,
	    FETCH_CURSOR_STR);
    }
    else if (status == CWORD_S_RDBENDOFSTREAM)
    {
	cword_messages_output_message ( 
	    CWORD_S_RDBNOTINDB,
	    (TRACE1 | SCREEN1),
	    CROSSWORD_STR,
	    xrd_name_str);

	EXEC SQL ROLLBACK;

	memset (crossword_record, 0, sizeof (CROSSWORD_RECORD));

	return (CWORD_S_RDBENDOFSTREAM);
    }
    else
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICERR,
	    (TRACE1 | SCREEN1),
	    FETCH_CURSOR_STR);

	EXEC SQL ROLLBACK;

	memset (crossword_record, 0, sizeof (CROSSWORD_RECORD));

	return (CWORD_S_RDBENDOFSTREAM);
    }

    strcpy (crossword_record->name, xrd_name_str);

    cword_word_truncate_str (temp1_str);    
    strcpy (crossword_record->remarks, temp1_str);

    crossword_record->clue_sum	    = temp1_int;
    crossword_record->x_dimensions  = temp2_int;
    crossword_record->y_dimensions  = temp3_int;
    crossword_record->grid_type     = temp4_int;
    crossword_record->symmetry      = temp5_int;

    EXEC SQL CLOSE DEFN_CURSOR;

    status = cword_rdb__check_status (SQLCA.SQLCODE);

    if (status == CWORD_S_NORMAL)
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICSUC,
	    TRACE3,
	    CLOSE_CURSOR_STR);
    }
    else
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICERR,
	    (TRACE1 | SCREEN1),
	    CLOSE_CURSOR_STR);

	EXEC SQL ROLLBACK;

	memset (crossword_record, 0, sizeof (CROSSWORD_RECORD));

	return (CWORD_S_RDBENDOFSTREAM);
    }

    /*
     *  Getting the clues for the crossword now 
     */

    EXEC SQL DECLARE CLUES_CURSOR CURSOR FOR
	SELECT
	    CLUS.CLUE_NUMBER,
	    CLUS.CLUE_DIRECTION,
	    CLUS.CLUE_X_COORD,
	    CLUS.CLUE_Y_COORD,
	    CLUS.CLUE_QUESTION,	
	    CLUS.CLUE_ANSWER
	FROM CLUE_DESCRIPTION CLUS
	WHERE
	    CLUS.CROSSWORD_NAME = :xrd_name_str
	ORDER BY
	    CLUS.CLUE_DIRECTION DESC,
	    CLUS.CLUE_NUMBER ASC;

    status = cword_rdb__check_status (SQLCA.SQLCODE);

    if (status == CWORD_S_NORMAL)
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICSUC,
	    TRACE3,
	    DECLARE_CURSOR_STR);
    }
    else
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICERR,
	    (TRACE1 | SCREEN1),
	    DECLARE_CURSOR_STR);

	EXEC SQL ROLLBACK;

	memset (crossword_record, 0, sizeof (CROSSWORD_RECORD));

	return (CWORD_S_RDBENDOFSTREAM);
    }
    
    EXEC SQL OPEN CLUES_CURSOR;

    status = cword_rdb__check_status (SQLCA.SQLCODE);

    if (status == CWORD_S_NORMAL)
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICSUC,
	    TRACE3,
	    OPEN_CURSOR_STR);
    }
    else
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICERR,
	    (TRACE1 | SCREEN1),
	    OPEN_CURSOR_STR);

	EXEC SQL ROLLBACK;

	memset (crossword_record, 0, sizeof (CROSSWORD_RECORD));

	return (CWORD_S_RDBENDOFSTREAM);
    }

    /*
     *  ?? Better error checking possible ??
     */

    EXEC SQL FETCH CLUES_CURSOR INTO
	:temp1_int,
	:temp2_int,
	:temp3_int,
	:temp4_int,
	:temp1_str,
	:temp2_str;

    count1 = 0;
    while (SQLCA.SQLCODE != SQL_EOS)
    {
	strcpy (crossword_record->clue_details[count1].clue_description, temp1_str);
	cword_word_truncate_str (crossword_record->clue_details[count1].clue_description);
	strcpy (crossword_record->clue_details[count1].clue_answer, temp2_str);
	cword_word_truncate_str (crossword_record->clue_details[count1].clue_answer);

	crossword_record->clue_details[count1].answer_length    = 
	    strlen (crossword_record->clue_details[count1].clue_answer);
	crossword_record->clue_details[count1].clue_number	= temp1_int;
	crossword_record->clue_details[count1].clue_direction	= temp2_int;
	crossword_record->clue_details[count1].x_coordinate	= temp3_int;
	crossword_record->clue_details[count1].y_coordinate	= temp4_int;

	count1++;
	
	EXEC SQL FETCH CLUES_CURSOR INTO
	    :temp1_int,
	    :temp2_int,
	    :temp3_int,
	    :temp4_int,
	    :temp1_str,
	    :temp2_str;
    }

    if (count1 != crossword_record->clue_sum)
    {
	printf (
	    "# Major error: number of clues found in database != clue_sum from definition - (%d != %d)\n",
	    count1,
	    crossword_record->clue_sum);

	memset (crossword_record, 0, sizeof (CROSSWORD_RECORD));

	EXEC SQL CLOSE CLUES_CURSOR;

	EXEC SQL ROLLBACK;

	return (CWORD_S_FAILURE);
    }

    cword_messages_output_message (
	CWORD_RDBGENERICSUC,
	TRACE3,
	FETCH_CURSOR_STR);

    EXEC SQL CLOSE CLUES_CURSOR;

    status = cword_rdb__check_status (SQLCA.SQLCODE);

    if (status == CWORD_S_NORMAL)
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICSUC,
	    TRACE3,
	    CLOSE_CURSOR_STR);
    }
    else
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICERR,
	    (TRACE1 | SCREEN1),
	    CLOSE_CURSOR_STR);

	EXEC SQL ROLLBACK;

	memset (crossword_record, 0, sizeof (CROSSWORD_RECORD));

	return (CWORD_S_RDBENDOFSTREAM);
    }

    EXEC SQL ROLLBACK;

    return (CWORD_S_NORMAL);

}



/*
 *  ?? Shouldn't try to delete a crossword if there are no crosswords to delete
 *  in the first place - get my drift ??
 */

PUBLIC int cword_rdb_delete_xrd (
	char			*xrd_name_str)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function to delete a crossword definition called *xrd_name_str from the
 *	Cword database.
 *
 *  FORMAL PARAMETERS:
 *
 *	xrd_name_str		(r) name of crossword to be deleted
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL		Crossword definition successfully deleted.
 *	other			Crossword definition not successfully deleted.
 *--
 */
{

    int status;

    status = cword_rdb__set_transaction (
		WRITE_TABLE,
		XRD_OPERATION);

    if (CWORD_S_NORMAL != status)
    {
	return (status);
    }

    /*
     *	By deleting this column it will set off a trigger in the database that
     *	will delete the corresponding CLUE_DESCRIPTION columns associated with
     *	the crossword.
     */

    EXEC SQL DELETE FROM CROSSWORD_DESCRIPTION WHERE
	CROSSWORD_NAME = :xrd_name_str;

    status = cword_rdb__check_status (SQLCA.SQLCODE);

    if (status == CWORD_S_NORMAL)
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICSUC,
	    TRACE3,
	    DELETE_STR,
	    CROSSWORD_DESC_STR);
    }
    /* 
     *  Needed to cover possibility of crossword not being in DB.
     */
    else if (status == CWORD_S_RDBENDOFSTREAM)
    {
	cword_messages_output_message ( 
	    CWORD_S_RDBNOTINDB,
	    (TRACE2 | SCREEN0),
	    CROSSWORD_STR,
	    xrd_name_str);

	EXEC SQL ROLLBACK;

	return (CWORD_S_FAILURE);
    }
    else
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICERR,
	    (TRACE1 | SCREEN1),
	    DELETE_STR,
	    CROSSWORD_DESC_STR);

	EXEC SQL ROLLBACK;

	return (CWORD_S_FAILURE);
    }

    EXEC SQL COMMIT;

    status = cword_rdb__check_status (SQLCA.SQLCODE);

    if (status == CWORD_S_NORMAL)
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICSUC,
	    TRACE1,
	    COMMIT_STR);
    }
    else
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICERR,
	    (TRACE3 | SCREEN3),
	    COMMIT_STR);

	EXEC SQL ROLLBACK;

	return (CWORD_S_FAILURE);
    }

    return (CWORD_S_NORMAL);

}



PUBLIC int cword_rdb_get_xrd_ll (
	CWORD_T_STR_LL		**xrd_ll)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function to search the Cword database looking for all valid crossword
 *	definitions.  Returns a list via **xrd_ll
 *
 *  FORMAL PARAMETERS:
 *
 *      **xrd_ll		(w) Link list of returned crossword definition names
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL		Normal Completion
 *      Rdb error 		Error getting crossword definition names
 *--
 */

{

    char xrd_name_str[151];

    int status;

    status = cword_rdb__set_transaction (
		READ_TABLE,
		XRD_OPERATION);

    if (CWORD_S_NORMAL != status)
    {
	return (status);
    }

    *xrd_ll = (CWORD_T_STR_LL *) 0;

    EXEC SQL DECLARE XRD_CURSOR CURSOR FOR
	SELECT
	    CD.CROSSWORD_NAME
	FROM CROSSWORD_DESCRIPTION CD;

    status = cword_rdb__check_status (SQLCA.SQLCODE);

    if (status == CWORD_S_NORMAL)
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICSUC,
	    TRACE3,
	    DECLARE_CURSOR_STR);
    }
    else
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICERR,
	    (TRACE1 | SCREEN1),
	    DECLARE_CURSOR_STR);
    }

    EXEC SQL OPEN XRD_CURSOR;

    status = cword_rdb__check_status (SQLCA.SQLCODE);

    if (status == CWORD_S_NORMAL)
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICSUC,
	    TRACE3,
	    OPEN_CURSOR_STR);
    }
    else
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICERR,
	    (TRACE1 | SCREEN1),
	    OPEN_CURSOR_STR);

	EXEC SQL ROLLBACK;

	return (CWORD_S_FAILURE);
    }

    /*
     *	This is the proper way to execute a cursor - IMO (Prove me wrong?).
     *
     *	Kick off the FETCH cursor.  Then have a while loop which will first
     *	check the return status of the previous FETCH and then execute another
     *	FETCH, if the previous one was OK.  Execute the while loop until the
     *	end of the stream or an error is recorded.
     */

    EXEC SQL FETCH XRD_CURSOR INTO
	:xrd_name_str;

    while (SQLCA.SQLCODE != SQL_EOS)
    {
	cword_word_truncate_str (xrd_name_str);

	cword_strll_add_str (
	    SORTED | UNIQUE,
	    xrd_name_str,
	    xrd_ll);

	if (*xrd_ll == (CWORD_T_STR_LL *) 0)
	{
	    EXEC SQL CLOSE XRD_CURSOR;

	    EXEC SQL ROLLBACK;

	    return (CWORD_S_FAILURE);
	}

	EXEC SQL FETCH XRD_CURSOR INTO
	    :xrd_name_str;
    }

    cword_messages_output_message (
	CWORD_RDBGENERICSUC,
	TRACE3,
	FETCH_CURSOR_STR);

    EXEC SQL CLOSE XRD_CURSOR;

    status = cword_rdb__check_status (SQLCA.SQLCODE);

    if (status == CWORD_S_NORMAL)
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICSUC,
	    TRACE3,
	    CLOSE_CURSOR_STR);
    }
    else
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICERR,
	    (TRACE1 | SCREEN1),
	    CLOSE_CURSOR_STR);
    }

    EXEC SQL ROLLBACK;

    return (CWORD_S_NORMAL);

}



PUBLIC int cword_rdb_add_grd (
	CROSSWORD_RECORD	*crossword_record)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Add a grid definition to the Cword database.
 *
 *  FORMAL PARAMETERS:
 *
 *	*crossword_record	(r) 
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL		Grid definition added.
 *	CWORD_S_FAILURE		Grid definition not added.
 *--
 */
{

    char temp_str[151];

    int status,
	count1,
	temp1_int,
	temp2_int,
	temp3_int,
	temp4_int,
	temp5_int;

    status = cword_rdb__set_transaction (
		WRITE_TABLE,
		GRD_OPERATION);

    if (CWORD_S_NORMAL != status)
    {
	return (status);
    }

    /*
     *  Make sure that a valid name is given to the crossword in question!
     *
     *	This is a bit non-intuitive but it's the best I can probably do given
     *	the constraints of the problem.
     */

    if (crossword_record->name[0] == (char) 0)
    {
	if (crossword_record->file_name[0] == (char) 0)
	{
	    strcpy (crossword_record->name, GRD_NO_NAME);
	}
	else
	{
	    strcpy (crossword_record->name, crossword_record->file_name);
	}

	printf (
	    "# Using grid name of %s since none provided in input definition\n",
	    crossword_record->name);
    }

    /* 
     *  temp_str same for all the INSERTS in this module.
     */

    strcpy (temp_str, crossword_record->name);

    temp1_int = crossword_record->clue_sum;
    temp2_int = crossword_record->x_dimensions;
    temp3_int = crossword_record->y_dimensions;
    temp4_int = crossword_record->grid_type;
    temp5_int = crossword_record->symmetry;

    EXEC SQL INSERT INTO GRID_DESCRIPTION
    (
	GRID_NAME,
	GRID_CLUES,
	GRID_LENGTH,
	GRID_WIDTH,
	GRID_TYPE,
	GRID_SYMMETRY
    )
    VALUES
    (
	:temp_str,
	:temp1_int,
	:temp2_int,
	:temp3_int,
	:temp4_int,
	:temp5_int
    );

    status = cword_rdb__check_status (SQLCA.SQLCODE);

    if (CWORD_S_NORMAL == status)
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICSUC,
	    TRACE3,
	    INSERT_STR,
	    GRID_DESC_STR);
    }
    else if (CWORD_RDBINVINDEX == status)
    {
	cword_messages_output_message (
	    CWORD_S_RDBDUPLICDEFN,
	    (TRACE0 | SCREEN0),
	    GRID_STR,
	    temp_str);

	EXEC SQL ROLLBACK;

	return (CWORD_S_FAILURE);
    }
    else
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICERR,
	    (TRACE1 | SCREEN1),
	    INSERT_STR,
	    GRID_DESC_STR);

	EXEC SQL ROLLBACK;

	return (CWORD_S_FAILURE);
    }

    cword_messages_output_message (
	CWORD_S_RDBADDGRIDSUC,
	(TRACE2 | SCREEN0));

    for (count1 = 0; count1 < crossword_record->clue_sum; count1++)
    {
	temp1_int = crossword_record->clue_details[count1].clue_direction;
	temp2_int = crossword_record->clue_details[count1].x_coordinate;
	temp3_int = crossword_record->clue_details[count1].y_coordinate;
	temp4_int = crossword_record->clue_details[count1].answer_length;

	EXEC SQL INSERT INTO LIGHT_DESCRIPTION
	(
	    GRID_NAME,
	    LIGHT_DIRECTION,
	    LIGHT_X_COORD,
	    LIGHT_Y_COORD,
	    LIGHT_LENGTH
	)
	VALUES	
	(
	    :temp_str,
	    :temp1_int,
	    :temp2_int,
	    :temp3_int,
	    :temp4_int
	);

	status = cword_rdb__check_status (SQLCA.SQLCODE);

	if (status != CWORD_S_NORMAL)
	{
	    cword_messages_output_message (
		CWORD_RDBGENERICERR,
		(TRACE3 | SCREEN3),
		INSERT_STR);

	    cword_messages_output_message (
		CWORD_S_RDBDUPDEFN,
		(TRACE1 | SCREEN1),
		GRID_STR);

	    EXEC SQL ROLLBACK;

	    return (CWORD_S_FAILURE);
	}
    }

    cword_messages_output_message (
	CWORD_RDBGENERICSUC,
	TRACE3,
	INSERT_STR,
	LIGHT_DESC_STR);

    cword_messages_output_message (
	CWORD_S_RDBADDLIGHTSUC,
	(TRACE2 | SCREEN0));

    EXEC SQL COMMIT;

    status = cword_rdb__check_status (SQLCA.SQLCODE);

    if (status == CWORD_S_NORMAL)
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICSUC,
	    TRACE3,
	    COMMIT_STR);
    }
    else
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICERR,
	    (TRACE1 | SCREEN1),
	    COMMIT_STR);

	EXEC SQL ROLLBACK;

	return (CWORD_S_FAILURE);
    }

    return (CWORD_S_NORMAL);

}



/*
 *  ?? Option on the UI to select a grid from the database and continue filling ??
 */

PUBLIC int cword_rdb_get_grd (
	char			*grd_name_str,
	CROSSWORD_RECORD	*crossword_record)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function to get grid definition from the database given its name
 *	*grd_name_str.
 *
 *  FORMAL PARAMETERS:
 *
 *	*grd_name_str		(r) Name of grid to grab from the database.
 *	*crossword_record	(w) Insert grid definition in here.
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL		grid definition found.
 *	CWORD_S_FAILURE		grid definition not found.
 *--
 */

{

    int status,
	count1,
	temp1_int,
	temp2_int,
	temp3_int,
	temp4_int,
	temp5_int;

    if (grd_name_str[0] == (char) 0)
    {
	printf ("# Error inputting valid grid name\n");

	return (CWORD_S_FAILURE); 
    }

    status = cword_rdb__set_transaction (
		READ_TABLE,
		GRD_OPERATION);

    if (CWORD_S_NORMAL != status)
    {
	return (status);
    }

    /*
     *  Getting the grid definition/statistics.
     */

    EXEC SQL DECLARE GRID_CURSOR CURSOR FOR
	SELECT
	    GRID.GRID_CLUES,
	    GRID.GRID_LENGTH,
	    GRID.GRID_WIDTH,
	    GRID.GRID_TYPE,
	    GRID.GRID_SYMMETRY
	FROM GRID_DESCRIPTION GRID
	WHERE
	    GRID.GRID_NAME = :grd_name_str;

    status = cword_rdb__check_status (SQLCA.SQLCODE);

    if (status == CWORD_S_NORMAL)
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICSUC,
	    TRACE3,
	    DECLARE_CURSOR_STR);
    }
    else
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICERR,
	    (TRACE3 | SCREEN3),
	    DECLARE_CURSOR_STR);

	EXEC SQL ROLLBACK;

	memset (crossword_record, 0, sizeof (CROSSWORD_RECORD));

	return (CWORD_S_RDBENDOFSTREAM);
    }

    EXEC SQL OPEN GRID_CURSOR;

    status = cword_rdb__check_status (SQLCA.SQLCODE);

    if (status == CWORD_S_NORMAL)
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICSUC,
	    TRACE3,
	    OPEN_CURSOR_STR);
    }
    else
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICERR,
	    (TRACE1 | SCREEN1),
	    OPEN_CURSOR_STR);

	EXEC SQL ROLLBACK;

	memset (crossword_record, 0, sizeof (CROSSWORD_RECORD));

	return (CWORD_S_RDBENDOFSTREAM);
    }

    EXEC SQL FETCH GRID_CURSOR INTO 
	:temp1_int,
	:temp2_int,
	:temp3_int,
	:temp4_int,
	:temp5_int;

 status = cword_rdb__check_status (SQLCA.SQLCODE);

    if (status == CWORD_S_NORMAL)
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICSUC,
	    TRACE3,
	    FETCH_CURSOR_STR);
    }
    else if (status == CWORD_S_RDBENDOFSTREAM)
    {
	cword_messages_output_message ( 
	    CWORD_S_RDBNOTINDB,
	    (TRACE1 | SCREEN1),
	    CROSSWORD_STR,
	    grd_name_str);

	EXEC SQL CLOSE GRID_CURSOR;

	EXEC SQL ROLLBACK;

	memset (crossword_record, 0, sizeof (CROSSWORD_RECORD));

	return (CWORD_S_RDBENDOFSTREAM);
    }
    else
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICERR,
	    (TRACE1 | SCREEN1),
	    FETCH_CURSOR_STR);

	EXEC SQL CLOSE GRID_CURSOR;

	EXEC SQL ROLLBACK;

	memset (crossword_record, 0, sizeof (CROSSWORD_RECORD));

	return (CWORD_S_RDBENDOFSTREAM);
    }

    strcpy (crossword_record->name, grd_name_str);

    crossword_record->clue_sum	    = temp1_int;
    crossword_record->x_dimensions  = temp2_int;
    crossword_record->y_dimensions  = temp3_int;
    crossword_record->grid_type     = temp4_int;
    crossword_record->symmetry      = temp5_int;

    EXEC SQL CLOSE GRID_CURSOR;

    status = cword_rdb__check_status (SQLCA.SQLCODE);

    if (status == CWORD_S_NORMAL)
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICSUC,
	    TRACE3,
	    CLOSE_CURSOR_STR);
    }
    else
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICERR,
	    (TRACE1 | SCREEN1),
	    CLOSE_CURSOR_STR);

	EXEC SQL ROLLBACK;

	memset (crossword_record, 0, sizeof (CROSSWORD_RECORD));

	return (CWORD_S_RDBENDOFSTREAM);
    }

    /*
     *  Getting the light definitions for the grid now.
     */

    EXEC SQL DECLARE LIGHTS_CURSOR CURSOR FOR
	SELECT
	    LIGHTS.LIGHT_DIRECTION,
	    LIGHTS.LIGHT_X_COORD,
	    LIGHTS.LIGHT_Y_COORD,
	    LIGHTS.LIGHT_LENGTH
	FROM LIGHT_DESCRIPTION LIGHTS
	WHERE
	    LIGHTS.GRID_NAME = :grd_name_str
	ORDER BY
	    LIGHTS.LIGHT_DIRECTION DESC,
	    LIGHTS.LIGHT_X_COORD DESC,
	    LIGHTS.LIGHT_Y_COORD ASC;

    status = cword_rdb__check_status (SQLCA.SQLCODE);

    if (status == CWORD_S_NORMAL)
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICSUC,
	    TRACE3,
	    DECLARE_CURSOR_STR);
    }
    else
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICERR,
	    (TRACE1 | SCREEN1),
	    DECLARE_CURSOR_STR);

	EXEC SQL ROLLBACK;

	memset (crossword_record, 0, sizeof (CROSSWORD_RECORD));

	return (CWORD_S_RDBENDOFSTREAM);
    }
    
    EXEC SQL OPEN LIGHTS_CURSOR;

    status = cword_rdb__check_status (SQLCA.SQLCODE);

    if (status == CWORD_S_NORMAL)
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICSUC,
	    TRACE3,
	    OPEN_CURSOR_STR);
    }
    else
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICERR,
	    (TRACE1 | SCREEN1),
	    OPEN_CURSOR_STR);

	EXEC SQL ROLLBACK;

	memset (crossword_record, 0, sizeof (CROSSWORD_RECORD));

	return (CWORD_S_RDBENDOFSTREAM);
    }

    EXEC SQL FETCH LIGHTS_CURSOR INTO
	:temp1_int,
	:temp2_int,
	:temp3_int,
	:temp4_int;

    count1 = 0;
    while (SQLCA.SQLCODE != SQL_EOS)
    {
	crossword_record->clue_details[count1].clue_direction	= temp1_int;
	crossword_record->clue_details[count1].x_coordinate	= temp2_int;
	crossword_record->clue_details[count1].y_coordinate	= temp3_int;
	crossword_record->clue_details[count1].answer_length    = temp4_int;

	count1++;
	
	EXEC SQL FETCH LIGHTS_CURSOR INTO
	    :temp1_int,
	    :temp2_int,
	    :temp3_int,
	    :temp4_int;
    }

    if (count1 != crossword_record->clue_sum)
    {
	printf (
	    "# Major error: number of light found in database != clue_sum from definition - (%d != %d)\n",
	    count1,
	    crossword_record->clue_sum);

	memset (crossword_record, 0, sizeof (CROSSWORD_RECORD));

	EXEC SQL CLOSE LIGHTS_CURSOR;

	EXEC SQL ROLLBACK;

	return (CWORD_S_FAILURE);
    }

    cword_messages_output_message (
	CWORD_RDBGENERICSUC,
	TRACE3,
	FETCH_CURSOR_STR);

    EXEC SQL CLOSE LIGHTS_CURSOR;

    status = cword_rdb__check_status (SQLCA.SQLCODE);

    if (status == CWORD_S_NORMAL)
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICSUC,
	    TRACE3,
	    CLOSE_CURSOR_STR);
    }
    else
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICERR,
	    (TRACE1 | SCREEN1),
	    CLOSE_CURSOR_STR);

	EXEC SQL ROLLBACK;

	memset (crossword_record, 0, sizeof (CROSSWORD_RECORD));

	return (CWORD_S_RDBENDOFSTREAM);
    }

    /*
     *  Sort the output crossword definition before going any further.
     */

    cword_utils_sort_crossword (crossword_record);

    EXEC SQL ROLLBACK;

    return (CWORD_S_NORMAL);

}



PUBLIC int cword_rdb_delete_grd (
	char			*grd_name_str)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Delete a grid definition called *grd_name_str from the Cword database.
 *
 *  FORMAL PARAMETERS:
 *
 *	*grd_name_str		(r)
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL		Grid definition deleted.
 *	CWORD_S_FAILURE		Grid definition not deleted.
 *--
 */
{

    int status;

    status = cword_rdb__set_transaction (
		WRITE_TABLE,
		GRD_OPERATION);

    if (CWORD_S_NORMAL != status)
    {
	return (status);
    }

    /*
     *	By deleting this column it will set off a trigger in the database that
     *	will delete the corresponding LIGHT_DESCRIPTION columns associated with
     *	the crossword.
     */

    EXEC SQL DELETE FROM GRID_DESCRIPTION WHERE
	GRID_NAME = :grd_name_str;

    status = cword_rdb__check_status (SQLCA.SQLCODE);

    if (status == CWORD_S_NORMAL)
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICSUC,
	    TRACE3,
	    DELETE_STR,
	    GRID_DESC_STR);
    }
    /* 
     *  Needed to cover possibility of crossword not being in DB.
     */
    else if (status == CWORD_S_RDBENDOFSTREAM)
    {
	cword_messages_output_message ( 
	    CWORD_S_RDBNOTINDB,
	    (TRACE2 | SCREEN0),
	    GRID_STR,
	    grd_name_str);

	EXEC SQL ROLLBACK;

	return (CWORD_S_FAILURE);
    }
    else
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICERR,
	    (TRACE1 | SCREEN1),
	    DELETE_STR,
	    GRID_DESC_STR);

	EXEC SQL ROLLBACK;

	return (CWORD_S_FAILURE);
    }

    EXEC SQL COMMIT;

    status = cword_rdb__check_status (SQLCA.SQLCODE);

    if (status == CWORD_S_NORMAL)
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICSUC,
	    TRACE1,
	    COMMIT_STR);
    }
    else
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICERR,
	    (TRACE3 | SCREEN3),
	    COMMIT_STR);

	EXEC SQL ROLLBACK;

	return (CWORD_S_FAILURE);
    }

    return (CWORD_S_NORMAL);

}



PUBLIC int cword_rdb_get_grd_ll (
	CWORD_T_STR_LL		**grd_ll)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function to search the Cword database looking for all valid grid
 *	definitions.  Returns a list via **grd_ll
 *
 *  FORMAL PARAMETERS:
 *
 *      **grd_ll		(w) Link list of returned grid definition names
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL		Normal Completion
 *      Rdb error 		Error getting crossword definition names
 *--
 */

{

    char grd_name_str[151];

    int status;

    status = cword_rdb__set_transaction (
		READ_TABLE,
		GRD_OPERATION);

    if (CWORD_S_NORMAL != status)
    {
	return (status);
    }

    *grd_ll = (CWORD_T_STR_LL *) 0;

    EXEC SQL DECLARE GRD_CURSOR CURSOR FOR
	SELECT
	    G.GRID_NAME
	FROM GRID_DESCRIPTION G;

    status = cword_rdb__check_status (SQLCA.SQLCODE);

    if (status == CWORD_S_NORMAL)
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICSUC,
	    TRACE3,
	    DECLARE_CURSOR_STR);
    }
    else
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICERR,
	    (TRACE1 | SCREEN1),
	    DECLARE_CURSOR_STR);
    }

    EXEC SQL OPEN GRD_CURSOR;

    status = cword_rdb__check_status (SQLCA.SQLCODE);

    if (status == CWORD_S_NORMAL)
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICSUC,
	    TRACE3,
	    OPEN_CURSOR_STR);
    }
    else
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICERR,
	    (TRACE1 | SCREEN1),
	    OPEN_CURSOR_STR);

	EXEC SQL ROLLBACK;

	return (CWORD_S_FAILURE);
    }

    /*
     *	This is the proper way to execute a cursor - IMO (Prove me wrong?).
     *
     *	Kick off the FETCH cursor.  Then have a while loop which will first
     *	check the return status of the previous FETCH and then execute another
     *	FETCH, if the previous one was OK.  Execute the while loop until the
     *	end of the stream or an error is recorded.
     */

    EXEC SQL FETCH GRD_CURSOR INTO
	:grd_name_str;

    while (SQLCA.SQLCODE != SQL_EOS)
    {
	cword_word_truncate_str (grd_name_str);

	cword_strll_add_str (
	    SORTED | UNIQUE,
	    grd_name_str,
	    grd_ll);

	if (*grd_ll == (CWORD_T_STR_LL *) 0)
	{
	    EXEC SQL CLOSE GRD_CURSOR;

	    EXEC SQL ROLLBACK;

	    return (CWORD_S_FAILURE);
	}

	EXEC SQL FETCH GRD_CURSOR INTO
	    :grd_name_str;
    }

    cword_messages_output_message (
	CWORD_RDBGENERICSUC,
	TRACE3,
	FETCH_CURSOR_STR);

    EXEC SQL CLOSE GRD_CURSOR;

    status = cword_rdb__check_status (SQLCA.SQLCODE);

    if (status == CWORD_S_NORMAL)
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICSUC,
	    TRACE3,
	    CLOSE_CURSOR_STR);
    }
    else
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICERR,
	    (TRACE1 | SCREEN1),
	    CLOSE_CURSOR_STR);
    }

    EXEC SQL ROLLBACK;

    return (CWORD_S_NORMAL);

}



PUBLIC int cword_rdb_get_anagram_ll (
	CWORD_T_STR_LL		**anagram_ll)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function to get a list of all available anagram indicators from the
 *	database.
 *
 *  FORMAL PARAMETERS:
 *
 *      **anagram_ll		(w) link list containing return anagram indicators
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL		Normal Completion
 *      Rdb error		Error
 *--
 */

{

    char anagram[151];

    int	status;

    status = cword_rdb__set_transaction (
		READ_TABLE,
		ANAGRAM_DESCRIPTION);

    if (CWORD_S_NORMAL != status)
    {
	return (status);
    }

    cword_messages_output_message (
	    CWORD_RDBGETANADESC,
	    (TRACE2 | SCREEN0));

    *anagram_ll = (CWORD_T_STR_LL *) 0;

    EXEC SQL DECLARE ANAGRAM_CURSOR CURSOR FOR
	SELECT
	    A.A_DESCRIPTION
	FROM ANAGRAM_DESCRIPTION A;

    status = cword_rdb__check_status (SQLCA.SQLCODE);

    if (status == CWORD_S_NORMAL)
    {
	cword_messages_output_message (
		CWORD_RDBGENERICSUC,
		TRACE3,
		DECLARE_CURSOR_STR);
    }
    else
    {
	cword_messages_output_message (
		CWORD_RDBGENERICERR,
		(TRACE1 | SCREEN1),
		DECLARE_CURSOR_STR);

	EXEC SQL ROLLBACK;

	return (CWORD_S_FAILURE);
    }

    EXEC SQL OPEN ANAGRAM_CURSOR;

    status = cword_rdb__check_status (SQLCA.SQLCODE);

    if (status == CWORD_S_NORMAL)
    {
	cword_messages_output_message (
		CWORD_RDBGENERICSUC,
		TRACE3,
		OPEN_CURSOR_STR);
    }
    else
    {
	cword_messages_output_message (
		CWORD_RDBGENERICERR,
		(TRACE1 | SCREEN1),
		OPEN_CURSOR_STR);

	EXEC SQL ROLLBACK;

	return (CWORD_S_FAILURE);
    }

    EXEC SQL FETCH ANAGRAM_CURSOR INTO
	:anagram;

    while (SQLCA.SQLCODE != SQL_EOS)
    {
	cword_word_truncate_str (anagram);

	cword_strll_add_str (
	    SORTED | UNIQUE,
	    anagram,
	    anagram_ll);

	if (*anagram_ll == (CWORD_T_STR_LL *) 0)
	{
	    EXEC SQL CLOSE ANAGRAM_CURSOR;

            EXEC SQL ROLLBACK;

	    return (CWORD_S_FAILURE);
	}

	EXEC SQL FETCH ANAGRAM_CURSOR INTO
	    :anagram;
    }

    cword_messages_output_message (
	    CWORD_RDBGENERICSUC,
	    TRACE3,
	    FETCH_CURSOR_STR);

    EXEC SQL CLOSE ANAGRAM_CURSOR;

    status = cword_rdb__check_status (SQLCA.SQLCODE);

    if (status == CWORD_S_NORMAL)
    {
	cword_messages_output_message (
		CWORD_RDBGENERICSUC,
		TRACE3,
		CLOSE_CURSOR_STR);
    }
    else
    {
	cword_messages_output_message (
		CWORD_RDBGENERICERR,
		(TRACE1 | SCREEN1),
		CLOSE_CURSOR_STR);

	EXEC SQL ROLLBACK;

	return (CWORD_S_FAILURE);
    }

    return (CWORD_S_NORMAL);

}



PUBLIC int cword_rdb_get_homophone_ll (
	CWORD_T_STR_LL		**homophone_ll)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function to return to the user a list of the homophones in the
 *	database.
 *
 *  FORMAL PARAMETERS:
 *
 *      **homophone_ll	(w) link list containing homophones found in DB
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL		Normal Completion
 *      Rdb Error		Error
 *--
 */

{

    char indicator[151],
	 description[151];

    int	status;

    status = cword_rdb__set_transaction (
		READ_TABLE,
		HOMOPHONE_DESCRIPTION);

    if (CWORD_S_NORMAL != status)
    {
	return (status);
    }

    cword_messages_output_message (
	    CWORD_RDBGETHOMODESC,
	    (TRACE2 | SCREEN0));

    *homophone_ll = (CWORD_T_STR_LL *) 0;

    EXEC SQL DECLARE HOMOPHONE_CURSOR CURSOR FOR
	SELECT
	    H.H_INDICATOR,
	    H.H_DESCRIPTION
	FROM HOMOPHONE_DESCRIPTION H;

    status = cword_rdb__check_status (SQLCA.SQLCODE);

    if (status == CWORD_S_NORMAL)
    {
	cword_messages_output_message (
		CWORD_RDBGENERICSUC,
		TRACE3,
		DECLARE_CURSOR_STR);
    }
    else
    {
	cword_messages_output_message (
		CWORD_RDBGENERICERR,
		(TRACE1 | SCREEN1),
		DECLARE_CURSOR_STR);

	EXEC SQL ROLLBACK;

	return (CWORD_S_FAILURE);
    }

    EXEC SQL OPEN HOMOPHONE_CURSOR;

    status = cword_rdb__check_status (SQLCA.SQLCODE);

    if (status == CWORD_S_NORMAL)
    {
	cword_messages_output_message (
		CWORD_RDBGENERICSUC,
		TRACE3,
		OPEN_CURSOR_STR);
    }
    else
    {
	cword_messages_output_message (
		CWORD_RDBGENERICERR,
		(TRACE1 | SCREEN1),
		OPEN_CURSOR_STR);

	EXEC SQL ROLLBACK;

	return (CWORD_S_FAILURE);
    }

    EXEC SQL FETCH HOMOPHONE_CURSOR INTO
	:indicator,
	:description;

    while (SQLCA.SQLCODE != SQL_EOS)
    {
	cword_word_truncate_str (indicator);
	cword_word_truncate_str (description);

	cword_strll_add_str (
	    SORTED | UNIQUE,
	    indicator,
	    homophone_ll);

	if (*homophone_ll == (CWORD_T_STR_LL *) 0)
	{
	    EXEC SQL CLOSE HOMOPHONE_CURSOR;

            EXEC SQL ROLLBACK;

	    return (CWORD_S_FAILURE);
	}

	cword_strll_add_str (
	    SORTED | UNIQUE,
	    description,
	    homophone_ll);

	if (*homophone_ll == (CWORD_T_STR_LL *) 0)
	{
	    EXEC SQL CLOSE HOMOPHONE_CURSOR;

            EXEC SQL ROLLBACK;

	    return (CWORD_S_FAILURE);
	}

	EXEC SQL FETCH HOMOPHONE_CURSOR INTO
	    :indicator,
	    :description;
    }

    cword_messages_output_message (
	    CWORD_RDBGENERICSUC,
	    TRACE3,
	    FETCH_CURSOR_STR);

    EXEC SQL CLOSE HOMOPHONE_CURSOR;

    status = cword_rdb__check_status (SQLCA.SQLCODE);

    if (status == CWORD_S_NORMAL)
    {
	cword_messages_output_message (
		CWORD_RDBGENERICSUC,
		TRACE3,
		CLOSE_CURSOR_STR);
    }
    else
    {
	cword_messages_output_message (
		CWORD_RDBGENERICERR,
		(TRACE1 | SCREEN1),
		CLOSE_CURSOR_STR);

	EXEC SQL ROLLBACK;

	return (CWORD_S_FAILURE);
    }

    return (CWORD_S_NORMAL);

}



PUBLIC int cword_rdb_get_synonym_ll (
	CWORD_T_STR_LL		**synonym_ll)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function to return to the user a list of synonyms found in the
 *	database.
 *
 *  FORMAL PARAMETERS:
 *
 *      **synonym_ll		(w) link list containing synonyms returned from database.
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL		Normal Completion
 *      Rdb error		Error
 *--
 */

{

    char indicator[151],
	 description[151];

    int status;

    status = cword_rdb__set_transaction (
		READ_TABLE,
		SYNONYM_DESCRIPTION);

    if (CWORD_S_NORMAL != status)
    {
	return (status);
    }

    cword_messages_output_message (
	    CWORD_RDBGETSYMDESC,
	    (TRACE2 | SCREEN0));

    *synonym_ll = (CWORD_T_STR_LL *) 0;

    EXEC SQL DECLARE SYNONYM_CURSOR CURSOR FOR
	SELECT
	    S.S_INDICATOR,
	    S.S_DESCRIPTION
	FROM SYNONYM_DESCRIPTION S;

    status = cword_rdb__check_status (SQLCA.SQLCODE);

    if (status == CWORD_S_NORMAL)
    {
	cword_messages_output_message (
		CWORD_RDBGENERICSUC,
		TRACE3,
		DECLARE_CURSOR_STR);
    }
    else
    {
	cword_messages_output_message (
		CWORD_RDBGENERICERR,
		(TRACE1 | SCREEN1),
		DECLARE_CURSOR_STR);

	EXEC SQL ROLLBACK;

	return (CWORD_S_FAILURE);
    }

    EXEC SQL OPEN SYNONYM_CURSOR;

    status = cword_rdb__check_status (SQLCA.SQLCODE);

    if (status == CWORD_S_NORMAL)
    {
	cword_messages_output_message (
		CWORD_RDBGENERICSUC,
		TRACE3,
		OPEN_CURSOR_STR);
    }
    else
    {
	cword_messages_output_message (
		CWORD_RDBGENERICERR,
		(TRACE1 | SCREEN1),
		OPEN_CURSOR_STR);

	EXEC SQL ROLLBACK;

	return (CWORD_S_FAILURE);
    }

    EXEC SQL FETCH SYNONYM_CURSOR INTO
	:indicator,
	:description;

    while (SQLCA.SQLCODE != SQL_EOS)
    {
	cword_word_truncate_str (indicator);
	cword_word_truncate_str (description);

	cword_strll_add_str (
	    SORTED | UNIQUE,
	    indicator,
	    synonym_ll);

	if (*synonym_ll == (CWORD_T_STR_LL *) 0)
	{
	    EXEC SQL CLOSE SYNONYM_CURSOR;

            EXEC SQL ROLLBACK;

	    return (CWORD_S_FAILURE);
	}

	cword_strll_add_str (
	    SORTED | UNIQUE,
	    description,
	    synonym_ll);

	if (*synonym_ll == (CWORD_T_STR_LL *) 0)
	{
	    EXEC SQL CLOSE SYNONYM_CURSOR;

            EXEC SQL ROLLBACK;

	    return (CWORD_S_FAILURE);
	}

	EXEC SQL FETCH SYNONYM_CURSOR INTO
	    :indicator,
	    :description;
    }

    cword_messages_output_message (
	    CWORD_RDBGENERICSUC,
	    TRACE3,
	    FETCH_CURSOR_STR);

    EXEC SQL CLOSE SYNONYM_CURSOR;

    status = cword_rdb__check_status (SQLCA.SQLCODE);

    if (status == CWORD_S_NORMAL)
    {
	cword_messages_output_message (
		CWORD_RDBGENERICSUC,
		TRACE3,
		CLOSE_CURSOR_STR);
    }
    else
    {
	cword_messages_output_message (
		CWORD_RDBGENERICERR,
		(TRACE1 | SCREEN1),
		CLOSE_CURSOR_STR);

	EXEC SQL ROLLBACK;

	return (CWORD_S_FAILURE);
    }

    return (CWORD_S_NORMAL);

}



/*
 *  ?? Flat file clue database could also be incorporated for those that don't
 *  have an SQL/rdb backend database - certainly will be kept in mind for the
 *  next version ??
 */

/*
 *  ?? Check the format of the clue before accepting it e.g. the answer is not
 *  in the question and that the light details are included in the clue.  Use
 *  CLUE.C ??
 */

/*
 *  ?? Make sure that the inputted clue combo doesn't already exist in the
 *  database before adding it - new function to check light_str/clue_str ??
 */

PUBLIC int cword_rdb_add_clue (
	char			*light_str,
	char			*clue_str)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Simple function to add clue/answer combination to the Cword database.
 *
 *  FORMAL PARAMETERS:
 *
 *	*light_str		(r) light to insert in database.
 *	*clue_str		(r) clue to insert in database.
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL		Clue/light inserted successfully.
 *	CWORD_S_FAILURE		Error adding clue/light.
 *--
 */

{

    /*
     *  ?? Is this correct programming practice ??
     */

    char *temp_str = UNDEFINED_XRD;

    int	status;

    if ( (light_str[0] == (char) 0) || (clue_str[0] == (char) 0) )
    {
	printf ("# Error: clue or light is NULL\n");

	return (CWORD_S_FAILURE);
    }

    cword_word_lower_str (light_str);

    /*
     *	This functionality cannot be accomplished in the database itself
     *	because the resulting unique index is too large fopr stogage.  This
     *	function is the next best thing.
     */

    if (cword_rdb_is_clue_defined (
	    light_str,
	    clue_str) == TRUE)
    {
	cword_messages_output_message (
	    CWORD_S_CLUEALREADYINDB,
	    (TRACE1 | SCREEN1));

	return (CWORD_S_FAILURE);
    }

    status = cword_rdb__set_transaction (
		WRITE_TABLE,
		CLUE_DESCRIPTION);

    if (CWORD_S_NORMAL != status)
    {
	return (CWORD_S_FAILURE);
    }

    /*
     *  Add clue with the crossword UNDEFINED_XRD!
     */

    EXEC SQL INSERT INTO CLUE_DESCRIPTION
    (
	CROSSWORD_NAME,
	CLUE_NUMBER,
	CLUE_DIRECTION,
	CLUE_X_COORD,
	CLUE_Y_COORD,
	CLUE_QUESTION,
	CLUE_ANSWER
    )
    VALUES	
    (
	:temp_str,
	1, /* ?? #define# */
	1, /* ?? #define# */
	1, /* ?? #define# */
	1, /* ?? #define# */
	:clue_str,
	:light_str
    );

    status = cword_rdb__check_status (SQLCA.SQLCODE);

    if (status != CWORD_S_NORMAL)
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICERR,
	    (TRACE1 | SCREEN1),
	    INSERT_STR,
	    CLUE_DESC_STR);

	cword_messages_output_message (
	    CWORD_S_RDBPUTANSERR,
	    (TRACE1 | SCREEN1),
	    clue_str);

	EXEC SQL ROLLBACK;

	return (CWORD_S_FAILURE);
    }

    EXEC SQL COMMIT;

    status = cword_rdb__check_status (SQLCA.SQLCODE);

    if (status != CWORD_S_NORMAL)
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICERR,
	    (TRACE1 | SCREEN1),
	    COMMIT_STR);

	EXEC SQL ROLLBACK;

	return (CWORD_S_FAILURE);
    }

    return (CWORD_S_NORMAL);

}



/*
 *  ?? What happens if there are multiple clues in the database for the
 *  inputted light ??
 */

PUBLIC int cword_rdb_get_clue (
	char			*light_str,
	char			*clue_str)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Query the database for clue(s) associated with the light *light_str.
 *	The first valid clue found is returned in *clue_str.
 *
 *  FORMAL PARAMETERS:
 *
 *	*light_str		(r) Light to search for in db
 *	*clue_str		(w) String to put clue when found
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL		Successfully found clue for inputted light.
 *	CWORD_S_FAILURE		Error finding clue for inputted light.
 *	Other			Database problem.
 *--
 */

{

    int status;

    if (light_str[0] == (char) 0)
    {
	return (CWORD_S_FAILURE);
    }

    clue_str[0] = (char) 0;

    status = cword_rdb__set_transaction (
		READ_TABLE,
		CLUE_DESCRIPTION);

    if (CWORD_S_NORMAL != status)
    {
	return (status);
    }

    cword_word_lower_str (light_str);

    EXEC SQL DECLARE CLUE_CURSOR CURSOR FOR
	SELECT
	    CLUES.CLUE_QUESTION
	FROM CLUE_DESCRIPTION CLUES
	WHERE CLUES.CLUE_ANSWER = :light_str
	ORDER BY CLUES.CLUE_ANSWER ASC;

    status = cword_rdb__check_status (SQLCA.SQLCODE);

    if (status != CWORD_S_NORMAL)
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICERR,
	    (TRACE1 | SCREEN1),
	    DECLARE_CURSOR_STR);

	EXEC SQL ROLLBACK;

	return (CWORD_S_RDBENDOFSTREAM);
    }

    EXEC SQL OPEN CLUE_CURSOR;

    status = cword_rdb__check_status (SQLCA.SQLCODE);

    if (status != CWORD_S_NORMAL)
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICERR,
	    (TRACE1 | SCREEN1),
	    OPEN_CURSOR_STR);

	EXEC SQL ROLLBACK;

	return (CWORD_S_RDBENDOFSTREAM);
    }

    /*
     *	?? Only returns the first instance of a clue string found to match
     *	CURSOR criteria.  This should change in the future ??
     */

    EXEC SQL FETCH CLUE_CURSOR INTO
	:clue_str;

    /*
     *  If the cursor is empty then return failure - no match made.
     */

    status = cword_rdb__check_status (SQLCA.SQLCODE);

    if (status != CWORD_S_NORMAL)
    {
	EXEC SQL CLOSE CLUE_CURSOR;

	EXEC SQL ROLLBACK;

	return (CWORD_S_FAILURE);
    }

    cword_word_truncate_str (clue_str);

    EXEC SQL CLOSE CLUE_CURSOR;

    status = cword_rdb__check_status (SQLCA.SQLCODE);

    if (status != CWORD_S_NORMAL)
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICERR,
	    (TRACE1 | SCREEN1),
	    CLOSE_CURSOR_STR);

	EXEC SQL ROLLBACK;

	return (CWORD_S_FAILURE);
    }

    EXEC SQL ROLLBACK;

    return (CWORD_S_NORMAL);

}



PUBLIC int cword_rdb_delete_clue (
	char			*light_str)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function to search the Cword database for a clue with the answer
 *	light_str.  Once found, it is deleted providing it is a clue that was
 *	added separately, and is not part of a crossword definition that is
 *	still in the database.
 *
 *  FORMAL PARAMETERS:
 *
 *	*light_str		(r) light to delete in the db.
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL		Successfully found clue for inputted light.
 *	CWORD_S_FAILURE		Error finfing clue for inputted light.
 *	Other			Database problem.
 *--
 */

{

    char *temp_str = UNDEFINED_XRD;

    int status;

    if (light_str[0] == (char) 0)
    {
	return (CWORD_S_FAILURE);
    }

    status = cword_rdb__set_transaction (
		WRITE_TABLE,
		CLUE_DESCRIPTION);

    if (CWORD_S_NORMAL != status)
    {
	return (status);
    }

    cword_word_lower_str (light_str);

    /*
     *  Delete a single clue (of crossword UNDEFINED_XRD) if 
     *  CLUE_ANSWER = :light_str.
     */

    EXEC SQL DELETE FROM CLUE_DESCRIPTION
	WHERE CLUE_ANSWER = :light_str AND
	    CROSSWORD_NAME = :temp_str;
/*    	LIMIT TO 1 ROW; */

    status = cword_rdb__check_status (SQLCA.SQLCODE);

    if (status == CWORD_S_NORMAL)
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICSUC,
	    TRACE3,
	    DECLARE_CURSOR_STR);
    }
    /*
     *	The cursor was not successful since the inputted clue wasn't found in
     *	the database relation.
     */
    else if (status == CWORD_S_RDBENDOFSTREAM)
    {
	EXEC SQL ROLLBACK;

	return (CWORD_S_FAILURE);
    }
    else
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICERR,
	    (TRACE1 | SCREEN1),
	    DECLARE_CURSOR_STR);

	EXEC SQL ROLLBACK;

	return (CWORD_S_RDBENDOFSTREAM);
    }

    EXEC SQL COMMIT;

    status = cword_rdb__check_status (SQLCA.SQLCODE);

    if (status != CWORD_S_NORMAL)
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICERR,
	    (TRACE1 | SCREEN1),
	    COMMIT_STR);

	EXEC SQL ROLLBACK;

	return (CWORD_S_RDBENDOFSTREAM);
    }

    return (CWORD_S_NORMAL);

}



/*
 *  ?? How to check an error condition during the execution of the SQL ??
 */

PUBLIC BOOLEAN cword_rdb_is_clue_defined (
	char			*light_str,
	char			*clue_str)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *      [@tbs@]
 *
 *  FORMAL PARAMETERS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      TRUE			Inputted clue not found in the database!
 *	FALSE			Inputted clue found in the database!
 *--
 */

{

    char *temp_str = UNDEFINED_XRD;

    long status;

    int count;

    if (light_str[0] == (char) 0)
    {
	return (CWORD_S_FAILURE);
    }

    status = cword_rdb__set_transaction (
		READ_TABLE,
		CLUE_DESCRIPTION);

    if (CWORD_S_NORMAL != status)
    {
	return (status);
    }

    cword_word_lower_str (light_str);

    count = -1;

    EXEC SQL SELECT COUNT(*) INTO :count
	FROM CLUE_DESCRIPTION
	    WHERE CLUE_QUESTION = :clue_str AND
		  CLUE_ANSWER = :light_str AND
		  CROSSWORD_NAME = :temp_str;

    DEBUG_TRACE (
	"Count(*) is %d for clue: \"%s\" and light: \"%s\" and crossword: \"%s\"\n",
	count,
	clue_str,
	light_str,
	temp_str);

    status = cword_rdb__check_status (SQLCA.SQLCODE);

    if (status == CWORD_S_NORMAL)
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICSUC,
	    TRACE3,
	    COUNTSTAR_STR);
    }
    else
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICERR,
	    (TRACE1 | SCREEN1),
	    COUNTSTAR_STR);

	EXEC SQL ROLLBACK;

	return (CWORD_S_RDBENDOFSTREAM);
    }

    EXEC SQL ROLLBACK;

    /*
     *  Don't bother checking SQLCODE!
     */

    if (count == 0)
    {
	return (FALSE);
    }
    else
    {
	if (count > 1)
	{
	    printf (
		"## More than one clue with the light \"%s\" has been matched\n",
		light_str);
	}

	return (TRUE);
    }

}



PUBLIC int cword_rdb_add_anagram (
	char			    *anagram)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function to add an anagram indicator to the database.
 *
 *  FORMAL PARAMETERS:
 *
 *      *anagram		    (r) string containing anagram to be added
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL		    Normal Completion
 *	CWORD_S_FAILURE		    Rdb error
 *--
 */

{

    long status;

    status = cword_rdb__set_transaction (
		WRITE_TABLE,
		ANAGRAM_DESCRIPTION);

    if (CWORD_S_NORMAL != status)
    {
	return (status);
    }

    /* 
     *  ?? Is it a good idea ??
     *
     *  Set the inputted string to uppercase
     */

    cword_word_lower_str  (anagram);

    EXEC SQL INSERT INTO ANAGRAM_DESCRIPTION
	(
	    A_DESCRIPTION
	)
	VALUES	
	(
	    :anagram
	);

    status = cword_rdb__check_status (SQLCA.SQLCODE);

    if (status == CWORD_S_NORMAL)
    {
	cword_messages_output_message (
		CWORD_RDBGENERICSUC,
		TRACE3,
		INSERT_STR,
		"## ANAGRAM_DESCRIPTION");
    }
    else
    {
	cword_messages_output_message (
		CWORD_RDBGENERICERR,
		(TRACE1 | SCREEN1),
		INSERT_STR,
		"## ANAGRAM_DESCRIPTION");

	cword_messages_output_message (
		CWORD_RDBADDANAERR,
		(TRACE2 | SCREEN0),
		anagram);

	EXEC SQL ROLLBACK;

	return (CWORD_S_FAILURE);
    }

    EXEC SQL COMMIT;

    status = cword_rdb__check_status (SQLCA.SQLCODE);

    if (status == CWORD_S_NORMAL)
    {
	cword_messages_output_message (
		CWORD_RDBGENERICSUC,
		TRACE3,
		COMMIT_STR);
    }
    else
    {
	cword_messages_output_message (
		CWORD_RDBGENERICERR,
		(TRACE1 | SCREEN1),
		COMMIT_STR);

	EXEC SQL ROLLBACK;

	return (CWORD_S_FAILURE);
    }

    cword_messages_output_message (
	    CWORD_RDBADDANASUC,
	    (TRACE2 | SCREEN0),
	    anagram);

    return (CWORD_S_NORMAL);

}



PUBLIC int cword_rdb_add_homophone (
	char			*indicator,
	char			*description)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *      Function to add a homophone couplet to the database.
 *
 *  FORMAL PARAMETERS:
 *
 *      *indicator		(r) string containing indicator
 *	*description		(r) string containing desciption
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL		Normal Completion
 *	CWORD_S_FAILURE		Rdb error
 *--
 */

{

    long status;

    status = cword_rdb__set_transaction (
		WRITE_TABLE,
		HOMOPHONE_DESCRIPTION);

    if (CWORD_S_NORMAL != status)
    {
	return (status);
    }

    /*
     *  Set all inputted strings to uppercase
     */

    cword_word_lower_str  (indicator);
    cword_word_lower_str  (description);

    EXEC SQL INSERT INTO HOMOPHONE_DESCRIPTION
	(
	    H_INDICATOR,
	    H_DESCRIPTION
	)
	VALUES	
	(
	    :indicator,
	    :description
	);

    status = cword_rdb__check_status (SQLCA.SQLCODE);

    if (status == CWORD_S_NORMAL)
    {
	cword_messages_output_message (
		CWORD_RDBGENERICSUC,
		TRACE3,
		INSERT_STR,
		"## HOMOPHONE_DESCRIPTION");
    }
    else
    {
	cword_messages_output_message (
		CWORD_RDBGENERICERR,
		(TRACE1 | SCREEN1),
		INSERT_STR,
		"## HOMOPHONE_DESCRIPTION");

	cword_messages_output_message (
		CWORD_RDBADDHOMERR,
		(TRACE2 | SCREEN0),
		indicator);

	EXEC SQL ROLLBACK;

	return (CWORD_S_FAILURE);
     }

    EXEC SQL COMMIT;

    status = cword_rdb__check_status (SQLCA.SQLCODE);

    if (status == CWORD_S_NORMAL)
    {
	cword_messages_output_message (
		CWORD_RDBGENERICSUC,
		TRACE3,
		COMMIT_STR);
    }
    else
    {
	cword_messages_output_message (
		CWORD_RDBGENERICERR,
		(TRACE1 | SCREEN1),
		COMMIT_STR);

	EXEC SQL ROLLBACK;

	return (CWORD_S_FAILURE);
    }

    cword_messages_output_message (
	    CWORD_RDBADDHOMSUC,
	    (TRACE2 | SCREEN0),
	    indicator);

    return (CWORD_S_NORMAL);

}



PUBLIC int cword_rdb_add_synonym (
	char			*indicator,
	char			*description)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *      Function to add a synonym couplet to the database.
 *
 *  FORMAL PARAMETERS:
 *
 *      *indicator		(r) indicator string 
 *	*description		(r) description string
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL		Normal Completion
 *	CWORD_S_FAILURE		Rdb Error
 *--
 */

{

    long status;

    status = cword_rdb__set_transaction (
		WRITE_TABLE,
		SYNONYM_DESCRIPTION);

    if (CWORD_S_NORMAL != status)
    {
	return (status);
    }

    /*
     *  Set all inputted strings to uppercase
     */

    cword_word_lower_str  (indicator);
    cword_word_lower_str  (description);

    EXEC SQL INSERT INTO SYNONYM_DESCRIPTION
	(
	    S_INDICATOR,
	    S_DESCRIPTION
	)
	VALUES	
	(
	    :indicator,
	    :description
	);

    status = cword_rdb__check_status (SQLCA.SQLCODE);

    if (status == CWORD_S_NORMAL)
    {
	cword_messages_output_message (
		CWORD_RDBGENERICSUC,
		TRACE3,
		INSERT_STR,
		"## SYNONYM_DESCRIPTION");
    }
    else
    {
	cword_messages_output_message (
		CWORD_RDBGENERICERR,
		(TRACE1 | SCREEN1),
		INSERT_STR,
		"## SYNONYM_DESCRIPTION");

	cword_messages_output_message (
		CWORD_RDBADDSYNERR,
		(TRACE2 | SCREEN0),
		indicator);

	EXEC SQL ROLLBACK;

	return (CWORD_S_FAILURE);
    }

    EXEC SQL COMMIT;

    status = cword_rdb__check_status (SQLCA.SQLCODE);

    if (status == CWORD_S_NORMAL)
    {
	cword_messages_output_message (
		CWORD_RDBGENERICSUC,
		TRACE3,
		COMMIT_STR);
    }
    else
    {
	cword_messages_output_message (
		CWORD_RDBGENERICERR,
		(TRACE1 | SCREEN1),
		COMMIT_STR);

	EXEC SQL ROLLBACK;

	return (CWORD_S_FAILURE);
    }

    cword_messages_output_message (
	CWORD_RDBADDSYNSUC,
	(TRACE2 | SCREEN0),
	indicator);

    return (CWORD_S_NORMAL);

}



PUBLIC int cword_rdb_output_stats ()

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	This function will output the number of crossword and grid definitions
 *	stored in the database.  It will also output the number of "loose"
 *	clues stored in the database.
 *
 *  FORMAL PARAMETERS:
 *
 *      None.
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL
 *	Other
 *--
 */

{

    char *undefined_str = UNDEFINED_XRD;

    int count,
	status;

    CWORD_OS_CLEAR_SCREEN ();

    status = cword_rdb__set_transaction (
		READ_TABLE,
		ALL);

    if (CWORD_S_NORMAL != status)
    {
	return (status);
    }

    cword_messages_output_message (
	CWORD_S_DBSTATSMSG,
	SCREEN0);

    /*
     *  Output:
     *
     *  1. Number of crosswords saved in the database.
     *  2. Number of grids saved in the database.
     *  3. Number of clues saved in the database.
     */

    EXEC SQL SELECT COUNT(*) into :count
	FROM CROSSWORD_DESCRIPTION;

    status = cword_rdb__check_status (SQLCA.SQLCODE);

    if (status != CWORD_S_NORMAL)
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICERR,
	    (TRACE3 | SCREEN0),
	    DECLARE_CURSOR_STR);

	EXEC SQL ROLLBACK;

	return (CWORD_S_FAILURE);
    }

    cword_messages_output_message (
	CWORD_WAITINGENDED,
	SCREEN0);
    cword_messages_output_message (
	CWORD_S_DEFNSINDNMSG,
	(TRACE3 | SCREEN0),
	count == 1 ? IS_STR : ARE_STR,
	count,
	count == 1 ? CROSSWORD_STR : CROSSWORDS_STR);

    EXEC SQL SELECT COUNT(*) into :count
	FROM GRID_DESCRIPTION;

    status = cword_rdb__check_status (SQLCA.SQLCODE);

    if (status != CWORD_S_NORMAL)
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICERR,
	    (TRACE3 | SCREEN0),
	    DECLARE_CURSOR_STR);

	EXEC SQL ROLLBACK;

	return (CWORD_S_FAILURE);
    }

    cword_messages_output_message (
	CWORD_WAITINGENDED,
	SCREEN0);
   cword_messages_output_message (
	CWORD_S_DEFNSINDNMSG,
	(TRACE3 | SCREEN0),
	count == 1 ? IS_STR : ARE_STR,
	count,
	count == 1 ? GRID_STR : GRIDS_STR);

    EXEC SQL SELECT COUNT(*) into :count
	FROM CLUE_DESCRIPTION
	    WHERE CROSSWORD_NAME = :undefined_str;

    status = cword_rdb__check_status (SQLCA.SQLCODE);

    if (status != CWORD_S_NORMAL)
    {
	cword_messages_output_message (
	    CWORD_RDBGENERICERR,
	    (TRACE3 | SCREEN0),
	    DECLARE_CURSOR_STR);

	EXEC SQL ROLLBACK;

	return (CWORD_S_FAILURE);
    }

    cword_messages_output_message (
	CWORD_WAITINGENDED,
	SCREEN0);
    cword_messages_output_message (
	CWORD_S_LOOSECLUESMSG,
	(TRACE3 | SCREEN0),
	count == 1 ? IS_STR : ARE_STR,
	count,
	count == 1 ? CLUE_STR : CLUES_STR);
    cword_messages_output_message (
	CWORD_WAITINGENDED,
	SCREEN0);

    EXEC SQL ROLLBACK;

    return (CWORD_S_NORMAL);

}



PUBLIC BOOLEAN cword_rdb_is_rdb_defined ()

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function to return whether the database code has been defined or not.
 *
 *  FORMAL PARAMETERS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      TRUE			Database defined.
 *	FALSE			Database not defined.
 *--
 */

{

    return (TRUE);

}

#else /* else DB */



/*
 *  The rest of the module is when DB is not defined!
 */

#include "main.h"
#include "messages.h"

PUBLIC BOOLEAN cword_rdb_is_rdb_defined ()

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function to return whether the database code has been defined or not.
 *
 *  FORMAL PARAMETERS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      TRUE			Database defined.
 *	FALSE			Database not defined.
 *--
 */

{

    return (FALSE);

}

#endif
