/*
 *++
 *  FACILITY:
 *
 *      soundex.c
 *
 *  ABSTRACT:
 *
 *	Module of functions implementing variants of the soundex algorithm.
 *
 *  AUTHORS:
 *
 *      James A. Lundon
 *
 *  CREATION DATE:
 *
 *      15 June 1993
 *--
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#ifdef SOUNDEX
#define MESSAGES_DEFINED 1
#endif

#include "main.h"
#include "utils.h"
#include "match.h"
#include "messages.h"
#include "memman.h"
#include "soundex.h"
#include "defaults.h"
#include "strll.h"

extern CWORD_G_DEFAULTS_RECORD cword_g_defaults_record;



#ifndef SOUNDEX

#define class(C) isalpha(C) ? soundex_str[ C - "aA"[isupper(C)] ] : '0'

PRIVATE int cword_soundex__compare_words (
	char			*soundex1,
	char			*soundex2);

PRIVATE char *cword_soundex__cut_soundex (
	char			*sound);



PUBLIC int cword_soundex_samelet_matches (
	char			*word,
	CWORD_T_STR_LL		**word_ll)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	This function executes one of the soundex algorithms and returns the
 *	matched list of words in the CWORD_T_STR_LL link list.  This function 
 *      only matches words which start with the same letter as the 
 *      reference word.
 *
 *  FORMAL PARAMETERS:
 *
 *	*word			(r) Word to Soundex 
 *	**word_ll		(r) Link list containing matches to be returned
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL		Successful completion
 *	CWORD_S_FAILURE		Error with word files or other error
 *--
 */

{

    char soundex_str[CWORD_SIZEOF_SOUNDEX + 1],
	 soundex_equiv[CWORD_SIZEOF_SOUNDEX + 1],
	 word_str[CWORD_SIZEOF_LINE + 1];

    int	status;

    *word_ll = (CWORD_T_STR_LL *) 0;

    cword_messages_output_message (
	CWORD_SOUNDEXSAME,
	(TRACE2 | SCREEN0),
	word);

#ifdef SOUNDEX_1

    cword_soundex_algorithm_1 (
	word,
	soundex_str);

    cword_messages_output_message (
	CWORD_S_SOUNDEX1MSG
	(TRACE0 | SCREEN0),
	word,
	soundex_str);

#else /* SOUNDEX_2 */

    cword_soundex_algorithm_2 (
	word,
	soundex_str);

    cword_messages_output_message (
	CWORD_S_SOUNDEX2MSG,
	(TRACE0 | SCREEN0),
	word,
	soundex_str);

#endif

    status = cword_match_read_word_files (
		1,
		CWORD_SIZEOF_WORD,
		word_str);

    if (CWORD_S_NORMAL != status)
    {
	return (CWORD_S_FAILURE);
    }

    cword_messages_output_message (
	CWORD_SOUNDEXNOW,
	SCREEN0);

    do 
    {
#ifdef SOUNDEX_1

	cword_soundex_algorithm_1 (
		word_str,
		soundex_equiv);	

	/* 
	 *  Clip soundex equivalent for comparison
	 */

	strcpy (soundex_equiv, cword_soundex__cut_soundex (soundex_equiv));

#else /* SOUNDEX_1 */

	cword_soundex_algorithm_2 (
		word_str,
		soundex_equiv);	

#endif

	/*
	 *  Pure comparison - only A42000 and A42000 will match.
	 */

	if (strcmp (soundex_str, soundex_equiv) == 0)
	{
	    cword_strll_add_str (
		SORTED | UNIQUE,
		word_str,
		word_ll);

	    if (*word_ll == (CWORD_T_STR_LL *) 0)
	    {
		return (CWORD_S_FAILURE);
	    }
	}
    }
    while (CWORD_S_NORMAL == cword_match_read_word_files (0, 0, word_str));

    cword_messages_output_message (
	    CWORD_WAITINGENDED,
	    SCREEN0);

    if (*word_ll == (CWORD_T_STR_LL *) 0)
    {
	cword_messages_output_message (
	    CWORD_S_NOSOUNDEXFOUND,
	    (SCREEN0 | TRACE0));

	return (CWORD_S_NORMAL);
    }

    return (CWORD_S_NORMAL);

}



PUBLIC int cword_soundex_getall_matches (
	char			*word,
	CWORD_T_STR_LL		**word_ll)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	This function executes one of the soundex algorithms and returns the
 *	matched list of word in the CWORD_T_STR_LL link list.  This function 
 *      matches words with same soundex equivalents, even if they start with 
 *      different letters.
 *
 *  FORMAL PARAMETERS:
 *
 *	*word			(r) Word to Soundex to
 *	**word_ll		(r) Link list containing matches
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL		Successful completion
 *	CWORD_S_FAILURE		Error with word file or other error
 *--
 */

{

    char soundex_str[CWORD_SIZEOF_SOUNDEX + 1],
	 soundex_equiv[CWORD_SIZEOF_SOUNDEX + 1],
	 word_str[CWORD_SIZEOF_LINE + 1];

    int status;

    *word_ll = (CWORD_T_STR_LL *) 0;

    cword_messages_output_message (
	CWORD_SOUNDEXDIFF,
	(TRACE2 | SCREEN0),
	word);

#ifdef SOUNDEX_1

    cword_soundex_algorithm_1 (
	word,
	soundex_str);

    cword_messages_output_message (
	CWORD_S_SOUNDEX1MSG
	(TRACE0 | SCREEN0),
	word,
	soundex_str);

#else /* SOUNDEX_1 */

    cword_soundex_algorithm_2 (
	word,
	soundex_str);

    cword_messages_output_message (
	CWORD_S_SOUNDEX2MSG,
	(TRACE0 | SCREEN0),
	word,
	soundex_str);

#endif

    status = cword_match_read_word_files (
		1,
		CWORD_SIZEOF_WORD,
		word_str);

    if (CWORD_S_NORMAL != status)
    {
	return (CWORD_S_FAILURE);
    }

    cword_messages_output_message (
	CWORD_SOUNDEXNOW,
	SCREEN0);

    do 
    {
#ifdef SOUNDEX_1

	cword_soundex_algorithm_1 (
	    word_str,
	    soundex_equiv);	

	/* 
	 *  Clip soundex equivalent for comparison
	 */

	strcpy (soundex_equiv, cword_soundex__cut_soundex (soundex_equiv));

#else /* SOUNDEX_1 */

	cword_soundex_algorithm_2 (
	    word_str,
	    soundex_equiv);	

#endif

	/*
	 *  Checks words for soundex equivalency with unequal starting letters
	 *  - A42000 and Z42000 will match successfully.
	 */

	if (CWORD_S_NORMAL == cword_soundex__compare_words (soundex_equiv, soundex_str))
	{
	    cword_strll_add_str (
		SORTED | UNIQUE,
		word_str,
		word_ll);

	    if (*word_ll == (CWORD_T_STR_LL *) 0)
	    {
		return (CWORD_S_FAILURE);
	    }
	}
    }
    while (CWORD_S_NORMAL == cword_match_read_word_files (0, 0, word_str));

    cword_messages_output_message (
	CWORD_WAITINGENDED,
	SCREEN0);

    if (*word_ll == (CWORD_T_STR_LL *) 0)
    {
	cword_messages_output_message (
	    CWORD_S_NOSOUNDEXFOUND,
	    (SCREEN0 | TRACE0));

	return (CWORD_S_NORMAL);
    }

    return (CWORD_S_NORMAL);

}



PUBLIC int cword_soundex_algorithm_1 (
	char			*name,
	char			*sound)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Accepts a name, performs the soundex algorithm on it, and returns the
 *	soundex version.  Does not alter original name.
 *
 *	?? References needed ??
 * 
 *  FORMAL PARAMETERS:
 *
 *	*name			(r) Input string to get soundex equivalent of
 *	*sound			(w) output soundex string
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  {@function value or completion codes@}
 *
 *      [@description or none@]
 *--
 */

{

    int x,
	y,
	count,
	    namelen;

    char *namecopy,
	 firstchar,
	 lastchar;

    static unsigned char replacetable[27] = {
				255, '0', '1', '2', '3', '0', '1', '2', 
				255, '0', '2', '2', '4', '5', '5',
				'0', '1', '2', '6', '2', '3', '0', '1', 
				255, '2', '0', '2'};

    y = 0;
    lastchar = 0;

    /* Fill string with the interger value 0 */

    memset (sound, (int) '0', CWORD_SIZEOF_SOUNDEX);

    sound[CWORD_SIZEOF_SOUNDEX] = (char) 0;

#if 0
    for (count = 0; count < CWORD_SIZEOF_SOUNDEX; count++)
	printf ("%d ", sound[count]);

    printf ("\n");
#endif

    if ((namecopy = (char *) malloc (strlen (name) + 1)) == 0) 
    {
	return (CWORD_S_FAILURE);
    }

    memset (namecopy, 0, strlen (name) + 1);

    /* Convert string to uppercase */

    count = 0;
    while (name[count] != (char) 0) 
    {
	namecopy[count] = (char) toupper(name[count]);
	count++;
    }

    if (strlen(name) != 0)
    {
	return (CWORD_S_FAILURE);
    }

    firstchar = namecopy[0];
    namelen = strlen (namecopy);

    for (x = 0; x < namelen; x++)
    {
	if (!isalpha(namecopy[x]))
	{
	    free(namecopy);
	    return (CWORD_S_FAILURE);
	}

	switch (namecopy[x])
	{
	case 'A' :
	case 'E' :
	case 'I' :
	case 'O' :
	case 'U' :
	case 'Y' : 
	    if (!x) 
		break;

	case 'W' :
	case 'H' : 
	    namecopy[x] = 64;
	    break;

	default:
	    break;
	}

	namecopy[x] = replacetable[namecopy[x] - 64];

	if (namecopy[x] == lastchar) 
	{
	    namecopy[x] = (unsigned char) 255;
	}
	else
	{
	    lastchar = namecopy[x];
	    if (y < 6) sound[y++] = namecopy[x];
	}
    }  

    sound[0] = firstchar;
    free (namecopy);

    return (CWORD_S_NORMAL);

}



PUBLIC int cword_soundex_algorithm_2 (
	char			*name,
	char			*sound)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	From Knuth's description of Odell & Russell's "Soundex" patents.
 *
 *	Examples:
 *
 *       input       output
 *       ----------- ------
 *       Euler       E460
 *       Gauss       G200
 *       Hilbert     H416
 *       Knuth       K530
 *       Lloyd       L300
 *       Lukasiewicz L222
 *
 *      ?? References needed ??
 *
 *  FORMAL PARAMETERS:
 *
 *	*name			(r) Input string to get soundex equivalent of
 *	*sound			(w) Output soundex string
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      [@description or none@]
 *--
 */

{

    /* Static veriable - taking the hit */
    static char soundex_str[] = "01230120022455012623010202";

    char *namecopy;

    register int a = 0,
		 c1,
		 c2;

    int count;

    memset (sound, 0, CWORD_SIZEOF_SOUNDEX);

    if ((namecopy = (char *) malloc (strlen (name) + 1)) == 0) 
	return (CWORD_S_FAILURE);

    memset (namecopy, (int) '0', strlen (name) + 1);

    count = 0;
    while (name[count] != (char) 0)
    {
	namecopy[count] = (char) toupper (name[count]);
	count++;
    }

    sound[0]= *namecopy;

    if ((c1 = (int) class((char) sound[0])) != 0)
	for (a++; a < 4 && *++namecopy; c1 = c2)
	    if (c1 != (c2 = class (*namecopy)) && c2 != (int) '0')
                sound[a++] = c2;

    while (a < 4) 
	sound[a++]= (char) '0';

    sound[CWORD_SIZEOF_SOUNDEX] = (char) 0;

    return (CWORD_S_NORMAL);

}



PRIVATE int cword_soundex__compare_words (
	char			*soundex1,
	char			*soundex2)
	
/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function matching soundex outputs, less the first letter signifying the
 *	letter of the alphabet the word has come from.  It is useful if one is
 *	trying to get alike sounding words not starting with the same letter as
 *	the input word. e.g
 *
 *	soundex1 = "A42000" and soundex2 = "Z42000" are equated successfully
 *	while the following pair soundex1 = "A42001" and soundex2 = "A42002"
 *	aren't equal.  This also compares successfully words wih the same
 *	starting letter.
 *
 *  FORMAL PARAMETERS:
 *
 *      *soundex1		(r) soundex string 1
 *	*soundex2		(r) soundex string 2
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL		
 *	CWORD_S_FAILURE		
 *--
 */

{

    soundex1++; 
    soundex2++;

    if (strcmp (soundex1, soundex2) != 0)
	return (CWORD_S_NORMAL);
    else
	return (CWORD_S_FAILURE);

}



PRIVATE char *cword_soundex__cut_soundex (
	char			*sound)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *      Function to take the zeros of the end of the inputted soundex value.
 *
 *	The question is: are there any difference between soundex output A1200
 *	and A12?  If there is then this function is redundant.  If there 
 *      isn't, this function is important as a further pre-processing tool
 *      used to obtain decent SOUNDEX equivalents.
 *
 *  FORMAL PARAMETERS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  {@function value or completion codes@}
 *
 *      [@description or none@]
 *--
 */

{

    int	count = strlen (sound) - 1;

    while ( (sound[count] == (char) '0') || (sound[count] == (char) '0') ) 
	count--;

    sound[++count] = (char) 0;

    return (sound);

}



#else

extern int main (
	int			argc, 
	char			*argv[])

{

    CWORD_T_STR_LL *word_ll;

    int status;

    if (argc < 2)
    {
	printf ("Not enough parameters were entered\n");
	exit (EXIT_FAILURE);
    }

    status = cword_soundex_samelet_matches (
		    argv[1],
		    &word_ll);

    while (word_ll != (CWORD_T_STR_LL *) 0)
    {
	printf ("%s\n", word_ll->word);
	word_ll = word_ll->next;
    }

    exit (EXIT_SUCCESS);
   
}

#endif
