/*
 *++
 *  FACILITY:
 *
 *      stats.c
 *
 *  ABSTRACT:
 *
 *	Module containing functions analysing crosswords and grids for
 *	statistics/reports/comments.
 *
 *  AUTHORS:
 *
 *      James A. Lundon
 *
 *  CREATION DATE:
 *
 *      27 May 1993
 *--
 */

/*
 *  Include files
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#if defined(VMS) || defined (__VMS)

#include <stat.h>
#include <unixio.h>

#endif

#if defined(__unix) || defined(__unix__)

#include <sys/types.h>
#include <sys/stat.h>

#endif

#if defined(__MSDOS__) || defined(MSDOS)

#include <sys\types.h>
#include <sys\stat.h>

#endif

#include "main.h"
#include "print.h"
#include "messages.h"
#include "verify.h"
#include "grid.h"
#include "charui.h"
#include "utils.h"
#include "match.h"
#include "os.h"
#include "stats.h"
#include "memman.h"
#include "fileio.h"
#include "strll.h"



#ifndef STATS

/*
 *  N.B. cword_stats_crosswrd_files_info (31 characters) not 
 *  cword_stats_crossword_files_info (32 characters).
 */

PUBLIC int cword_stats_crosswrd_files_info (
	char			*file_spec)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function to return to the screen a small amount of information about
 *	the crossword file(s) specified in *file_spec.
 *
 *  FORMAL PARAMETERS:
 *
 *	*file_spec		(r) Wildcarded file specification
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL		Normal completion.
 *--
 */

{

    CWORD_T_STR_LL *file_ll,
                   *temp_ll;

    char file_name_str[CWORD_SIZEOF_WORD + 1],
	 temp1_str[CWORD_SIZEOF_WORD + 1],
	 temp2_str[CWORD_SIZEOF_WORD + 1],
	 temp3_str[CWORD_SIZEOF_WORD + 1];

    int x_dims,
        y_dims,
	clue_num,
	line_num,
	status;

    BOOLEAN is_fully_specified = FALSE;

    cword_messages_output_message (
	CWORD_S_XRDFILEINFO,
	SCREEN0,
	file_spec);

    status = CWORD_OS_LIST_FILES (
		(char *) 0,
		file_spec,
		&file_ll);

    /*
     *	Use fileio functon to read first few lines of crossword files and
     *	return relevant info to caller (me).
     */

    temp_ll = file_ll;

    if (temp_ll == (CWORD_T_STR_LL *) 0)
    {
	cword_messages_output_message (
	    CWORD_NOFILESFOUND,
	    SCREEN0);
    }
    else
    {
	cword_messages_output_message (
	    CWORD_S_DASHES,
	    SCREEN0);

	cword_messages_output_message (
	    CWORD_S_XRDFILESTATMSG,
	    SCREEN0,
	    CWORD_X_CHAR,
	    CWORD_Y_CHAR);

	cword_messages_output_message (
	    CWORD_S_DASHES,
	    SCREEN0);

	/*
	 *  Number of lines that are scrolled gone before and file information
	 *  is outputted!
	 */

	line_num = 4;
	while (temp_ll != (CWORD_T_STR_LL *) 0)
	{
	    status = cword_fileio_read_crossword_top (
			temp_ll->str,
			file_name_str,
			&x_dims,
			&y_dims,
			&clue_num);

	    if (CWORD_S_NORMAL != status)
	    {
		cword_messages_output_message (
		    CWORD_S_DETAILSFILEERR,
		    SCREEN0,
		    temp_ll->str);
	    }
	    else
	    {
		/*
		 *  Reformat temp_ll->str and file_name_str so that they
		 *  will fit properly with existing structure if they are too
		 *  long.
		 */

		if (strlen (temp_ll->str) > 17)
		{
		    strncpy (temp3_str, temp_ll->str, 14);
		    temp3_str[14] = (char) 0;

		    sprintf (
			temp1_str,
			"%s...",
			temp3_str);
		}
		else
		{
		    strcpy (temp1_str, temp_ll->str);
		}

		if (strlen (file_name_str) > 15)
		{
		    strncpy (temp3_str, file_name_str, 12);
		    temp3_str[12] = (char) 0;

		    sprintf (
			temp2_str,
			"%s...",
			temp3_str);
		}
		else if ( (strlen (file_name_str) <= 15) && (strlen (file_name_str) > 0) )
		{
		    strcpy (temp2_str, file_name_str);
		}
		else
		{
		    strcpy (temp2_str, CWORD_NONE_STR);
		}

		is_fully_specified = cword_stats_is_fully_speced (temp_ll->str);

		cword_messages_output_message (
		    CWORD_S_XRDFILEDETAILS,
		    SCREEN0,
		    temp1_str,
		    temp2_str,
		    x_dims,
		    y_dims,
		    clue_num,
		    is_fully_specified ? CWORD_YES_STR : CWORD_NO_STR);
	    }

	    temp_ll = temp_ll->next;

	    line_num++;

	    if (cword_charui_check_page_break (
		    (void *) temp_ll,
		    CWORD_S_PRESSRETURNQUIT,
		    &line_num) == CWORD_S_FAILURE)
	    {
		/*
		 *  Stop displaying data about crossword files.
		 */

		break;
	    }
	}
    }

    cword_strll_free_str_ll (&file_ll);

    return (CWORD_S_NORMAL);

}



PUBLIC int cword_stats_get_grid_type (
	char			*file_spec)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function to return the type of grid of the input grid file!  This is
 *	used in the case where the grid type of the file is not known or is
 *	lost.  This function is useful when working on old grid file.
 *
 *  FORMAL PARAMETERS:
 *
 *      *file_spec		(r) grid file spec
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      grid_type		Valid grid type
 *	INVALID_GRID		Invalid grid file or grid type found!
 *--
 */

{

    CROSSWORD_RECORD *crossword_record;

    int status,
	grid_type;

    crossword_record = cword_memman_make_crossword_rec ();

    status = cword_fileio_read_grid (
		file_spec,
		crossword_record);

    if (CWORD_S_NORMAL != status)
    {
	return (INVALID_GRID);
    }

    grid_type = crossword_record->grid_type;

    cword_memman_free_cross_rec (crossword_record);

    return (grid_type);

}



PUBLIC int cword_stats_get_symmetry (
	char			*file_spec)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function to obtain type of symmetry of inputted grid file *file_spec.
 *
 *  FORMAL PARAMETERS:
 *
 *	*file_spec		(r) Grid file specification
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      symmetry		Valid type of symmetry.
 *	INVALID_SYMMETRY	Invalid grid file read or other problem.
 *--
 */

{

    CROSSWORD_RECORD *crossword_record;

    int status,
	symmetry;

    crossword_record = cword_memman_make_crossword_rec ();

    status = cword_fileio_read_grid ( 
		    file_spec,
		    crossword_record);

    if (CWORD_S_NORMAL != status)
    {
	return (INVALID_SYMMETRY);
    }

    symmetry = crossword_record->symmetry;

    cword_memman_free_cross_rec (crossword_record);

    return (symmetry);

}



PUBLIC BOOLEAN cword_stats_is_fully_speced (
	char			*file_spec)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function to checkwhether the inputted crossword denoted by *file_spec
 *	is fully specified.  By specified I mean that there are both lights and
 *	distinct user defined clues defined in the definition.
 *
 *  FORMAL PARAMETERS:
 *
 *      *file_spec		(r) Crossword definition file spec 
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      TRUE			Crossword is fully specified.
 *	FALSE			Crossword is not fully specified.
 *--
 */

{

    CROSSWORD_RECORD *crossword_record;

    int status;

    BOOLEAN is_fully_specified = FALSE;

    crossword_record = cword_memman_make_crossword_rec ();

    status = cword_fileio_read_crossword ( 
		file_spec,
		crossword_record);

    if (CWORD_S_NORMAL != status)
    {
	return (FALSE);
    }

    is_fully_specified = cword_verify_is_fully_speced (crossword_record);

    cword_memman_free_cross_rec (crossword_record);

    return (is_fully_specified);

}



PUBLIC int cword_stats_grid_files_info (
	char			*file_spec)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function to return to the screen a small amount of information about
 *	the grid files specified in *file_spec.
 *
 *  FORMAL PARAMETERS:
 *
 *	*file_spec		(r) Wildcarded file specification
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL		Normal completion.
 *--
 */

{

    CWORD_T_STR_LL *file_ll,
                   *temp_ll;

    char file_name_str[CWORD_SIZEOF_WORD + 1],
	 temp1_str[CWORD_SIZEOF_WORD + 1],
	 temp2_str[CWORD_SIZEOF_WORD + 1],
	 temp3_str[CWORD_SIZEOF_WORD + 1],
	 symmetry_str[CWORD_SIZEOF_LINE + 1];

    int x_dims,
        y_dims,
	clue_num,
	line_num,
	status,
	symmetry,
	grid_type;

    cword_messages_output_message (
	CWORD_S_GRDFILEINFO,
	SCREEN0,
	file_spec);

    status = CWORD_OS_LIST_FILES (
		(char *) 0,
		file_spec,
		&file_ll);

    /*
     *	Use fileio functon to read first few lines of crossword files and
     *	return relevant info to caller (me).
     */

    temp_ll = file_ll;

    if (temp_ll == (CWORD_T_STR_LL *) 0)
    {
	cword_messages_output_message (
	    CWORD_NOFILESFOUND,
	    SCREEN0);
    }
    else
    {
	cword_messages_output_message (
	    CWORD_S_DASHES,
	    SCREEN0);

	cword_messages_output_message (
	    CWORD_S_GRDFILESTATMSG,
	    SCREEN0,
	    CWORD_X_CHAR,
	    CWORD_Y_CHAR);

	cword_messages_output_message (
	    CWORD_S_DASHES,
	    SCREEN0);

	line_num = 3;
	while (temp_ll != (CWORD_T_STR_LL *) 0)
	{
	    status = cword_fileio_read_grid_top (
		temp_ll->str,
		file_name_str,
		&x_dims,
		&y_dims,
		&clue_num,
		&grid_type,
		&symmetry);

	    if (CWORD_S_NORMAL != status)
	    {
		cword_messages_output_message (
		    CWORD_S_DETAILSFILEERR,
		    SCREEN0,
		    temp_ll->str);
	    }
	    else
	    {
		/*
		 *  Reformat temp_ll->str and file_name_str so that they will
		 *  fit properly with existing structure if they are too long.
		 */

		if (strlen (temp_ll->str) > 17)
		{
		    strncpy (temp3_str, temp_ll->str, 14);
		    temp3_str[14] = (char) 0;

		    sprintf (
			temp1_str,
			"%s...",
			temp3_str);
		}
		else
		{
		    strcpy (temp1_str, temp_ll->str);
		}

		if (strlen (file_name_str) > 12)
		{
		    strncpy (temp3_str, file_name_str, 9);
		    temp3_str[9] = (char) 0;

		    sprintf (
			temp2_str,
			"%s...",
			temp3_str);
		}
		else if ( (strlen (file_name_str) <= 12) && (strlen (file_name_str) > 0) )
		{
		    strcpy (temp2_str, file_name_str);
		}
		else
		{
		    strcpy (temp2_str, CWORD_NONE_STR);
		}

		/*
		 *  ?? Proper messaging needed ??
		 */

		if (symmetry == DEG_90_SYMMETRY)
		    strcpy (symmetry_str, "90 ");
		else if (symmetry == DEG_180_SYMMETRY)
		    strcpy (symmetry_str, "180 ");
		else if (symmetry == AXIAL_LR_SYMMETRY)
		    strcpy (symmetry_str, "LR ");
		else if (symmetry == AXIAL_RL_SYMMETRY)
		    strcpy (symmetry_str, "RL ");
		else if (symmetry == NO_SYMMETRY)
		    strcpy (symmetry_str, "No ");
		else
		    strcpy (symmetry_str, "Invalid ");

		if (grid_type == BLOCK_GRID)
		    strcat (symmetry_str, CWORD_BLOCK_GRID_STR);
		else if (grid_type == BARRED_GRID)
		    strcat (symmetry_str, CWORD_BARRED_GRID_STR);
		else if (grid_type == AMERICAN_GRID)
		    strcat (symmetry_str, CWORD_AMERICAN_GRID_STR);
		else if (grid_type == NOVELTY_GRID)
		    strcat (symmetry_str, CWORD_NOVELTY_GRID_STR);
		else
		    strcat (symmetry_str, CWORD_INVALID_GRID_STR);

		cword_messages_output_message (
		    CWORD_S_GRDFILEDETAILS,
		    SCREEN0,
		    temp1_str,
		    temp2_str,
		    x_dims,
		    y_dims,
		    clue_num,
		    symmetry_str);
	    }

	    temp_ll = temp_ll->next;

	    line_num++;

	    if (cword_charui_check_page_break (
		    (void *) temp_ll,
		    CWORD_S_PRESSRETURNQUIT,
		    &line_num) == CWORD_S_FAILURE)
	    {
		/*
		 *  Stop displaying data about grid files.
		 */

		break;
	    }
	}
    }

    cword_strll_free_str_ll (&file_ll);

    return (CWORD_S_NORMAL);

}



PUBLIC void cword_stats_get_grid_percents (
	CROSSWORD_RECORD	*crossword_record,
	int			*blanks,
	int			*unches,
	int			*checks)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function to return using *crossword_record the percentage of blank,
 *	unched and checked cells in the defined grid.
 *
 *  FORMAL PARAMETERS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      None!
 *--
 */

{

    CROSSWORD_GRID check_contents;

    float percent_filled,
	  percent_checked,
	  percent_unchecked,
	  percent_blank;

    int	count1,
	count2,
	cells_filled,
	cells_checked;

    /*
     *  Analyse the grid and return:
     *
     *	1. number of cells in the grid that are blank (blanks)
     *	2. number of cells in the grid that are filled but are unchecked
     *	   (unches)
     *	3. number of cells in the grid that are checkes i.e. overlapped by two
     *	   clues (checks)
     */

    /*
     *	Get the property of each cell in the crossword grid.  It important that
     *	Cword uses get_block_props in deference to get_barred_props.
     */

    cword_verify_get_block_props (
	    crossword_record,
	    check_contents);

    cells_filled = cells_checked = 0;
    for (count1 = 0; count1 < crossword_record->x_dimensions; count1++)
	for (count2 = 0; count2 < crossword_record->y_dimensions; count2++)
	{
	    /*
	     *	If the cell in question is not UNFILLED i.e. it's either an
	     *	UNCH or CHECKED.
	     */

	    if (check_contents[count1][count2] != CWORD_UNFILLED_CHAR)
		cells_filled++;

	    /*
	     *	If the cell in question is CHECKED.
	     */

	    if (check_contents[count1][count2] == CWORD_CHECKED_CHAR)
		cells_checked++;
	}

    *blanks = (crossword_record->x_dimensions * crossword_record->y_dimensions) - cells_filled;
    *unches = cells_filled - cells_checked;
    *checks = cells_checked++;

    DEBUG_TRACE (
	"grid percentages: blanks %d - unches %d - checks %d\n",
	*blanks,
	*unches,
	*checks);

    return;

}



PUBLIC int cword_stats_output_percents (
	CROSSWORD_RECORD	*crossword_record)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function that outputs - using cword_stats_get_grid_percents the
 *	percentages obtained therein.
 *
 *  FORMAL PARAMETERS:
 *
 *      *crossword_record	(r) Pointer to a crossword definition.
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL		Normal completion.
 *--
 */

{

    int blanks,
	unches,
	checks;

    float checked_fl,
	  unchecked_fl,
	  blank_fl;

    cword_stats_get_grid_percents (
	    crossword_record,
	    &blanks,
	    &unches,
	    &checks);

    if (crossword_record->grid_type == BLOCK_GRID)
	cword_messages_output_message (
	    CWORD_S_GRIDFOUNDMSG,
	    SCREEN0,
	    CWORD_BLOCK_GRID_STR);
    else if (crossword_record->grid_type == BARRED_GRID)
	cword_messages_output_message (
	    CWORD_S_GRIDFOUNDMSG,
	    SCREEN0,
	    CWORD_BARRED_GRID_STR);
    else if (crossword_record->grid_type == AMERICAN_GRID)
	cword_messages_output_message (
	    CWORD_S_GRIDFOUNDMSG,
	    SCREEN0,
	    CWORD_AMERICAN_GRID_STR);
    else
	cword_messages_output_message (
	    CWORD_S_GRIDFOUNDMSG,
	    SCREEN0,
	    CWORD_NOVELTY_GRID_STR);

    cword_messages_output_message (
	CWORD_WAITINGENDED,
	SCREEN0);

    checked_fl = (float) ((float) (checks * 100) / (float) (crossword_record->x_dimensions * 
                                                   crossword_record->y_dimensions));

    unchecked_fl = (float) ((float) (unches * 100) / (float) (crossword_record->x_dimensions * 
                                                   crossword_record->y_dimensions));

    blank_fl = (float) ((float) (blanks * 100) / (float) (crossword_record->x_dimensions * 
                                                   crossword_record->y_dimensions));

    if (blanks == 0)
	cword_messages_output_message (
		CWORD_S_CHKUNCHKMSG,
		SCREEN0,
		checked_fl > 0.1 ? (int) (checked_fl + 0.5) : 0,
		unchecked_fl > 0.1 ? (int) (unchecked_fl + 0.5) : 0);
    else
	cword_messages_output_message (
		CWORD_S_CHKUNCHKBLKMSG,
		SCREEN0,
		checked_fl > 0.1 ? (int) (checked_fl + 0.5) : 0,
		unchecked_fl > 0.1 ? (int) (unchecked_fl + 0.5) : 0,
		blank_fl > 0.01 ? (int) (blank_fl + 0.5) : 0);

    return (CWORD_S_NORMAL);

}



PUBLIC int cword_stats_output_light_freq (
	CROSSWORD_RECORD	*crossword_record)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *      Function to output the light length frequency data.
 *
 *  FORMAL PARAMETERS:
 *
 *      *crossword_record	(r) Pointer to a crossword definition.
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL		Normal completion.
 *--
 */

{

    CWORD_T_LENFREQ_STRUCT clue_lengths;

    char temp_str[CWORD_SIZEOF_MESSAGE + 1],
	 temp1_str[CWORD_SIZEOF_MESSAGE + 1],
	 temp2_str[CWORD_SIZEOF_MESSAGE + 1];

    int	count1;

    /*
     *  Get light length distributions for outputting
     */

    cword_stats_get_clue_dist (
	crossword_record,
	clue_lengths);

    /*
     *  Output light length frequencies like so:
     *
     *  length     :  2  3  4  7  8  9 etc.
     *  frequency  :  1  4  7 10 12 11 etc.
     */

    cword_messages_output_message (
	CWORD_WAITINGENDED,
	SCREEN0);

    cword_messages_get_message (CWORD_S_LIGHTLENS, temp1_str);
    cword_messages_get_message (CWORD_S_LIGHTFREQS, temp2_str);

    count1 = 0;
    while (count1 < CWORD_SIZEOF_CELLDIM)
    {
	if (clue_lengths[count1] != 0)
	{
	    sprintf (
		temp_str,
		" %2d",
		count1);

	    strcat (temp1_str, temp_str);

	    sprintf (
		temp_str, 
		" %2d",
		clue_lengths[count1]);

	    strcat (temp2_str, temp_str);
	}

	count1++;
    }

    strcat (temp1_str, "\n");
    strcat (temp2_str, "\n");

    cword_messages_output_message (
	CWORD_USERDEFINEDSTRING,
	SCREEN0,
	temp1_str);

    cword_messages_output_message (
	CWORD_USERDEFINEDSTRING,
	SCREEN0,
	temp2_str);

    cword_messages_output_message (
	CWORD_WAITINGENDED,
	SCREEN0);

    cword_messages_output_message (
	CWORD_S_TOTALLIGHTSMSG,
	SCREEN0,
	crossword_record->clue_sum);

    return (CWORD_S_NORMAL);

}



PUBLIC int cword_stats_report_on_crossword (
	CROSSWORD_RECORD	*crossword_record)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function that reports to the user on an inputted crossword definition.
 *
 *  FORMAL PARAMETERS:
 *
 *	*crossword_record	(r) Pointer to crossword definition.
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL		Normal completion.
 *--
 */

{

    /*
     *  ?? Things that could be done on crossword definition ??
     *
     *	1. Letter analysis of words in the grid?
     *  2. Answers in clues?
     */

    /*
     *	All sorts of statistical information is found and reported about the
     *	crossword definition in question.
     */

    CWORD_OS_CLEAR_SCREEN ();

    /*
     *  ?? If the crossword has no name ??
     */

    cword_messages_output_message (
	CWORD_S_STATREPTCROSS,
	SCREEN0,
	crossword_record->name);

    cword_messages_output_message (
	CWORD_S_STATDIMSARE,
	SCREEN0,
	crossword_record->x_dimensions,
	crossword_record->y_dimensions);

    /*
     *	Get the amount of cells in the crossword grid that are filled and
     *	checked!
     */

    cword_stats_output_percents (crossword_record);

    /*
     *  Output the light length frequencies.
     */

    cword_stats_output_light_freq (crossword_record);

    /*
     *	Check are all the clues in the grid fully user specified.
     */

    if (TRUE == cword_verify_is_fully_speced (crossword_record))
    {
	cword_messages_output_message (
	    CWORD_S_GRIDFULLSPEC,
	    SCREEN0);
    }
    else
    {
	cword_messages_output_message (
	    CWORD_S_GRIDNOTFULLSPEC,
	    SCREEN0);
    }

    if (crossword_record->grid_type == BARRED_GRID)
    {
	cword_verify_check_ximenes (crossword_record);

	cword_messages_output_message (
	    CWORD_WAITINGENDED,
	    SCREEN0);
    }

    /*
     *	?? Would like to have the crossword outputted on the same screen as the
     *	other data, if possible ??
     */

    cword_messages_output_message (
	CWORD_S_XRDAFTERGRTURN,
	SCREEN0);

    cword_charui_pause_ui (0, FALSE);

    /*
     *  Tell the function to handle crossword_record as a crossword!
     */

    cword_charui_display_definition (
	OUTPUT_CROSSWORD,
	crossword_record);

    return (CWORD_S_NORMAL);

}



PUBLIC void cword_stats_get_clue_dist (
	CROSSWORD_RECORD	*crossword_record,
	CWORD_T_LENFREQ_STRUCT 	light_freq)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function to return into clue_dist the number and lengths of lights in
 *	the inputted grid.
 *
 *      There will be no clue frequencies in the 0th location of the array!
 *
 *  FORMAL PARAMETERS:
 *
 *	*crossword_record	(r) crossword or grid description
 *	light_freq		(w) array to be written with details of light 
 *                                  density and lengths
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      None!
 *--
 */

{

    int	count1;

    for (count1 = 0; count1 < CWORD_SIZEOF_CELLDIM; count1++)
	light_freq[count1] = 0;

    for (count1 = 0; count1 < crossword_record->clue_sum; count1++)
    {
	light_freq[strlen (crossword_record->clue_details[count1].clue_answer)] += 1;
    }

    return;

}




PUBLIC int cword_stats_report_on_grid (
	CROSSWORD_RECORD	*crossword_record)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function that reports the vital stats of the grid definition entered
 *	as a CROSSWORD_RECORD through the command line.
 *
 *  FORMAL PARAMETERS:
 *
 *      *crossword_record	(r) Pointer to crossword definition.
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL		Normal completion
 *--
 */

{

    /*
     *	Fill-in the remaining crossword definition fields not filled by the
     *	original grid definiton
     */

    if (crossword_record->is_completely_specified == FALSE)
	cword_utils_complete_definition (crossword_record);

    /*
     *	All sorts of things should be found out and reported about the
     *	crossword in question.
     */

    CWORD_OS_CLEAR_SCREEN ();

    /*
     *  ?? If the grid has no name ??
     */

    cword_messages_output_message (
	CWORD_S_STATREPTGRID,
	SCREEN0,
	crossword_record->name);

    cword_messages_output_message (
	CWORD_S_STATDIMSARE,
	SCREEN0,
	crossword_record->x_dimensions,
	crossword_record->y_dimensions);

    cword_stats_output_percents (crossword_record);

    /*
     *  Output the light length frequencies
     */

    cword_stats_output_light_freq (crossword_record);

    if (TRUE == cword_verify_is_grid_connected (crossword_record))
    {
	cword_messages_output_message (
	    CWORD_S_GRIDCONNECT,
	    SCREEN0);
    }
    else
    {
	cword_messages_output_message (
	    CWORD_S_GRIDNOTCONNECT,
	    SCREEN0);
    }

    cword_messages_output_message (
	CWORD_WAITINGENDED,
	SCREEN0);

    cword_messages_output_message (
	CWORD_S_GRDAFTERGRTURN,
	SCREEN0);

    cword_charui_pause_ui (0, FALSE);

    /*
     *  Tell the function to handle crossword_record as a grid only!
     */

    cword_charui_display_definition (
	OUTPUT_GRID,
	crossword_record);

    return (CWORD_S_NORMAL);

}



PUBLIC int cword_stats_get_clue_sum_for (
	int			direction,
	CROSSWORD_RECORD	*crossword_record)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function that returns the number of clues in direction defined in the
 *	variable direction.
 *
 *  FORMAL PARAMETERS:
 *
 *	direction		(r)
 *	*crossword_record	(r)
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *	Number of clues in the direction as defined in the variable direction
 *	in the function parameter list.
 *--
 */

{

    int num = 0,
	count;

    if ( (direction != ACROSS) && (direction != DOWN) && (direction != BOTH) )
    {
	printf ("# Error calling cword_stats_get_clue_sum_for - no clue direction given\n");

	return (0);
    }	

    if (direction == ACROSS)
    {
	for (count = 0; count < crossword_record->clue_sum; count++)
	    if (crossword_record->clue_details[count].clue_direction == ACROSS)
		num++;
    }
    else
    {
	for (count = 0; count < crossword_record->clue_sum; count++)
	    if (crossword_record->clue_details[count].clue_direction == DOWN)
		num++;
    }

    return (num);

}



PUBLIC unsigned short cword_stats_get_high_clue_num (
	int			direction_type,
	CROSSWORD_RECORD	*crossword_record)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function to get the highest values clue number in the inputted
 *	crossword record definition.  This is used, for example, during the
 *	printing phase to correct format the crossword grid output.
 *
 *  FORMAL PARAMETERS:
 *
 *	direction_type		(r) ACROSS or DOWN
 *      *crossword_record	(r) Pointer to CROSSWORD_RECORD struct
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      Value of highest found clue in crossword definition.
 *--
 */

{

    unsigned short highest_clue = 0;

    int count;

    if (direction_type == BOTH)
    {
	for (count = 0; count < crossword_record->clue_sum; count++)
	{
	    if (crossword_record->clue_details[count].clue_number > highest_clue)
		highest_clue = crossword_record->clue_details[count].clue_number;
	}
    }
    else if (direction_type == ACROSS)
    {
	for (count = 0; count < crossword_record->clue_sum; count++)
	{
	    if ( (crossword_record->clue_details[count].clue_number > highest_clue) &&
		  (crossword_record->clue_details[count].clue_direction == ACROSS) )
		highest_clue = crossword_record->clue_details[count].clue_number;
	}
    }
    else if (direction_type == DOWN)
    {
	for (count = 0; count < crossword_record->clue_sum; count++)
	{
	    if ( (crossword_record->clue_details[count].clue_number > highest_clue) &&
		  (crossword_record->clue_details[count].clue_direction == DOWN) )
		highest_clue = crossword_record->clue_details[count].clue_number;
	}
    }
    else
    {
        printf ("# Error calling get_highest_value_clue\n");
    }

    return (highest_clue);

}



PUBLIC unsigned long cword_stats_get_file_size (
	char			*file_str)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function that returns the total size of the inputted file in *bytes*.
 *
 *  FORMAL PARAMETERS:
 *
 *      *file_str		(r) Spec of file to be stat'ed.
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      Size of file found with stat () otherwise 0 if error found.
 *--
 */

{

    /*
     *	Non-standard way of defining struct to contain file information.
     */

#if defined(VMS) || defined (__VMS)
    stat_t file_info;
#else
    struct stat file_info;
#endif

    int status;

    if (file_str[0] == (char) 0)
    {
	printf ("# NULL file spec entered for sizing\n");

	return (0);
    }

    status = stat (
		file_str, 
		&file_info);

    if (status != 0)
    {
	/*
	 *  Error getting stat information.
	 */

	return (0);
    }
    else
    {
	/*
	 *  Returning size of file found with stat.
	 */

	return (file_info.st_size);
    }

}



PUBLIC int cword_stats_get_word_file_stats ()

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function to output to the user an idea of the size of the word files
 *	currently in use in cword.  Other cumulative statistics are also
 *	provided for completeness.
 *
 *  FORMAL PARAMETERS:
 *
 *      None!
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL		Normal completion.
 *	CWORD_S_FAILURE		Error found!
 *--
 */

{

    char wordfile_spec[CWORD_SIZEOF_FILESPEC + 1];

    unsigned long word_num = 0,
                  total_word_num = 0,
		  grand_word_num = 0;

    int count,
	line_num;

    cword_messages_output_message (
	CWORD_S_WORDFILESTATS,
	SCREEN0);

    cword_match_get_word_file_name (RESET, 0);

    line_num = 2;
    for (count = 1; count <= CWORD_SIZEOF_WORD; count++)
    {
	wordfile_spec[0] = (char) 0;

	/*
	 *  Continue getting word files until cword_match_get_word_file_name
	 *  returns FALSE.
	 */

	while (cword_match_get_word_file_name (
		    count,
		    wordfile_spec) == TRUE)
	{
	    if (cword_utils_does_file_exist (wordfile_spec) == TRUE)
	    {
		/*
		 *  File is OK for reading I will now assume.
		 */

		word_num = (unsigned long)
                            ((float) cword_stats_get_file_size (wordfile_spec) / 
                              (float) (count + 1));

		if (word_num > 1)
		{
		    cword_messages_output_message (
			CWORD_S_SINGLEWORDSTATS,
			SCREEN0,
			wordfile_spec,
			word_num);
		}
		else if (word_num == 1)
		{
		    cword_messages_output_message (
			CWORD_S_SINGLEWORDSTAT,
			SCREEN0,
			wordfile_spec,
			word_num);
		}
		else
		{
		    /*
		     *  Should not happen!
		     */

		    printf (
			"# Word file %20s has no words\n", 
			wordfile_spec);
		}

		/*
		 *  ?? Would use CWORD_S_PRESSRETURNQUIT but how do I determine
		 *  whether there are more valid word files on the way for
		 *  display ??
		 */

		line_num++;
		cword_charui_check_page_break (
		    NOT_FINISHED,
		    CWORD_S_PRESSRETURN,
		    &line_num);
	    }
	    else
	    {
		DEBUG_TRACE (
		    "Word file %s is not accessible\n",
		    wordfile_spec);
	    }

	    /*
	     *	Add word_num to total_word_num before blanking word_num again -
	     *	obvious.
	     */

	    total_word_num += word_num;
	    word_num = 0;
	}

	if (total_word_num != 0)
	{
	    cword_messages_output_message (
		CWORD_S_TOTALWORDSTAT,
		SCREEN0,
		count,
		total_word_num);

	    line_num++;
	    cword_charui_check_page_break (
		NOT_FINISHED,
		CWORD_S_PRESSRETURN,
		&line_num);

	    /*
	     *  Keeping a track on grand_word_num total.
	     */

	    grand_word_num += total_word_num;
	    total_word_num = 0;
	}
    }

    if (grand_word_num != 0)
    {
	cword_messages_output_message (
	    CWORD_WAITINGENDED,
	    SCREEN0);

	cword_messages_output_message (
	    CWORD_S_GRANDWORDSTAT,
	    (SCREEN0 | TRACE0),
	    grand_word_num);

	cword_messages_output_message (
	    CWORD_WAITINGENDED,
	    SCREEN0);
    }

    cword_match_get_word_file_name (RESET, 0);

    return (CWORD_S_NORMAL);

}



PUBLIC int cword_stats_gen_clue_nums (
	CROSSWORD_RECORD	*crossword_record)


/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function to generate clue numbers from a grid/crossword definition
 *	provided through *crossword_record.  This function is useful when
 *	initialising an empty grid.
 *
 *  FORMAL PARAMETERS:
 *
 *      *crossword_record	(w) pointer to an instance of a crossword_record
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL		Successful completion
 *--
 */

{

    int	count1,
	count2,
	count3,
	clue_counter = 1,
	clue_num = 0;

    /*
     *	?? Major memory hog that I will have to look into changeing when
     *	porting on DOS/Windows ??
     */

    short clue_start_locs[CWORD_SIZEOF_CELLDIM][CWORD_SIZEOF_CELLDIM][2];

    BOOLEAN clue_found;

    cword_messages_output_message (
	CWORD_GENCLUENUMS, 
	TRACE3);

    if (crossword_record->clue_sum == 0)
	return (CWORD_S_NORMAL);

    for (count1 = 0; count1 < CWORD_SIZEOF_CELLDIM; count1++)
	for (count2 = 0; count2 < CWORD_SIZEOF_CELLDIM; count2++)
	    for (count3 = 0; count3 < 2; count3++)
		clue_start_locs[count1][count2][count3] = 0;

    /*
     *  Obtaining the starting location of the clues in the completed grid
     */

    for (count1 = 0; count1 < crossword_record->clue_sum; count1++)
    {
	if (crossword_record->clue_details[count1].clue_direction == ACROSS)
	{
	    clue_start_locs[crossword_record->clue_details[count1].x_coordinate - 1]
			    [crossword_record->clue_details[count1].y_coordinate - 1][ACROSS] = count1 + 1;

	    DEBUG_TRACE (
		"clue_start_locs Across (%d,%d) [real] (%d,%d) [array]\n", 
		crossword_record->clue_details[count1].x_coordinate,
		crossword_record->clue_details[count1].y_coordinate,
		crossword_record->clue_details[count1].x_coordinate - 1,
		crossword_record->clue_details[count1].y_coordinate - 1);
	}
	else
	{
	    clue_start_locs[crossword_record->clue_details[count1].x_coordinate - 1]
			    [crossword_record->clue_details[count1].y_coordinate - 1][DOWN] = count1 + 1;

	    DEBUG_TRACE (
		"clue_start_locs Down (%d,%d) [real] (%d,%d) [array]\n", 
		crossword_record->clue_details[count1].x_coordinate,
		crossword_record->clue_details[count1].y_coordinate,
		crossword_record->clue_details[count1].x_coordinate - 1,
		crossword_record->clue_details[count1].y_coordinate - 1);
	}
    }
    
    /*
     *  Check validity
     */

    DEBUG_TRACE (
	"count1 to %d and count2 to %d\n",
	 crossword_record->x_dimensions,
	 crossword_record->y_dimensions);

    /*
     *	Remember to start at the top left hand corner (Y = y_max, X = 0) and go
     *	across (X -> x_max) and then down along the ranks (y -> 0).
     */

    for (count2 = crossword_record->y_dimensions - 1; count2 >= 0; count2--)
    {
	for (count1 = 0; count1 <= crossword_record->x_dimensions - 1; count1++)
	{
	    clue_found = FALSE;

	    if (clue_start_locs[count1][count2][ACROSS] != 0)
	    {
		clue_found = TRUE;

		crossword_record->clue_details[clue_start_locs[count1][count2][ACROSS] - 1].clue_number = 
		    clue_counter;

		DEBUG_TRACE (
			"Across (%d,%d) [real]\n", 
			count1, 
			count2);

		clue_num++;
	    }

	    if (clue_start_locs[count1][count2][DOWN] != 0)
	    {
		clue_found = TRUE;

		crossword_record->clue_details[clue_start_locs[count1][count2][DOWN] - 1].clue_number = 
		    clue_counter;

		DEBUG_TRACE (
			"Down (%d,%d) [real]\n",
			count1,
			count2);

		clue_num++;
	    }

	    if (TRUE == clue_found)
		clue_counter++;
	}
    }

    if (clue_num != crossword_record->clue_sum)
    {
	printf (
	    "# Error getting all clue number information.  Only %u clues found\n",
	    clue_num);

	return (CWORD_S_FAILURE);
    }

    return (CWORD_S_NORMAL);

}



#else

extern int main (int argc, char *argv[])

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *      Main function driving the program to completion - STATS HARNESS
 *
 *  FORMAL PARAMETERS:
 *
 *      argc		(r) number of parameters inputted
 *	*argv[]		(r) contains the parameters inputted at command line
 *--
 */

{

    CROSSWORD_RECORD	    *crossword_record;
    int			    status;

    if (argc < 2)
    {
	exit (0);
    }

    status = cword_fileio_read_crossword (
			    argv[1],
			    crossword_record);

    if (FALSE == status)
	exit (EXIT_FAILURE);

    status = cword_stats_report_on_grid (
			    crossword_record);

    exit (CWORD_S_NORMAL);

}

#endif
