/*
 *  ?? Function that will return whether file inputted is of defined type i.e.
 * 
 *  cword_utils_is_file_of_type (
 *	int			file_type,
 *	char			*file_str);
 *
 *  where file_type could be any of the following: XRD_FILE, GRD_FILE,
 *  LOG_FILE, SED_FILE ??
 */

/*
 *++
 *  FACILITY:
 *
 *      utils.c
 *
 *  ABSTRACT:
 *
 *	Module containing many utility functions used by Cword.
 *
 *  AUTHORS:
 *
 *      James A. Lundon
 *
 *  CREATION DATE:     
 *
 *	22 February 1993
 *--
 */

/*
 *  Include files
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <time.h>
#include <ctype.h>
#include <math.h>
#include <stdlib.h>
#include <limits.h>
#include <signal.h>

/* 
 *  ?? Problem with varargs.h not being included to prototype the variable
 *  argument list functions ??
 */

#if defined(__unix) || defined(__unix__)

#include <sys/file.h>
#include <sys/uio.h>
#include <unistd.h>

#endif

#if defined(VMS) || defined(__VMS)

#include <unixio.h>
#include <file.h>

#endif

#if defined(__MSDOS__) || defined(MSDOS)

#include <io.h>
#include <fcntl.h>
#include <sys\stat.h>

#endif

#ifdef UTILS
#define MESSAGES_DEFINED 1 
#endif

#include "utils.h"
#include "fileio.h"
#include "main.h"
#include "os.h"
#include "messages.h"
#include "charui.h"
#include "memman.h"
#include "generate.h"
#include "word.h"
#include "stats.h"
#include "match.h"
#include "strll.h"
#include "defaults.h"

extern CWORD_G_DEFAULTS_RECORD	cword_g_defaults_record;



#ifndef UTILS

/*
 *  Only used within this module to reference the trace file 
 */

PRIVATE void cword_utils__id_system ();

PRIVATE void cword_utils__switch_clue_details (
	CLUE_DETAILS		*clue_1,
	CLUE_DETAILS		*clue_2);

/*
 *  As per C style guidelines => local global used to keep track of the
 *  trace log file.  Only used in this module, no where else! 
 */

static FILE *cword_utils_l_trace_file = (FILE *) 0;



PUBLIC int cword_utils_trim_filespec (
	char			*filespec)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 * 
 *	Truncates filespec string up to the first space found.
 *
 *  FORMAL PARAMETERS:
 *
 *      *filespec		(r/w) String containing file specification
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL		Successful completion
 *--
 */

{

    int count = 0;

    while (filespec[count] != ' ')
    	count++;

    filespec[count] = (char) 0;

    return (CWORD_S_NORMAL);

}



PUBLIC int cword_utils_trim_string (
	char			*trim_str)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 * 
 *      Trims the input string - AKA SQL/Rdb trim.
 *
 *  FORMAL PARAMETERS:
 *
 *      *trim_str		(r/w) String containing string to be trimmed
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL		Successful completion
 *--
 */

{

    int count = 0;

    while (trim_str[count] != (char) 0)
        count++;

    count--;
    while (trim_str[count] == (char) ' ')
        count--;

    count++;
    trim_str[count] = (char) 0;

    return (CWORD_S_NORMAL);

}



/* 
 *  ?? Maybe break into two ??
 */

PUBLIC int cword_utils_is_file_ok (
	int			read_or_write,
	char			*file_suffix,
	char			*file_spec)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function to determine whether a file exists before it is written
 *	to/read from.  This function is especially useful finding out whether a
 *	file exists for reading or whether we are overwriting a file which
 *	already exists.
 *
 *	This function is mostly used when reading or writing crossword/grid
 *	files but could have wider application if necessary.
 *
 *	The function cword_utils_is_filename_legal is called from within this
 *	to check whether the input file_spec is valid.
 *
 *  FORMAL PARAMETERS:
 *
 *	read_or_write		(r) Whether the file is to be written to or read from
 *      *file_spec		(r) file specification to be search for
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL		File is present
 *	CWORD_S_FAILURE		File in not present
 *--
 */

{

    char    temp_ch;

    int	    status;

    static int times_checked = 0;

    status = cword_utils_is_filename_legal (file_spec);

    if (CWORD_S_NORMAL != status)
    {
	return (CWORD_S_FAILURE);
    }

    if (read_or_write == WRITE_FILE)
    {
	/*
	 *  Want to write to the file in question.
	 */

	if (cword_utils_does_file_exist (file_spec) == TRUE)
	{
	    if (times_checked == 3)
	    {
		cword_messages_output_message (
			    CWORD_S_FILEINVALID,
			    SCREEN0);

		return (CWORD_S_FAILURE);
	    }
    
	    /*
	     *  File already exists - what to do then?
	     */

	    cword_messages_output_message (
			CWORD_S_OVERWRITEFILE,
			SCREEN0,
			file_spec,
			CWORD_Y_CHAR,
			CWORD_N_CHAR,
			CWORD_Y_CHAR);

	    temp_ch = cword_charui_get_ch (FALSE);

	    if (toupper (temp_ch) == CWORD_N_CHAR)
	    {
		cword_messages_output_message (
			    CWORD_S_INPUTFILESPEC,
			    SCREEN0,
			    file_spec,
			    file_suffix);

		if (cword_charui_get_file_spec (
			    TRUE,
			    file_suffix,
			    CWORD_SIZEOF_FILESPEC,
			    file_spec) == CWORD_S_NORMAL)
		{
		    times_checked++;

		    status = cword_utils_is_file_ok (
			    read_or_write,
			    file_suffix,
			    file_spec);

		    times_checked--;

		    if (CWORD_S_NORMAL != status)
		    {
			return (CWORD_S_FAILURE);
		    }
		}
		else
		{
		    cword_messages_output_message (
				CWORD_S_FILEINVALID,
				SCREEN0);

		    return (CWORD_S_FAILURE);
		}
	    }
	    else
	    {
		cword_messages_output_message (
			    CWORD_S_FILEOVERWRITE,
			    SCREEN0,
			    file_spec);
	    }

	    return (CWORD_S_NORMAL);
	}
	else
	{
	    /*
	     *  Trying to write to a file that doesn't exists is OK!
	     */

	    return (CWORD_S_NORMAL);
	}
    }
    else
    {
	/*
	 *  Want to read the file in question?
	 */

	if (cword_utils_does_file_exist (file_spec) == FALSE)
	{
	    /*
	     *  Trying to read from a file that doesn't exist - error!
	     */
    
	    cword_messages_output_message (
			CWORD_S_FILENOTEXIST,
			(TRACE1 | SCREEN0),
			file_spec);

	    return (CWORD_S_FAILURE);
	}
	else
	{
	    /*
	     *  Trying to read from a file that exists is OK!
	     */

	    return (CWORD_S_NORMAL);
	}
    }

    /* 
     *  Shouldn't get down to here
     */

}



PUBLIC long cword_utils_get_random_number ()

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function to get a random number.  Have to watch implementation across
 *	platforms for portability.  
 *
 *	This function will be possibly used when selecting a random word from a
 *	list, if all other characteristics of the word/light are equal.
 *
 *  FORMAL PARAMETERS:
 *
 *      None
 *
 *  FUNCTION VALUE:
 *
 *      Random number between 0 and 100
 *--
 */
{

    MY_DIV div_time;

    time_t temp_time;

    int count;

    time (&temp_time);

    /*
     *	Time returned as a long so cword_utils_div has to be able to handle
     *	them.
     */

    div_time = cword_utils_div (temp_time, 10);

    for (count = 0; count < (int) div_time.rem; count++)
	rand();

#if defined (__osf__) || defined (hpux) || defined (__hpux)

    /*
     *	OSF/1 and HP-UX have a different interpretation of what the rand ()
     *	function should return than all other systesm I've ported to.  In most
     *	systems rand () will return an integer from 0 to INT_MAX but on OSF/1
     *	and HP-UX rand () returns an integer from 0 to RAND_MAX (or 2**15 - 1).
     */

    return ((long) ((float) rand () / ((float) RAND_MAX/(float) 100)));

#else

    /*
     *  ?? Is there a definition for RAND_MAX on these other systems too ??
     */

    return ((long) ((float) rand () / ((float) INT_MAX/(float) 100)));

#endif

}



/*
 *  ?? ID user and system too (assuming TCP/IP is installed ??
 */

PRIVATE void cword_utils__id_system ()

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Not a very powerful.  It's used to roughly identify the OS running
 *	Cword (or its real name) to the log file!
 *
 *	Mostly an #ifdef job - there must be a better way of handling this than
 *	the way I'm doing it at the moment.
 *
 *  FORMAL PARAMETERS:
 *
 *      None
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  {@function value or completion codes@}
 *
 *      [@description or none@]
 *--
 */

{

#if defined (VMS) || defined (__VMS)

/*
 *  Split into VAX (__VAX) /AXP (__AXP) options.
 */

#if defined(vax)
    WRITE_TRACE ("# Executing on an OpenVMS/VAX system\n");
#elif defined(__ALPHA)
    WRITE_TRACE ("# Executing on an OpenVMS/AXP system\n");
#else
    WRITE_TRACE ("# Executing on a VAX system -- CPU type undefined\n");
#endif

#endif

#if defined (__unix) || defined (__unix__)

    WRITE_TRACE ("# Using a generic U*X system\n");

#ifdef __osf__
    WRITE_TRACE ("# Specifically using an Alpha OSF/1 system\n");
#endif

#ifdef __ultrix
    WRITE_TRACE ("# Specifically using an ULTRIX system\n");
#endif

#if defined (__hpux) || defined (hpux)
    WRITE_TRACE ("# Specifically using an HP-UX system\n");
#endif

#endif

#if defined(__MSDOS__) || defined(MSDOS)

    WRITE_TRACE ("# Using a generic DOS system\n");

#ifdef __TURBOC__
    WRITE_TRACE ("# Specifically using a BORLAND/TURBO C compiler\n");
#endif

#ifdef _MSC_VER
    WRITE_TRACE ("# Specifically using a Microsoft C compiler\n");
#endif

#endif

    return;

}



/*
 *  Could move trace file functionality into the message module as I 
 *  would then have access from within the output_message function to 
 *  whether the trace file is open or not ??
 */


PUBLIC int cword_utils_open_trace_file (
	char			*trace_file_name)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *      Function that kicks off trace file processing.
 *
 *  FORMAL PARAMETERS:
 *
 *      *trace_file_name	(r) O/S name of trace file
 *
 *  IMPLICIT INPUTS:
 *
 *      cword_utils_l_trace_file Trace file FILE pointer.
 *      cword_g_defaults_record.write_log_file 
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL		Trace file opened without errors
 *	CWORD_NOTRACE		Trace file not successfully created
 *--
 */
{

    char trace_text[CWORD_SIZEOF_MESSAGE + 1],
	 message_text[CWORD_SIZEOF_MESSAGE + 1],
	 temp_file_str[CWORD_SIZEOF_FILESPEC + 1],
	 time_str[CWORD_SIZEOF_TIME + 1];

    int	temp_val;

    /* 
     *  If the option no log file is chosen then return normall without doing
     *  anything.
     */

    if (cword_g_defaults_record.write_log_file == FALSE)
	return (CWORD_S_NORMAL);

    if (trace_file_name == (char *) 0)
    {
	cword_messages_output_message (
		    CWORD_TRACENOTOPEN,
		    SCREEN0);

	return (CWORD_NOTRACE);
    }

    /*
     *	By uniquely naming the trace file, it guarantees that no log file will
     *	be written over by mistake - particularly the case on DOS and U*X
     *
     *	First trace file is temp_file_str01.log (for example) and the others
     *	will be named temp_file_str[02-99].log.
     *
     *	?? Could make this a user default i.e. whether the log file is to be
     *	over written ??
     */

    temp_val = 0;
    do
    {
	temp_val++;

	/*
	 *  Looking for a trace log file that doesn't exist.
	 */

        if (temp_val < 99)
    	{
	    sprintf (
	        temp_file_str,
	        "%s%02d%s",
	        trace_file_name,
	        temp_val,
	        CWORD_SUFFIX_TRACE_FILE);
        }
    	else
    	{
	    sprintf (
	        temp_file_str,
	        "%s%03d%s",
	        trace_file_name,
	        temp_val,
	        CWORD_SUFFIX_TRACE_FILE);
    	}
    }
    while (cword_utils_does_file_exist (temp_file_str) == TRUE);    

    /*
     *  Open trace log file for writing.
     */

    cword_utils_l_trace_file = fopen (temp_file_str, "w");
	
    if (cword_utils_l_trace_file == (FILE *) 0)
    {
	cword_messages_output_message (
		CWORD_NOTRACE,
		SCREEN0);

	return (CWORD_NOTRACE);
    }
    else
    {
	CWORD_OS_GET_TIME (time_str);

	cword_messages_get_message (
	    CWORD_TRACEFILEOPEN,
	    message_text);

	sprintf (
	    trace_text,
	    message_text,
	    temp_file_str, 
	    time_str);

	if (fprintf (cword_utils_l_trace_file, trace_text) == -1)
	{
	    cword_messages_output_message (
		    CWORD_TRACEWRITEERR,
		    SCREEN0);

	    return (CWORD_S_FAILURE);
	}
	else
	{
	    cword_messages_output_message (
		    CWORD_TRACEFILEOPEN,
		    SCREEN0,
		    temp_file_str, 
		    time_str);
	}
    
	cword_utils__id_system ();

	return (CWORD_S_NORMAL);	
    }

}



PUBLIC int cword_utils_write_trace (
	char			*trace_str,
	...)
	    
/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *      Writes a line of Cword output to the trace file.
 *
 *  FORMAL PARAMETERS:
 *
 *      *trace_str		(r) Line of text to be written to trace file.
 *
 *  IMPLICIT INPUTS:
 *
 *      cword_utils_l_trace_file    trace file pointer
 *      cword_g_defaults_record.write_log_file 
 *
 *  FUNCTION VALUE:
 *
 *	CWORD_S_NORMAL		Successfuly wrote to trace file/trace file not enabled		    
 *	CWORD_S_FAILURE		Error outputting to trace file
 *--
 */
{

    va_list args;

    /* 
     *	If the option write_log_file is disabled *or* the log file is not
     *	opened yet then return CWORD_S_NORMAL without doing anything.
     */

    if ( (cword_g_defaults_record.write_log_file == FALSE) || 
	  (cword_utils_l_trace_file == (FILE *) 0) )
	return (CWORD_S_NORMAL);

    /*
     *  ?? Slick idea about how to improve trace file processing ?? :
     *
     *	Have a global varaible cword_global_trace_level? which will be
     *	incremented at the start of the function and then at the end of the
     *	same function decremented.  If WRITE_TRACE is called during the
     *	function then it will know how far down the call tree the program is at
     *	the time and portray this in its output to the trace file.  Thus you
     *	can get a DECforms like trace file which is somewhat similar to a
     *	function call tree.
     *
     *  The new write call will look like this then:
     *
     *  vsprintf (
     *		write_str,
     *		"%*c %s"
     *		cword_global_trace_level,
     *		trace_str);
     *
     *	fprintf (blah-blah-blah);
     *
     */

    va_start (args, trace_str);

    if (cword_utils_l_trace_file == (FILE *) 0)
    {
        /*
         *  If the trace file is not yet opened or just closed then return
         *  as if nothing has happened.
         *
         *  This is useful when Control-C'ing out of the application.
         */

        return (CWORD_S_NORMAL);
    }

    if (vfprintf (
	    cword_utils_l_trace_file, 
	    trace_str, 
	    args) == EOF)
    {
	cword_messages_output_message (
		CWORD_TRACEWRITEERR,
		SCREEN0);

	va_end (args);

	return (CWORD_S_FAILURE);
    }

    va_end (args);

    /*
     *	?? Maybe not the best to output an error message if fflush () doesn't
     *	work properly ??
     */

    if (fflush (cword_utils_l_trace_file) == EOF)
    {
	cword_messages_output_message (
		CWORD_TRACEFLUSHERR,
		SCREEN0);

	return (CWORD_S_FAILURE);
    }

    return (CWORD_S_NORMAL);

}



PUBLIC int cword_utils_write_trace_varargs (
	char			*trace_str,
	va_list			args)
	    
/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *      Writes a line of Cword output to the trace file.
 *
 *	Similar to cword_utils_write_trace except that this function accepts a
 *	va_list data type and immediately pushes it into a vfprintf.
 *
 *	Any architecture/design changes made in here should be reflected in the
 *	function cword_utils_write_trace too.
 *
 *  FORMAL PARAMETERS:
 *
 *      *trace_str		(r) Line of text to be written to trace file.
 *	args			(r) Variable argument list.
 *
 *  IMPLICIT INPUTS:
 *
 *      cword_utils_l_trace_file    trace file pointer
 *      cword_g_defaults_record.write_log_file 
 *
 *  FUNCTION VALUE:
 *
 *	CWORD_S_NORMAL		Successfuly wrote to trace file/trace file not enabled		    
 *	CWORD_S_FAILURE		Error outputting to trace file
 *--
 */
{

    /* 
     *	If the option write_log_file is disabled *or* the log file is not
     *	opened yet then return CWORD_S_NORMAL without doing anything.
     */

    if ( (cword_g_defaults_record.write_log_file == FALSE) || 
	  (cword_utils_l_trace_file == (FILE *) 0) )
	return (CWORD_S_NORMAL);

    if (vfprintf (
	    cword_utils_l_trace_file, 
	    trace_str, 
	    args) == EOF)
    {
	cword_messages_output_message (
		CWORD_TRACEWRITEERR,
		SCREEN0);

	va_end (args);

	return (CWORD_S_FAILURE);
    }

    if (fflush (cword_utils_l_trace_file) == EOF)
    {
	cword_messages_output_message (
		CWORD_TRACEFLUSHERR,
		SCREEN0);

	return (CWORD_S_FAILURE);
    }

    return (CWORD_S_NORMAL);

}



PUBLIC void cword_utils_close_trace_file ()

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *      Close the trace file.
 *
 *  FORMAL PARAMETERS:
 *
 *      None
 *
 *  IMPLICIT INPUTS:
 *
 *      cword_utils_l_trace_file    trace file pointer
 *      cword_g_defaults_record.write_log_file 
 *
 *  FUNCTION VALUE:
 *
 *	CWORD_S_NORMAL		Successfuly closed trace file		    
 *	CWORD_NOTRACE		Error in trace file
 *--
 */
{

    char time_str[CWORD_SIZEOF_TIME + 1];

    /* 
     *  If the option no log file is chosen then return normall without doing
     *  anything.
     */

    if (cword_g_defaults_record.write_log_file == FALSE)
	return;

    if (cword_utils_l_trace_file == (FILE *) 0)
    {
	return;
    }

    CWORD_OS_GET_TIME (time_str);

    cword_messages_output_message (
		CWORD_TRACEFILECLOSE,
		TRACE1,
		time_str);

    fclose (cword_utils_l_trace_file);

    cword_utils_l_trace_file = (FILE *) 0;

    return;

}



PUBLIC int cword_utils_debug_output (
	char			*debug_str,
	...)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function that is used for conditional message outputting.  It checks
 *	for the existance of the environment variable CWORD_DEBUG and if it
 *	find it, it will output to stdout the contents of the input function
 *	parameters plus a little header saying that it is debuggable output.
 *
 *  FORMAL PARAMETERS:
 *
 *	*debug_str		(r) String to output 
 *	...			(r) Variable argument list used with debug_str
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL		Successful outputting debug information
 *	CWORD_S_FAILURE		Unsuccessful outputting debug information
 *--
 */

{

    static BOOLEAN is_cword_debug_set = FALSE,
		   is_cword_debug_checked = FALSE;

    if (is_cword_debug_checked == FALSE)
    {
	if (getenv ("CWORD_DEBUG") != (char *) 0)
	    is_cword_debug_set = TRUE;

	is_cword_debug_checked = TRUE;
    }

    if (is_cword_debug_set == TRUE)
    {
	va_list args;

	char output_str[CWORD_SIZEOF_MESSAGE + 1];

	sprintf (
		output_str,
		"%s: %s",
		CWORD_DEBUG_STR,
		debug_str);

	va_start (args, debug_str);

	if (vfprintf (
		    stdout,
		    output_str,
		    args) == EOF)
	{
	    printf ("# Error using vfprintf\n");

	    va_end (args);

	    return (CWORD_S_FAILURE);
	}

	va_end (args);
    }

    return (CWORD_S_NORMAL);

}



PUBLIC int cword_utils_complete_definition (
	CROSSWORD_RECORD	*crossword_record)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	The function will complete the filling of a crossword record so that it
 *	will be outputted to a file in a state recognisable for printing.
 *
 *  IMPLICIT INPUTS:
 *
 *      cword_g_defaults_record.temp_crossword_name
 *
 *  FORMAL PARAMETERS:
 *
 *      *crossword_record	    (w)	pointer to an instance of a crossword_record
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL		Successful completion
 *--
 */

{

    int	count1,
	count2,
	answer = 0,
	description = 0;

    if ((char) 0 == crossword_record->name[0])
    {
	/*
	 *  No crossword name needed -- default none!
	 */
    }

    if (crossword_record->name[0] != (char) 0)
	cword_messages_output_message (
	    CWORD_NEWSVALSINXWORD,
	    TRACE2,
	    crossword_record->name);
    else
	cword_messages_output_message (
	    CWORD_NEWSVALSINXWORD,
	    TRACE2,
	    CWORD_UNDEFINED_STR);

    if ((char) 0 == crossword_record->remarks[0]) 
    {
	/*
	 *  No crossword remarks needed -- default none!
	 */
    }

    for (count1 = 0; count1 < crossword_record->clue_sum; count1++ )
    {
	if (crossword_record->clue_details[count1].clue_answer[0] == (char) 0)
	{
	    count2 = 0;

	    /*
	     *  There is only a question mark in the clue answer when not inputted
	     */

	    answer++;

	    while (count2 < crossword_record->clue_details[count1].answer_length)
	    {
		crossword_record->clue_details[count1].clue_answer[count2] = '?';
		count2++;
	    }
	}
    
	if (crossword_record->clue_details[count1].clue_description[0] == (char) 0)
	{
	    description++;

	    strcpy (
		crossword_record->clue_details[count1].clue_description, 
		crossword_record->clue_details[count1].clue_answer);
	}
    }

    if (description > 0)
    {
	DEBUG_TRACE (
	    "New clue descriptions required for %2d clues\n",
	    description);
    }

    if (answer > 0)
    {
	DEBUG_TRACE (
	    "New answers required for %2d clues\n",
	    answer);
    }

    crossword_record->is_completely_specified = TRUE;

    return (CWORD_S_NORMAL);

}



PRIVATE void cword_utils__switch_clue_details (
	CLUE_DETAILS		*clue_1,
	CLUE_DETAILS		*clue_2)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function to switch the contents of two clues in a crossword record.
 *
 *	Used for sorting a crossword/grid definition.
 *
 *  FORMAL PARAMETERS:
 *
 *      *clue1			(w) Contents of clue1 to be switched with clue2.
 *	*clue2			(w) Contents of clue2 to be switched with clue1.
 *
 *  FUNCTION VALUE:
 *
 *      None!
 *--
 */

{

    CLUE_DETAILS temp_clue;

    temp_clue.clue_number   = clue_1->clue_number;
    temp_clue.clue_direction= clue_1->clue_direction;
    temp_clue.x_coordinate  = clue_1->x_coordinate;
    temp_clue.y_coordinate  = clue_1->y_coordinate;
    temp_clue.answer_length = clue_1->answer_length;
    strcpy (temp_clue.clue_description, clue_1->clue_description);
    strcpy (temp_clue.clue_answer, clue_1->clue_answer);

    clue_1->clue_number	    = clue_2->clue_number;
    clue_1->clue_direction  = clue_2->clue_direction;
    clue_1->x_coordinate    = clue_2->x_coordinate;
    clue_1->y_coordinate    = clue_2->y_coordinate;
    clue_1->answer_length   = clue_2->answer_length;
    strcpy (clue_1->clue_description, clue_2->clue_description);
    strcpy (clue_1->clue_answer, clue_2->clue_answer);

    clue_2->clue_number	    = temp_clue.clue_number;
    clue_2->clue_direction  = temp_clue.clue_direction;
    clue_2->x_coordinate    = temp_clue.x_coordinate;
    clue_2->y_coordinate    = temp_clue.y_coordinate;
    clue_2->answer_length   = temp_clue.answer_length;
    strcpy (clue_2->clue_description, temp_clue.clue_description);
    strcpy (clue_2->clue_answer, temp_clue.clue_answer);

    return;

}



PUBLIC void cword_utils_sort_crossword (
	CROSSWORD_RECORD	*crossword_record)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	This function will sort the inputted CROSSWORD_RECORD so that it can
 *	then be processed more effectively - either to an output grid/crossword
 *	file or further within cword.
 *
 *	Sorting - in this case - involves the clustering into ACROSS and DOWN
 *	the clues together with the correct sequence in which the clues will be
 *	encountered in the crossword from top left to bottom right hand side.
 *
 *	N.B. Can be use for both crossword and grid definitions.
 *
 *  FORMAL PARAMETERS:
 *
 *      *crossword_record	(r/w) pointer to an instance of a CROSSWORD_RECORD
 *
 *  FUNCTION VALUE:
 *
 *	CWORD_S_NORMAL		Successful completion
 *--
 */
{

    int count1;

    BOOLEAN is_sorted = FALSE;

    if (crossword_record->clue_sum == 0)
    {
	printf ("# Error with input crossword/grid definition for sorting\n");

	return;
    }

    /*
     *	Generate clue numbers for the crossword before trying to sorting it.
     */

    cword_stats_gen_clue_nums (crossword_record);

    if (crossword_record->name[0] != (char) 0)
    {
	cword_messages_output_message (
	    CWORD_SORTCROSSWORD,
	    TRACE3,
	    crossword_record->name);
    }
    else
    {
	cword_messages_output_message (
	    CWORD_SORTCROSSWORD,
	    TRACE3,
	    CWORD_UNDEFINED_STR);
    }

    /*
     *	?? Bubble sort - yuck...I know. ??
     */

    while (is_sorted == FALSE)
    {
	is_sorted = TRUE;

	for (count1 = 0; count1 < crossword_record->clue_sum - 1; count1++)
	{
	    if (crossword_record->clue_details[count1].clue_direction < 
		  crossword_record->clue_details[count1 + 1].clue_direction)
	    {
		/*
		 *  Switch anyhow!
		 */

		cword_utils__switch_clue_details (
		    &crossword_record->clue_details[count1],
		    &crossword_record->clue_details[count1 + 1]);

		is_sorted = FALSE;
	    }
	    else if (crossword_record->clue_details[count1].clue_direction == 
	  	      crossword_record->clue_details[count1 + 1].clue_direction)
	    {
		/*
		 *  Check the clue numbers and if greater then switch else
		 *  don't.
		 */

		if (crossword_record->clue_details[count1].clue_number >
		     crossword_record->clue_details[count1 + 1].clue_number)
		{
		    cword_utils__switch_clue_details (
			&crossword_record->clue_details[count1],
			&crossword_record->clue_details[count1 + 1]);

		    is_sorted = FALSE;
		}
	    }
	}
    }

    return;

}



PUBLIC int cword_utils_manage_sigint (
	int			sigint_val)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function to enable/disable Control-C processing.
 *
 *  FORMAL PARAMETERS:
 *
 *      sigint_val		(r) ENABLE/DISABLE value.
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL		Normal completion.
 */

{

    /*
     *  Enable/Disable SIGINT depending on sigint_val.
     */

    if (sigint_val == ENABLE)
    {
	/*
	 *  Establish/re-establish proper Control-C processing using
	 *  cword_utils_grab_signal.
	 */

	signal (SIGINT, cword_utils_grab_signal);
    }
    else if (sigint_val == DISABLE)
    {
	/*
	 *  Ignore Control-C
	 */

        signal (SIGINT, SIG_IGN);
    }
    else
    {
	printf ("# sigint_val not valid\n");
    }

    return (CWORD_S_NORMAL);

}



/*
 *  This function should not be used at all when in Windows (Quickwin).
 */

#if !defined (_WINDOWS)

#if defined(VMS) || defined(__VMS)

PUBLIC void cword_utils_grab_signal (
	int			signal_no,
    	...)

#else

PUBLIC void cword_utils_grab_signal (
	int			signal_no)

#endif

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Simple function to grab SIGINT signals.  print error message,
 *      close the trace file, and exit with error.
 *
 *  FORMAL PARAMETERS:
 *
 *      signal_no		(r) signal number generated
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  COMPLETION CODES:
 *
 *      EXIT_FAILURE
 *
 *  SIDE EFFECTS:
 *
 *      Program dies :-(
 *--
 */

{

    cword_messages_output_message (
	    CWORD_SIGNALGENERATED,
	    (SCREEN0 | TRACE0));

    /*
     *  ?? Be more careful about re-enterancy problems ??
     */

    /*
     *	?? In strict ANSI C mode using DEC C on OpenVMS, there is a problem
     *	with the lack of a prototype for signal ??
     *
     *			signal (SIGINT, cword_utils_grab_signal);
     *		........^
     *	%CC-W-PTRMISMATCH, In this statement, the referenced type of the pointer value
     *	 "cword_utils_grab_signal" is "function (...) returning void", which is not
     *	 compatible with "function (...) returning void".
     *			At line number ???? in UTILS.C
     *
     *	?? I don't understand the message - will have to investigate this
     *	further ??
     */ 

    if (signal_no == SIGINT)
    {
	signal (SIGINT, cword_utils_grab_signal);

	cword_messages_output_message (CWORD_SIGINTMSG, (TRACE0 | SCREEN0));

        /*
         *  Close the trace log file now to prevent spurious trace 
         *  messages being outputted.
         *
         *  ?? It isn't possible to easily prevent CWORD_TRACEWRITEERR
         *  messages from being outputted other than to use a static 
         *  boolean ??
         */

        CLOSE_TRACE ();
    }

    if (signal_no == SIGFPE)
    {
	signal (SIGFPE, cword_utils_grab_signal);

	cword_messages_output_message (CWORD_SIGFPEMSG, (TRACE0 | SCREEN0));
    }

    /*
     *	This signal should exist on Visual C++ 1.5 if we create a project with
     *	a DOS exe target.
     */

    if (signal_no == SIGSEGV)
    {
	signal (SIGSEGV, cword_utils_grab_signal);

	cword_messages_output_message (CWORD_SIGSEGVMSG, (TRACE0 | SCREEN0));
    }

#if !(defined(__MSDOS__) || defined(MSDOS))

    if (signal_no == SIGBUS)
    {
	signal (SIGBUS, cword_utils_grab_signal);

	cword_messages_output_message (CWORD_SIGBUSMSG, (TRACE0 | SCREEN0));
    }

#endif

    if (signal_no != SIGINT)
    {
	cword_messages_output_message (
	    CWORD_S_REPORTPROBLEMS, 
	    (TRACE0 | SCREEN0),
	    CWORD_VERSION_ID);

	/*
	 *  Really make sure that the user is awake when I'm outputting this
	 *  message :-(.
	 */

	cword_charui_ring_bell (3);
    }	

    cword_messages_output_message (CWORD_SIGEXITMSG, (TRACE0 | SCREEN0));

    CLOSE_TRACE ();

    exit (EXIT_FAILURE | NO_SCREEN);

}

#endif



PUBLIC MY_DIV cword_utils_div (
	long			numer,
	long			denom)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	My own div() function as this isn't supported on some of the systems I
 *	am porting Cword to.
 *
 *  FORMAL PARAMETERS:
 *
 *      numer			(r) numerator
 *	denom			(r) denominator
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      temp			Instance of MY_DIV data structure! 
 *--
 */

{

    MY_DIV temp;

    if (denom == 0)
    {
	temp.quot = 0;
	temp.rem = 0;
	return (temp);
    }

    temp.quot = numer / denom;
    temp.rem = numer - denom*temp.quot;

    return (temp);

}



PUBLIC int cword_utils_is_filename_legal (
	char			*legalfile_str)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function to check whether the inputted file spec is a valid one for the
 *	system Cword is being executed on.
 *
 *  FORMAL PARAMETERS:
 *
 *	*legalfile_str		(r) Check the inputted string for filename validity
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  {@function value or completion codes@}
 *
 *      [@description or none@]
 *--
 */

{

    /*
     *  ?? 8.3 on DOS ??
     */

    /*
     *  ?? More detail is needed ??
     *
     *  ?? What about U*X style file names ??
     *
     *	?? Taken out ".." as a part of an invalid file specification since our
     *	old friend U*X uses it ??
     */

    if ( (strstr (legalfile_str, "*") != (char *) 0) || 
	  (strstr (legalfile_str, "%") != (char *) 0) ||
	   (strstr (legalfile_str, "#") != (char *) 0) )
    {
	cword_messages_output_message (
		CWORD_ILLFILENAME,
		SCREEN0,
		legalfile_str);

	return (CWORD_S_FAILURE);
    }

    return (CWORD_S_NORMAL);

}



/*
 *  ?? Badly named function ??
 *
 *  ?? cword_utils_sleep () ??
 */

PUBLIC void cword_utils_sleep_for (
	unsigned int		time_val)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *      Function to sleep for a predefined period of time.  Need some work for DOS?
 *
 *  FORMAL PARAMETERS:
 *
 *      time_val		(r) Time in seconds to sleep
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      None!
 *--
 */

{

#if defined(__MSDOS__) || defined(MSDOS)

    /*
     *	?? Real kludge but it seems that there is no sleep function defined in
     *	Visual C++ 1.5 ??
     */

    long index,
	 count;

    index = 0;
    for (count = 0; count < (time_val * 1000); count++)
    	index = count;
    index = 0;
    	
#else

    /*
     *	In OpenVMS, in strict ANSI C mode, using DEC C there is a problem
     *	to do with absence of a prototype for sleep.  This is solved by adding
     *	the following:
     *  
     *	    /undefine="__HIDE_FORBIDDEN_NAMES"
     *
     *	This has the effect of releasing, to the compiler, a prototype
     *	definition for sleep.
     */

    sleep (time_val);

#endif

    return;

}



PUBLIC BOOLEAN cword_utils_remove_file (
	char			*file_str)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *      [@tbs@]
 *
 *  FORMAL PARAMETERS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      TRUE			File deleted.
 *	FALSE			File not deleted or not found.
 *--
 */

{

    /*
     *  ?? VMS -> file.txt;* for all revisions ??
     */

    if (cword_utils_does_file_exist (file_str) == TRUE)
    {
	/*
	 *  File found!
	 */

	if (remove (file_str) != 0)
	{
	    cword_messages_output_message (
		CWORD_S_DELETEFILEERR,
		(TRACE0 | SCREEN0),
		file_str);

	    return (FALSE);
	}
	else
	{
	    DEBUG_TRACE (
		"Deleted %s\n",
		file_str);

	    return (TRUE);
	}
    }
    else
    {
	/*
	 *  Deleting a file that doesn't exist is usually an error.
	 */

	if (file_str[0] != (char) 0)
	{
	    cword_messages_output_message (
		CWORD_S_DELETEFILEERR,
		(TRACE0 | SCREEN0),
		file_str);
	}

	return (FALSE);
    }

}



PUBLIC BOOLEAN cword_utils_rename_file (
	char			*from_str,
	char			*to_str)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *      [@tbs@]
 *
 *  FORMAL PARAMETERS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      TRUE			*from_str renamed to *to_str.
 *	FALSE			Error renaming *from_str to *to_str.
 *--
 */

{

    if (cword_utils_does_file_exist (from_str) == TRUE)
    {
	/*
	 *  The file from_str is found!
	 */

	if (cword_utils_does_file_exist (to_str) == TRUE)
	{
	    /*
	     *  The file to_str is found!
	     *
	     *  Delete before continuing with renaming process.
	     */

	    if (cword_utils_remove_file (to_str) == FALSE)
	    {
		return (FALSE);
	    }
	    else
	    {
		if (rename (from_str, to_str) != 0)
		{
		    cword_messages_output_message (
			CWORD_S_RENAMEFILEERR,
			(TRACE0 | SCREEN0),
			from_str,
			to_str);

		    return (FALSE);
		}
		else
		{
		    DEBUG_TRACE (
			"Renamed %s to %s\n",
			from_str,
			to_str);

		    return (TRUE);
		}
	    }
	}
	else
	{
	    if (rename (from_str, to_str) != 0)
	    {
		cword_messages_output_message (
		    CWORD_S_RENAMEFILEERR,
		    (TRACE0 | SCREEN0),
		    from_str,
		    to_str);

		return (FALSE);
	    }
	    else
	    {
		DEBUG_TRACE (
		    "Renamed %s to %s\n",
		    from_str,
		    to_str);

		return (TRUE);
	    }
	}
    }
    else
    {
	cword_messages_output_message (
	    CWORD_S_RENAMEFILEERR,
	    (TRACE0 | SCREEN0),
	    from_str,
	    to_str);

	return (FALSE);
    }

}



/*
 *  ?? Badly named function ??
 *
 *  ?? cword_utils_access () ??
 */

PUBLIC BOOLEAN cword_utils_does_file_exist (
	char			*file_str)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Returns TRUE if file exists, return FALSE if it doesn't!
 *
 *  FORMAL PARAMETERS:
 *
 *      *file_str		(r) File specification to be checked for existance.
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      TRUE			File exists!
 *	FALSE			File doesn't exist!
 *--
 */

{

    DEBUG_TRACE (
	"Checking existance of %s using cword_utils_does_file_exist\n",
	file_str);

    /*
     *	?? Just checking that the file file_str exists - nothing more.  Could
     *	be more effective to check whether the file is writable, readable, or
     *	just plain exists (within cword) in later versions ??
     */

#if defined(__MSDOS__) || defined(MSDOS)
    if (_access (file_str, 0) != 0)
#else
    if (access (file_str, 0) == 0)
#endif
    {
	/*
	 *  File exists!
	 */

	return (TRUE);
    }
    else
    {
	/*
	 *  File doesn't exist!
	 */

	return (FALSE);
    }

}



#else

extern int main (
        int			argc, 
	char			*argv[])

{
}

#endif
