/*
 *++
 *  FACILITY:
 *
 *      verify.c
 *
 *  ABSTRACT:
 *
 *	Module containing functions to verify the validity of an inputted
 *	crossword definition.
 *
 *	Much of the following functionality is almost obsolete as crossword and
 *	grid definitions are now created via Cword itself, and not by hand as
 *	was the case in the early life of the project.  Much of this
 *	functionality will continue to be executed as it's still possible to
 *	input an ill-defined hand crafted grid, or crossword definition, to
 *	cword.  Not that you should be doing things like that now!
 *
 *  AUTHORS:
 *
 *      James A. Lundon
 *
 *  CREATION DATE:     
 *
 *      12 April 1993
 *--
 */

/*
 *  Include Files
 */

#include <stdio.h>
#include <string.h>
#include <math.h>

#ifdef VERIFY
#define MESSAGES_DEFINED 1
#endif

#include "main.h"
#include "verify.h"
#include "messages.h"
#include "print.h"
#include "fileio.h"
#include "generate.h"
#include "defaults.h"
#include "utils.h"
#include "stats.h"

extern CWORD_G_DEFAULTS_RECORD	cword_g_defaults_record;

/*
 *  Proportion of grid blank that will constitute, or not, an American grid.
 */

#define AMERICAN_BLANKS	    17



#ifndef VERIFY

PRIVATE int cword_verify__check_inputs (
	CROSSWORD_RECORD	*crossword_record);

PRIVATE void cword_verify__change_symmetry (
	int			new_symmetry,
	CROSSWORD_RECORD	*crossword_record);

PRIVATE int cword_verify__isloated_grid (
	int			initialised,
	int			clue_subscript,
	CROSSWORD_RECORD	*crossword_record);

PRIVATE BOOLEAN cword_verify__free_cell_found (
	CROSSWORD_RECORD	*crossword_record);



PRIVATE int cword_verify__check_inputs (
	CROSSWORD_RECORD	*crossword_record)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function that the crossword definition is valid.  What this function
 *	checks includes:
 *
 *	    1. answer length are equal to strlen (answer)
 *	    2, grid co-ordinates of lights are within the grid
 *	    3. ...
 *
 *  FORMAL PARAMETERS:
 *
 *      *crossword_record	(r) crossword details
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL		Crossword defiintion parameters verified
 *	CWORD_S_FAILURE		Problems with crossword definition
 *--
 */

{

    int count1;

    /*
     *	Check the crossword definition light lengths against the defined
     *	quantities?
     */

    for (count1 = 0; count1 < crossword_record->clue_sum; count1++)
    {
	if ( (crossword_record->clue_details[count1].answer_length 
		    != (int) strlen (crossword_record->clue_details[count1].clue_answer))
			    &&
	     (crossword_record->clue_details[count1].answer_length < crossword_record->x_dimensions + 1) )
	{   
	    if (crossword_record->clue_details[count1].clue_direction == ACROSS)
	    {
		cword_messages_output_message (
			    CWORD_VERINVANSNUM,
			    (TRACE2 | SCREEN2),
			    crossword_record->clue_details[count1].clue_number, 
			    CWORD_ACROSS_STR);
	    }
	    else
	    {	
		cword_messages_output_message (
			    CWORD_VERINVANSNUM,
			    (TRACE2 | SCREEN2),
			    crossword_record->clue_details[count1].clue_number, 
			    CWORD_DOWN_STR);
	    }

	    return (CWORD_S_FAILURE);
	}
    }

    /*
     *  Check that the inputted co-ordinates of the crossword lights are
     *  within the defined grid dimensions?
     */

    for (count1 = 0; count1 < crossword_record->clue_sum; count1++)
    {
	if ( (crossword_record->clue_details[count1].x_coordinate > crossword_record->x_dimensions)
				    ||
	      (crossword_record->clue_details[count1].y_coordinate > crossword_record->y_dimensions) )
	{
	    cword_messages_output_message (
			CWORD_VERINVANSLOC,
			(TRACE2 | SCREEN2),
			crossword_record->clue_details[count1].clue_number, 
			crossword_record->clue_details[count1].clue_direction ?
							CWORD_ACROSS_STR : CWORD_DOWN_STR);

	    return (CWORD_S_FAILURE);
	}
    }

    return (CWORD_S_NORMAL);

}



PUBLIC int cword_verify_check_ximenes (
	CROSSWORD_RECORD	*crossword_record)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *      [@tbs@]
 *
 *  FORMAL PARAMETERS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  {@function value or completion codes@}
 *
 *      [@description or none@]
 *--
 */

{

    printf ("## NYI: checking Ximenes conventions on barred grid\n");

    return (CWORD_S_NORMAL);

}



PUBLIC BOOLEAN cword_verify_is_fully_speced (
	CROSSWORD_RECORD	*crossword_record)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function that reports whether the inputted crossword is fully user
 *	specified.  By this I mean are all the clues valid for the crossword
 *	inputted.
 *
 *  FORMAL PARAMETERS:
 *
 *      *crossword_record	(r) crossword details
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      TRUE			Crossword is fully specified.
 *	FALSE			Crossword inputted isn't fully specified.
 *--
 */

{

    int count;

    BOOLEAN is_grid_speced = TRUE;

    for (count = 0; count < crossword_record->clue_sum; count++)
    {
	if (strcmp (
		crossword_record->clue_details[count].clue_description,
		crossword_record->clue_details[count].clue_answer) == 0)
	{
	    is_grid_speced = FALSE;

	    break;
	}
    }

    return (is_grid_speced);

}



PRIVATE void cword_verify__change_symmetry (
	int			new_symmetry,
	CROSSWORD_RECORD	*crossword_record)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *      [@tbs@]
 *
 *  FORMAL PARAMETERS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      None!
 *--
 */

{

    char temp_str[CWORD_SIZEOF_LINE + 1];

    cword_messages_output_message (
		CWORD_S_SYMMNOTCORRECT,
		(TRACE2 | SCREEN2));

    crossword_record->symmetry = new_symmetry;

    if (crossword_record->symmetry == DEG_90_SYMMETRY)
	cword_messages_get_message (CWORD_90SYMMETRYMSG, temp_str);
    else if (crossword_record->symmetry == DEG_180_SYMMETRY)
	cword_messages_get_message (CWORD_180SYMMETRYMSG, temp_str);
    else if (crossword_record->symmetry == AXIAL_LR_SYMMETRY)
	cword_messages_get_message (CWORD_AXIALLEFTRIGHTMSG, temp_str);
    else if (crossword_record->symmetry == AXIAL_RL_SYMMETRY)
	cword_messages_get_message (CWORD_AXIALRIGHTLEFTMSG, temp_str);
    else if (crossword_record->symmetry == NO_SYMMETRY)
	cword_messages_get_message (CWORD_NOSYMMETRYMSG, temp_str);
    else
	cword_messages_get_message (CWORD_NOSYMMETRYMSG, temp_str);

    cword_messages_output_message (
		CWORD_S_SYMMCHANGE,
		(TRACE2 | SCREEN2),
		temp_str);

    return;

}



PUBLIC int cword_verify_get_block_symm (
	CROSSWORD_RECORD	*crossword_record)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function to check whether symmetry is present in a crossword grid.  If
 *	symmetry isn't present the crossword is rejected as incorrectly
 *	compiled.  There are a few different measures of symmetry as it
 *	pertains to crossword compilation.   A crossword can be symmetric around
 *	a positive 90 degree rotation or, 180, or 270, degree rotation.
 *
 *  FORMAL PARAMETERS:
 *
 *      *crossword_record	(r) Pointer to crossword struct crossword definition
 *
 *  FUNCTION VALUE:
 *
 *	Symmetry constant	DEG_90_SYMMETRY, DEG_180_SYMMETRY, AXIAL_LR_SYMMETRY,
 *				AXIAL_RL_SYMMETRY or NO_SYMMETRY
 *--
 */

{

    CROSSWORD_GRID block_grid,
		   mirror_grid;

    int	count1,
	count2,
	status;

    BOOLEAN is_symmetry_found;

    /*
     *	I'm assuming that we have a blocked grid before getting block symmetry
     *	props.
     */

    status = cword_verify_get_block_props (
		crossword_record,
		block_grid);

    if (CWORD_S_NORMAL != status)
    {
	/*
	 *  Return invalid symmetry type rather than CWORD_S_FAILURE.
	 */

	return (INVALID_SYMMETRY);
    }

    /*
     *  Check 90 degree symmetry by moving 90 degrees anti-clockwise!
     */

    is_symmetry_found = TRUE;
    for (count2 = 0; count2 < crossword_record->x_dimensions; count2++)
	for (count1 = 0; count1 < crossword_record->y_dimensions; count1++)
	    mirror_grid[count1][count2] = 
		    block_grid[crossword_record->y_dimensions - count2 - 1]
				  [count1];

    for (count1 = 0; count1 < crossword_record->x_dimensions; count1++)
	for (count2 = 0; count2 < crossword_record->y_dimensions; count2++)
	{
	    if (mirror_grid[count1][count2] != block_grid[count1][count2])
	    {
		is_symmetry_found = FALSE;
		break;
	    }
	}

    if (is_symmetry_found == TRUE)
    {
	return (DEG_90_SYMMETRY);
    }

    /*
     *	If 90 degree symmetry isn't found then check for 180 degree next!
     */

    is_symmetry_found = TRUE;
    for (count1 = 0; count1 < crossword_record->x_dimensions; count1++)
	for (count2 = 0; count2 < crossword_record->y_dimensions; count2++)
	    mirror_grid[count1][count2] = 
		    block_grid[crossword_record->x_dimensions - count1 - 1]
				  [crossword_record->y_dimensions - count2 - 1];

    for (count1 = 0; count1 < crossword_record->x_dimensions; count1++)
	for (count2 = 0; count2 < crossword_record->y_dimensions; count2++)
	{
	    if (mirror_grid[count1][count2] != block_grid[count1][count2])
	    {
		is_symmetry_found = FALSE;
		break;
	    }
	}

    if (is_symmetry_found == TRUE)
    {
	return (DEG_180_SYMMETRY);
    }

    /*
     *	If 90 degree/180 degree symmetry not found then check for other axial
     *	symmetries - Left-top to Right-bottom!
     */

    is_symmetry_found = TRUE;
    for (count1 = 0; count1 < crossword_record->x_dimensions; count1++)
	for (count2 = 0; count2 < crossword_record->y_dimensions; count2++)
	    mirror_grid[count1][count2] = 
		    block_grid[crossword_record->y_dimensions - count2 - 1]
				  [crossword_record->x_dimensions - count1 - 1];

    for (count1 = 0; count1 < crossword_record->x_dimensions; count1++)
	for (count2 = 0; count2 < crossword_record->y_dimensions; count2++)
	{
	    if (mirror_grid[count1][count2] != block_grid[count1][count2])
	    {
		is_symmetry_found = FALSE;
		break;
	    }
	}

    if (is_symmetry_found == TRUE)
    {
	return (AXIAL_LR_SYMMETRY);
    }

    /*
     *	If 90 degree/180 degree/Left-top Right-bottom symmetry not found then
     *	check for other axial symmetries - Right-top to Left-bottom!
     */

    is_symmetry_found = TRUE;
    for (count2 = 0; count2 < crossword_record->x_dimensions; count2++)
	for (count1 = 0; count1 < crossword_record->y_dimensions; count1++)
	    mirror_grid[count1][count2] = 
		    block_grid[count2][count1];

    for (count1 = 0; count1 < crossword_record->x_dimensions; count1++)
	for (count2 = 0; count2 < crossword_record->y_dimensions; count2++)
	{
	    if (mirror_grid[count1][count2] != block_grid[count1][count2])
	    {
		is_symmetry_found = FALSE;
		break;
	    }
	}

    if (is_symmetry_found == TRUE)
    {
	return (AXIAL_RL_SYMMETRY);
    }

    /*
     *	If we get down to here then no symmetry has been found so the only
     *	symmetry left is: No Symmetry!
     */

    return (NO_SYMMETRY);

}



#define BARRED_GRID_PRINT_CHECK(barred_grid)    if (barred_grid & LEFT) \
						    printf ("barred LEFT defined\n"); \
						if (barred_grid & RIGHT) \
						    printf ("barred RIGHT defined\n"); \
						if (barred_grid & TOP) \
						    printf ("barred TOP defined\n"); \
						if (barred_grid & BOTTOM) \
						    printf ("barred BOTTOM defined\n");

#define MIRROR_GRID_PRINT_CHECK(mirror_grid)    if (mirror_grid & LEFT) \
						    printf ("mirror LEFT defined\n"); \
						if (mirror_grid & RIGHT) \
						    printf ("mirror RIGHT defined\n"); \
						if (mirror_grid & TOP) \
						    printf ("mirror TOP defined\n"); \
						if (mirror_grid & BOTTOM) \
						    printf ("mirror BOTTOM defined\n"); \



PUBLIC int cword_verify_get_barred_symm (
	CROSSWORD_RECORD	*crossword_record)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	This function will get the symmetry type of an inputted barred grid.
 *
 *  FORMAL PARAMETERS:
 *
 *      *crossword_record	(r) Pointer to crossword definition
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      [@description or none@]
 *--
 */

{

    CROSSWORD_GRID barred_grid,
		   mirror_grid;

    int count1,
        count2,
	status;

    BOOLEAN is_symmetry_found;

    status = cword_verify_get_barred_props (
		crossword_record,
		barred_grid);

    if (CWORD_S_NORMAL != status)
    {
	/*
	 *  Return invalid symmetry type rather than CWORD_S_FAILURE.
	 */

	return (INVALID_SYMMETRY);
    }

    /*
     *  Check 90 degree symmetry by moving 90 degrees anti-clockwise!
     */

    is_symmetry_found = TRUE;
    for (count2 = 0; count2 < crossword_record->x_dimensions; count2++)
	for (count1 = 0; count1 < crossword_record->y_dimensions; count1++)
	    mirror_grid[count1][count2] = 
		    barred_grid[crossword_record->y_dimensions - count2 - 1]
				  [count1];

    /*
     *  left -> top
     *  right -> bottom
     *  top -> right
     *  bottom -> left
     *
     *	This check is very much dependent on the 90 degree anti-clockwise
     *	transformation.
     *
     *	FYI: If the transformation is going 90 degree clockwise then
     *	mirror_grid and barred_grid in the check below should be switched
     *	around.
     */

    for (count1 = 0; count1 < crossword_record->x_dimensions; count1++)
    {
	for (count2 = 0; count2 < crossword_record->y_dimensions; count2++)
	{
	    if ( (!(mirror_grid[count1][count2] & LEFT)) && 
		  (barred_grid[count1][count2] & TOP) &&
		   (count1 != 0) )
	    {
#if 0
		printf ("# LEFT count1 is %u and count2 is %u\n", count1, count2);
		BARRED_GRID_PRINT_CHECK (barred_grid[count1][count2]);
		MIRROR_GRID_PRINT_CHECK (mirror_grid[count1][count2]);
#endif

		is_symmetry_found = FALSE;
		break;
	    }

	    if ( (!(mirror_grid[count1][count2] & RIGHT)) &&
		  (barred_grid[count1][count2] & BOTTOM) &&
		   (count1 != (crossword_record->x_dimensions - 1)) )
	    {
#if 0
		printf ("# RIGHT count1 is %u and count2 is %u\n", count1, count2);
		BARRED_GRID_PRINT_CHECK (barred_grid[count1][count2]);
		MIRROR_GRID_PRINT_CHECK (mirror_grid[count1][count2]);
#endif

		is_symmetry_found = FALSE;
		break;
	    }

	    if ( (!(mirror_grid[count1][count2] & TOP)) &&
	          (barred_grid[count1][count2] & RIGHT) &&
		   (count2 != (crossword_record->y_dimensions - 1)) )
	    {
#if 0
		printf ("# TOP count1 is %u and count2 is %u\n", count1, count2);
		BARRED_GRID_PRINT_CHECK (barred_grid[count1][count2]);
		MIRROR_GRID_PRINT_CHECK (mirror_grid[count1][count2]);
#endif

		is_symmetry_found = FALSE;
		break;
	    }

	    if ( (!(mirror_grid[count1][count2] & BOTTOM)) &&
		  (barred_grid[count1][count2] & LEFT) &&
                   (count2 != 0) )
	    {
#if 0
		printf ("# BOTTOM count1 is %u and count2 is %u\n", count1, count2);
		BARRED_GRID_PRINT_CHECK (barred_grid[count1][count2]);
		MIRROR_GRID_PRINT_CHECK (mirror_grid[count1][count2]);
#endif

		is_symmetry_found = FALSE;
		break;
	    }
	}

	if (is_symmetry_found == FALSE)
	{
	    break;
	}
    }

    if (is_symmetry_found == TRUE)
    {
	return (DEG_90_SYMMETRY);
    }

    /*
     *	If 90 degree symmetry isn't found then check for 180 degree next!
     *
     *	It doesn't matter whether the transformation is going clockwise or
     *	anti-clockwise.
     */

    is_symmetry_found = TRUE;
    for (count1 = 0; count1 < crossword_record->x_dimensions; count1++)
	for (count2 = 0; count2 < crossword_record->y_dimensions; count2++)
	    mirror_grid[count1][count2] = 
		    barred_grid[crossword_record->x_dimensions - count1 - 1]
				  [crossword_record->y_dimensions - count2 - 1];

    /*
     *  left -> right
     *  right -> left
     *  top -> bottom
     *  bottom -> top
     */

    for (count1 = 0; count1 < crossword_record->x_dimensions; count1++)
    {
	for (count2 = 0; count2 < crossword_record->y_dimensions; count2++)
	{
	    if ( (!(barred_grid[count1][count2] & LEFT)) && (mirror_grid[count1][count2] & RIGHT) )
	    {
		is_symmetry_found = FALSE;
		break;
	    }

	    if ( (!(barred_grid[count1][count2] & RIGHT)) && (mirror_grid[count1][count2] & LEFT) )
	    {
		is_symmetry_found = FALSE;
		break;
	    }

	    if ( (!(barred_grid[count1][count2] & TOP)) && (mirror_grid[count1][count2] & BOTTOM) )
	    {
		is_symmetry_found = FALSE;
		break;
	    }

	    if ( (!(barred_grid[count1][count2] & BOTTOM)) && (mirror_grid[count1][count2] & TOP) )
	    {
		is_symmetry_found = FALSE;
		break;
	    }
	}

	if (is_symmetry_found == FALSE)
	{
	    break;
	}
    }

    if (is_symmetry_found == TRUE)
    {
	return (DEG_180_SYMMETRY);
    }

    /*
     *	?? Keep this code in for the time being until some barred grid using
     *	these symmetry types have been created ??
     */

    printf ("## Code not been tested on LR/RL grids yet\n");

    /*
     *	If 90 degree/180 degree symmetry not found then check for other axial
     *	symmetries - Left-top to Right-bottom!
     */

    is_symmetry_found = TRUE;
    for (count1 = 0; count1 < crossword_record->x_dimensions; count1++)
	for (count2 = 0; count2 < crossword_record->y_dimensions; count2++)
	    mirror_grid[count1][count2] = 
		    barred_grid[crossword_record->y_dimensions - count2 - 1]
				  [crossword_record->x_dimensions - count1 - 1];

    /*
     *  left -> top
     *  right -> bottom
     *  top -> left
     *  bottom -> right
     */

    for (count1 = 0; count1 < crossword_record->x_dimensions; count1++)
    {
	for (count2 = 0; count2 < crossword_record->y_dimensions; count2++)
	{
	    if ( (barred_grid[count1][count2] & LEFT) && (!(mirror_grid[count1][count2] & TOP)) )
	    {
		is_symmetry_found = FALSE;
		break;
	    }

	    if ( (barred_grid[count1][count2] & RIGHT) && (!(mirror_grid[count1][count2] & BOTTOM)) )
	    {
		is_symmetry_found = FALSE;
		break;
	    }

	    if ( (barred_grid[count1][count2] & TOP) && (!(mirror_grid[count1][count2] & LEFT)) )
	    {
		is_symmetry_found = FALSE;
		break;
	    }

	    if ( (barred_grid[count1][count2] & BOTTOM) && (!(mirror_grid[count1][count2] & RIGHT)) )
	    {
		is_symmetry_found = FALSE;
		break;
	    }
	}

	if (is_symmetry_found == FALSE)
	{
	    break;
	}
    }

    if (is_symmetry_found == TRUE)
    {
	return (AXIAL_LR_SYMMETRY);
    }

    /*
     *	If 90 degree/180 degree/Left-top Right-bottom symmetry not found then
     *	check for other axial symmetries - Right-top to Left-bottom!
     */

    is_symmetry_found = TRUE;
    for (count2 = 0; count2 < crossword_record->x_dimensions; count2++)
	for (count1 = 0; count1 < crossword_record->y_dimensions; count1++)
	    mirror_grid[count1][count2] = 
		    barred_grid[count2][count1];

    /*
     *  left -> bottom
     *  right -> top
     *  top -> right
     *  bottom -> left
     */

    for (count1 = 0; count1 < crossword_record->x_dimensions; count1++)
    {
	for (count2 = 0; count2 < crossword_record->y_dimensions; count2++)
	{
	    if ( (barred_grid[count1][count2] & LEFT) && (!(mirror_grid[count1][count2] & BOTTOM)) )
	    {
		is_symmetry_found = FALSE;
		break;
	    }

	    if ( (barred_grid[count1][count2] & RIGHT) && (!(mirror_grid[count1][count2] & TOP)) )
	    {
		is_symmetry_found = FALSE;
		break;
	    }

	    if ( (barred_grid[count1][count2] & TOP) && (!(mirror_grid[count1][count2] & RIGHT)) )
	    {
		is_symmetry_found = FALSE;
		break;
	    }

	    if ( (barred_grid[count1][count2] & BOTTOM) && (!(mirror_grid[count1][count2] & LEFT)) )
	    {
		is_symmetry_found = FALSE;
		break;
	    }
	}

	if (is_symmetry_found == FALSE)
	{
	    break;
	}
    }

    if (is_symmetry_found == TRUE)
    {
	return (AXIAL_RL_SYMMETRY);
    }

    /*
     *	If we get down to here then no symmetry has been found so the only
     *	symmetry left is: No Symmetry!
     */

    return (NO_SYMMETRY);

}



PUBLIC int cword_verify_get_grid_symmetry (
	CROSSWORD_RECORD	*crossword_record)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function to check the symmetry of an inputted crossword grid.  The
 *	symmetry_type variable will determine what symmetry to check for.
 *
 *  FORMAL PARAMETERS:
 *
 *      *crossword_record	(r) Pointer to crossword definition
 *
 *  FUNCTION VALUE:
 *
 *      Valid symmetry type
 *--
 */

{

    int symmetry_type = INVALID_SYMMETRY;

    if (crossword_record->is_completely_specified == FALSE)
	cword_utils_complete_definition (crossword_record);

    /*
     *	Select the grid type before continuing with the verification of the
     *	cell symmetry.
     */

    if (cword_verify__free_cell_found (crossword_record) == TRUE)
    {
	return (cword_verify_get_block_symm (crossword_record));
    }
    else
    {
	return (cword_verify_get_barred_symm (crossword_record));
    }

}



PUBLIC int cword_verify_get_letters_grid (
	CROSSWORD_RECORD	*crossword_record,
	CROSSWORD_GRID		letters_grid)
	    
/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function to obtain the cell/letter properties of the inputted crossword
 *	definition.
 *
 *	Note: it does no error checking of the input crossword definition.
 *
 *	The possible value of any element of letters_grid can be one of the
 *	following:
 *
 *	    1. a-z - taken from crossword_record's clue_answer member.
 *	    2. CWORD_UNCHED_CHAR.
 *
 *  FORMAL PARAMETERS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  {@function value or completion codes@}
 *
 *      [@description or none@]
 *--
 */

{

    int count1,
	count2,
	temp1,
	temp2,
	letter1,
	letter2;

    for (count1 = 0; count1 < CWORD_SIZEOF_CELLDIM; count1++)
	for (count2 = 0; count2 < CWORD_SIZEOF_CELLDIM; count2++)
	    letters_grid[count1][count2] = NOT_TAKEN;    

    for (count1 = 0; count1 < crossword_record->clue_sum; count1++)
    {
	if (crossword_record->clue_details[count1].clue_direction == ACROSS)
	{
	    letter1 = 0;
	    for (count2 = crossword_record->clue_details[count1].x_coordinate;
		  count2 < (crossword_record->clue_details[count1].x_coordinate + 
			     crossword_record->clue_details[count1].answer_length);
		   count2++)
	    {
		temp2 = crossword_record->clue_details[count1].y_coordinate;

		letters_grid[count2 - 1][temp2 - 1] = 
			(int) crossword_record->clue_details[count1].clue_answer[letter1];

		letter1++;
		
	    }
	}
	else
	{
	    letter2 = 0;
	    for (count2 = crossword_record->clue_details[count1].y_coordinate;
		  count2 > (crossword_record->clue_details[count1].y_coordinate -
			     crossword_record->clue_details[count1].answer_length);
		   count2--)
	    {
		temp1 = crossword_record->clue_details[count1].x_coordinate;

		letters_grid[temp1 - 1][count2 - 1] =
			(int) crossword_record->clue_details[count1].clue_answer[letter2];

		letter2++;
	    }
	}
    }

    /*
     *	Change NOT_TAKEN to CWORD_UNCHED_CHAR, otherwise leave alone.
     */

    for (count1 = 0; count1 < crossword_record->x_dimensions; count1++)
	for (count2 = 0; count2 < crossword_record->y_dimensions; count2++)
	{
	    if (letters_grid[count1][count2] == NOT_TAKEN)
		letters_grid[count1][count2] = CWORD_UNFILLED_CHAR;
	}


#ifdef DEBUG
    printf ("# Results of cword_verify_get_letters_grid are:\n");
    for (count2 = 0; count2 < crossword_record->y_dimensions; count2++)
    {
	printf ("# -> ");
	for (count1 = 0; count1 < crossword_record->x_dimensions; count1++)
	{
	    printf(
		"%c",
		letters_grid[count1][crossword_record->y_dimensions - 1 - count2]);
	}
	printf (" <- ##\n");
    }
    printf ("\n");
#endif

    return (CWORD_S_NORMAL);

}



PUBLIC int cword_verify_get_block_props (
	CROSSWORD_RECORD	*crossword_record,
	CROSSWORD_GRID		blocked_grid)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Calculate the individual cell properties of a block [Block, American,
 *	or Novelty] grid.
 *
 *	The possible value of any element of blocked_grid can be one of the
 *	following:
 *
 *	    1. CWORD_UNFILLED_CHAR 
 *	    2. CWORD_UNCHED_CHAR
 *	    3. CWORD_CHECKED_CHAR
 *          4. CWORD_UNKNOWN_CHAR
 *
 *  FORMAL PARAMETERS:
 *
 *      *crossword_record	(r) Pointer to crossword definition
 *	blocked_grid		(w) array containing the completed checked grid
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL		Normal Completion
 *--
 */

{

    int count1,
	count2,
	temp1,
	temp2,
	status = CWORD_S_NORMAL;

    for (count1 = 0; count1 < CWORD_SIZEOF_CELLDIM; count1++)
	for (count2 = 0; count2 < CWORD_SIZEOF_CELLDIM; count2++)
	    blocked_grid[count1][count2] = NOT_TAKEN;

    for (count1 = 0; count1 < crossword_record->clue_sum; count1++)
    {
	if (crossword_record->clue_details[count1].clue_direction == ACROSS)
	{
	    for (count2 = crossword_record->clue_details[count1].x_coordinate;
		  count2 < (crossword_record->clue_details[count1].x_coordinate + 
			     crossword_record->clue_details[count1].answer_length);
		   count2++)
	    {
		temp2 = crossword_record->clue_details[count1].y_coordinate;

		/*
		 *  Checks the grid for locations on the grid where three of
		 *  more clues either originate or cross through. Illegal!
		 *
		 *  Obtains information about unches for use in validating the
		 *  grid against adjacent unches. Same below!
		 */

		switch (blocked_grid[count2 - 1][temp2 - 1])
		{
		case NOT_TAKEN : 
		    blocked_grid[count2 - 1][temp2 - 1] = UNCH;
		    break;

		case UNCH : 
		    blocked_grid[count2 - 1][temp2 - 1] = CHECKED;
		    break;

		case CHECKED: 
		default:
		    cword_messages_output_message (
				CWORD_VERBLOCKPROP,
				(TRACE2 | SCREEN2),
				temp2, 
				count2);

		    status = CWORD_S_FAILURE;
		    break;
		}

		DEBUG_TRACE (
			"Across - now at location %2d %2d clue %s\n",
			count2 - 1,
			temp2 - 1,
			crossword_record->clue_details[count1].clue_answer);

		if (status == CWORD_S_FAILURE)
		{   
		    break;
		}

	    }
	}
	else
	{
	    for (count2 = crossword_record->clue_details[count1].y_coordinate;
		  count2 > (crossword_record->clue_details[count1].y_coordinate -
			     crossword_record->clue_details[count1].answer_length);
		   count2--)
	    {
		temp1 = crossword_record->clue_details[count1].x_coordinate;

		switch (blocked_grid[temp1 - 1][count2 - 1])
		{
		case NOT_TAKEN : 
		    blocked_grid[temp1 - 1][count2 - 1] = UNCH;
		    break;

		case UNCH : 
		    blocked_grid[temp1 - 1][count2 - 1] = CHECKED;
		    break;

		case CHECKED:
		default:
		    cword_messages_output_message (
				CWORD_VERBLOCKPROP,
				(TRACE2 | SCREEN2),
				temp1, 
				count2);

		    status = CWORD_S_FAILURE;
		    break;
		}

		DEBUG_TRACE (
			"Down - now at location %2d %2d clue %s \n",
			temp1,
			count2,
			crossword_record->clue_details[count1].clue_answer);

		if (status == CWORD_S_FAILURE)
		{   
		    break;
		}

	    }
	}
    }

    /*
     *	Change NOT_TAKEN to CWORD_UNFILLED_CHAR.
     *	Change UNCH to CWORD_UNCHED_CHAR.
     *	Change CHECKED to CWORD_CHECKED_CHAR.
     *  Otherwise CWORD_UNKNOWN_CHAR.
     */

    for (count1 = 0; count1 < crossword_record->x_dimensions; count1++)
	for (count2 = 0; count2 < crossword_record->y_dimensions; count2++)
	{
	    if (blocked_grid[count1][count2] == NOT_TAKEN)
		blocked_grid[count1][count2] = CWORD_UNFILLED_CHAR;
	    else if (blocked_grid[count1][count2] == UNCH)
		blocked_grid[count1][count2] = CWORD_UNCHED_CHAR;
	    else if (blocked_grid[count1][count2] == CHECKED)
		blocked_grid[count1][count2] = CWORD_CHECKED_CHAR;
	    else
		blocked_grid[count1][count2] = CWORD_UNKNOWN_CHAR;    		
	}

#ifdef DEBUG
    printf ("# Results of cword_verify_get_block_props are:\n");
    for (count2 = 0; count2 < crossword_record->y_dimensions; count2++)
    {
	printf ("# -> ");
	for (count1 = 0; count1 < crossword_record->x_dimensions; count1++)
	{
	    printf (
		"%c",
		blocked_grid[count1][crossword_record->y_dimensions - 1 - count2]);
	}
	printf (" <- ##\n");
    }
    printf ("\n");
#endif

    return (status);

}



PUBLIC int cword_verify_get_barred_props (
        CROSSWORD_RECORD	*crossword_record,
	CROSSWORD_GRID		barred_grid)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Calculate the individual cell properties of a proven barred grid
 *	crossword.
 *
 *	The properties of each cell of the barred_grid array can be one of the
 *	following:
 *
 *	    1. NOT_TAKEN
 *	    2. LEFT
 *	    3. RIGHT
 *	    4. TOP
 *	    5. BOTTOM
 *	    6. TOP_BOTTOM
 *	    7. LEFT_RIGHT
 *	    8. ALL
 *
 *  FORMAL PARAMETERS:
 *
 *      *crossword_record	(r) pointer to crossword struct containing the clues
 *	barred_grid		(w) array containing the completed checked grid
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL		Successful completion
 *--
 */

{

    char temp_str[CWORD_SIZEOF_LINE + 1];

    int	count1, 
	count2,
	temp1,
	temp2,
	letter1,
	letter2;

    for (count1 = 0; count1 < CWORD_SIZEOF_CELLDIM; count1++)
	for (count2 = 0; count2 < CWORD_SIZEOF_CELLDIM; count2++)
	    barred_grid[count1][count2] = NOT_TAKEN;    
    
    /*
     *  Calculate the individual crossword cells properties
     *
     *  1. NOT_TAKEN    blank
     *
     *			-----
     *  2. LEFT	        ||  |
     *			-----
     *
     *			-----
     *  3. RIGHT	|  ||
     *			-----
     *
     *			====
     *  4. TOP		|  |
     *			----
     *
     *			----
     *  5. BOTTOM	|  |
     *			====
     *
     *			====
     *  6. TOP_BOTTOM	|  |
     *			====
     *
     *			------
     *  7. LEFT_RIGHT	||  ||
     *			------
     *
     *			======
     *  8. ALL		||  ||
     *			======
     */

    /*
     *  Calculate the various properties of each of the individual cells in the crossword
     */ 

    for (count1 = 0; count1 < crossword_record->clue_sum; count1++)
    {
	if (crossword_record->clue_details[count1].clue_direction == ACROSS)
	{
	    letter1 = 0;
	    temp2 = crossword_record->clue_details[count1].y_coordinate;

	    for (count2 = crossword_record->clue_details[count1].x_coordinate;
		  count2 < (crossword_record->clue_details[count1].x_coordinate + 
			     crossword_record->clue_details[count1].answer_length);
		   count2++)
	    {
		if (crossword_record->clue_details[count1].clue_answer[++letter1] == (char) '\0') 
		    barred_grid[count2 - 1][temp2 - 1] |= LEFT;
		else
		{
		    if (letter1 != 1)
			barred_grid[count2 - 1][temp2 - 1] |= LEFT_RIGHT;
		    else
			barred_grid[count2 - 1][temp2 - 1] |= RIGHT;			
		}
	    }
	}
	else
	{
	    letter2 = 0;
	    temp1 = crossword_record->clue_details[count1].x_coordinate;

	    for (count2 = crossword_record->clue_details[count1].y_coordinate;
		  count2 > (crossword_record->clue_details[count1].y_coordinate -
			     crossword_record->clue_details[count1].answer_length);
		   count2--)
	    {
		if (crossword_record->clue_details[count1].clue_answer[++letter2] == '\0') 
		    barred_grid[temp1 - 1][count2 - 1] |= TOP;
		else
		    if (letter2 != 1)
			barred_grid[temp1 - 1][count2 - 1] |= TOP_BOTTOM;
		    else
			barred_grid[temp1 - 1][count2 - 1] |= BOTTOM;
	    }
	}
    }

    (void) temp_str;

#if DEBUG

    /*
     *  Useful when debugging!
     */

    printf ("# Sides of the individual cells in the barred grid which are **thin** :\n");

    for (count1 = 0; count1 < crossword_record->x_dimensions; count1++)
    {
	printf ("-----");
    }
    printf ("-\n");

    /*
     *  ?? This debug outputting has not tested for non-square grids ??
     */

    for (count2 = 0; count2 < crossword_record->y_dimensions; count2++)
    {
	strcpy (temp_str, "|");

	for (count1 = 0; count1 < crossword_record->x_dimensions; count1++)
	{
	    /*
	     *	Output the positions which are adjacent to the borders of the
	     *	grid.  This doesn't include the outer border characteristics.
	     */

	    if (barred_grid[count1][crossword_record->y_dimensions - 1 - count2] & LEFT)
		    strcat (temp_str, "L");
	    else
		    strcat (temp_str, " ");
	    if (barred_grid[count1][crossword_record->y_dimensions - 1 - count2] & RIGHT)
		    strcat (temp_str, "R");
	    else
		    strcat (temp_str, " ");
	    if (barred_grid[count1][crossword_record->y_dimensions - 1 - count2] & TOP)
		    strcat (temp_str, "T");
	    else
		    strcat (temp_str, " ");
	    if (barred_grid[count1][crossword_record->y_dimensions - 1 - count2] & BOTTOM)
		    strcat (temp_str, "B");
	    else
		    strcat (temp_str, " ");

	    strcat (temp_str, "|");
	}

	strcat (temp_str, "\n");
	printf (temp_str);
    }

    for (count1 = 0; count1 < crossword_record->x_dimensions; count1++)
    {
	printf ("-----");
    }
    printf ("-\n");

    printf ("\n# Values correspondings to grid\n\n");

    for (count2 = 0; count2 < crossword_record->y_dimensions; count2++)
    {
	for (count1 = 0; count1 < crossword_record->x_dimensions; count1++)
	{
	    printf (
		"%3u ",
		barred_grid[count1][crossword_record->x_dimensions - 1 - count2]); 
	}
	printf ("\n");
    }

#endif

    return (CWORD_S_NORMAL);

}



/*
 *  ?? This will only work with fully specified crosswords.  Make work with
 *  partly specified crosswords ??
 */

PUBLIC int cword_verify_check_letters (
	CROSSWORD_RECORD	*crossword_record)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function which checks that there are no conflicting letters in the
 *	lights/answers of the crossword clues - only works where lights/answers
 *	cells overlap, of course.
 *
 *  FORMAL PARAMETERS:
 *
 *      *crossword_record	(r) Pointer to crossword definition
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL		None of the crossword light/answer letters are unequal
 *	CWORD_S_FAILURE		Some crossword light/answers are unequal
 *--
 */

{

    CROSSWORD_GRID overlap_props;

    int count1,
	count2,
	temp1,
	temp2,
	letter1,
	letter2,
	check_letters;

    check_letters = TRUE;

    for (count1 = 0; count1 < CWORD_SIZEOF_CELLDIM; count1++)
	for (count2 = 0; count2 < CWORD_SIZEOF_CELLDIM; count2++)
	    overlap_props[count1][count2] = NOT_TAKEN;    

    for (count1 = 0; count1 < crossword_record->clue_sum; count1++)
    {
	if (crossword_record->clue_details[count1].clue_direction == ACROSS)
	{
	    letter1 = 0;
	    for (count2 = crossword_record->clue_details[count1].x_coordinate;
		  count2 < (crossword_record->clue_details[count1].x_coordinate + 
			     crossword_record->clue_details[count1].answer_length);
		   count2++)
	    {
		temp2 = crossword_record->clue_details[count1].y_coordinate;

		if (overlap_props[count2 - 1][temp2 - 1] == NOT_TAKEN)
		{
		    overlap_props[count2 - 1][temp2 - 1] = 
			    (int) crossword_record->clue_details[count1].clue_answer[letter1];
		}
		else
		{
		    if (overlap_props[count2 - 1][temp2 - 1] != 
			    (int) crossword_record->clue_details[count1].clue_answer[letter1])
		    {
			check_letters = FALSE;

			cword_messages_output_message (
				CWORD_S_CHECKLETTERPROB,
				(SCREEN2 | TRACE2),
				crossword_record->clue_details[count1].clue_answer[letter1],
				crossword_record->clue_details[count1].clue_answer);

			break;
		    }
		}

		letter1++;
	    }
	}
	else
	{
	    letter2 = 0;
	    for (count2 = crossword_record->clue_details[count1].y_coordinate;
		  count2 > (crossword_record->clue_details[count1].y_coordinate -
			     crossword_record->clue_details[count1].answer_length);
		   count2--)
	    {
		temp1 = crossword_record->clue_details[count1].x_coordinate;

		if (overlap_props[temp1 - 1][count2 - 1] == NOT_TAKEN)
		{
		    overlap_props[temp1 - 1][count2 - 1] =
			    (int) crossword_record->clue_details[count1].clue_answer[letter2];
		}
		else
		{
		    if (overlap_props[temp1 - 1][count2 - 1] != 
			    (int) crossword_record->clue_details[count1].clue_answer[letter2])
		    {
			check_letters = FALSE;

			cword_messages_output_message (
				CWORD_S_CHECKLETTERPROB,
				(SCREEN2 | TRACE2),
				crossword_record->clue_details[count1].clue_answer[letter2],
				crossword_record->clue_details[count1].clue_answer);

			break;
		    }
		}

		letter2++;
	    }
	}
    }

#ifdef DEBUG

    printf ("# Elements of the cell in the grid:\n");

    for (count1 = 0; count1 < crossword_record->x_dimensions; count1++)
    {
	printf ("--");
    }
    printf ("--\n");

    for (count2 = 0; count2 < crossword_record->x_dimensions; count2++)
    {
	printf ("|");
	for (count1 = 0; count1 < crossword_record->x_dimensions; count1++)
	{
	    printf (
		"%c ", 
		(char) overlap_props[count1][crossword_record->x_dimensions - 1 - count2]); 
	}
	printf ("|\n");
    }

    for (count1 = 0; count1 < crossword_record->x_dimensions; count1++)
    {
	printf ("--");
    }
    printf ("--\n");

#endif

    if (check_letters == FALSE)
    {
	return (CWORD_S_FAILURE);
    }
    else
    {
	return (CWORD_S_NORMAL);
    }

}



PUBLIC int cword_verify_check_unches (
        CROSSWORD_RECORD	*crossword_record)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function which finds all adjacent unches, if any, and report the status
 *	back to the calling function.  If this function finds any adjacent
 *	unches, CWORD_S_FAILURE is returned.
 *
 *  FORMAL PARAMETERS:
 *
 *      *crossword_record	(r) Pointer to crossword definition
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL		No adjacent unches found in the crossword
 *	CWORD_S_FAILURE		Adjacent unches found in the crossword
 *--
 */

{

    CROSSWORD_GRID unches_grid;

    int count1,
	count2,
	status;

    /*
     *  Get the cell properties.
     */

    status = cword_verify_get_block_props (
		    crossword_record,
		    unches_grid);

    if (CWORD_S_NORMAL != status)
    {
	return (CWORD_S_FAILURE);
    }

    /* 
     *  Checks across clues 
     */

    for (count2 = crossword_record->x_dimensions - 1; count2 >= 0; count2--)
    {
	for (count1 = 0; count1 < crossword_record->y_dimensions - 1; count1++)
	{
	    if ( (unches_grid[count1][count2] == CWORD_UNCHED_CHAR) &&
		  (unches_grid[count1 + 1][count2] == CWORD_UNCHED_CHAR) )
	    {
		cword_messages_output_message (
			    CWORD_VERADJACUNCHES,
			    (TRACE2 | SCREEN2),
			    count1 + 1,
			    count2 + 1,
			    count1 + 2,
			    count2 + 1);

		status = CWORD_S_FAILURE;

		break;
	    }
	}

	if (status == CWORD_S_FAILURE)
	    break;
    }

    if (status != CWORD_S_NORMAL)    
    {
	return (status);
    }
	
    /* 
     *  Checks down clues 
     */

    for (count1 = 0; count1 < crossword_record->x_dimensions; count1++)
    {
	for (count2 = crossword_record->y_dimensions - 1; count2 > 0; count2--)
	{
	    if ( (unches_grid[count1][count2] == CWORD_UNCHED_CHAR) &&
		  (unches_grid[count1][count2 - 1] == CWORD_UNCHED_CHAR) )
	    {
		cword_messages_output_message (
			    CWORD_VERADJACUNCHES,
			    (TRACE2 | SCREEN2),
			    count1 + 1,
			    count2 + 1,
			    count1 + 1,
			    count2);

		status = CWORD_S_FAILURE;

		break;
	    }
	}

	if (status == CWORD_S_FAILURE)
	    break;
    }

    return (status);

}



PRIVATE int cword_verify__isloated_grid (
        int			initialised,
	int			clue_subscript,
	CROSSWORD_RECORD	*crossword_record)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Function to check whether a crossword grid is fully connected or not.
 *	What I mean by fully connected is that it is possible to follow a path
 *	from one light to another all the way around the puzzle.   The function
 *	will call itself until it exhausts all possibilities or finds a
 *	connected path.
 *
 *  FORMAL PARAMETERS:
 *
 *      initialised		(r) TRUE/FALSE
 *      clue_subscript		(r) Subscript of clue in crossword_record
 *      *crossword_record	(r) Pointer to crossword definition
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL		Grid is properly connected
 *	CWORD_S_FAILURE		Grid is not properly connected
 *	CWORD_CONTINUE		The jury is still out
 *--
 */

{

    /*
     *	Array to store the clue numbers that have been successfully checked.
     */

    static short connected_answers[CWORD_SIZEOF_CLUENUM];

    int count1,	
	count2,
	status;

    BOOLEAN is_found;

    if (crossword_record->clue_sum > 255)
    {
	printf ("# Maximum number of clues that this function can handle is 255.\n");

	return (CWORD_S_FAILURE);
    }

#ifdef DEBUG

    if (initialised == FALSE)
	printf ("# Isolated grid check");

    printf (
	"# clue %d %s is being added to the list\n", 
	crossword_record->clue_details[0].clue_number,
	crossword_record->clue_details[0].clue_direction ? CWORD_ACROSS_STR : CWORD_DOWN_STR);

#endif

    if (initialised == FALSE)
    {
	for (count1 = 0; count1 < CWORD_SIZEOF_CLUENUM; count1++)
	    connected_answers[count1] = NOT_DEFINED;

	connected_answers[0] = clue_subscript;
	
	DEBUG_TRACE (
	    "Clue %d or %d %s is added to the list\n", 
	    clue_subscript,
	    crossword_record->clue_details[clue_subscript].clue_number,
	    crossword_record->clue_details[clue_subscript].clue_direction ? CWORD_ACROSS_STR : CWORD_DOWN_STR);
    }

    if (crossword_record->clue_details[clue_subscript].clue_direction == ACROSS)
    {

	for (count1 = 0; count1 < crossword_record->clue_sum; count1++)
	{
	    /*
	     *  Does:
	     *  1. [count1].x_coordinate cut (&&) the range [sub].x_coordinate <--> [sub].x_coordinate + [sub].length - 1 (&&)
	     *  2. [sub].y_coordinate cut (&&) the range [count1].y_coordinate <--> [count1].y_coordinate - [count1].length + 1 
	     */

	    if (crossword_record->clue_details[count1].clue_direction == DOWN)
	    {
		if ( ( (crossword_record->clue_details[count1].x_coordinate >= 
		         crossword_record->clue_details[clue_subscript].x_coordinate) &&
		       (crossword_record->clue_details[count1].x_coordinate <= 
		         (crossword_record->clue_details[clue_subscript].x_coordinate + 
                          crossword_record->clue_details[clue_subscript].answer_length - 1)) ) 
						&&
                     ( (crossword_record->clue_details[clue_subscript].y_coordinate <= 
		         crossword_record->clue_details[count1].y_coordinate) &&
		       (crossword_record->clue_details[clue_subscript].y_coordinate >= 
		         (crossword_record->clue_details[count1].y_coordinate - 
                          crossword_record->clue_details[count1].answer_length + 1)) ) )
		{

		    /*
		     *	Check whether the clue found to check the
		     *	clue_subscript clue isn't already in the
		     *	connect_answers array 
		     */

		    count2 = 0; is_found = FALSE;
		    while (connected_answers[count2] != NOT_DEFINED)
		    {
			if (connected_answers[count2] == count1)
			{
			    is_found = TRUE;
			    break;
			}
			else
			{
			    count2++;
			}
		    }

		    if (is_found == FALSE)
		    {
			DEBUG_TRACE (
			    "clue %d or %d %s is added to the list\n", 
			    count1,
			    crossword_record->clue_details[count1].clue_number,
			    crossword_record->clue_details[count1].clue_direction ? CWORD_ACROSS_STR : CWORD_DOWN_STR);

			connected_answers[count2] = count1;

			/*
			 *  ?? Check the return value ??
			 */

			status = cword_verify__isloated_grid (
					TRUE,
					count1,
					crossword_record);

			(void) status;
		    }
		}
	    }
	}
    }	
    else    /* start of clue_direction == DOWN */
    {   
	for (count1 = 0; count1 < crossword_record->clue_sum; count1++)
	{
	    /*
	     *  DOWN clues can only check ACROSS clues thus this check
	     */

	    if (crossword_record->clue_details[count1].clue_direction == ACROSS)
	    {
		/*
		 *  Does:
		 *  1. [count1].x_coordinate cut (||) the range [sub].x_coordinate <--> [sub].x_coordinate + [sub].length - 1 (&&)
		 *  2. [sub].y_coordinate cut (||) the range [count1].y_coordinate <--> [count1].y_coordinate - [count1].length + 1 
		 */

		if ( ( (crossword_record->clue_details[clue_subscript].x_coordinate >= 
		         crossword_record->clue_details[count1].x_coordinate) &&
		       (crossword_record->clue_details[clue_subscript].x_coordinate <= 
		         (crossword_record->clue_details[count1].x_coordinate + 
                          crossword_record->clue_details[count1].answer_length - 1)) )
						&&
                     ( (crossword_record->clue_details[count1].y_coordinate <= 
		         crossword_record->clue_details[clue_subscript].y_coordinate) &&
		       (crossword_record->clue_details[count1].y_coordinate >= 
		         (crossword_record->clue_details[clue_subscript].y_coordinate - 
                          crossword_record->clue_details[clue_subscript].answer_length + 1)) ) )

		{
		    /*
		     *	Check whether the clue found to check the
		     *	clue_subscript clue isn't already in the
		     *	connect_answers array 
		     */

		    count2 = 0; is_found = FALSE;
		    while (connected_answers[count2] != NOT_DEFINED)
		    {
			if (connected_answers[count2] == count1)
			{
			    is_found = TRUE;
			    break;
			}
			else
			{
			    count2++;
			}
		    }

		    if (is_found == FALSE)
		    {
			DEBUG_TRACE (
			    "Clue %d or %d %s is added to the list\n", 
			    count1,
			    crossword_record->clue_details[count1].clue_number,
			    crossword_record->clue_details[count1].clue_direction ? CWORD_ACROSS_STR : CWORD_DOWN_STR);

			connected_answers[count2] = count1;

			/*
			 *  ?? Check the return value ??
			 */

			status = cword_verify__isloated_grid (
					TRUE,
					count1,
					crossword_record);

			(void) status;
		    }
		}
	    }
	}
    }   /* end of clue_direction == DOWN */

    count1 = 0;
    while (connected_answers[count1] != NOT_DEFINED)
	count1++;

    if (count1 == crossword_record->clue_sum)
    {
	return (CWORD_S_NORMAL);	
    }
    else
    {
	return (CWORD_CONTINUE);
    }

}



PUBLIC BOOLEAN cword_verify_is_grid_connected (
	CROSSWORD_RECORD	*crossword_record)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Main calling function that checks whether a defined grid is fully
 *	connected or not.  This function will fail when there is an isolated
 *	region which cannot be checked by another part of the grid.  These
 *	crossword types are non standard.
 *
 *	Equation to check is:
 *
 *	    A_X <= D_X <= A-X + A_Length - 1
 *
 *	    D_Y >= A_Y >= D_Y - D_length + 1
 *
 *  FORMAL PARAMETERS:
 *
 *      *crossword_record	(r) Pointer to crossword definition
 *
 *  FUNCTION VALUE:
 *
 *      TRUE			Fully connected grid was found.
 *	FALSE			Fully connected grid was *not* found.
 *--
 */
{

    int status;

    /*
     *  Call cword_verify__isloated_grid with the first element of the grid usually 1 across
     */

    status = cword_verify__isloated_grid (
		FALSE,
		0, /* Kick the verification off */
		crossword_record);

    if (status == CWORD_CONTINUE)
    {
	/*
	 *  Grid is *not* fully specified.
	 */

	return (FALSE);
    }
    else
    {
	/*
	 *  Grid is fully connected.
	 */

	return (TRUE);
    }

}



PRIVATE BOOLEAN cword_verify__free_cell_found (
	CROSSWORD_RECORD	*crossword_record)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	Checks for any free cells in the inputted crossword grid definition.
 *	Only the cell properties are analysed in the grid, nothing else.
 *
 *  FORMAL PARAMETERS:
 *
 *      *crossword_record	(r) Pointer to crossword definition
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      TRUE			Found no free cells => barred grid
 *	FALSE			Found at least one free cell => blocked grid
 *--
 */

{

    CROSSWORD_GRID barred_props;

    int count1,
	count2,
	status;

    BOOLEAN is_cell_free = FALSE;

    /*
     *	Doesn't matter whether we are getting stats on barred of block grid!
     */

    status = cword_verify_get_block_props (
		    crossword_record,
		    barred_props);

    if (CWORD_S_NORMAL != status)
    {
	return (FALSE);
    }

    for (count2 = 0; count2 < crossword_record->x_dimensions; count2++)
    {
	for (count1 = 0; count1 < crossword_record->y_dimensions; count1++)
	{
	    if (CWORD_UNFILLED_CHAR == barred_props[count1][count2])    
	    {
		is_cell_free = TRUE;

		DEBUG_TRACE (
			"Free cell found at location (%d,%d)\n", 
			count1 + 1, 
			count2 + 1);

		break;
	    }
	}

	if (TRUE == is_cell_free)
	    break;
    }

    /*
     *	If a free cell found then it implies that we have a blocked grid.  If
     *	no free cells found then a barred grid has been found.
     */

    if (TRUE == is_cell_free)
	return (TRUE);
    else
	return (FALSE);

}



PRIVATE int cword_verify__common_grid_checks (
	CROSSWORD_RECORD	*crossword_record)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *      [@tbs@]
 *
 *  FORMAL PARAMETERS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  {@function value or completion codes@}
 *
 *      [@description or none@]
 *--
 */

{

    /*
     *  All common checks to the grid should be made here!
     */

    /* 
     *	Check the length of the answers with the inputted length of answers
     *	provided.
     */

    if (CWORD_S_NORMAL != cword_verify__check_inputs (crossword_record))
    {
	return (CWORD_S_FAILURE);
    }

    /* 
     *	Verify that there are no overlapping letters in the crossword defintion
     *	lights.
     */

    if (CWORD_S_NORMAL != cword_verify_check_letters (crossword_record))
    {
	cword_messages_output_message (
		CWORD_VERINVANSONGRID,
		(TRACE2 | SCREEN2));

	return (CWORD_S_FAILURE);
    }

    return (CWORD_S_NORMAL);

}



PUBLIC int cword_verify_novelty_grid_defn (
	CROSSWORD_RECORD	*crossword_record)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *      Function to check the validity of a Novelty grid type.
 *
 *  FORMAL PARAMETERS:
 *
 *      *crossword_record	(r) grid/crossword definition
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL		Crossword inputted is valid for this grid type
 *      CWORD_S_FAILURE		Crossword inputted is not valid for this grid type
 *--
 */

{

    int status;

    if (CWORD_S_NORMAL != cword_verify__common_grid_checks (crossword_record))
    {
	return (CWORD_S_FAILURE);
    }

    /*
     *	?? Everything else here -- if there is any other check to be made ??
     *
     *  With a novelty grid anything goes i.e.
     *
     *	1. No unch check
     *  2. No symmetry check
     *	3. Any degree of adjacency is allowed
     */

    return (CWORD_S_NORMAL);

}



PUBLIC int cword_verify_american_grid_defn (
	CROSSWORD_RECORD	*crossword_record)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *      Function to check the validity of an American grid type.
 *
 *  FORMAL PARAMETERS:
 *
 *      *crossword_record	(r) grid/crossword definition
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL		Crossword inputted is valid AMERICAN_GRID type
 *      CWORD_S_FAILURE		Crossword inputted is not a valid AMERICAN_GRID type
 *--
 */

{

    int status;

    /*
     *	An American style grid is square in shape and has all filled cells
     *	checked; no unches allowed.  It isn't barred though.  It follows one of
     *	the main symmetry types too.
     */

    if (CWORD_S_NORMAL != cword_verify__common_grid_checks (crossword_record))
    {
	return (CWORD_S_FAILURE);
    }

    /*
     *  ?? Symmetry check -- how to handle it most effectively ??
     */

    if (cword_g_defaults_record.symmetry_check == TRUE)    
    {
	int symmetry;

	if ( (crossword_record->x_dimensions != crossword_record->y_dimensions) && 
	      (crossword_record->symmetry != NO_SYMMETRY) )
	{
	    cword_messages_output_message (
			CWORD_VERINVDIMSYM,
			(TRACE2 | SCREEN2),
			CWORD_X_CHAR,
			CWORD_Y_CHAR);
	}

	symmetry = cword_verify_get_grid_symmetry (crossword_record);

	if (INVALID_SYMMETRY == symmetry)
	{
	    cword_messages_output_message (
			CWORD_S_INVLDSYMMMSG,
			(TRACE2 | SCREEN2));

	    return (CWORD_S_FAILURE);
	}


	if (crossword_record->symmetry == NOTDEFINED_SYMMETRY)
	{
	    /*
	     *  If no symmetry is defined then define it!
	     */

	    crossword_record->symmetry = symmetry;
	}
	else if (symmetry != crossword_record->symmetry)
	{
	    cword_verify__change_symmetry (
		    symmetry,
		    crossword_record);
	}
    }

    /*
     *	Check the cell percentage configurations i.e. % of checked/unchecked
     *	and blank cells.
     */

    if (TRUE == cword_verify__free_cell_found (crossword_record))
    {
	int blanks,
	    unches,
	    checks;

	/*
	 *  ?? Not currently using *checks* ??
	 */

	cword_stats_get_grid_percents (
		crossword_record,
		&blanks,
		&unches,
		&checks);

	if ( ((float) AMERICAN_BLANKS / (float) 100) < 
               (float) ((float) blanks / (float) (crossword_record->x_dimensions * 
			                           crossword_record->y_dimensions)) )
	{
	    cword_messages_output_message (
		    CWORD_S_AMERICANBLANKS,
		    SCREEN2,
		    AMERICAN_BLANKS);

	    return (CWORD_S_FAILURE);
	}

	if (unches != 0)
	{
	    cword_messages_output_message (
		    CWORD_S_UNCHESFOUND,
		    SCREEN2);

	    return (CWORD_S_FAILURE);
	}
    }

    /*
     *	If we get downto here then the American grid type is valid for the
     *	crossword/grid in question.
     */

    return (CWORD_S_NORMAL);

}



PUBLIC int cword_verify_block_grid_defn (
	CROSSWORD_RECORD	*crossword_record)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	This function will check *only* whether the inputted crossword
 *	definition is a valid BLOCKED_GRID grid definition or not.
 *
 *  FORMAL PARAMETERS:
 *
 *      *crossword_record	(r) Pointer to crossword definition
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL		BLOCK_GRID successfully found
 *	CWORD_S_FAILURE		BLOCK_GRID not found!
 *--
 */

{

    int	status;

    if (CWORD_S_NORMAL != cword_verify__common_grid_checks (crossword_record))
    {
	return (CWORD_S_FAILURE);
    }

    /*
     *	Check for free cells - if none found then the selected grid type not
     *	valid!
     */

    if (FALSE == cword_verify__free_cell_found (crossword_record))
    {
	cword_messages_output_message (
		    CWORD_S_NOFREECELLFOUND,
		    (TRACE2 | SCREEN2));

	return (CWORD_S_FAILURE);
    }

    /* 
     *  Verify grid symmetry
     */

    if (cword_g_defaults_record.symmetry_check == TRUE)    
    {
	int symmetry;

	if (crossword_record->x_dimensions != crossword_record->y_dimensions)
	{
	    cword_messages_output_message (
			CWORD_VERINVDIMSYM,
			(TRACE2 | SCREEN2),
			CWORD_X_CHAR,
			CWORD_Y_CHAR);
	}

	symmetry = cword_verify_get_grid_symmetry (crossword_record);

	if (INVALID_SYMMETRY == symmetry)
	{
	    cword_messages_output_message (
			CWORD_S_INVLDSYMMMSG,
			(TRACE2 | SCREEN2));

	    return (CWORD_S_FAILURE);
	}

	if (crossword_record->symmetry == NOTDEFINED_SYMMETRY)
	{
	    /*
	     *  If no symmetry is defined then define it!
	     */

	    crossword_record->symmetry = symmetry;
	}
	if (symmetry != crossword_record->symmetry)
	{
	    cword_verify__change_symmetry (
		    symmetry,
		    crossword_record);
	}
    }

    /* 
     *	Verify the block crossword grid unches - the lack of them adjoining
     *	each other is usually the sign of a well designed grid definition.
     */

    if (cword_g_defaults_record.unches_check == TRUE)
    {
	status = cword_verify_check_unches (crossword_record);

	if (CWORD_S_NORMAL != status)
	{
	    return (CWORD_S_FAILURE);
	}
    }

    /*
     *  BLOCK_GRID defintion found if we get down to here
     */

    return (CWORD_S_NORMAL);

}



PUBLIC int cword_verify_barred_grid_defn (
	CROSSWORD_RECORD	*crossword_record)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	This function will check whether the inputted crossword definition is a
 *	valid BARRED_GRID grid definition or not.
 *
 *  FORMAL PARAMETERS:
 *
 *      *crossword_record	(r) Pointer to crossword definition
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      CWORD_S_NORMAL		BARRED_GRID successfully found
 *      CWORD_S_FAILURE		BARRED_GRID not found!
 *--
 */

{

    if (CWORD_S_NORMAL != cword_verify__common_grid_checks (crossword_record))
    {
	return (CWORD_S_FAILURE);
    }

    /*
     *	Check for free cells in the grid.  If some are found then the grid type
     *	is not valid for this example!
     */

    if (TRUE == cword_verify__free_cell_found (crossword_record))
    {
	cword_messages_output_message (
		    CWORD_S_FREECELLFOUND,
		    (TRACE2 | SCREEN2));

	return (CWORD_S_FAILURE);
    }

    /*
     *	Check barred grid symmetry.
     */

    if (cword_g_defaults_record.symmetry_check == TRUE)    
    {
	int symmetry;

	if (crossword_record->x_dimensions != crossword_record->y_dimensions)
	{
	    cword_messages_output_message (
			CWORD_VERINVDIMSYM,
			(TRACE2 | SCREEN2),
			CWORD_X_CHAR,
			CWORD_Y_CHAR);
	}

	symmetry = cword_verify_get_grid_symmetry (crossword_record);

	if (INVALID_SYMMETRY == symmetry)
	{
	    cword_messages_output_message (
			CWORD_S_INVLDSYMMMSG,
			(TRACE2 | SCREEN2));

	    return (CWORD_S_FAILURE);
	}

	if (crossword_record->symmetry == NOTDEFINED_SYMMETRY)
	{
	    /*
	     *  If no symmetry is defined then define it!
	     */

	    crossword_record->symmetry = symmetry;
	}
	if (symmetry != crossword_record->symmetry)
	{
	    cword_verify__change_symmetry (
		    symmetry,
		    crossword_record);
	}
    }

    /* 
     *	Verify that there are no overlapping letters in the crossword defintion
     *	lights.
     */

    if (CWORD_S_NORMAL != cword_verify_check_letters (crossword_record))
    {
	cword_messages_output_message (
		CWORD_VERINVANSONGRID,
		(TRACE2 | SCREEN2));

	return (CWORD_S_FAILURE);
    }

    /*
     *  If we get down to here then a BARRED_GRID has been found successfully
     */

    return (CWORD_S_NORMAL);

}



PUBLIC int cword_verify_get_grid_type (
	CROSSWORD_RECORD	*crossword_record)

/*
 *++
 *  FUNCTIONAL DESCRIPTION:
 *
 *	This function checks whether the inputted crossword definition is one
 *	of the following grid types:
 *
 *	    1. Barred grid
 *	    2. American style grid
 *	    3. Block grid
 *	    4. Novelty grid
 *
 *	This function shold only be called if the grid type is not already
 *	known.
 *
 *  FORMAL PARAMETERS:
 *
 *	*crossword_record	(r) Pointer to crossword definition
 *
 *  IMPLICIT INPUTS:
 *
 *      [@description or none@]
 *
 *  IMPLICIT OUTPUTS:
 *
 *      [@description or none@]
 *
 *  FUNCTION VALUE:
 *
 *      BARRED_GRID		Barred grid found
 *	BLOCK_GRID              Block grid found
 *	AMERICAN_GRID		American grid found
 *	NOVELTY_GRID		Nolvelty grid found
 *	INVALID_GRID		Invalid grid type found
 *--
 */

{

    int status;

    if (crossword_record->is_completely_specified == FALSE)
	cword_utils_complete_definition (crossword_record);

    status = cword_verify_barred_grid_defn (crossword_record);

    if (CWORD_S_NORMAL == status)
    {
	/*
	 *  Valid barred grid found
	 */

	return (BARRED_GRID);
    }

    status = cword_verify_american_grid_defn (crossword_record);

    if (CWORD_S_NORMAL == status)
    {
	/*
	 *  Valid American type grid found
	 */

	return (AMERICAN_GRID);
    }

    status = cword_verify_block_grid_defn (crossword_record);
    
    if (CWORD_S_NORMAL == status)
    {
	/*
	 *  Valid block grid found
	 */

	return (BLOCK_GRID);
    }

    status = cword_verify_novelty_grid_defn (crossword_record);
    
    if (CWORD_S_NORMAL == status)
    {
	/*
	 *  Valid novelty type grid found
	 */

	return (NOVELTY_GRID);
    }

    return (INVALID_GRID);

}



#else

extern int main (
	int			argc,
	char			*argv[])

{

    CROSSWORD_RECORD	*crossword_record;

    int			status;

    crossword_record = (CROSSWORD_RECORD *) malloc (sizeof (CROSSWORD_RECORD));
    memset (crossword_record, 0, sizeof (crossword_record));

    status = cword_utils_generate_grid (
		    crossword_record);

    status = cword_verify_connected_grid (crossword_record);

    if (CWORD_CONTINUE == status)
    {
	status = cword_messages_output_message (
		status,
		(TRACE2 | SCREEN2));
    }

    return (CWORD_S_NORMAL);

}

#endif
